#ifndef CutApplicationProcessor_h
#define CutApplicationProcessor_h 1

// C++
#include <string>

// LCIO
#include "lcio.h"
#include <EVENT/TrackerHit.h>

// Marlin
#include "marlin/Processor.h"

namespace AIDA{
  class IHistogram1D;
  class IHistogram2D;
}

namespace marlintpc{

    /**  Applies cuts on all relevant track and hit parameters.
     *   If one hit on the track does not apply to all cut the complete track is skipped.
     *
     *   All cut parameters are optional parameters. If you don't give a value in the steering file
     *   the cut is not applied. The default values are meaningles.
     *
     *   The output collection is a subset of the input collection, which means only the pointers
     *   to the tracks are stored. Do not remove the input collection from the file.
     *   If the output collection is non transient also the input collection has to be non transient.
     *
     *  <h4>Input - Prerequisites</h4>
     *  \li A collection of tracks (lcio::Track). <br>
     *  
     *
     *  <h4>Output</h4> 
     *  \li A collection of tracks that fulfill the cut criteria
     * 
     * @param InputCollectionName Name of input collection of tracks, default TPCTracks
     *
     * !!!FIXME!!! List all parameters
     *
     * @author Simone Zimmermann, Martin Killenberg
     */
  
  
    class CutApplicationProcessor : public marlin::Processor {
	
    public:
	
	/* This method will be called by the marlin package
	 * It returns a processor of the currend type (here CutApplicationProcessor)
	 */
	virtual Processor*  newProcessor() { return new CutApplicationProcessor ; }
	
	CutApplicationProcessor() ;
	
	virtual void init() ;
	
	virtual void processRunHeader( lcio::LCRunHeader* run ) ;
	
	virtual void processEvent( lcio::LCEvent * evt ) ; 
	
    protected:
	
	std::string _inputColName ;  ///< the name of the input collection
	std::string _outputColName ;  ///< the name of the output collection

	/** give the status which is set to the transient flag of the output collection
	 * it is an int instead of a boolean, because the processor parameter cannot be a boolean
	 */
	int _outputIsTransient;

	// event specific cuts
	int _lowcutNTracks; ///<- lower cut on number of tracks
	int _highcutNTracks;///<- upper cut on number of tracks

	// track specific cuts
	int _lowcutNHitsOnTrack;///<- lower cut on number of hits on track
	int _highcutNHitsOnTrack;///<- upper cut on number of hits on track
	float _lowcutPhi;///<- lower cut on phi
	float _highcutPhi;///<- upper cut on phi
	float _lowcutTanLambda;///<- lower cut on tan lambda
	float _highcutTanLambda;///<- upper cut on tan lambda
	float _lowcutdEdx;///<- lower cut on dEdx (track)
	float _highcutdEdx;///<- upper cut on dEdx (track)
	float _lowcutOmega;///<- lower cut on omega
	float _highcutOmega;///<- upper cut on omega
	float _lowcutD0;///<- lower cut on D0
	float _highcutD0;///<- upper cut on D0
	float _lowcutZ0;///<- lower cut on Z0
	float _highcutZ0;///<- upper cut on Z0

	// hit specific cuts
	float _lowcutCharge;///<- lower cut on charge of hit
	float _highcutCharge;///<- upper cut on charge of hit
	double _lowcutX;///<- lower cut on x position of hit
	double _highcutX;///<- upper cut on x position of hit
	double _lowcutY;///<- lower cut on y position of hit
	double _highcutY;///<- upper cut on y position of hit
	double _lowcutZ;///<- lower cut on z position of hit
	double _highcutZ;///<- upper cut on z position of hit

//private:

    } ;
    
}

#endif
