#ifndef Z_BIN_TEMPLATE_PROCESSOR_H
#define Z_BIN_TEMPLATE_PROCESSOR_H 1

//LCIO
#include <lcio.h>

//MARLIN
#include "marlin/Processor.h"

namespace AIDA{
  class IHistogram1D;
  class IHistogram2D;
}


namespace marlintpc
{

    /** A template to create processors wich create 1D or 2D distribution histograms, one
     *  histogram per z bin.
     *  In addition one histogram is filled without any z cut.
     *  
     *  Usage: 
     *  \li Derrive your processor from this class. The number of 1D and 2D histogram sets
     *  are the template parameters. (Example: Two 1D histograms, x-Distribution and y-Distribution,
     *  and one 2D-Histogram xy-Distribtuion. For each of these a set of z-binned histograms is created,
     *  plus one histogram without z-cut each.)\n
     *  For convenience it is recommended to define index constants for the histogram sets:
     *  \code
     *  static const  xDistributionHistoIndex 0; // The templated set of x  histograms is the 1D index 0
     *  static const  yDistributionHistoIndex 1; // The templated set of y  histograms is the 1D index 1
     *  static const xyDistributionHistoIndex 0; // The templated set of xy histograms is the 2D index 0
     *  \endcode
     *  \li As this is a template class you have to include the \c ZBinTemplateProcessor.cc in your .cc file:
     *  \code
     *  #include "ZBinTemplateProcessor.cc"
     *  \endcode
     *  \li Set the histgram names, titles, ranges and binning in your init()
     *  \li <b> Call \c ZBinTemplateProcessor::init() in your init function! </b>
     *  This is needed because the AIDA histograms are registered here. The fillHistogram will
     *  throw and \c lcio::Exception if ZBinTemplateProcessor::init() has not been called.
     *  \li Fill the histograms using the fillHistogram method.
     *
     *  @param MinZ Minimum z value of lowest bin in case of z binning (default: 0)
     *  @param MaxZ Maximum z value of highest bin in case of z binning (default: 260)
     *  @param nZBins Number ob bins in z (default: 10)<br>
     *         z binning switched of if 0, only the histogram without z cut is created
     *
     *  As an example you can use HitAndTrackCounterProcessor.
     *
     *  @author Martin Killenberg, University of Bonn
     */

    template<unsigned int N1DHISTOS, unsigned int N2DHISTOS>
	class ZBinTemplateProcessor : public marlin::Processor
    {
      public:
	    
	/** The newProcessor() function does not make sense for this template class, so it throws an
	 *  \c lcio::Exception.
	 *
	 */
	virtual Processor*  newProcessor() { throw Exception("You cannot create an instance of ZBinTemplateProcessor"); }
	
	virtual ~ZBinTemplateProcessor();

	/// Register the AIDA histograms at the AIDAProcessor
	virtual void init();

	
	/** A nested convenience struct which contains the minimal and maximal value of a (1D) histogram rage.
	 */
	struct HistoRange
	{
	    float min;
	    float max;
	};


	/** A nested convenience struct which contains two 1D HistoRange objects.
	 */
	struct HistoRange2D
	{
	    HistoRange xRange;
	    HistoRange yRange;
	};

	/** A nested convenience struct which contains the number of bins for both directions of a 2D histogram.
	 */
	struct HistoNBins2D
	{
	    int nBinsX;
	    int nBinsY;
	};

      protected:

	ZBinTemplateProcessor(const std::string & typeName );

	/** Fill value \c x into the histogram set \c histoIndex. The correct z histogram is chosen according
	 *  to the  \c z value. In addition the histogram without z cut is filled.
	 *  Throws an \c lcio::Exception in case the histograms have not been initialised.
	 */
	void fillHistogram(unsigned int histoIndex, float x, float z);
	/** Fill the \c x and \c y value pair into the 2D histogram set \c histoIndex.
	 *  The correct z histogram is chosen according
	 *  to the  \c z value. In addition the histogram without z cut is filled.
	 *  Throws an \c lcio::Exception in case the histograms have not been initialised.
	 */
	void fillHistogram(unsigned int histoIndex, float x, float y, float z);
	    
	std::vector< std::vector<AIDA::IHistogram1D *> > _1DHistos; ///< Vector of 1D histogram vectors
	std::vector< std::vector<AIDA::IHistogram2D *> > _2DHistos; ///< Vector of 2D histogram vectors
	/** Vector of base names for the 1D histograms. For the z-binned histograms an index is appended,
	 *  the histogram without z-cut is left wothout index
	 */
	std::vector< std::string >  _histo1DBaseNames;
	/** Vector of base names for the 2D histograms. For the z-binned histograms an index is appended,
	 *  the histogram without z-cut is left wothout index
	 */
	std::vector< std::string >  _histo2DBaseNames;
	/** Vector of titles for the 1D histograms. 
	 *  For the z-binned histograms the z-range is appeded to the title.
	 */
	std::vector< std::string >  _histo1DTitles;
	/** Vector of titles for the 2D histograms. 
	 *  For the z-binned histograms the z-range is appeded to the title.
	 */
	std::vector< std::string >  _histo2DTitles;
	std::vector< HistoRange >   _histoRanges1D; ///< Vector of ranges for the 1D histograms
	std::vector< HistoRange2D > _histoRanges2D; ///< Vector of ranges for the 2D histograms
	std::vector< int > _nBins1D; ///< Vector with number of bins for the 1D histograms
	std::vector< HistoNBins2D > _nBins2D; ///< Vector with number of bins for the 1D histograms

	float _minZ; ///< Minimum z value of lowest bin
	float _maxZ; ///< Maximum z value of highest bin
	
	int _nZBins; ///< Number of z bins
	
	float _zBinWidth; ///< Width of a bin, calculated as (_maxZ-_minZ)/_nZBins;

      private:
	bool _isInitialised;
 
    };
}

#endif // Z_BIN_TEMLATE_PROCESSOR_H
