#include "BiasedResidualsProcessor.h"

//C++
#include <iostream>
using namespace std;

//LCIO
#include <IMPL/LCCollectionVec.h>
#include <EVENT/Track.h>
#include <EVENT/TrackerHit.h>
#include "IntersectionCalculator.h"

// Aida
//#ifdef MARLIN_USE_AIDA // the whole processor does not make sense without aida
#include <AIDA/AIDA.h>
#include <marlin/AIDAProcessor.h>
#include <AIDA/IHistogramFactory.h>
#include <AIDA/IHistogram1D.h>
// endif

#include <marlin/Global.h>

#include "TrackFitterFactory.h"
#include "TrackFitterBase.h"

using namespace lcio ;
using namespace marlin ;

namespace marlintpc
{

BiasedResidualsProcessor aBiasedResidualsProcessor;
    
BiasedResidualsProcessor::BiasedResidualsProcessor(): Processor("BiasedResidualsProcessor") , _fitterFactory(0)
{
    // register steering parameters: name, description, class-variable, default value
    registerInputCollection(    LCIO::TRACK,
				"InputTracks" , 
				"Name of the input tracks collection"  ,
				_inputColName ,
				std::string("TPCSeedTracks") ) ;
    // modify processor description
    _description = "BiasedResidualsProcessor: Fill biased residuals of hits in tracks into AIDA histgram" ;

  registerProcessorParameter( "HistogramMinValueXY" , 
			      "Minimum value in xy histogram (default: -0.3 mm)"  ,
			      _histogramMinValueXY ,
			      float(-0.3) ) ;

  registerProcessorParameter( "HistogramMaxValueXY" , 
			      "Maximal value in xy histogram (default:  0.3 mm)"  ,
			      _histogramMaxValueXY ,
			      float(0.3) ) ;

  registerProcessorParameter( "HistogramMinValueZ" , 
			      "Minimum value in z histogram (default: -1.0 mm)"  ,
			      _histogramMinValueZ ,
			      float(-1.0) ) ;

  registerProcessorParameter( "HistogramMaxValueZ" , 
			      "Maximal value in z histogram (default:  1.0 mm)"  ,
			      _histogramMaxValueZ ,
			      float(1.0) ) ;
}

BiasedResidualsProcessor::~BiasedResidualsProcessor()
{
    delete _fitterFactory;
}

void BiasedResidualsProcessor::init()
{ 
	// usually a good idea to
	printParameters() ;

	// get a histograms from the AIDAProcessor
	_residualsXYHisto =   AIDAProcessor::histogramFactory(this)->
	  createHistogram1D( "residualsXYHisto", 
			     "residualsXYHisto", 
			     100, -1., 1.);
	_residualsZHisto =   AIDAProcessor::histogramFactory(this)->
	  createHistogram1D( "residualsZHisto", 
			     "residualsZHisto", 
			     100, -1., 1.);
    }

void BiasedResidualsProcessor::processRunHeader( EVENT::LCRunHeader* run )
{
    // delete the old fitter factory (of none had been initialised it is 0, so it's save to call delete
    delete _fitterFactory;

    // instantiate new fitter factory with parameters from run header
//    _fitterFactory = new _TrackFitterFactory(run); // not implemented yet, initialising with fitter type

    _fitterFactory = new TrackFitterFactory( TrackFitterBase::FitterTypes::CHISQUARE );
}
    
void BiasedResidualsProcessor::processEvent( LCEvent * evt ) 
{ 
    //read input data
    streamlog_out(MESSAGE3) << "Event: " << evt->getEventNumber() << std::endl;

    // we need a LCCollectionVec to use iterators, as LCCollectio is not a std::vector
    LCCollectionVec* InputCollection=0;
    try{
	InputCollection = dynamic_cast<LCCollectionVec*>(evt->getCollection( _inputColName )) ;
    }
    catch(DataNotAvailableException &e)
    {
	// no input data
	// just continue ...
	return;
    }


    if( InputCollection != 0 )
    {
	int trackNr=0;

	TrackFitterBase *trackFitter = _fitterFactory->createTrackFitter();

	for(LCCollectionVec::iterator it = InputCollection->begin() ; 
	    it < InputCollection->end(); it ++ )
	{
	    
	    streamlog_out(DEBUG4) <<"\t TrackNumber "<<trackNr++<<std::endl;

	    Track * track = dynamic_cast<Track *>(*it) ;

	    TrackerHitVec hitVector = track->getTrackerHits();
	    
//	    cout << "nr of hits " << hitVector.size()<< endl;

//	    int hitNr=0;

//	    cout << "\t\t hit\t x \t y \t z \t residual_x/r*phi" << endl;
	    for (TrackerHitVec::iterator hit = hitVector.begin();
		 hit <  hitVector.end(); hit ++)
	    {
//		const double *pos = (*hit)->getPosition();

		try
		{
		    EVENT::DoubleVec residuals = trackFitter->calculateResiduals(*hit, track);
		    _residualsXYHisto->fill( residuals[0] );
		    _residualsZHisto->fill(  residuals[1] );

//		    cout << residuals[0] << "\t" << residuals[1]  << std::endl;
		}
		catch ( IntersectionCalculator::NoIntersectionException &e )
		{
		    cout << e.what() << endl;
		}
	
	    }
        }

    }
}


}// namespace 
