#include "CutApplicationProcessor.h"


// LCIO
#include <EVENT/LCCollection.h>
#include <EVENT/Track.h>
#include <EVENT/TrackerHit.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/TrackImpl.h>

using namespace lcio ;
using namespace marlin ;

namespace marlintpc{

    CutApplicationProcessor aCutApplicationProcessor ;
    

    CutApplicationProcessor::CutApplicationProcessor() : 
      Processor("CutApplicationProcessor") {

	/* Description of the processor which will be displayed in the
	 * steering file automatically generated by 
	 * marlin (use MarlinTPC -x )
	 */
	_description = "CutApplicationProcessor applies cuts on all parameters for further analysis";
	
	
	/* register steering parameters: name, description, class-variable, default value
	 * the type will by definded by the type of the default value
	 * string, double, float and int are possible
	 * plaese note to get a float you must cast in C++: (float) 1.3 
	 */
	registerInputCollection( LCIO::TRACK,
				 "InputCollectionName" , 
				 "Name of the input collection"  ,
				 _inputColName ,
				 std::string("TPCTracks") ) ;

	registerOutputCollection( LCIO::TRACK,
				 "OutputCollectionName" , 
				 "Name of the output collection"  ,
				 _outputColName ,
				 std::string("TPCTracksCut") ) ;
	
	registerProcessorParameter( "SetOutputTransient" , 
				    "if not 0 the output collection is set transient, default 0s"  ,
				    _outputIsTransient,
				    0);

	registerOptionalParameter( "lowCutNTracks" , 
				    "lower cut on number of tracks"  ,
				    _lowcutNTracks ,
				    (int) 0. ) ;

	registerOptionalParameter( "highCutNTracks" , 
				    "upper cut on number of tracks"  ,
				    _highcutNTracks ,
				    (int) 10. ) ;

	registerOptionalParameter( "lowCutNHitsOnTrack" , 
				    "lower cut on number of hits on track"  ,
				    _lowcutNHitsOnTrack ,
				    (int) 0. ) ;

	registerOptionalParameter( "highCutNHitsOnTrack" , 
				    "upper cut on number of hits on track"  ,
				    _highcutNHitsOnTrack ,
				    (int) 10. ) ;

	registerOptionalParameter( "lowCutPhi" , 
				    "lower cut on phi"  ,
				    _lowcutPhi ,
				    (float) -1. ) ;

	registerOptionalParameter( "highCutPhi" , 
				    "upper cut on phi"  ,
				    _highcutPhi ,
				    (float) 1. ) ;

	registerOptionalParameter( "lowCutTanLambda" , 
				    "lower cut on tan lambda"  ,
				    _lowcutTanLambda ,
				    (float) -1.7 ) ;

	registerOptionalParameter( "highCutTanLambda" , 
				    "upper cut on tan lambda"  ,
				    _highcutTanLambda ,
				    (float) 1.7 ) ;

	registerOptionalParameter( "lowCutdEdx" , 
				    "lower cut on dEdx"  ,
				    _lowcutdEdx ,
				    (float) 0. ) ;

	registerOptionalParameter( "highCutdEdx" , 
				    "upper cut on dEdx"  ,
				    _highcutdEdx ,
				    (float) 100000. ) ;

	registerOptionalParameter( "lowCutCharge" , 
				    "lower cut on charge of cluster"  ,
				    _lowcutCharge ,
				    (float) 0. ) ;

	registerOptionalParameter( "highCutCharge" , 
				    "upper cut on charge of cluster"  ,
				    _highcutCharge ,
				    (float) 100000. ) ;

	registerOptionalParameter( "lowCutD0" , 
				    "lower cut on D0"  ,
				    _lowcutD0 ,
				    (float) 0. ) ;

	registerOptionalParameter( "highCutD0" , 
				    "upper cut on D0"  ,
				    _highcutD0 ,
				    (float) 0. ) ;

	registerOptionalParameter( "lowCutZ0" , 
				    "lower cut on Z0"  ,
				    _lowcutZ0 ,
				    (float) 0. ) ;

	registerOptionalParameter( "highCutZ0" , 
				    "upper cut on Z0"  ,
				    _highcutZ0 ,
				    (float) 0. ) ;

	registerOptionalParameter( "lowCutOmega" , 
				    "lower cut on Omega"  ,
				    _lowcutOmega ,
				    (float) 0. ) ;

	registerOptionalParameter( "highCutOmega" , 
				    "upper cut on Omega"  ,
				    _highcutOmega ,
				    (float) 0. ) ;

	registerOptionalParameter( "lowCutX" , 
				    "lower cut on x position of hit"  ,
				    _lowcutX ,
				    (double) 0. ) ;

	registerOptionalParameter( "highCutX" , 
				    "upper cut on x position of hit"  ,
				    _highcutX ,
				    (double) 14.08 ) ;

	registerOptionalParameter( "lowCutY" , 
				    "lower cut on y position of hit"  ,
				    _lowcutY ,
				    (double) 0. ) ;

	registerOptionalParameter( "highCutY" , 
				    "upper cut on y position of hit"  ,
				    _highcutY ,
				    (double) 14.08 ) ;

	registerOptionalParameter( "lowCutZ" , 
				    "lower cut on z position of hit"  ,
				    _lowcutZ ,
				    (double) 0. ) ;

	registerOptionalParameter( "highCutZ" , 
				    "upper cut on z position of hit"  ,
				    _highcutZ ,
				    (double) 270. ) ;

	
    }
    
    
    void CutApplicationProcessor::init() { 
	
	// usually a good idea to
	printParameters() ;


	
    }
    
    void CutApplicationProcessor::processRunHeader( LCRunHeader* run) { 
	
      
      run->parameters().setValue( _processorName + "_revision", "$Rev: 1297 $");

      for ( ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++ ) {
	if ( ! i->second->isOptional() || i->second->valueSet() ) {
	  run->parameters().setValue( _processorName + "_" + i->second->name(), 
				      i->second->value() );
	}
      }

	
    } 
    
void CutApplicationProcessor::processEvent( LCEvent * evt ) 
{ 
    
    //read input data
    streamlog_out(MESSAGE3) << "Event: " << evt->getEventNumber() << std::endl;

    LCCollectionVec* inputCol = 0; 
    try{
	inputCol = dynamic_cast<LCCollectionVec*>(evt->getCollection( _inputColName )) ;
    } //end of try {
	
    catch(DataNotAvailableException){
	// if there is not input collection give a message to the user
	streamlog_out(MESSAGE2)<<" processEvent:"
			       <<" input collection "<<_inputColName
				   <<" cannot be found in event "<<evt->getEventNumber()<<std::endl;
	return;
    }
    
    
	// first perform the event specific cuts
    if ( parameterSet("lowCutNTracks") && inputCol->getNumberOfElements() < _lowcutNTracks)
	return;
    if ( parameterSet("highCutNTracks") && inputCol->getNumberOfElements() > _highcutNTracks)
	return;
    
    
    // create the output collection with a string defining the TYPE (see lcio.h)
    LCCollectionVec* outputCol = new LCCollectionVec( LCIO::TRACK ); 
    // set the transient flag as definded by the processor parameter
    outputCol->setTransient( (_outputIsTransient != 0) );
    // the output collection is a subset of the input collection
    // only the pointers are stored, the track is not copied
    outputCol->setSubset();
    
    for ( LCCollectionVec::iterator trackIter = inputCol->begin();
	  trackIter < inputCol->end(); trackIter++) 
    {
	streamlog_out(DEBUG1)<<"looping tracks"<<std::endl;
	
	Track* track = dynamic_cast<Track*>( *trackIter ) ;
	
	//cast returns 0-pointer if cast fails
	if ( track == 0 ) continue; 
	
	// now check all cuts and continue without saving the track if cut is not fulfilled
	
	if ( parameterSet("lowCutNHitsOnTrack") &&  track->getPhi() < _lowcutNHitsOnTrack ) 
	    continue;
	if ( parameterSet("highCutNHitsOnTrack") &&  track->getPhi() > _highcutNHitsOnTrack ) 
	    continue;
	
	if ( parameterSet("lowCutPhi") &&  track->getPhi() < _lowcutPhi ) 
	    continue;
	if ( parameterSet("highCutPhi") &&  track->getPhi() > _highcutPhi ) 
	    continue;
	
	if ( parameterSet("lowCutTanLambda") &&  track->getTanLambda() < _lowcutTanLambda ) 
	    continue;
	if ( parameterSet("highCutTanLambda") &&  track->getTanLambda() > _highcutTanLambda ) 
	    continue;
	
	if ( parameterSet("lowCutD0") &&  track->getD0() < _lowcutD0 ) 
	    continue;
	if ( parameterSet("highCutD0") &&  track->getD0() > _highcutD0 ) 
	    continue;
	
	if ( parameterSet("lowCutZ0") &&  track->getZ0() < _lowcutZ0 ) 
	    continue;
	if ( parameterSet("highCutZ0") &&  track->getZ0() > _highcutZ0 ) 
	    continue;
	
	if ( parameterSet("lowCutOmega") &&  track->getOmega() < _lowcutOmega ) 
	    continue;
	if ( parameterSet("highCutOmega") &&  track->getOmega() > _highcutOmega ) 
	    continue;
	    
	if ( parameterSet("lowCutdEdx") &&  track->getdEdx() < _lowcutdEdx ) 
	    continue;
	if ( parameterSet("highCutdEdx") &&  track->getdEdx() > _highcutdEdx ) 
	    continue;
	
	streamlog_out(DEBUG1)<<"looping hits"<<std::endl;

	TrackerHitVec hitVec = track->getTrackerHits();
	bool trackIsInvalid=false; // flag to know if cut has not been fulfilled
	
	for ( TrackerHitVec::iterator hitIter = hitVec.begin(); hitIter < hitVec.end(); hitIter++ ) 
	{
	    const double *pos = (*hitIter)->getPosition();
	    
	    streamlog_out(DEBUG1)<<"D"<<std::endl;
			    
	    if ( parameterSet("lowCutX") && pos[0] < _lowcutX ) 
		trackIsInvalid=true;
	    if ( parameterSet("highCutX") &&  pos[0] > _highcutX) 
		trackIsInvalid=true;
	    
	    if ( parameterSet("lowCutY") && pos[1] < _lowcutY ) 
		trackIsInvalid=true;
	    if ( parameterSet("highCutY") &&  pos[1] > _highcutY) 
		trackIsInvalid=true;

	    if ( parameterSet("lowCutZ") && pos[2] < _lowcutZ ) 
		trackIsInvalid=true;
	    if ( parameterSet("highCutZ") &&  pos[2] > _highcutZ) 
		trackIsInvalid=true;
	    
	    if ( parameterSet("lowCutCharge") && (*hitIter)->getdEdx() < _lowcutCharge ) 
		trackIsInvalid=true;
	    if ( parameterSet("highCutCharge") && (*hitIter)->getdEdx() > _highcutCharge   )
		trackIsInvalid=true;
	}// for hitIter
	
	if (!trackIsInvalid)
	    outputCol->addElement( *trackIter );
    }// for trackIter
    
    // only add non-empty collections to the event
    if ( outputCol->empty() ) 
	delete outputCol;
    else
	evt->addCollection(outputCol, _outputColName);

}// process event
    
}// namecpace marlintpc
