#include "TrackParametersDistributionProcessor.h"

//C++
#include <iostream>
#include <cmath>
using namespace std;

//LCIO
#include <IMPL/LCCollectionVec.h>
#include <EVENT/Track.h>

// Aida
//#ifdef MARLIN_USE_AIDA // the whole processor does not make sense without aida
#include <AIDA/AIDA.h>
#include <marlin/AIDAProcessor.h>
#include <AIDA/IHistogramFactory.h>
#include <AIDA/IHistogram1D.h>
// endif

#include <marlin/Global.h>

using namespace lcio ;
using namespace marlin ;

namespace marlintpc
{

TrackParametersDistributionProcessor aTrackParametersDistributionProcessor;
    
TrackParametersDistributionProcessor::TrackParametersDistributionProcessor()
    : Processor("TrackParametersDistributionProcessor"),
      _distributionPhiHisto(0) , _distributionLambdaHisto(0),
      _distributionTanLambdaHisto(0), _distributionD0Histo(0),
      _distributionZ0Histo(0), _distributionOmegaHisto(0)
{
    // register steering parameters: name, description, class-variable, default value
    registerInputCollection(    LCIO::TRACK,
				"InputTracks" , 
				"Name of the input tracks collection"  ,
				_inputColName ,
				std::string("TPCSeedTracks") ) ;
    // modify processor description
    _description = "TrackParametersDistributionProcessor: Fill phi, lambda, and z distribution of hits in tracks into AIDA histgram" ;

  registerProcessorParameter( "HistogramMinValuePhi" , 
			      "Minimum value in phi histogram (default: -50. mrad)"  ,
			      _histogramMinValuePhi ,
			      float(-50.) ) ;

  registerProcessorParameter( "HistogramMaxValuePhi" , 
			      "Maximal value in phi histogram (default: 50. mrad)"  ,
			      _histogramMaxValuePhi ,
			      float(50.) ) ;


  registerProcessorParameter( "HistogramMinValueLambda" , 
			      "Minimum value in lambda histogram (default: -50. mrad)"  ,
			      _histogramMinValueLambda ,
			      float(-50.) ) ;

  registerProcessorParameter( "HistogramMaxValueLambda" , 
			      "Maximal value in lambda histogram (default:  50. mrad)"  ,
			      _histogramMaxValueLambda ,
			      float(50.) ) ;


  registerProcessorParameter( "HistogramMinValueTanLambda" , 
			      "Minimum value in tan(lambda) histogram (default: -20.)"  ,
			      _histogramMinValueTanLambda ,
			      float(-20.) ) ;

  registerProcessorParameter( "HistogramMaxValueTanLambda" , 
			      "Maximal value in tan(lambda) histogram (default:  20. )"  ,
			      _histogramMaxValueTanLambda ,
			      float(20.) ) ;


  registerProcessorParameter( "HistogramMinValueD0" , 
			      "Minimum value in d0 histogram (default: -50. mm)"  ,
			      _histogramMinValueD0 ,
			      float(-50.) ) ;

  registerProcessorParameter( "HistogramMaxValueD0" , 
			      "Maximal value in d0 histogram (default:  50. mm)"  ,
			      _histogramMaxValueD0 ,
			      float(50.) ) ;

 registerProcessorParameter( "HistogramMinValueZ0" , 
			      "Minimum value in z0 histogram (default: -10. mm)"  ,
			      _histogramMinValueZ0 ,
			      float(-10.) ) ;

  registerProcessorParameter( "HistogramMaxValueZ0" , 
			      "Maximal value in z0 histogram (default:  300. mm)"  ,
			      _histogramMaxValueZ0 ,
			      float(300.) ) ;

 registerProcessorParameter( "HistogramMinValueOmega" , 
			      "Minimum value in Omega histogram (default: -0.01 mm^-1)"  ,
			      _histogramMinValueOmega ,
			      float(-0.01) ) ;

  registerProcessorParameter( "HistogramMaxValueOmega" , 
			      "Maximal value in Omega histogram (default:  0.01 mm^-1)"  ,
			      _histogramMaxValueOmega ,
			      float(0.01) ) ;


}

TrackParametersDistributionProcessor::~TrackParametersDistributionProcessor()
{
}

void TrackParametersDistributionProcessor::init()
{ 
	// usually a good idea to
	printParameters() ;

	// get a histograms from the AIDAProcessor
	_distributionPhiHisto =   AIDAProcessor::histogramFactory(this)->
	  createHistogram1D( "phiDistributionHisto", 
			     "phiDistributionHisto", 
			     100, _histogramMinValuePhi, _histogramMaxValuePhi);

	_distributionLambdaHisto =   AIDAProcessor::histogramFactory(this)->
	  createHistogram1D( "lambdaDistributionHisto", 
			     "lambdaDistributionHisto", 
			     100, _histogramMinValueLambda, _histogramMaxValueLambda);

	_distributionTanLambdaHisto =   AIDAProcessor::histogramFactory(this)->
	  createHistogram1D( "tanLambdaDistributionHisto", 
			     "tanLambdaDistributionHisto", 
			     100, _histogramMinValueTanLambda, _histogramMaxValueTanLambda);

	_distributionD0Histo =   AIDAProcessor::histogramFactory(this)->
	  createHistogram1D( "d0DistributionHisto", 
			     "d0DistributionHisto", 
			     100, _histogramMinValueD0, _histogramMaxValueD0);

	_distributionZ0Histo =   AIDAProcessor::histogramFactory(this)->
	  createHistogram1D( "z0DistributionHisto", 
			     "z0DistributionHisto", 
			     100, _histogramMinValueZ0, _histogramMaxValueZ0);

	_distributionOmegaHisto =   AIDAProcessor::histogramFactory(this)->
	  createHistogram1D( "OmegaDistributionHisto", 
			     "OmegaDistributionHisto", 
			     100, _histogramMinValueOmega, _histogramMaxValueOmega);
    }

void TrackParametersDistributionProcessor::processRunHeader( EVENT::LCRunHeader* run )
{
}
    
void TrackParametersDistributionProcessor::processEvent( LCEvent * evt ) 
{ 
    //read input data
    streamlog_out(MESSAGE3) << "Event: " << evt->getEventNumber() << std::endl;

    // we need a LCCollectionVec to use iterators, as LCCollectio is not a std::vector
    LCCollectionVec* InputCollection=0;
    try{
	InputCollection = dynamic_cast<LCCollectionVec*>(evt->getCollection( _inputColName )) ;
    }
    catch(DataNotAvailableException &e)
    {
	// no input data
	// just continue ...
	streamlog_out(WARNING1)<<"inputCollection: "<< _inputColName <<"  could not be found in event "
			       << evt->getEventNumber() <<std::endl;
	return;
    }


    // cast did not succeed
    if( InputCollection == 0 )
    {
	streamlog_out(ERROR2) << "InputCollection vector is 0" << std::endl;
	return;
    }

    for(LCCollectionVec::iterator it = InputCollection->begin() ; 
	it < InputCollection->end(); it ++ )
    {
	
	Track * track = dynamic_cast<Track *>(*it) ;
	
	_distributionPhiHisto->fill(track->getPhi());
	_distributionLambdaHisto->fill( std::atan( track->getTanLambda() ));
	_distributionTanLambdaHisto->fill( track->getTanLambda() );
	_distributionD0Histo->fill( track->getD0() ) ;
	_distributionZ0Histo->fill( track->getZ0() ) ;
	_distributionOmegaHisto->fill( track->getOmega() ) ;

	streamlog_out(DEBUG0)<<"phi = "<<track->getPhi()<<"\t lambda = "<<std::atan( track->getTanLambda() )
			     <<"\nd0 = "<<track->getD0()<<"\tz0 = "<<track->getZ0()<<std::endl;

    }
    
}


}// namespace 
