/* PedestalCalculator.cc
 * $Id: PedestalCalculator.cc,v 1.1 2006/07/07 08:58:52 mjanssen Exp $
 */
#include "PedestalCalculator.h"

//C++
#include <iostream>
#include <map>
#include <cmath>

//LCIO
#include <EVENT/LCCollection.h>
#include <EVENT/TPCHit.h>
#include <EVENT/TrackerData.h>
#include <EVENT/LCGenericObject.h>
#include <IMPL/LCCollectionVec.h>

using namespace lcio ;
using namespace marlin ;

namespace marlintpc
{

    PedestalCalculator aPedestalCalculator;
    
    PedestalCalculator::PedestalCalculator() : DBEntryMaker(std::string("PedestalCalculator"),
                                               std::string("TPCConvertedRawData"),
                                               std::string("TPCPedestal") ) 
    {
        // modify processor description
        _description = "PedestalCalculator: calculate the pedestal (mean and width)" ;
    }
    
    void PedestalCalculator::init() 
    { 
        makerInit();
        printParameters();
        _ChannelPedestalCollection = new ChannelPedestalCollection_t();
        _ChannelPedestalCollection->clear();
    }

    void PedestalCalculator::processRunHeader( LCRunHeader* run) 
    { 
        //std::cout<<"PedestalCalculator::processRunHeader"<<std::endl;
        run->parameters().setValue( _processorName + "_revision", "$Rev: 509 $");
	for ( ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++ ) {
	    if ( ! i->second->isOptional() || i->second->valueSet() ) {
	      run->parameters().setValue( _processorName + "_" + i->second->name(), i->second->value() );
	    }
	}

        _NRun++ ;
    } 

    void PedestalCalculator::processEvent( LCEvent * evt ) 
    { 
        //read input data

        LCCollection* InputCollection = evt->getCollection( _InputColName ) ;

        if( InputCollection != 0 )
        {
            int nPulse = InputCollection->getNumberOfElements();
            FloatVec* Data = new FloatVec();
            for(int iPulse=0; iPulse< nPulse ; iPulse++)
            {
                TrackerData* thisPulse = dynamic_cast<TrackerData*>( InputCollection->getElementAt( iPulse ) ) ;
                if (thisPulse)
                {
                    ChannelPedestal* thisChannelPedestal = 0;
                    for (ChannelPedestalCollection_t::iterator iChannelPedestal = 
                         _ChannelPedestalCollection->begin();
                         iChannelPedestal != _ChannelPedestalCollection->end();
                         iChannelPedestal++)
                    {
                        //std::cout<<"channel ID:"<<(*iChannelPedestal)->ChannelID<<std::endl;
                        if ( (*iChannelPedestal)->ChannelID == thisPulse->getCellID0() )
                        {   
                            //std::cout<<"found ChannelPedestal"<<std::endl;
                            thisChannelPedestal = (*iChannelPedestal);
                        }
                    }      
                    if (!thisChannelPedestal) 
                    {
                        // pad does not exist
                        //std::cout<<"create new ChannelPedestal"<<std::endl;
                        thisChannelPedestal = new ChannelPedestal( );
                        thisChannelPedestal->ChannelID   = thisPulse->getCellID0();
                        thisChannelPedestal->PedestalCol = new PedestalVector_t();
                        _ChannelPedestalCollection->push_back( thisChannelPedestal );
                    }
                    // clear the data from data of older pulses
                    ///std::cout<<"clear data"<<std::endl;
                    Data->clear();
                    // fill in new data
                    FloatVec TrackerDataSpectrum = thisPulse->getChargeValues();
                    //std::cout<<"fill data"<<std::endl;
                    for (FloatVec::iterator iSlice=TrackerDataSpectrum.begin();
                         iSlice!=TrackerDataSpectrum.end(); iSlice++)
                    {
                        // fill the data vector with 
                        Data->push_back( *iSlice );
                    }
                    //std::cout<<"calculate pedestal"<<std::endl;
                    thisChannelPedestal->PedestalCol->push_back( calcPedestal( Data ) );
                }
            }        
        }
        //std::cout<<"makerProcessEvent"<<std::endl;
        makerProcessEvent( evt );
    }

    void PedestalCalculator::check( LCEvent * evt ) 
    { 
        // nothing to check here - could be used to fill checkplots in reconstruction processor
    }


    void PedestalCalculator::end()
    { 
        makerEnd();
    }
 
    StcPedestal* PedestalCalculator::calcPedestal(FloatVec* Data)
    {
        int    ingoreSurrounding = 5;   // # time slices before and after the currend slice must not be a signal
        double signalThreshold   = 5.0; // slice is signal, if is bigger than ? * pedestal width 
        double precision         = 1.0E-5; // precision of pedestal calculation
        float calcValue = 0.0;
        float calcWidth = 512.0;
        int   cycleNorm  = 0;
        float cycleValue = 0.0;
        float cycleWidth = 0.0;  

        for (int iRun=0; iRun < 5; iRun++)
        {
            // makes 5 cycles 
            //cout<<"cycle "<<iRun<<endl;
            cycleNorm  = 0;
            cycleValue = 0.0;
            cycleWidth = 0.0;
            FloatVec::iterator iData = Data->begin();

            // makes shure, that pedestal calculation starts in area with no signal
            int NoSignalCounter = 0;
            while  ( (iData!=Data->end()) && (NoSignalCounter< 2*ingoreSurrounding+1) ) 
            {
                if ( ( (*iData)-calcValue< signalThreshold*calcWidth)&&
                     ( (*iData)-calcValue>-signalThreshold*calcWidth) )
                {
                    // is no signal
                    NoSignalCounter++;
                }
                else
                {
                    // is signal
                    NoSignalCounter = 0;
                }
                iData++;
            }
            iData-=ingoreSurrounding;

            while ( (iData+ingoreSurrounding)<Data->end() )
            {
                if ( ((*iData+ingoreSurrounding)-calcValue< signalThreshold*calcWidth)&&
                     ((*iData+ingoreSurrounding)-calcValue>-signalThreshold*calcWidth) )
                {
                    // no signal

                    cycleNorm++ ;  
                    cycleValue += (*iData);
                    cycleWidth += (*iData)*(*iData);
                    iData++;
                }
                else
                {
                    //signal: find next region without signal
                    NoSignalCounter = 0;
                    while  ( (iData!=Data->end()) && (NoSignalCounter < 2*ingoreSurrounding+1) ) 
                    {
                        if ( ( (*iData)-calcValue< signalThreshold*calcWidth)&&
                             ( (*iData)-calcValue>-signalThreshold*calcWidth) )
                        {
                            // is no signal
                            NoSignalCounter++;
                        }
                        else
                        {
                            // is signal
                            NoSignalCounter = 0;
                        }
                        iData++;
                    }
                    iData-=ingoreSurrounding;
                }
            }
            if (cycleNorm>0)
            {
                if (fabs(calcValue-cycleValue/cycleNorm) < precision) break;
                calcValue = cycleValue/cycleNorm;
                calcWidth = sqrt(cycleWidth/cycleNorm - calcValue*calcValue);
                // s = 1/n*sum(x-<x>)^2=<x^2> - <x>^2
            }
        }    
        StcPedestal* returnPedestal= new StcPedestal();
        returnPedestal->Mean  = calcValue;
        returnPedestal->Width = calcWidth;
        return returnPedestal;
    }

    void PedestalCalculator::calcAverage(ChannelPedestal* thisChannelPedestal)
    {
        StcPedestal*  Average      = new StcPedestal();
        StcPedestal*  AverageError = new StcPedestal();

        // calculate value and its error, for pedestal and width
        Average->Mean       = 0;
        AverageError->Mean  = 0;
        Average->Width      = 0;
        AverageError->Width = 0;
        if ( thisChannelPedestal->PedestalCol->size() > 0) 
        {
            for (PedestalVector_t::iterator iChannelPedestal = thisChannelPedestal->PedestalCol->begin();
                 iChannelPedestal != thisChannelPedestal->PedestalCol->end(); iChannelPedestal++)
            {
                Average->Mean       += (*iChannelPedestal)->Mean;
                AverageError->Mean  += ( (*iChannelPedestal)->Mean )* ( (*iChannelPedestal)->Mean );
                Average->Width      += (*iChannelPedestal)->Width;
                AverageError->Width += ( (*iChannelPedestal)->Width ) * ( (*iChannelPedestal)->Width );
            }
            Average->Mean  /= thisChannelPedestal->PedestalCol->size();
            Average->Width /= thisChannelPedestal->PedestalCol->size();
            if ( thisChannelPedestal->PedestalCol->size() > 1 )
            {
                AverageError->Mean = sqrt( (AverageError->Mean / thisChannelPedestal->PedestalCol->size()
                                - (Average->Mean * Average->Mean ) ) 
                               / ( thisChannelPedestal->PedestalCol->size()-1) );
                AverageError->Width = sqrt( (AverageError->Width / thisChannelPedestal->PedestalCol->size()
                                 - (Average->Width * Average->Width ) ) 
                                / ( thisChannelPedestal->PedestalCol->size()-1) );
                // s^2 = <x^2> - <x>^2 => sigma = s / sqrt(n-1) 
            }
            else 
            {
                AverageError->Mean  = Average->Mean;
                AverageError->Width = Average->Width;
            }
        }
        else
        {
            // no values to calculate
            Average->Mean       = 0;
            Average->Width      = 0;
            AverageError->Mean  = 0;
            AverageError->Width = 0;
        }
        thisChannelPedestal->Average      = Average;
        thisChannelPedestal->AverageError = AverageError;
    }

    LCCollectionVec* PedestalCalculator::CreatePedestalCollection( ChannelPedestalCollection_t* 
                                          thisChannelPedestalCollection )
    {
        LCCollectionVec* returnCollection = new LCCollectionVec( LCIO::LCGENERICOBJECT );
        // calculate pedestal and store it
        for (ChannelPedestalCollection_t::iterator iChannelPedestal = thisChannelPedestalCollection->begin();
             iChannelPedestal != thisChannelPedestalCollection->end(); iChannelPedestal++)
        {
            calcAverage( *iChannelPedestal );
            returnCollection->addElement( new tpcconddata::Pedestal( (*iChannelPedestal)->ChannelID,
                                         (*iChannelPedestal)->Average->Mean,
                                         (*iChannelPedestal)->Average->Width ) );
        }
        return returnCollection;
    }

    LCCollection* PedestalCalculator::makeLCCDData()
    {
        LCCollectionVec* returnCol = 0;
        if ( (_ChannelPedestalCollection->size()>0)&&( _lastSinceTime<_lastEventTime ) )
        {
            std::cout<<std::endl<<std::endl;
            std::cout<<"store pedestals: valid "<<_lastSinceTime<<" - "<<_lastEventTime<<std::endl;
            printPedestalShort( _ChannelPedestalCollection );
            returnCol = CreatePedestalCollection( _ChannelPedestalCollection );
            returnCol->parameters().setValue("Revision",tpcconddata::Pedestal::getRevision() );
            //clean up of pedestal collection
            for (ChannelPedestalCollection_t::iterator iChannelPedestal = _ChannelPedestalCollection->begin();
                 iChannelPedestal != _ChannelPedestalCollection->end(); iChannelPedestal++)
            {
                (*iChannelPedestal)->PedestalCol->clear();
                delete ( (*iChannelPedestal)->Average );
                delete ( (*iChannelPedestal)->AverageError );
            }
        }
        return returnCol;
    }

    void PedestalCalculator::printPedestal(ChannelPedestalCollection_t* thisChannelPedestalCollection)
    {
        // output
        std::cout<<std::endl;
        for (ChannelPedestalCollection_t::iterator iChannelPedestal = thisChannelPedestalCollection->begin();
             iChannelPedestal != thisChannelPedestalCollection->end(); iChannelPedestal++)
        {
            calcAverage( *iChannelPedestal );
            printf( "PadID: %5.0i,  ",(*iChannelPedestal)->ChannelID );
            printf( "pedestal: mean = %6.4f+-%6.4f",
            (*iChannelPedestal)->Average->Mean, (*iChannelPedestal)->AverageError->Mean);
            if ( (*iChannelPedestal)->Average->Width == 0 )
            {
                std::cout<<" width calculation failed";
            }
            else
            {
                printf("   width = %6.4f+-%6.4f",
                        (*iChannelPedestal)->Average->Width, (*iChannelPedestal)->AverageError->Width);
            }
            std::cout<<std::endl;
        }
    } 

    void PedestalCalculator::printPedestalShort(ChannelPedestalCollection_t* 
                           thisChannelPedestalCollection)
    {
        // output
        ChannelPedestal*  firstChannelPedestal = *(thisChannelPedestalCollection->begin());
        calcAverage( firstChannelPedestal );
        int firstChannel = firstChannelPedestal->ChannelID;
        int lastChannel  = firstChannel;
        int smallestPedestalMeanChannel  = firstChannel;
        int largestPedestalMeanChannel   = firstChannel;
        int smallestPedestalWidthChannel = firstChannel;
        int largestPedestalWidthChannel  = firstChannel;
        float smallestPedestalMean  = firstChannelPedestal->Average->Mean;
        float largestPedestalMean   = smallestPedestalMean;
        float smallestPedestalWidth = firstChannelPedestal->Average->Width;
        float largestPedestalWidth  = smallestPedestalWidth;
        bool  noWidthCalFailed = true;

        for (ChannelPedestalCollection_t::iterator iChannelPedestal = thisChannelPedestalCollection->begin();
             iChannelPedestal != thisChannelPedestalCollection->end(); iChannelPedestal++)
        {
            calcAverage( *iChannelPedestal );
            int Channel = (*iChannelPedestal)->ChannelID;
            if (Channel<firstChannel) firstChannel = Channel;
            if (Channel>lastChannel)  lastChannel  = Channel;
            float PedestalMean = (*iChannelPedestal)->Average->Mean;
            if (PedestalMean<smallestPedestalMean)
            {
                smallestPedestalMean = PedestalMean;
                smallestPedestalMeanChannel = Channel;
            }
            if (PedestalMean>largestPedestalMean)
            {
                largestPedestalMean  = PedestalMean;
                largestPedestalMeanChannel = Channel;
            }
            float PedestalWidth = (*iChannelPedestal)->Average->Width;
            if ( PedestalWidth == 0 )
            {
                if (noWidthCalFailed)
                {
                    std::cout<<"width calculation failed for channel: ";
                    noWidthCalFailed=false;
                }
                std::cout<<Channel<<", ";
            }
            else
            {
                if (PedestalWidth<smallestPedestalWidth)
                {
                    smallestPedestalWidth = PedestalWidth;
                    smallestPedestalWidthChannel = Channel;
                }
                if (PedestalWidth>largestPedestalWidth)  
                {
                    largestPedestalWidth  = PedestalWidth;
                    largestPedestalWidthChannel = Channel;
                }
            }
        }
        if (!noWidthCalFailed)
        {
            std::cout<<std::endl;
        }
        std::cout<<"calculate pedstals for channel "<<firstChannel<<" to "<<lastChannel<<std::endl;
        std::cout<<"pedestal mean between "
                 <<smallestPedestalMean<<" (channel "<<smallestPedestalMeanChannel<<") and "
                 <<largestPedestalMean<<" (channel "<<largestPedestalMeanChannel<<")"<<std::endl;
        std::cout<<"pedestal width between "
                 <<smallestPedestalWidth<<" (channel "<<smallestPedestalWidthChannel<<") and "
                 <<largestPedestalWidth<<" (channel "<<largestPedestalWidthChannel<<")"<<std::endl;
    } 

    void PedestalCalculator::printPedestal(LCCollection* thisCollection)
    {
        // output
        int         nPedestal    = thisCollection->getNumberOfElements();
        tpcconddata::Pedestal* thisPedestal = 0;   
        std::cout<<std::endl;
        for (int iPedestal = 0; iPedestal<nPedestal; iPedestal++)
        {
            thisPedestal = new tpcconddata::Pedestal( dynamic_cast<LCGenericObject*>
                                  ( thisCollection->getElementAt( iPedestal ) ) );
            printf( "PadID: %5.0i,  ", thisPedestal->getChannel() );
            printf( "pedestal: mean = %6.4f", thisPedestal->getPedestalValue() );
            if ( thisPedestal->getPedestalWidth() == 0 )
            {
                std::cout<<" width calculation failed";
            }
            else
            {
                printf( "   width = %6.4f", thisPedestal->getPedestalWidth() );
            }
            std::cout<<std::endl;
        }
    } 
}
