#ifndef GEMProcessor_h
#define GEMProcessor_h 1

//LCIO
#include "IMPL/LCCollectionVec.h"
#include "marlin/Processor.h"
#include "lcio.h"

//C++
#include <string>

//GEAR
#include <gear/PadRowLayout2D.h>


/** This Processor takes drifted electrons and uses a triple GEM structure for amplification. The amplified
 *  electrons are safed as SimTrackerHits with charge safed in dEdx. 
 *  ATTENTION!This processor checks also for electrons arriving on padplane. Those not on padplane are dumped and
 *  not amplified. Correctly this should be done in the digitisation part but it is done here for reasons of 
 *  performance. This has to be kept in mind when the processor is not used in combination with the 
 *  ChargeDistributionProcessor.  
 */

class GEMProcessor : public marlin::Processor {
  
 public:
  
  virtual Processor*  newProcessor() { return new GEMProcessor ; }
   
  GEMProcessor() ;
  virtual ~GEMProcessor(){}
  
  /** Called at the begin of the job before anything is read. Use to initialize the processor, e.g. book histograms.
   */
  virtual void init() ;
  
  /** Called for every run.
   */
  virtual void processRunHeader( EVENT::LCRunHeader* run ) ;
  
  /** Called for every event - the working horse.
   */
  virtual void processEvent( EVENT::LCEvent * evt ) ; 
    
  virtual void check( EVENT::LCEvent * evt ) ; 
    
  /** Called after data processing for clean up.
   */
  virtual void end() ;

   void LoadGEMSettings();

  /** Gives the Collection into GEM i=GEMNr for enumber of electrons in front of GEM hole.
   */
  int GEM_Collection(int GEMNr, int enumber);

  /** Gives the gain in GEM i=GEMNr using an exponential distribution for the gain in one GEM. 
   */
  double GEM_Gain(int GEMNr, int enumber);

  /** Gives the Extraction out of GEM i=GEMNr for enumber of electrons indside GEM hole.
   */
  int GEM_Extraction(int GEMNr, int enumber);

  /** Claculates the attachment between GEM 
   */
  double CalculateAttachment(double E);

  double AttachmentParam(double E,double a0,double a1,double a2,double a3, double a4,double a5);

  /** Function to claculate number of electrons after passing through the GEM structure.
   *  This function makes use of the parametrisation of the charge transfer through the GEMs.
   */
  int CalculateNrElectrons(double attachment1, double attachment2, double attachment3);


 
protected:

  /** Pad layout from gear
   */
  gear::PadRowLayout2D const * _padLayout;

  /** Input/output collection names.
   */
  std::string _inputCollectionName;
  std::string _outputCollectionName;

  /** Proccessor parameter
   */
  std::string paramfile;
  std::string gas;
  const char* Gas;
  float UGEM1,UGEM2,UGEM3;
  float ETrans1,ETrans2,EInd;
  float _gainFactor;
  int _Transient;

  float E;
  double Extr[3];
  double Coll[3];
  double Gain[3];

  /** Attachment 
   */
  double A1,A2,A3;

  bool isCurler;

  /** Internal event counter. Independent of the number the event has. 
   *  If _nEvt is 3 this is the third event to get processed. The number 
   *  you get with evt->getEventNumber() may differ!
   */
  int _nEvt;

};

#endif 
