#ifndef VOXELTPC_H
#define VOXELTPC_H 1

//MARLINTPC
#include "TPCVoxel.h"
#include "HitDistributor.h"

//C++
#include <map> 
#include <vector>
#include <stdexcept>

//LCIO
#include <EVENT/SimTrackerHit.h>
#include <IMPL/LCCollectionVec.h>

//GEAR
#include <gear/PadRowLayout2D.h>
#include <gearimpl/RectangularPadRowLayout.h>
#include <gearimpl/FixedPadSizeDiskLayout.h>

//MARLIN
#include <marlin/Global.h>


namespace marlintpc
{

class TPCVoxel;

/** The VoxelTPC is a helper class containing a std::map of all 
 *  TPCVoxels which contain charge. It uses the
 *  GEAR pad geometry and the length of a time bin to
 *  calculate the pad index and time index of the voxels.
 *  One time bin does not necessarily have to be one sample of the
 *  ADC. It can also be a nanosecond (useful to have good time resolution
 *  for a realistic electronics shaping) or one bunch crossing (to 
 *  store one bunch train on a per event basis).
 */


class VoxelTPC
{
 public:
    VoxelTPC( gear::PadRowLayout2D const * padLayout , double lengthOfTimeBin );

    ~VoxelTPC();

    /** Returns an LCCollectionVec of the stored TPCVoxels.
     *  This vector is newly taken from the heap so 
     *  it can be given to an event, which takes over the ownership.
     *  Be careful to delete it and all the voxels it contains 
     *  if you don't add it to the event! 
     */
     LCCollectionVec * getVoxelCollection( double outputLength = -1 );

    /** Does the same as getCollection() but erases the voxels 
     *  from the map.
     */
    LCCollectionVec * fetchVoxelCollection( double outputLength = -1 );
    
    /** Distributes the charge onto the pads assuming a 2D Gaussian 
     *  distribution with the width sigmaTrans. The charge is deposited 
     *  in the according voxels with  the zBin in arbitrry units, which 
     *  can be chosen by the user e.g. drift time or a length scale. 
     *  The zBin can be smeared with a Gaussian with width sigmaLong.
     *  By default sigmaLong is 0 and the charge is not distributed in z.
     *  This is useful in case of single electrons which have been amplified 
     *  in a GEM. Here charge is distributed in xy, but the distribution in
     *  z is negligible.
     */
    void putChargeInVoxels( SimTrackerHit *hit, double sigmaTrans, double sigmaLong);

    /** Puts the charge into the voxel. If the voxel does not have
     *  charge yet it is added to the amp, otherwise the charge is
     *  added to the existing entry for the voxel.
     */
    void putChargeInVoxel( TPCVoxel const &voxel);

    /** Convenience function for putChargeInVoxel( TPCVoxel const &voxel)
     */
    void putChargeInVoxel( int padIndex, double zLength, int charge );

    /** Get the voxel with a given index.
     *  In case the voxel is not in the map it throws an
     *  EVENT::DataNotAvailableException
     */ 
    TPCVoxel getVoxel( int padIndex, int timeIndex );

    /** Returns a vextor with all voxels with a specific pad index
     */
    std::vector<TPCVoxel> getVoxelsOnPad(int padIndex);

    /** Returns a vector of all voxels with a specific time index
     */
    std::vector<TPCVoxel> getVoxelsWithZBin(int zBin);

    /** Adds the given collection of TPCVoxels to the map
     */
    void addCollection(LCCollection* voxelCol);


    /** Removes a voxel from the set.
     *  It removes the corresponding voxel in the map,
     *  not the object which is passed to this function.
     */
    void removeVoxel(TPCVoxel const &voxel);

    /** Removes a voxel with given pad index and zBin from the map.
     */
    void removeVoxel(int padIndex, int zBin);

    /** Removes voxels in vector from the set.
     *  It removes the corresponding voxels in the map,
     *  not the objects in the vector.
     */
    void removeVoxels(std::vector<TPCVoxel> const &voxelVec);

    /** Removes all voxels with given pad index.
     */
    void removeVoxelsOnPad(int padIndex);

    /** Removes all voxels with given time index.
     */
    void removeVoxelsWithZBin(int zBin);

    /** Takes a length in z and returns the number of the corresponding zBin.
     */
    int getZBin( double zLength );



 protected:
   
  /** This structure is for the sorting of the voxels. They are sorted first by 
   *  zBin second by padIndex. The charge of the TPCVoxel is the data for the map.
   */
  struct VoxelIndex {

    VoxelIndex(int _padIndex, int _zBin);
    ~VoxelIndex();

    int zBin;
    int padIndex;

    bool operator()(VoxelIndex* vi1, VoxelIndex* vi2) {

      return(vi1->zBin = vi2->zBin ? vi1->padIndex < vi2->padIndex : vi1->zBin < vi2->zBin);
    }
    bool operator<(VoxelIndex const vi) const {

      return(zBin == vi.zBin ? padIndex < vi.padIndex : zBin < vi.zBin);
    }
  };

  /** Map to store the information contained in the voxels. padIndex and zBin are 
   *  used in a struct as key. The map sorted first by the zBin then by the padIndex.
   *  the charge contained in the voxel is the map data.
   */
  std::map<VoxelIndex, int> voxelMap;

  /** Pointer on the gear pad layout. 
   */
  gear::PadRowLayout2D const * _padLayout;

  /** Pointer on HitSmearer for the distribution of the hits.
   */
  HitDistributor * _distributeHitPtr;

  /** Instantiation parameters.
   */
  double _lengthOfZBin;

  
};

} //namespace marlintpc 

#endif // VOXELTPC_H

