#include <HitDistributor.h>

//C++
#include <cmath>
#include <complex>

//MARLIN
#include <streamlog/streamlog.h>


namespace marlintpc{

  /** Constructor for FixedPadSizeDiskLayout.
   */
  HitDistributor::HitDistributor(gear::PadRowLayout2D  const* padLayout, 
                                 double zBinLength) {

    _padLayout     = padLayout;
    _zBinLength    = zBinLength;
    _padLayoutType = _padLayout->getPadLayoutType();
  }


  HitDistributor::~HitDistributor() {}

  /** Steering function. The only function accessible from the outside.
   *  Takes a hit for input and calls all other functios for the smearing
   *  depending on the pad layout of the readout plane.
   */
  std::vector<TPCVoxel> HitDistributor::distributeHit(SimTrackerHit* hit, 
                                                  double sigmaTrans, double sigmaLong){

    _xHit = hit->getPosition()[0];
    _yHit = hit->getPosition()[1];   

    switch( _padLayoutType ){
    case gear::PadRowLayout2D::POLAR: {
      gear::FixedPadSizeDiskLayout const* circLayout = dynamic_cast<gear::FixedPadSizeDiskLayout const*>( _padLayout );
      std::vector<int> padIndexVec = getPads( circLayout, sigmaTrans );
      if( padIndexVec.empty() )
	return std::vector<TPCVoxel>();
      else
        return distributeHitOnPads( circLayout, padIndexVec, hit, sigmaTrans, sigmaLong );
      break;
    }
    case gear::PadRowLayout2D::CARTESIAN: {
      gear::RectangularPadRowLayout const* rectLayout = dynamic_cast<gear::RectangularPadRowLayout const*>( _padLayout );
      std::vector<int> padIndexVec = getPads( rectLayout, sigmaTrans );
      if( padIndexVec.empty() )
	return std::vector<TPCVoxel>();
      else
        return distributeHitOnPads( rectLayout, padIndexVec, hit, sigmaTrans, sigmaLong );
      break;
    }
    default:
      throw gear::Exception("Unknown pad layout type");
    };

    return std::vector<TPCVoxel>();    
  }

  /** Gets all pads within the smearing radius and stores their indices 
   *  in vector for circular pad layout.
   */  
  std::vector<int> HitDistributor::getPads(gear::FixedPadSizeDiskLayout  const* circPL, 
                                       double sigmaTrans){
    std::vector<int> padIndexVec;

    std::complex<double> compHit( _xHit, _yHit );
    double rHit   = std::abs( compHit );
    double phiHit = std::arg( compHit );
    if( phiHit < 0 )
      phiHit = ( 2 * M_PI )+phiHit;

    int hitIndex;
    try{
       hitIndex = circPL->getNearestPad( rHit , phiHit );
    }catch(std::out_of_range& e){
      return padIndexVec;
    }

    int cursor = hitIndex;

    if( checkDistance( circPL, hitIndex, sigmaTrans ) ){

      padIndexVec.push_back( hitIndex );
      padIndexVec = getPadsOnTheLeft( circPL, padIndexVec, hitIndex, sigmaTrans );
      padIndexVec = getPadsOnTheRight( circPL, padIndexVec, hitIndex, sigmaTrans );
    }else{
      return padIndexVec;
    }

    /** step up */
    int row = circPL->getRowNumber( cursor ) ;

    while(1) {
      row++;
      if( row < circPL->getNRows() ) {
	try{
	  cursor = circPL->getNearestPad( circPL->getPadCenter( circPL->getPadIndex(row,0) )[0] , phiHit);
	}catch(std::out_of_range& e){
	  break;
	}
	if( checkDistance(circPL, cursor, sigmaTrans) ) {
          padIndexVec.push_back( cursor );
	  padIndexVec = getPadsOnTheLeft( circPL, padIndexVec, cursor, sigmaTrans );
	  padIndexVec = getPadsOnTheRight( circPL, padIndexVec, cursor, sigmaTrans );
	}else
	  break;
      }else
	break;
    }
    cursor = hitIndex;
    row = circPL->getRowNumber( cursor );

    /** step down */
    while(1) {
      row--;
      if( row >= 0 ) {
	try{
	  cursor = circPL->getNearestPad(circPL->getPadCenter(circPL->getPadIndex(row,0))[0], phiHit);
	}catch(std::out_of_range& e){
	  break;
	}
	if( checkDistance( circPL, cursor, sigmaTrans ) ) {
          padIndexVec.push_back( cursor );
	  padIndexVec = getPadsOnTheLeft( circPL, padIndexVec, cursor, sigmaTrans );
	  padIndexVec = getPadsOnTheRight( circPL, padIndexVec, cursor, sigmaTrans );
	}else
	  break;
      }
      else
	break;
    }
    return padIndexVec;
  }

  /** Gets all pads within the smearing radius and stores their indices 
   *  in vector for rectangular pad layout.
   */  
  std::vector<int> HitDistributor::getPads(gear::RectangularPadRowLayout const* rectPL, 
                                       double sigmaTrans){
    std::vector<int> padIndexVec;

    int hitIndex;
    try{
       hitIndex = rectPL->getNearestPad( _xHit , _yHit );
    }catch(std::out_of_range& e){
      return padIndexVec;
    }

    streamlog_out(DEBUG2) << "_xHit      " << _xHit <<" ; _yHit      " << _yHit << std::endl;
    streamlog_out(DEBUG2) << "hitcenterX " << rectPL->getPadCenter(hitIndex) [0]
			  <<" ; hitcenterY " << rectPL->getPadCenter(hitIndex) [1] << std::endl;

    int cursor = hitIndex;

    if( checkDistance( rectPL, hitIndex, sigmaTrans ) ){

      padIndexVec.push_back( hitIndex );
      padIndexVec = getPadsOnTheLeft( rectPL, padIndexVec, hitIndex, sigmaTrans );
      padIndexVec = getPadsOnTheRight( rectPL, padIndexVec, hitIndex, sigmaTrans );

    }else
      return padIndexVec;

    int row = rectPL->getRowNumber( cursor );
    streamlog_out(DEBUG2) << "row "<< row << " ; hitIndex " << hitIndex << " which is pad "
			  << (hitIndex & 0xFFFF) << " in row " << (( hitIndex & 0xFFFF0000 )>>16) << std::endl;

    /* step up */
    while(1) {
      row++;
      if( row < rectPL->getNRows() ) {
	try{
	  cursor = rectPL->getNearestPad( _xHit , rectPL->getPadCenter( rectPL->getPadIndex(row,0) )[1] );
	}catch(std::out_of_range& e){
	  break;
	}
	if( checkDistance(rectPL, cursor, sigmaTrans) ) {
          padIndexVec.push_back( cursor );
	  padIndexVec = getPadsOnTheLeft( rectPL, padIndexVec, cursor, sigmaTrans );
	  padIndexVec = getPadsOnTheRight( rectPL, padIndexVec, cursor, sigmaTrans );
	}else
	  break;
      }else
	break;
    }

    cursor = hitIndex;
    row = rectPL->getRowNumber(cursor);

    /* step down */
    while(1) {
      row--;
      if( row >= 0 ){
	try{
	    streamlog_out(DEBUG2) <<  "rectPL->getPadIndex(row,0) " << rectPL->getPadIndex(row,0)
				  << " which is row " << ((rectPL->getPadIndex(row,0)& 0xFFFF0000 )>>16) << std::endl;
	    streamlog_out(DEBUG2) << " center is " << rectPL->getPadCenter( rectPL->getPadIndex(row,0)  )[1] 
				  << " , -yHit is " << _yHit << std::endl;
	  cursor = rectPL->getNearestPad(_xHit, rectPL->getPadCenter( rectPL->getPadIndex(row,0)  )[1] );
	}catch(std::out_of_range& e){
	  break;
	}
	if( checkDistance( rectPL, cursor, sigmaTrans ) ){
	  padIndexVec.push_back( cursor );
	  padIndexVec = getPadsOnTheLeft( rectPL, padIndexVec, cursor, sigmaTrans );
	  padIndexVec = getPadsOnTheRight( rectPL, padIndexVec, cursor, sigmaTrans );
	}else
	  break;
      }else
	break;
    }

    return padIndexVec;
  }

  /** Gets the indices of all pads in the row, which are left from the given pad 
   *  and wthin the smearing radius, and adds them to the vector which contains
   *  all pad indices of hit. (for circular pad layout)
   */
  std::vector<int> HitDistributor::getPadsOnTheLeft(gear::FixedPadSizeDiskLayout  const* circPL, 
                                                std::vector<int> padVec, int centerIndex, 
                                                double sigmaTrans){
    int np = centerIndex;

    while(1){
      np = circPL->getLeftNeighbour( np );

      if( checkDistance(circPL, np, sigmaTrans) && np != centerIndex )
	padVec.push_back( np );     
      else 
        return padVec;
    }
  }

  /** Gets the indices of all pads in the row, which are left from the given pad 
   *  and wthin the smearing radius, and adds them to the vector which contains
   *  all pad indices of hit. (for rectangular pad layout)
   */
  std::vector<int> HitDistributor::getPadsOnTheLeft(gear::RectangularPadRowLayout const* rectPL, 
                                                std::vector<int> padVec, int centerIndex, 
                                                double sigmaTrans){
    int np = centerIndex;

    while(1){
      try{
        np = rectPL->getLeftNeighbour( np );
      }catch(std::exception& e){   //exception handling has to be adjusted when the new gear version is installed!
	return padVec;
      }
      if( checkDistance( rectPL, np, sigmaTrans ) )
	padVec.push_back( np );
      else
	return padVec;
    }
  }

  /** Gets the indices of all pads in the row, which are right from the given pad 
   *  and wthin the smearing radius, and adds them to the vector which contains
   *  all pad indices of hit. (for circular pad layout)
   */
  std::vector<int> HitDistributor::getPadsOnTheRight(gear::FixedPadSizeDiskLayout  const* circPL, 
                                                 std::vector<int> padVec, int centerIndex, 
                                                 double sigmaTrans){
    int np = centerIndex;

    while(1){
      np = circPL->getRightNeighbour( np );

      if( checkDistance(circPL, np, sigmaTrans) && np != centerIndex )
	padVec.push_back( np );
      else
        return padVec;
    }
  }

  /** Gets the indices of all pads in the row, which are right from the given pad 
   *  and wthin the smearing radius, and adds them to the vector which contains
   *  all pad indices of hit. (for rectangular pad layout)
   */
  std::vector<int> HitDistributor::getPadsOnTheRight(gear::RectangularPadRowLayout const* rectPL, 
                                                 std::vector<int> padVec, int centerIndex, 
                                                 double sigmaTrans){
    int np = centerIndex;

    while(1){
      try{
	np = rectPL->getRightNeighbour( np );
      }catch(std::exception& e){
	return padVec;
      }
      if( checkDistance( rectPL, np, sigmaTrans) )
	padVec.push_back( np );
      else
	return padVec;
    }
  }

  /** Checks whether the distance between the given pad and the hit is lower than the smearing 
   *  range. Returns true when any part of the pad is within the smearing radius, even when it 
   *  is only a little corner of the pad. (for circular pad layout)
   */
  bool HitDistributor::checkDistance(gear::FixedPadSizeDiskLayout  const* circPL, int padIndex, double sigmaTrans ){

    const double SMEAR = 4;

    double rad = circPL->getPadCenter( padIndex )[0];
    double phi = circPL->getPadCenter( padIndex )[1];

    std::complex<double> zPad = std::polar( rad, phi);
    double xPad = zPad.real();
    double yPad = zPad.imag();

    double halfPadWidth = circPL->getFixedPadWidth() / 2;
    double halfPadHeight = circPL->getRowHeight( circPL->getRowNumber(padIndex) ) / 2 ;

    double uncorrDist = sqrt( pow(_xHit - xPad,2) + pow( _yHit-yPad,2) );

    if( (uncorrDist-std::min(halfPadWidth,halfPadHeight)) < SMEAR * sigmaTrans)
      return true;
     
    double xCloser = 0;
    double yCloser = 0;

    if( (xPad*yPad) >= 0 ){
      double xHeight = fabs(halfPadHeight * cos( phi ) );
      double xWidth  = fabs(halfPadWidth  * cos( ((M_PI) /2)-phi) );
      double yHeight = fabs(halfPadHeight * sin( phi ) );
      double yWidth  = fabs(halfPadWidth  * sin( (M_PI/2)-phi) );

      xCloser = std::max( xHeight, xWidth );
      yCloser = std::max( yHeight, yWidth );

    }else{
      double teta = M_PI - phi;

      double xHeight = fabs( halfPadHeight * cos( teta ) );
      double xWidth  = fabs( halfPadWidth * cos( (M_PI/2) - teta ) );
      double yHeight = fabs( halfPadHeight * sin( teta ) );
      double yWidth  = fabs( halfPadWidth * sin( (M_PI/2) - teta) );

      xCloser = std::max( xHeight, xWidth );
      yCloser = std::max( yHeight, yWidth );
    }

    if( xPad > _xHit )
      xPad = xPad - xCloser;
    else
      xPad = xPad + xCloser;

    if( yPad > _yHit )
      yPad = yPad - yCloser;
    else
      yPad = yPad + yCloser;

    double dist = sqrt( pow( _xHit-xPad, 2) + pow( _yHit-yPad,2) );

    if( dist < SMEAR * sigmaTrans )
      return true;
    else
      return false;
    
  }


  /** Checks whether the distance between the given pad and the hit is lower than the smearing 
   *  range. Returns true when any part of the pad is within the smearing radius, even when it 
   *  is only a little corner of the pad. (for rectangular pad layout)
   */
  bool HitDistributor::checkDistance(gear::RectangularPadRowLayout const* rectPL, int padIndex, double sigmaTrans ){

    const double SMEAR = 4;
    
    double xPad = rectPL->getPadCenter( padIndex )[0];
    double yPad = rectPL->getPadCenter( padIndex )[1];

    //claculate half pad width (with pitch!)
    int row = rectPL->getRowNumber( padIndex );
    double halfPadWidth  = rectPL->getPadWidth( padIndex ) / 2 ;
    if( rectPL->getPadsInRow( row ).size() >= 2 ) {
      int one = rectPL->getPadIndex( row, 0 );
      int two = rectPL->getPadIndex( row, 1 );
      halfPadWidth = fabs( rectPL->getPadCenter( one )[0] - rectPL->getPadCenter( two )[0])/2.;
    }
    double halfPadHeight = rectPL->getRowHeight( rectPL->getRowNumber(padIndex) ) /2. ;

    if( sqrt( pow(fabs( _xHit - xPad )- halfPadWidth , 2)
             +pow(fabs( _yHit - yPad )- halfPadHeight ,2)) < (SMEAR*sigmaTrans) )
      return true;
    else
      return false;
  }


  /** Smears the distributes the charge of the given hit in voxels according to the given pads, 
   *  transversal and longitudinal diffusion for circular pad layout. 
   */
  std::vector<TPCVoxel> HitDistributor::distributeHitOnPads(gear::FixedPadSizeDiskLayout  const* circPL, 
                                                    std::vector<int> padIndexVec, SimTrackerHit* hit, 
                                                    double sigmaTrans, double sigmaLong){

    std::vector<TPCVoxel> voxelVec;

    if( hit->getdEdx() > 0 ) {

      while(padIndexVec.begin()!=padIndexVec.end())  {

        int i = padIndexVec.front();

        //claculate the x,y coordinates of the pad center
        double xPad = circPL->getPadCenter(i)[0]*cos(circPL->getPadCenter(i)[1]);
        double yPad = circPL->getPadCenter(i)[0]*sin(circPL->getPadCenter(i)[1]);

        //rotate coordinate system so that the new x,y axis are parallel to pad widht,height
        double xHitRot =   _xHit*sin(circPL->getPadCenter(i)[1]) 
                         + _yHit*cos(circPL->getPadCenter(i)[1]);
        double yHitRot =   _yHit*sin(circPL->getPadCenter(i)[1])
                       - _xHit*cos(circPL->getPadCenter(i)[1]);
        double xPadRot =    xPad*sin(circPL->getPadCenter(i)[1]) 
                        + yPad*cos(circPL->getPadCenter(i)[1]);
        double yPadRot = -  xPad*cos(circPL->getPadCenter(i)[1]) 
                          + yPad*sin(circPL->getPadCenter(i)[1]);
      
        double halfPadWidth  = circPL->getFixedPadWidth()/2;
        double halfRowHeight = circPL->getRowHeight(circPL->getRowNumber(i))/2;

        //start and end values for the integration
        double xStart, xEnd, yStart, yEnd;
        if(xPadRot<xHitRot) {
          xEnd =   xPadRot - halfPadWidth;
	  xStart = xPadRot + halfPadWidth;
        }else{
	  xEnd =   xPadRot + halfPadWidth;
	  xStart = xPadRot - halfPadWidth;
        }
        if(yPadRot<yHitRot) {
          yEnd =   yPadRot - halfRowHeight;
	  yStart = yPadRot + halfRowHeight;
        }else {
	  yEnd =   yPadRot + halfRowHeight;
	  yStart = yPadRot - halfRowHeight;
        }

        // integration over a 2d gaussian
        int charge= static_cast<int>(round((hit->getdEdx())/4 *((erf((xEnd  -xHitRot)/(sqrt(2.)*sigmaTrans)) 
                                                                -erf((xStart-xHitRot)/(sqrt(2.)*sigmaTrans)))
                                                              * (erf((yEnd  -yHitRot)/(sqrt(2.)*sigmaTrans)) 
                                                                -erf((yStart-yHitRot)/(sqrt(2.)*sigmaTrans))))));
	
        charge = static_cast<int>(fabs(charge));
	
        int zBin = static_cast<int>( hit->getTime()  /_zBinLength ); 

	//sigmaLong = 5;

        if( sigmaLong ){

	  const double tDiff = 4;

	  int minBin = static_cast<int>( (hit->getTime() - (tDiff*sigmaLong)) / _zBinLength  );
	  int maxBin = static_cast<int>( (hit->getTime() + (tDiff*sigmaLong)) / _zBinLength  );

	  if( minBin < 0 )
	    minBin = 0;

	  for(int j = minBin; j <= maxBin; j++){ 

	    int tDiffCharge = static_cast<int> ( charge / 2  * ( erf(((j+1)*_zBinLength-hit->getTime())/(sqrt(2.)*sigmaLong))-
                                                                erf(( j   *_zBinLength-hit->getTime())/(sqrt(2.)*sigmaLong))  ));

	    voxelVec.push_back( TPCVoxel(i,j,tDiffCharge) );
	  }
	  padIndexVec.erase( padIndexVec.begin() );
	}else{ 
          voxelVec.push_back( TPCVoxel(i,zBin,charge)); 
          padIndexVec.erase(padIndexVec.begin());
	}
      }
    }
    return voxelVec;
  }
  /** Smears the distributes the charge of the given hit in voxels according to the given pads, 
   *  transversal and longitudinal diffusion for rectangular pad layout. 
   */
  std::vector<TPCVoxel> HitDistributor::distributeHitOnPads(gear::RectangularPadRowLayout const* rectPL, 
                                                    std::vector<int> padIndexVec, SimTrackerHit* hit, 
                                                    double sigmaTrans, double sigmaLong){

    std::vector<TPCVoxel> voxelVec;

    if( hit->getdEdx() > 0 ){

      //integration for all the stored pad indices
      while(padIndexVec.begin()!=padIndexVec.end()) {

        int i = padIndexVec.front();

//	streamlog_out(DEBUG2) << "padIndex: " << i << std::endl;

	//calculate the width of the pad (with pitch!)
	int row = rectPL->getRowNumber( i );
	double halfPadWidth = rectPL->getPadWidth(i)/2;
	if( rectPL->getPadsInRow( row ).size() >= 2) {
	  int one = rectPL->getPadIndex( row, 0);
	  int two = rectPL->getPadIndex( row, 1);
	  halfPadWidth = fabs( rectPL->getPadCenter(one)[0] - rectPL->getPadCenter(two)[0] ) /2.;
	}

        //start and end values for the integration
        double xStart, xEnd, yStart, yEnd; 
        if( rectPL->getPadCenter(i)[0] < _xHit ) {
          xStart = rectPL->getPadCenter(i)[0]+ halfPadWidth;
    	  xEnd   = rectPL->getPadCenter(i)[0]- halfPadWidth;
        }else {
          xStart = rectPL->getPadCenter(i)[0]- halfPadWidth;
	  xEnd   = rectPL->getPadCenter(i)[0]+ halfPadWidth;
        }  
        if( rectPL->getPadCenter(i)[1]<hit->getPosition()[1]) {
          yStart = rectPL->getPadCenter(i)[1]+rectPL->getRowHeight(rectPL->getRowNumber(i))/2;
	  yEnd   = rectPL->getPadCenter(i)[1]-rectPL->getRowHeight(rectPL->getRowNumber(i))/2;
        }else {
	  yStart = rectPL->getPadCenter(i)[1]-rectPL->getRowHeight(rectPL->getRowNumber(i))/2;
	  yEnd   = rectPL->getPadCenter(i)[1]+rectPL->getRowHeight(rectPL->getRowNumber(i))/2;
        }

        // integration over a 2d gaussian
        int charge= static_cast<int>(round(hit->getdEdx()/4 *((erf((xEnd  - _xHit)/(sqrt(2.)*sigmaTrans))
                                                              -erf((xStart- _xHit)/(sqrt(2.)*sigmaTrans)))
                                                            * (erf((yEnd  - _yHit)/(sqrt(2.)*sigmaTrans))	
                                                              -erf((yStart- _yHit)/(sqrt(2.)*sigmaTrans))))));
	charge = static_cast<int>(fabs(charge));   


        int zBin = static_cast<int>( hit->getTime()  /_zBinLength ); 

	streamlog_out(DEBUG0) << "sigma= " << sigmaLong << std::endl;
	streamlog_out(DEBUG0) << "time= " << hit->getTime()  << std::endl;
	streamlog_out(DEBUG0) << "_zBinLength= "<<_zBinLength<<std::endl;
	streamlog_out(DEBUG0) << "time/_zBinLength= "<<hit->getTime()  /_zBinLength<<std::endl;
	streamlog_out(DEBUG0) << "zBin= " << zBin << std::endl;

        if( sigmaLong ){

	  const double tDiff = 4;

	  int minBin = static_cast<int>( (hit->getTime() - (tDiff*sigmaLong)) / _zBinLength  );
	  int maxBin = static_cast<int>( (hit->getTime() + (tDiff*sigmaLong)) / _zBinLength  );

	  if( minBin < 0 )
	    minBin = 0;

	  for(int j = minBin; j <= maxBin; j++){ 

	    int tDiffCharge = static_cast<int> ( charge * ( erf(((j+1)*_zBinLength-hit->getTime())/(sqrt(2.)*sigmaLong))-
                                                            erf(( j   *_zBinLength-hit->getTime())/(sqrt(2.)*sigmaLong))  ));


	    voxelVec.push_back( TPCVoxel(i,j,tDiffCharge) );
	  }
	  padIndexVec.erase( padIndexVec.begin() );
	}else{ 
          voxelVec.push_back( TPCVoxel(i,zBin,charge)); 
          padIndexVec.erase(padIndexVec.begin());
	}
      } 
    }
    return voxelVec;

  }







} //namespace marlintpc
