#include "IonBackDriftProcessor.h"


//MARLINTPC
#include "TPCVoxel.h"

//C++
#include "iostream"
#include "math.h"
#include "libgem.h"
#include "vector"

//LCIO
#include "IO/LCWriter.h"
#include "IMPL/LCEventImpl.h" 
#include "IMPL/LCCollectionVec.h"

//GEAR
#include "gear/PadRowLayout2D.h"
#include "gear/TPCParameters.h"

///MARLIN
#include "marlin/Global.h"



using namespace lcio ;
using namespace marlin ;
using namespace std;
using namespace marlintpc;



IonBackDriftProcessor aIonBackDriftProcessor;


IonBackDriftProcessor::IonBackDriftProcessor() : Processor("IonBackDriftProcessor"),_ionDisc(0) {

  _description = "IonBackDriftProcessor gives coordinates of ions escaping GEM structure.";


  //registers steering parameters, name, description, class-variable, default value

  registerInputCollection(    LCIO::LCGENERICOBJECT,
                              "InputCollectionName" ,
                              "Name of the input collection with amplified electrons",
                              _inputCollectionName,
                              std::string("DistributedElectrons") );
                          
  registerOutputCollection(   LCIO::LCGENERICOBJECT,
			      "OutputCollectionName",
                              "Name of the output collection with backdrifting ions",
                              _outputCollectionName,
                              std::string("BackdriftingIons") );

  registerProcessorParameter( "ParamFile" , 
			      "Name of the File containing Parametrisation of GEMs"  ,
			      _paramFile ,
			      std::string("parameterfile") ) ;

  registerProcessorParameter( "Transient" , 
			      "switch off writing of output collection"  ,
			      _transient,
			      (int) 0) ;

  registerProcessorParameter( "UGEM1" , 
			      "Voltage on GEM 1"  ,
			      _UGEM1 ,
			      (double)315) ;

  registerProcessorParameter( "UGEM2" , 
			      "Voltage on GEM 2"  ,
			      _UGEM2 ,
			      (double)315) ;

  registerProcessorParameter( "UGEM3" , 
			      "Voltage on GEM 3"  ,
			      _UGEM3 ,
			      (double)315) ;

  registerProcessorParameter( "ETrans1" , 
			      "Value of Transfer Field 1"  ,
			      _ETrans1 ,
			      (double)2500) ;

  registerProcessorParameter( "ETrans2" , 
			      "Value of Transfer Field 2"  ,
			      _ETrans2 ,
			      (double)2500) ;

  registerProcessorParameter( "EInd" , 
			      "Value of Induction Field"  ,
			      _EInd ,
			      (double)5000) ;

  registerProcessorParameter( "DriftField",
                              "Value of the Drift Field" ,
			      _E ,
                              (double) 240);

}

void IonBackDriftProcessor::init() {

  streamlog_out(MESSAGE2) <<_description << endl;

  //always a good idea....
  printParameters();

  _nRun = 0 ;
  _nEvt = 0 ;

  //gear pad layout
  _padLayout = &(Global::GEAR->getTPCParameters().getPadLayout());

  //VoxelTPC for management of the build up of the ion disc
  if( _ionDisc )
    delete _ionDisc;
  _ionDisc = new marlintpc::VoxelTPC( _padLayout, 1);

    LoadGEMSettings();
}


void IonBackDriftProcessor::processRunHeader( LCRunHeader* run) { 

  run->parameters().setValue( _processorName + "_revision", "$Rev: 1048 $" );

  for ( ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++ ) {
    if ( ! i->second->isOptional() || i->second->valueSet() ) {
      run->parameters().setValue( _processorName + "_" + i->second->name(), i->second->value() );
    }
  }

  _nRun++ ;
} 


void IonBackDriftProcessor::processEvent( LCEvent * evt ) {

  LCCollection* pads = NULL;

  int entries = 1;

  try {
    pads = evt->getCollection( _inputCollectionName )  ;
  }catch(EventException &exception){
    streamlog_out(WARNING3) << exception.what() << endl;
    entries=0;
  }

  if(entries!=0){
    entries = pads->getNumberOfElements() ;
    streamlog_out(MESSAGE0) << "        " << entries << " objects in collection " << endl;
  }


  for(int i=0;i<entries;i++) {

    TPCVoxel* tempVoxel = dynamic_cast<TPCVoxel*>( pads->getElementAt(i) );
    _ionDisc->putChargeInVoxel( tempVoxel->getPadIndex(), 0, tempVoxel->getCharge() );
  }

  LCCollectionVec* ionVec = _ionDisc->getVoxelCollection();

  if (!ionVec){
    streamlog_out(DEBUG2) << "Empty ion voxel collection while input was not empty." << std::endl;
    return;
  }

 if( _transient )
    ionVec->setTransient( true );

  if (ionVec->empty() && _nEvt!=0) 
    delete ionVec;
  else //adds output collection to output file
    evt->addCollection( ionVec , _outputCollectionName ) ;
  
  //if(_nEvt%100==0)
    streamlog_out(MESSAGE2) <<"        Event "<< _nEvt <<" done!"<<endl;

  _nEvt ++ ;

}


//loads GEM settings from conditions data file
void IonBackDriftProcessor::LoadGEMSettings() {

  char *infile = const_cast<char*>(_paramFile.c_str());
  if( ! readpar( infile ) ) {
    streamlog_out(WARNING2) << "Not all parameters could be read from parameter file " 
                            << _paramFile << "!" << endl << "Exiting.";
    exit(1);
  }
  else 
    streamlog_out(DEBUG4) << "Parameter file read successfully " << endl;

  double params[7] = {_E, _UGEM1, _ETrans1, _UGEM2, _ETrans2, _UGEM3, _EInd}; 
  _ionBackDrift = ionf( params );

  streamlog_out(MESSAGE0) <<"Settings: "<< _E <<" "<< _UGEM1<<" "<< _ETrans1
                          <<" "<< _UGEM2 <<" "<< _ETrans2<<" "<< _UGEM3<<" "
                          << _EInd << endl << "Ionbackdrift: "<< _ionBackDrift << endl;

}


void IonBackDriftProcessor::check( LCEvent * evt ) { 
  // nothing to check here - could be used to fill checkplots in reconstruction processor
}


void IonBackDriftProcessor::end(){ 

}
