//*****************************************************
// Part of program to simulate a TPC with GEM readout *
//                                                    *
// Author: Astrid Muennich                            *
//                                                    *
//   Astrid Muennich                                  *
//   RWTH Aachen, III. Phys. Inst. B                  * 
//   Office: 28 A 112, Phone: +49/(0)241/8            *
//   Email: muennich@physik.rwth-aachen.de            * 
//                                                    *
//*****************************************************

/** Programm to apply electronic effect like shaping and binning of chareg into ADCCounts.
 */

#include "TPCElectronicsProcessor.h"
#include <iostream>
#include <math.h>
#include <IO/LCWriter.h>
#include <IMPL/LCRunHeaderImpl.h>
#include <EVENT/LCCollection.h>
#include <EVENT/MCParticle.h>
#include <EVENT/TrackerData.h>
#include <EVENT/SimTrackerHit.h>
#include <IMPL/TrackerRawDataImpl.h>
#include <IMPL/LCEventImpl.h>
#include <IMPL/LCCollectionVec.h>
#include <TPCVoxel.h>

using namespace lcio ;
using namespace marlin ;
using namespace std;
using namespace marlintpc;


TPCElectronicsProcessor aTPCElectronicsProcessor;

TPCElectronicsProcessor::TPCElectronicsProcessor() : Processor("TPCElectronicsProcessor") 
{
  
  _description = "TPCElectronicsProcessor takes care of binning the times and charge of electrons into an ADC.";
  
  registerInputCollection(    LCIO::LCGENERICOBJECT,
                              "CollectionName" , 
			      "Name of the Input collection with TPCPads"  ,
			      tpcpads ,
			      std::string("DistributedElectrons") ) ;

  registerInputCollection(    LCIO::SIMTRACKERHIT,
                              "CollectionName1" , 
			      "Name of the Input collection with drifted electrons"  ,
			      driftelectrons ,
			      std::string("DriftedElectrons") ) ;

  registerInputCollection(    LCIO::SIMTRACKERHIT,
                              "CollectionName2" , 
			      "Name of the Input collection with primary ionisation"  ,
			      primaries ,
			      std::string("PrimaryIonisation") ) ;

  registerOutputCollection(   LCIO::TRACKERRAWDATA,
                              "OutputCollectionName",
                              "Name of the output collection with adc values",
                              _outputCollectionName,
                              std::string("TPCRawData") ),
  
  registerProcessorParameter( "ReadoutFrequency" , 
			      "Readout Frequency in Mhz"  ,
			      frequency ,
			      (float)12.5) ;
  
  registerProcessorParameter( "RiseTime" , 
			      "Rise Time of Shaper in ns"  ,
			      risetime ,
			      (float)120.) ;
  
  registerProcessorParameter( "ADCResolution" , 
			      "Resolution of ADC in bit"  ,
			      adcres ,
			      (int)8) ;
  
  registerProcessorParameter( "ADCMaxCharge" , 
			      "Maximal Charge of ADC in number of electrons"  ,
			      adcmaxcharge ,
			      (int)33500) ;
  
  registerProcessorParameter( "ADCThreshold",
			      "Threshold for ADC in ADC bins",
			      adcthreshold,
			      (int)5);
   registerProcessorParameter( "GainFactor",
			       "Correction for gain due to temperature etc.",
			       gainfactor,
			       (double)1.0);
   registerProcessorParameter( "IntCutOff",
			       "Integration Cut Off of Gaussian Shape",
			       cutoff,
			       (double)3.0);

  registerProcessorParameter( "Transient" , 
			      "switch off writing of output collection"  ,
			      is_transient,
			      (int) 0) ;
}

void TPCElectronicsProcessor::init()
{
  cout<<_description<<endl;
  printParameters() ;
  _nRun = 0 ;
  _nEvt = 0 ;
  
}

void TPCElectronicsProcessor::processRunHeader( LCRunHeader* run) 
{ 
  run->parameters().setValue( _processorName + "_revision", "$Rev: 1111 $" );

  for ( ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++ ) {
    if ( ! i->second->isOptional() || i->second->valueSet() ) {
      run->parameters().setValue( _processorName + "_" + i->second->name(), i->second->value() );
    }
  }

  _nRun++ ;
} 

void TPCElectronicsProcessor::processEvent( LCEvent * evt ) 
{
  LCCollection* pads=0;
  int events=1;
  try {
    pads = evt->getCollection( tpcpads )  ;
  } catch(EventException &exception){
    events=0;
  }
  if(events!=0)
    events=pads->getNumberOfElements();

 
  LCCollectionVec* tpcdataVec = new LCCollectionVec( LCIO::TRACKERRAWDATA )  ;
  if(is_transient)
    tpcdataVec->setTransient(true);

    
  if(_nEvt==0)
    {
      tpcdataVec->parameters().setValue("ReadoutFrequency",frequency);
      tpcdataVec->parameters().setValue("RiseTime",risetime);
      tpcdataVec->parameters().setValue("ADCResolution",adcres);
      tpcdataVec->parameters().setValue("ADCMaxCharge",adcmaxcharge);
      tpcdataVec->parameters().setValue("ADCThreshold",adcthreshold);

      LCCollection* drift_el = evt->getCollection( driftelectrons )  ;     
      LCCollection* prim_el = evt->getCollection( primaries )  ;
      float difflong = drift_el->parameters().getFloatVal("DiffLong");
      float vdrift = drift_el->parameters().getFloatVal( "VDrift" ) ;
      double  TPCLength=prim_el->parameters().getFloatVal( "TPCLength") ;
      timeslices = (int)(ceil)(TPCLength/vdrift*frequency);
      timejump = (int)(3.*difflong*sqrt(TPCLength)/vdrift*sqrt(10.)); 
      
      adcbins=1;
      for(int i=0;i<adcres;i++) adcbins=adcbins*2;
      adcchargefactor = (double)adcbins/(double)adcmaxcharge;
      slicewidth = 1000./frequency;
    
    }    
  int sumq=0;
  int edgetime=0;
  int lasttime=0;
  int lastchannr=0;
  double sumTimeQ=0;
    
  for(int i=0;i<events;i++)
    {
      TPCVoxel *temphit=dynamic_cast<TPCVoxel*>(pads->getElementAt(i));
      if(i==0)
	{
	  lasttime = temphit->getTimeIndex();
	  lastchannr= temphit->getPadIndex();
	}
      int charge =  temphit->getCharge();
      int time = (int)temphit->getTimeIndex();
      int channr=(int)temphit->getPadIndex();
      if(channr==lastchannr && abs(time-lasttime)<=timejump)
	{
	  
		  edgetime=time;
		  sumq+=charge;
		  sumTimeQ+=(double)charge*(double)time;
		  lasttime=time; 
	}
      else   // peak has ended
	{
	  WritePeak(sumTimeQ,sumq,edgetime,lastchannr,tpcdataVec,gainfactor);

	  if(sumTimeQ  / (double)(sumq) <0) cout<<"Warning: AvTime<0"<<endl;
	  //set first valus of new peak
	  sumq=charge;
	  sumTimeQ=(double)(charge)*(double)time;
	  edgetime=time;
	  lasttime=time;
	  lastchannr=channr;
	}//peak ended
      if(i==events-1) //last peak in event must be written
	{
	  WritePeak(sumTimeQ,sumq,edgetime,lastchannr,tpcdataVec,gainfactor);
	}
	
    }
  if (tpcdataVec->empty() && _nEvt!=0 ) 
    delete tpcdataVec;
  else
    evt->addCollection( tpcdataVec , _outputCollectionName) ;
  

  if(_nEvt%100==0)
    cout<<"TPCElectronicsProcessor:    Event "<<_nEvt<<" done!"<<endl;
  _nEvt ++ ;
      
}

void TPCElectronicsProcessor::WritePeak(double sumTimeQ,int sumq,int edgetime,
			       int lastchannr, LCCollectionVec* tpcdataVec, 
			       double gainfactor)
{
  double Q=(double)sumq/gainfactor;
  double sigma = risetime/3.0;
  float tav = (double)sumTimeQ / (double)(sumq); 
  double starttime = tav-risetime;
  double stoptime= tav+cutoff*sigma;
  int timeslice = (int)(starttime/1000.0*frequency);
  double t1=starttime;
  double t2=(timeslice+1)*1000.0/frequency;
  double f=0.;
  int firstbin=0;
  bool notyetinpeak=true;
  int slicecounter=0;
  
  EVENT::ShortVec adc;
  while(f>0 || notyetinpeak)
    {
      if(t2<=stoptime)
	f=Q/2.*(erf((t2-tav)/(sqrt(2.)*sigma))-erf((t1-tav)/(sqrt(2.)*sigma)));
      else
	f=Q/2.*(erf((stoptime-tav)/(sqrt(2.)*sigma))-erf((t1-tav)/(sqrt(2.)*sigma)));
      
      if(f>0) notyetinpeak=false;    
      if(static_cast<unsigned long int>(f*adcchargefactor) > static_cast<unsigned int>(adcthreshold)  
	 && static_cast<int>(t1/1000.0*frequency)<=timeslices)
	{
	  if((long int)(f*adcchargefactor)>=adcbins) 
	    f=adcmaxcharge;
	  
	  adc.push_back((unsigned short int)(f*adcchargefactor));
	  if(adc.size()==1)
	    firstbin=(int)(t1/1000.0*frequency+0.5);
	  
	  slicecounter++;
	}
      t1=t2;
      t2+=slicewidth;
      if(t1>=stoptime)
	break;
    }//while
  if(adc.size()>0)
    {
      TrackerRawDataImpl* tpcpeak = new TrackerRawDataImpl;
      tpcpeak->setADCValues(adc);
      tpcpeak->setCellID0(lastchannr);
      tpcpeak->setTime(firstbin);
      tpcdataVec->addElement(tpcpeak);
      adc.clear();
      
    }
}


void TPCElectronicsProcessor::check( LCEvent * evt ) { 
}

void TPCElectronicsProcessor::end(){ 
  
   std::cout << name() 
 	    << " processed " << _nEvt << " events in " << _nRun << " runs "
 	    << std::endl ;

}
