/** Processor to sort amplified electrons into voxel
 */
#include "TimePixMapHandlerProcessor.h"
#include "TimePixPixelMode.h"
#include "TimePixDigitisationProcessor.h"

//C++
#include <iostream>
#include <cmath>

//LCIO
#include "IMPL/LCRunHeaderImpl.h"
#include "IMPL/LCEventImpl.h" 
#include "IMPL/LCCollectionVec.h"
#include "IMPL/TrackerRawDataImpl.h"
#include <EVENT/LCCollection.h>

// Aida
//#ifdef MARLIN_USE_AIDA // the whole processor does not make sense without aida
#include <AIDA/AIDA.h>
#include <marlin/AIDAProcessor.h>
#include <AIDA/IHistogramFactory.h>
#include <AIDA/IHistogram1D.h>
#include <AIDA/IHistogram2D.h>
// endif

//GEAR
#include <gear/TPCParameters.h>

//MARLIN
#include <marlin/Global.h>
#include "TPCVoxel.h"


using namespace lcio ;
using namespace marlin ;
using namespace std;
using namespace tpcconddata;

namespace marlintpc{ 


TimePixDigitisationProcessor aTimePixDigitisationProcessor ;


TimePixDigitisationProcessor::TimePixDigitisationProcessor() : Processor("TimePixDigitisationProcessor"), 
							       _tpc(0) {
  
  _description = "TimePixDigitisationProcessor maps charge to padplane ..." ;
  
  
  // register steering parameters: name, description, class-variable, default value


  registerInputCollection(    LCIO::LCGENERICOBJECT,
                              "InputCollectionName" , 
			      "Name of the Input collection "  ,
			      _inputCollectionName ,
			      string("DistributedElectrons") ) ;

  registerOutputCollection(   LCIO::TRACKERRAWDATA ,
                              "OutputCollectionName" ,
                              "Name of the Output collection" ,
                              _outputCollectionName ,
			      string("SimTimePixRawData") );


  registerProcessorParameter( "Transient" , 
			      "switch off writing of output collection"  ,
			      _transient,
			      (int) 0) ;

  registerProcessorParameter( "NColumn" , 
			      "Number of columns"  ,
			      _nColumn,
			      (int) 256 ) ;

  registerProcessorParameter( "NRow" , 
			      "Number of rows"  ,
			      _nRow,
			      (int) 256 ) ;

  registerProcessorParameter( "ChipID" , 
			      "chipID of current chip"  ,
			      _chipID,
			      (int) 0 ) ;

  registerProcessorParameter( "ReadOutFrequency" , 
			      "readout frequency of chip in MHz"  ,
			      _readoutFrequency,
			      (float) 56. ) ;

  registerProcessorParameter( "ShutterTime" , 
			      "shutter time of the measurement in mus"  ,
			      _shutter,
			      (float) 100. ) ;

  registerProcessorParameter( "PropFactorElectronsADC" , 
			      "conversion factor from ADC counts to electrons"  ,
			      _propFactorElectronsADC,
			      (float) 29.6 ) ;

  registerProcessorParameter( "AppliedThreshold" , 
			      "threshold of the TOT counts, applied by experimenter",
			      _appliedThreshold,
			      (int) 335 ) ;

  registerProcessorParameter( "HardwareThreshold" , 
			      "internal ADCcounts threshold of the chip",
			      _hardwareThreshold,
			      (int) 332 ) ;

  registerProcessorParameter( "ParameterTOTequation_a" , 
			      "Parameter a of equation TOT = c/(ln(Q-a)-b)-d",
			      _a,
			      (float) 560.57) ;

  registerProcessorParameter( "ParameterTOTequation_b" , 
			      "Parameter b of equation TOT = c/(ln(Q-a)-b)-d",
			      _b,
			      (float) 11.47) ;


  registerProcessorParameter( "ParameterTOTequation_c" , 
			      "Parameter c of equation TOT = c/(ln(Q-a)-b)-d",
			      _c,
			      (float) -408.13) ;


  

  _timePixMap = 0;
  
 
}

void TimePixDigitisationProcessor::init() { 

  streamlog_out(MESSAGE2) << _description << endl;

  _padLayout = &(Global::GEAR->getTPCParameters().getPadLayout());
 
 
  if( _tpc )
    delete _tpc;
  _tpc = new marlintpc::VoxelTPC( _padLayout, 1);

  printParameters() ;

  _timePixMap = TimePixMapHandlerProcessor::getTimePixMap();

  _distributionTimeValues =   AIDAProcessor::histogramFactory(this)->
      createHistogram1D( "VoxelTimeValues", 
			 "VoxelTimeValues", 
			 10000000, 0, 10000000);

}


 
void TimePixDigitisationProcessor::processRunHeader( LCRunHeader* run) {

  run->parameters().setValue( _processorName + "_revision", "$Rev: 1134 $" );

  for ( ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++ ) {
    if ( ! i->second->isOptional() || i->second->valueSet() ) {
      run->parameters().setValue( _processorName + "_" + i->second->name(), i->second->value() );
    }
  }

} 

 
void TimePixDigitisationProcessor::processEvent( LCEvent * evt ) {

    streamlog_out(MESSAGE1)<<" Event "<<evt->getEventNumber()<<std::endl;
    
  LCCollection* inputCol = NULL; 

  int objects = 1;

  try{
    inputCol = evt->getCollection( _inputCollectionName ); 
  }catch(EventException &exception) {
    streamlog_out(WARNING3) << "Exception: " << exception.what() << endl;
    objects = 0;
  }

  streamlog_out(DEBUG0)<<"_tpc->addCollection( inputCol );"<<std::endl;
  //add input collection to VoxelTPC
  _tpc->addCollection( inputCol );


  LCCollectionVec* outputCol = new LCCollectionVec( LCIO::TRACKERRAWDATA );

  if(_transient)   outputCol->setTransient(true);

  streamlog_out(DEBUG4)<<"ADCVector : resize & initialize "<<std::endl;
  EVENT::ShortVec ADCVector;
  ADCVector.resize( _nColumn*_nRow );
  for ( int i = 0; i < _nColumn*_nRow; i++ ) {
      ADCVector[i] = -1;
  }


  for ( int i = 0; i < inputCol->getNumberOfElements(); i++ ) {


      TPCVoxel* tpcVoxel = dynamic_cast<TPCVoxel*>(inputCol->getElementAt( i ) );

      if ( tpcVoxel != 0 ) { //dynamic cast gives nullpointer back if cast failed

	  //GEAR parameters
	  int padIndex = tpcVoxel->getPadIndex();
	  int rowNr = _padLayout->getRowNumber( padIndex );
	  int padNr = _padLayout->getPadNumber( padIndex );

	  //LCIO cellid0
	  int cellid0 = rowNr * _nColumn + padNr;

          //only proceed if this padIndex had not been dealt with yet
	  if ( ADCVector[cellid0] == -1 ) {

	      //determine mode of pixel:
	      int mode = ((_timePixMap->find( TimePixPixelMode::PixelKey( _chipID, cellid0 ) ))
			  ->second)->getPixelMode();
	  
	      streamlog_out(DEBUG2)<<"getVoxelsOnPad("<<padIndex<<")"<<std::endl;
	      std::vector< TPCVoxel >  voxelVector = _tpc->getVoxelsOnPad( padIndex );


	      if ( mode == TimePixPixelMode::TOT )
		  ADCVector[ cellid0 ] = getTOT( voxelVector );

	      if ( mode == TimePixPixelMode::TIMEPIX )
		  ADCVector[ cellid0 ] = getTIME( voxelVector );

	      if ( mode == TimePixPixelMode::MEDIPIX ) 
		  ADCVector[ cellid0 ] = getMEDIPIX( voxelVector );

	      if ( mode == TimePixPixelMode::ONEHIT )
		  ADCVector[ cellid0 ] = 1;
		  
	      streamlog_out(DEBUG2)<<std::endl;
	      streamlog_out(DEBUG2)<<"*******  EVENT NR : "<< evt->getEventNumber() <<" **********"<<std::endl;
	      streamlog_out(DEBUG2)<<"         PADINDEX : "<<padIndex<<std::endl;
	      streamlog_out(DEBUG2)<<"         CELLID0 :  "<<cellid0<<std::endl;
	      

	      streamlog_out(DEBUG0)<<"ADCVector["<<cellid0<<"] =  "<<ADCVector[cellid0]<<std::endl;

	      

	  }//if ( ADCVector[cellid0] == -1 ) 

      }//if ( tpcVoxel != 0 )
  
  } //for ( int i = 0; i < inputCol->getNumberOfElements(); i++ ) 
 
  streamlog_out(DEBUG4)<<"create new TrackerRawData"<<std::endl;
  //only write out the zerosuppresed rawdata
  for ( unsigned int cellid = 0; cellid < ADCVector.size(); cellid++ ) {
      
      if ( ADCVector[cellid] > 0 ) { //create new TrackerRawData
	  
	  TrackerRawDataImpl* rawData = new TrackerRawDataImpl();
	  rawData->setCellID1( _chipID );
	  rawData->setCellID0( cellid );
	  ShortVec ADCValues;

	  while ( ADCVector[cellid] > 0 ){
	      ADCValues.push_back( ADCVector[cellid] );
	      cellid++;
	      if ( cellid > ADCVector.size()-1 ) break;
	  }
	  
	  rawData->setADCValues( ADCValues );
	  outputCol->addElement( rawData );
	  
      }
      
  }

  streamlog_out(DEBUG4)<<"outputCol has "<<outputCol->getNumberOfElements()<<" elements"<<std::endl;

  if ( outputCol->getNumberOfElements() > 0 ) 
      evt->addCollection( outputCol , _outputCollectionName );
  else 
      delete outputCol;

  //  if(evt->getEventNumber()%100==0)
    streamlog_out(MESSAGE3) << "  Event "<< evt->getEventNumber() <<" done!"<<endl;

    LCCollectionVec* removeCol = _tpc->fetchVoxelCollection( );
    delete removeCol;


}//processEvent


void TimePixDigitisationProcessor::check( LCEvent * evt ) { 

}


void TimePixDigitisationProcessor::end(){ 
//    int mode = 
//	((_timePixMap->find( TimePixPixelMode::PixelKey( _chipID, cellid0 ) ))
//	 ->second)->getPixelMode();

}

int TimePixDigitisationProcessor::getTOT( TPCVoxel voxel ) {
/*
  TOT = c/(ln(Q-a)-b)-d
  
  a    560.57 ±    55.00
  b    11.47  ±     0.026
  c    -408.13   ± 10.75
  d    -2.40    ±    2.81 
*/
    int tot = 0;

    if ( voxel.getCharge() > _a ) {
    
	tot = static_cast<int>(_c / ( std::log( voxel.getCharge() - _a ) - _b ) );
    
	//number of electrons below threshold
	if ( tot < 0 ) tot = 0;

    }

    return tot;
}

int TimePixDigitisationProcessor::getTOT( std::vector< TPCVoxel > VoxelVec ) {

    int tot = 0;
    for ( std::vector< TPCVoxel >::iterator i = VoxelVec.begin(); i < VoxelVec.end(); i++ ) {

	tot += getTOT( *i );

    }

    return tot;

}

int TimePixDigitisationProcessor::getMEDIPIX( std::vector< TPCVoxel > VoxelVec ){

    int medipix = 0;
    int tot = 0;

    for ( std::vector< TPCVoxel >::iterator i = VoxelVec.begin(); i < VoxelVec.end(); i++ ) {
	
	if ( i == VoxelVec.begin() ) medipix = 1;

	else {//only if this voxel is not the first one
              //test if this voxel vanishes in signal of last voxel
	      //this is the case if tot-counts / _readoutFrequency is a greater time than TimeIndex
	      //BEWARE: _readoutFrequency is in MHz and TimeIndex in ns !!
 
	    if ( (tot/_readoutFrequency)*1E3 < i->getTimeIndex() ) //yes this hit is a second medipix count
		medipix++;
	}
	    
	tot = getTOT( *i );


    }

    return medipix;

}

int TimePixDigitisationProcessor::getTIME( std::vector< TPCVoxel > VoxelVec ){

    int t = static_cast<int>( VoxelVec.begin()->getTimeIndex()*1E-3 );//mus
    
    int time = static_cast<int>( ( _shutter - t )*_readoutFrequency );

    if ( time < 0  || getTOT(VoxelVec) == 0 ) time = 0;

    return time;

}

} //namespace marlintpc



