#ifndef PedestalSubtractorProcessor_h
#define PedestalSubtractorProcessor_h 1

// C++
#include <string>

// LCIO
#include "lcio.h"

// Marlin
#include "marlin/Processor.h"

// MarlinTPC
#include "PedestalHandler.h"


/// the MarlinTPC Package: reconstruction of TPC data
namespace marlintpc{

    /** Pedestal subtracting processor of the MarlinTPC package.
     * 
     *  pubstract pedestals provided via LCCD
     *
     *  <h4>Input - Prerequisites</h4>
     *  Collection of TrackerData
     *
     *  <h4>Output</h4> 
     *  Collection of TrackerData.
     * 
     * @param InputCollectionName Name of input data collection
     * @param OutputCollectionName Name of output data collection
     * @param SetOutputTransient If not 0 the output collection is set transient
     * @param PedestalCollectionName Name of the collection containing the pedestals
     *
     * if used without LCCD in addition
     * @param PedestalFile Name of the file where the pedestal are stored
     * is needed
     *
     * @author M. E. Janssen, DESY
     * @version $Id: PedestalSubtractorProcessor.h,v 1.1 2006/07/07 08:52:04 mjanssen Exp $ 
     */
    class PedestalSubtractorProcessor : public marlin::Processor {
	
    public:
	
	virtual Processor*  newProcessor() { return new PedestalSubtractorProcessor ; }
	
	PedestalSubtractorProcessor() ;
	
	virtual void init() ;
	
	virtual void processRunHeader( lcio::LCRunHeader* run ) ;
	
	virtual void processEvent( lcio::LCEvent * evt ) ; 
	
	virtual void check( lcio::LCEvent * evt ) ; 
	
	virtual void end() ;
	
	
    protected:
	
	std::string _inputColName ;    ///<- the name of the input collection
	std::string _outputColName ;   ///<- the name of the output collection
	std::string _pedestalColName ; ///<- the name of the pedestal collection

	// the following is only needed if compiled without LCCD
#ifndef USE_LCCD
	std::string _pedestalFileName ; ///<- the name of the pedestal file	
#endif
	
	/** give the status which is set to the transient flag of the output collection
	 *  it is an int instead of a boolean, because the processor parameter cannot be a boolean
	 */
	int _outputIsTransient;
	
	/** is true till the first output collection is wirten to an event
	 *  used to to store parameters only in the first collection
	 *  (also done, if collection is transient)
	 */
	bool isFirstCollection;

	int _nRun ; ///<- counts the runs processed
	int _nEvt ; ///<- counts the events processed
	
	PedestalHandler* _PedestalHandler; ///<- Handler which proviede the pedestals

    } ;

}

#endif



