#ifndef PulseFinderProcessor_h
#define PulseFinderProcessor_h 1

//C++
#include <string>

//LCIO
#include "lcio.h"
#include "EVENT/LCCollection.h"
#include "EVENT/TrackerData.h"
#include "EVENT/TrackerPulse.h"

//Marlin
#include "marlin/Processor.h"

// MarlinTPC
#include "PedestalHandler.h"


namespace marlintpc{

    /** ceate pulse form corrected raw data
     *
     *  first the pulses are found using a threshold based method (zero supression)
     *  second the time and charge information are calculated
     *  FIXME: the quality is not determined yet
     *
     *  <h4>Input - Prerequisites</h4>
     *  Collection of TrackerData (can be zero supressed)
     *
     *  <h4>Output</h4> 
     *  Collection of TrackerPulses
     *  Collection of TrackerData which belong to a pulse
     * 
     *  @param InputCollectionName name of the input TrackerData collection
     *  @param OutputTrackerPulseCollectionName name of the output TrackerPulse collection
     *  @param OutputTrackerDataCollectionName  name of the output TrackerData collection
     *  @param PedestalCollectionName name of the collection containing the pedestal information
     *  @param PedestalCollectionFile name of the file containing the pedestals (if not provided via LCCD)
     *  @param ForceSpectrumSave if not 0 the ADC spectrum will be stored for all pulses
     *  @param PulseStartThreshold a pulse start if ADC value is the given value times width of the pedestal
     *  @param SaveNBinsBeforeStart # bins saved before the start bin (exept there is another pulse)
     *  @param SaveNBinsAfterStart # bins saved after the start bin (exept there is another pulse)
     *  @param ReadoutFrequency freqenzy of the readout in Hz (only needed if not povided by GEAR
     *
     * @author M. E. Janssen, DESY
     * @version $Id: PulseFinderProcessor.h,v 1.1 2006/07/07 08:52:04 mjanssen Exp $ 
     */
    class PulseFinderProcessor : public marlin::Processor {
	
    public:
	
	virtual Processor*  newProcessor() { return new PulseFinderProcessor ; }
	
	/// constructor
	PulseFinderProcessor() ;
	
	virtual void init() ;
	
	virtual void processRunHeader( lcio::LCRunHeader* run ) ;
	
	virtual void processEvent( lcio::LCEvent * evt ) ; 
	
	virtual void check( lcio::LCEvent * evt ) ; 
	
	virtual void end() ;
	
	
    protected:
		
	int _nRun ; ///< run counter
	int _nEvt ; ///< event counter
	
	std::string  _inputColName;         ///< name of the collection, where input data are stored
	std::string  _outputDataColName;    ///< name of the collection, where TrackerData are stored
	std::string  _outputPulseColName;   ///< name of the collection, where TrackerPulse are stored
	std::string  _pedestalColName ;     ///<- the name of the pedestal collection
	
	// the following is only needed if compiled without LCCD
#ifndef USE_LCCD
	std::string _pedestalFileName ; ///<- the name of the pedestal file	
#endif
	
	/** set the status which is set to the transient flag of the output collection
	 *  it is an int instead of a boolean, because the processor parameter cannot be a boolean
	 */
	int _outputIsTransient;
	
	/** if not 0 the ADC spectrum will be saved for all pulses 
	 *  instead of only for pulses with quality !=0
	 *  it is an int instead of a boolean, bcause the processor parameter cannot be a boolean
	 */
	int _forceSpectrumSave;

	/** is true till the first output collection is wirten to an event
	 *  used to to store parameters only in the first collection
	 *  (also done, if collection is transient)
	 */
	bool isFirstCollection;

	float _pulseStartThreshold;  ///< start threshold to detect a pulse
	int   _saveNBinsBeforeStart; ///< # of bins stored before the threshold
	int   _saveNBinsAfterStart;  ///< # of bins stored after the threshold 
        /* the minmal bin per pulse 
	 * = _saveNBinsBeforeStart + _saveNBinsAfterStart
	 */
	int   _minPulseDataSize;
	
	double _ReadoutFrequency;    ///< readout frequency in Hz (node: the time will be stored in ns)
	
	PedestalHandler* _PedestalHandler; ///< will provide the acutal pedestal width
	
	/** find pulses by threshold in the ADC information
	 *  @param thisData 'raw' data (non zero supressed, can also handle zero supressed data)
	 */
	int   findPulses( lcio::TrackerData* thisData, lcio::LCCollectionVec* thisCollection,
			  float StartThreshold );
	
	/** return the polarity of the pulse spectrum
	 *  @param thisData the ADC spectrum to ba analyzed
	 *  @return true if polarity is negative
	 */
	bool  checkForNegPolarity( lcio::TrackerData* thisData, float noiseValue );

	/** calculate the time information of the given pulse spectrum
	 *  the time information will the mean of the (positive) derivative
	 *  one the first part will be taken into acount
	 *  if there is a second seperated (positive) part of the derivative, it will be ingnored
	 *  if the pulse has negative polarity the frist negative part 
	 *  of the derivative will be taken for calculation
	 *  @param thisPulse ADC spectrum from which the time will be calculated
	 *  @param NoiseValue variation smaller than this value will be ingnored during the calculation
	 *  @param negPolarity the polarity of the pulse (true if pulse have negative polarity)
	 *  @return the time information of the pulse in ns
	 */
	float calcPulseTime(lcio::TrackerData* thisPulse, bool negPolarity, float NoiseValue = 0.0);
	
	/** calculate the integrated charge of the given pulse
	 *  by sum up all bins of the ADC spectrum
	 * polarity of the pulse will be taken into acount
	 *  @param thisPulse ADC spectrum for charge calculation
	 *  @param negPolarity the polarity of the pulse (true if pulse have negative polarity)
	 */
	float calcPulseCharge(lcio::TrackerData* thisPulse, bool negPolarity);
	
    };
    
}
#endif
    


