#ifndef TimePixClusterFinderProcessor_h
#define TimePixClusterFinderProcessor_h 1

// C++
#include <string>
#include <vector>

// LCIO
#include "lcio.h"
#include <EVENT/TrackerData.h>

// Marlin
#include "marlin/Processor.h"

//Root
#include "TROOT.h"
#include "THStack.h"

namespace AIDA{
  class IHistogram1D;
}


namespace marlintpc{

    /* the following comment is tranfered to doxygen documentation
     * which begins with a second asterix, or in case of on line comment an extra third slash 
     */
    /**  Clusterfinding Processor of the MarlinTPC package .
     * 
     *   converts zero suppressed TrackerData of a TimePix to TrackerHits
     *
     *  <h4>Input - Prerequisites</h4>
     *  a collection of zero suppressed TrackerData
     *
     *  <h4>Output</h4> 
     *  Collection containing TrackerHits which represent clusters of adjacent pixels.
     * 
     * @param InputCollectionName Name of input collection, zero suppressed TrackerData
     * @param OutputCollectionName Name of output collection, unseparated TrackerHits
     * @param SetOutputTransient If not 0 the output collection is set transient
     * @param nColumn number of columns on the chip
     * @param nRow number of rows on the chip
     * 
     * @author Simone Zimmermann
     */
    class TimePixClusterFinderProcessor : public marlin::Processor {
	
    public:
	
	/* This method will be called by the marlin package
	 * It returns a processor of the currend type (here TimePixClusterFinderProcessor)
	 */
	virtual Processor*  newProcessor() { return new TimePixClusterFinderProcessor ; }
	
	/* the default constructor
	 * here the processor parameters are registered to the marlin package
	 * other initialisation should be placed in the init method
	 */
	TimePixClusterFinderProcessor() ;
	
	/* Called at the beginning of the job before anything is read.
	 * Use to initialize the processor, e.g. book histograms
	 */
	virtual void init() ;
	
	/* Called for every run.
	 * in this processor it is not used
	 */
	virtual void processRunHeader( lcio::LCRunHeader* run ) ;
	
	/* Called for every event - the working horse.
	 * Here the real conversion take place
	 */
	virtual void processEvent( lcio::LCEvent * evt ) ; 
	
	/* This method is only called if the check flag is set (default) in the main processor
	 * (where the input files are given)
	 * It should be used to create check plots which are not needed in a complete analysis
	 * (where normaly the check flag is deactivated)
	 */
	virtual void check( lcio::LCEvent * evt ) ; 
	
	
	/* Called after data processing for clean up.
	 * e.g. saving control histogramms, delete created objects
	 * (which are not stored in a collection/event), etc.
	 */
	virtual void end() ;
	
	
    protected:
	
	/* normally the doxygen documentation is palced before the method/variable
	 * but for one line comments in case of variables it may be more convenied for you
	 * to place it after the variable. In this case you use an arrow to tell doxygen
	 * to which variable the commet belongs to.
	 */
	
	std::string _inputColName ;  ///<- the name of the input collection
	std::string _outputColName ; ///<- the name of the output collection
	std::string _outputColNameData ; ///<- the name of the output collection
	int _nColumn; ///<- number of columns on the chip
	int _nRow; ///<- number of rows on the chip
	int _minNoPixel; ///<- minimum number of pixel in cluster
	
	/** give the status which is set to the transient flag of the output collection
	 * it is an int instead of a boolean, because the processor parameter cannot be a boolean
	 */
	int _outputIsTransient;

	/** is true till the first output collection is wirten to an event
	 *  used to to store parameters only in the first collection
	 *  (also done, if collection is transient)
	 */
	bool isFirstCollection;

	int _nRun ; ///<- counts the processed runs
	int _nEvt ; ///<- counts the processed events

    private:

 	std::vector<TrackerData*> DataVec;
	AIDA::IHistogram1D * histoncl;
//	THStack** hs;

    } ;
    
}

#endif
    
    

