#ifndef TimePixMapHandlerProcessor_h
#define TimePixMapHandlerProcessor_h 1

#include "TimePixMapListener.h"
//#include "TimePixHighTOTAnalyserProcessor.h"

// C++
#include <string>
#include <map>
#include <vector>

// LCIO
#include "lcio.h"

// Marlin
#include "marlin/Processor.h"
#include "TimePixPixelMode.h"
#include "lccd/ConditionsMap.hh"


//AIDA
//#include <AIDA/AIDA.h>

//ROOT
//#include "TFile.h"
//#include "THStack.h"

namespace AIDA{

  class IHistogram1D;
  class IHistogram2D;

}

namespace marlintpc{

    class TimePixHighTOTAnalyserProcessor;

    /* the following comment is tranfered to doxygen documentation
     * which begins with a second asterix, or in case of on line comment an extra third slash 
     */
    /**  
     *  reads in conditionsdata objects TimePixModeMap and TimePixStatusMap and unifies them to the TimePixMap
     *
     *  <h4>Input - Prerequisites</h4>
     *  a collection of TimePixModeMap and a collection of TimePixStatusMap
     *
     *  <h4>Output</h4> 
     *  TimePixMap
     * 
     * @param DataCollectionName Name of the collection containing the  data 
     * @param HitCollectionName Name of the collection containing the TrackerHits
     * @param ModeMapCollectionName Name of mode map collection
     * @param SetOutputTransient If not 0 the output collection is set transient
     * @param nColumn number of columns on the chip
     * @param nRow number of rows on the chip
     * @param clusterSizeToSep number of pixels in a cluster for it to be separated
     * @param SetOutputTransient If not 0 the output collection is set transient
     * 
     * @author Simone Zimmermann
     */


    class TimePixMapHandlerProcessor : public marlin::Processor {
	
    public:
	
	/* This method will be called by the marlin package
	 * It returns a processor of the currend type (here TimePixMapHandlerProcessor)
	 */
	virtual Processor*  newProcessor() { return new TimePixMapHandlerProcessor ; }
	
	/* the default constructor
	 * here the processor parameters are registered to the marlin package
	 * other initialisation should be placed in the init method
	 */
	TimePixMapHandlerProcessor() ;
	
	/* Called at the beginning of the job before anything is read.
	 * Use to initialize the processor, e.g. book histograms
	 */
	virtual void init() ;
	
	/* Called for every run.
	 * in this processor it is not used
	 */
	virtual void processRunHeader( lcio::LCRunHeader* run ) ;
	
	/* Called for every event - the working horse.
	 * Here the real conversion take place
	 */
	virtual void processEvent( lcio::LCEvent * evt ) ; 
	
	/* This method is only called if the check flag is set (default) in the main processor
	 * (where the input files are given)
	 * It should be used to create check plots which are not needed in a complete analysis
	 * (where normaly the check flag is deactivated)
	 */
	virtual void check( lcio::LCEvent * evt ) ; 
	
	
	/* Called after data processing for clean up.
	 * e.g. saving control histogramms, delete created objects
	 * (which are not stored in a collection/event), etc.
	 */
	virtual void end() ;


	/** This is the function that can be called from all following processors.
	 *  It returns the current TimePixMap.
	 */
	static tpcconddata::TimePixMap * getTimePixMap();

	/** This function is called within the TimePixHighTOTAnalyserProcessor.
	 *  It returns the Nullpointer if there is no TimePixMapHandlerProcessor active in 
	 *  the reconstruction chain.
	 *  It initializes the _HighTOTProcessor pointer.
	 *  If this pointer has been initialised, getTimePixMap() returns the TimePixMap
	 *  of the current event corrected to high TOT values
	 */	
	static TimePixMapHandlerProcessor* registerHighTOTProcessor( TimePixHighTOTAnalyserProcessor* htp );

	friend class TimePixHighTOTAnalyserProcessor;


    protected:
	
	/* normally the doxygen documentation is palced before the method/variable
	 * but for one line comments in case of variables it may be more convenied for you
	 * to place it after the variable. In this case you use an arrow to tell doxygen
	 * to which variable the commet belongs to.
	 */
	
	std::string _inputColName ;  ///<- the name of the input collection
	std::string _outputColName ; ///<- the name of the output collection
	std::string _modeMapColName ; ///<- the name of the mode map collection
	std::string _statusMapColName ;///<- the name of the correction to mode map collection

	int _nColumn; ///<- number of columns on the chip
	int _nRow; ///<- number of rows on the chip
	int _clusterSizeToSep; ///<- number of pixels in a cluster in order for it to be separated
	int _chipID; ///<- number of current chip, 0 if single chip

	/** give the status which is set to the transient flag of the output collection
	 * it is an int instead of a boolean, because the processor parameter cannot be a boolean
	 */
	int _outputIsTransient;

	/** is true till the first output collection is writen to an event
	 *  used to store parameters only in the first collection
	 *  (also done, if collection is transient)
	 */
	bool isFirstCollection;

	int _nRun ; ///<- counts the processed runs

	double a; ///<- parameter of axis of projection y = a*x + b
	double b; ///<- parameter of axis of projection y = a*x + b

	/**  pointer to TimePixMapListener
	 *   TimePixMapListener controlls the TimePixMap.
	 */
	static tpcconddata::TimePixMapListener* _timePixMapListener;

	/**  pointer to TimePixHighTOTAnalyserProcessor
	 *   in case that this processor is active, the TimePixMap must be corrected
	 *   further by this processor on event to event basis.
	 */
	static TimePixHighTOTAnalyserProcessor* _HighTOTProcessor;

	/**  pointer to TimePixMapHandlerProcessor
	 *   needed as return value for registerHighTOTProcessor.
	 */
	static TimePixMapHandlerProcessor* _MapHandlerProcessor;

    private:
    	/**  pointer to unmodified TimePixMap
	 *   as handeld by TimePixMapListener
	 *   consisting of TimePixModeMap and TimePixStatusMap only. 
	 */
	tpcconddata::TimePixMap* getUnmodifiedTimePixMap();

	std::vector<AIDA::IHistogram1D *> hProjectXVec;
	AIDA::IHistogram1D * highTOT;
	AIDA::IHistogram2D * aidahistocluster;
	AIDA::IHistogram2D ** aidaonehit;
	

    } ;
    
}

#endif
    
    

