#ifndef TOPO_FINDER_H
#define TOPO_FINDER_H 1

#include <map>

#include <lcio.h>
#include <EVENT/TrackerHit.h>
#include <EVENT/TrackerPulse.h>
#include <gear/TPCParameters.h>
#include <gear/PadRowLayout2D.h>

namespace marlintpc{

/** TopoFinder contains finding algorithms for clusters of TrackerPulses.
 *  It operates on a std::multimap<int, TrackerPulse*> which contains
 *  the pulses to search in a multimap sorted by the corresponding pad index.
 *  \attention !! Be careful with the iterators! They are deleted from the map 
 *     and are invalid afterwards!!
 *  
 *  @author Martin Killenberg, University of Bonn
 *
 *  This class is not a processor, but called from them.
 */

class TopoFinder {

 protected:

    /// The constructor.
    TopoFinder();

//    void setPulseMap(std::multimap<int, EVENT::TrackerPulse*> *map);

    /** Find the hit which contains the given pulse. 
     *	It looks for pulses in neighbouroing pads.
     */
    EVENT::TrackerHit* findHit(std::multimap<int, EVENT::TrackerPulse*>::iterator currentPulse);

    /** Find Hits in the neighbouring rows. If empty rows are allowed this does not necessarrily 
     *	mean next neighbouring rows. (Empty row means there is no hit in this row which matches the
     *	given hit. It does not have to be completely empty.)
     */
    EVENT::TrackerHitVec *findHitsInNeighbouringRows(EVENT::TrackerHit* hit);

    /** Find all the hits in the given pad range. There might be more than one hit in 
     *	case of a looper or an interaction. 
     */
    EVENT::TrackerHitVec *findHitsInRange(int index1, int index2, float time);
		 
    int _maxEmptyPads;   ///< Maximum number of consecuitve empty pads in cluster
    int _minHitSize; ///< Minimum number of pads per hit
    float _maxTimeSpread;///< Maximum time between pulses in a cluster in ??
    int _minTrackHits; ///< Minimum number of hits on track
    int _maxSkipRows;    ///< Maximum number of empty rows

    float _vDrift; ///< Optional: Set drift velocity in case there is no conditions data

    /// The pad layout as defined in GEAR
    int _padLayoutType; ///< Pad layout type (PadRowLayout2D::CARTESIAN or PadRowLayout2D::POLAR)
    gear::PadRowLayout2D const *_padLayout;

    /// Drift distance as described in gear
    float _maxDriftDistance;

    /** Flag to determine if the second half TPC is reconstructed, which looks in the
     *	negative z direction.
     *	In this case the x and z coordinate are multiplied with -1 (endplate rotated by
     *	180 degress around y axis wrt. first endplate)
     */
    int _negativeZDirection;

    /** A multimap containing the pad index and the corresponding pulse.
     *  Sorted by pad index.
     */
    std::multimap<int, EVENT::TrackerPulse*> *_pulsesByIndex;

 public:
    /** A helper function to add the contents of one TrackerHitVec to another.
     *  Be aware that you need handlers of TrackerHitVec because the pointers are modified!
     *  \attention vectorToBeAdded is 0 after calling this function!
     */
    static void addTrackerHitVecs(EVENT::TrackerHitVec **vectorToAddTo, EVENT::TrackerHitVec **vectorToBeAdded);

    /** Get the revision of the TopoFinder.
     *  Use this information in the processor which uses or is derived from this class:
     *  \verbatim     run->parameters().setValue( _processorName + "_revision", 
     *                          std::string("$Rev:$  using TopoFinder " + myTopoFinder->getRevision()));
     *
     */
    std::string getRevision() ;
};
    
}// namespace marlintpc

#endif //TOPO_FINDER_H
