#ifndef TRACK_FITTER_BASE_H
#define TRACK_FITTER_BASE_H

#include <string>
#include <marlin/StringParameters.h>
#include <EVENT/LCRunHeader.h>
#include <EVENT/Track.h>
#include <IMPL/TrackImpl.h>
#include <EVENT/TrackerHit.h>

namespace gear
{
    class TPCParameters;
    class PadRowLayout2D;
}//namespace gear

namespace marlintpc
{
	/** 
	 *   The TrackFitterBase is a virtual class from which the
	 *   actual track fitters are derived. It provides a common interface
	 *   which is accessed by the track fitter processors and analysis 
	 *   processors.                                                  <br>
	 *   Most members are purely virtual.
	 *   They have to be overloaded/implemented in the
	 *   derived classes.
	 *
	 *   To calculate something like a single point resolution it
	 *   is required that these "points" are accessed via the 
	 *   EVENT::TrackerHit as an interface. Some methods like the global
	 *   likelihood method don't use the xy-information from hits but
	 *   refit a x-position from the charge on the pads using the given 
	 *   track fit . This requires a calculateResiduals method
	 *   individually for each fitting method.
	 *   To have a common interface for all methods the TrackerHit
	 *   is used, through which the more low level information like 
	 *   pulses can be accessed.                                     <br>
	 *   The calculateResiduals methods are not purely virtual. They are
	 *   implemented for the generic case which calculates the distance
	 *   to the point of closest approach (pca) in the readout plane 
	 *   and the z-distance  to the pca in the s-z plane.
	 */

    class TrackFitterBase
    {
	public:

	/** Defines constants to determine which fitter type was used
	 *  to perform the fit and calculate the residuals.
	 *  When starting to implement a fitter please add a constant 
	 *  and commit this file to the trunk  so your constant is 
	 *  reserved.
	 */
	// These variables are static, don't forget to instantiate them in TrackFitterBase.cc
	struct FitterTypes
	{
	    //static const int BASE = 0;
	    static const int LIKELIHOOD = 1; ///< Global likelihood method
	    static const int CHISQUARE = 2; ///< \f$\chi^2\f$ without geometry information
	    static const int CHISQUAREPADS = 3; ///< \f$\chi^2\f$ unsing pad geometry information
	    static const int LINEARREGRESSION = 4; ///< Linear regression for a straight line
	};

	/** The Constructor, nothing to do here. The implementations will get some parameters here.
	 */
	TrackFitterBase(){}

	virtual ~TrackFitterBase(){}

	/// Get the revision of the actual fitter implementation.
	/// This is needed for revision logging.
        /// Implementation should look like:
        /// \verbatim return std::string( "$Rev$" ); \endverbatim
	/// In case you don't reimplement the calculateResidulas methods you 
	/// should also state the revision of this base class:
        /// \verbatim return std::string( "$Rev$ ; TrackFitterBase ") + TrackFitterBase::getRevision(); \endverbatim
        /// Don't forget to switch on the 'svn Rev' property for this file:
	/// \verbatim svn propset svn:keywords Rev TrackFitterImpl.cc \endverbatim
	virtual std::string getRevision();

	/// Information what fitter implementation this is.
	virtual int getFitterType() = 0;

	/** The function which does the actual fitting.
	 *  The returned TrackImpl is created from heap. Be careful to 
	 *  either add it to an event/collection or to delete it otherwise
	 *  to avoid memory leaks.
	 */
	virtual IMPL::TrackImpl *fitTrack(EVENT::Track *seedTrack) = 0;

	/** Return the distance of the hit \c testHitNumber of the testTrack 
	 *  to the point of closest approach on the referenceTrack wrt.\ the 
	 *  hit.
	 *  If no referenceTrack is given the
	 *  distance to the testTrack is returned.
	 * 
	 *  Throws EVENT::DataNotAvailableException in case there is no hit number 
	 *  "testHitNumber" on testHit.
	 *
	 *  Calls  calculateResidual(EVENT::TrackerHit *testHit, EVENT::Track *trackWithoutTestHit)
	 *  in this implementation.
	 */
	virtual EVENT::DoubleVec calculateResiduals(EVENT::Track *testTrack,
					     uint testHitNumber,
					     EVENT::Track *referenceTrack = NULL);

	/** Return the distance of the given hit to the test track 
	 *  where this hit has been removed.
	 *  The residual is calculated as the distance 
	 *  to the point of closest approach on the referenceTrack wrt.\ the  hit.
	 */
	virtual EVENT::DoubleVec calculateResiduals(EVENT::TrackerHit *testHit,
					     EVENT::Track *trackWithoutTestHit);

    protected:
	// calculates the s of the pca wrt. the hit in the readout plane

	// calculate the pca wrt. the hit in the readout plane
	
	

	/// Calculate the angular distance on a circle.
	/// The maximum can be \f$\pi\f$.
	double distanceOnCircle(double phi1, double phi2);
    };

}// namespace marlintpc

#endif // TRACK_FITTER_BASE_H
