#ifndef TRACK_FITTER_CHI_SQUARE_PADS_H
#define TRACK_FITTER_CHI_SQUARE_PADS_H

#include <string>
#include <marlin/StringParameters.h>
#include <EVENT/LCRunHeader.h>
#include <EVENT/Track.h>
#include <IMPL/TrackImpl.h>
#include <EVENT/TrackerHit.h>
#include "TrackFitterChiSquare.h"

namespace gear
{
    class TPCParameters;
    class PadRowLayout2D;
}//namespace gear

namespace marlintpc
{
    /** g\f$\chi^2\f$ fitter which takes into account the pad geometry.
     *  It reimplements the \c getResiduals functions to calculate the residuals
     *  along the pad rows and not perpendicular to the track.
     *
     *  Currently the actual fitter is not implemented yet.
     *  But one can instanciate the fitter to use the
     *  \c getResiduals functions, which is reimplemented to caculate the residuals 
     *  along the pad rows.
     *
     *  \todo: Implement the fitter function.
     */

    class TrackFitterChiSquarePads : public TrackFitterChiSquare
    {
	public:

	/** The constructor needs the pad geometry, which is taken grom GEAR
	 */
	TrackFitterChiSquarePads(gear::TPCParameters const *gearTPCParameters);

	virtual ~TrackFitterChiSquarePads(){}

	virtual std::string getRevision();

	/// Information what fitter implementation this is.
	virtual int getFitterType();


	/** Return the distance of the hit \c testHitNumber of the testTrack 
	 *  to the point of closest approach on the referenceTrack wrt.\ the 
	 *  hit.
	 *  If no referenceTrack is given the
	 *  distance to the testTrack is returned.
	 * 
	 *  Throws EVENT::DataNotAvailableException in case there is no hit number 
	 *  "testHitNumber" on testHit.
	 *
	 *  Calls  calculateResidual(EVENT::TrackerHit *testHit, EVENT::Track *trackWithoutTestHit)
	 *  in this implementation.
	 */
	virtual DoubleVec calculateResiduals(EVENT::Track *testTrack,
					     uint testHitNumber,
					     EVENT::Track *referenceTrack = NULL);

	/** Return the distance of the given hit to the test track 
	 *  where this hit has been removed.
	 *  The residual is calculated as the distance 
	 *  along the pad row.
	 *  
	 */
	virtual DoubleVec calculateResiduals(EVENT::TrackerHit *testHit,
					     EVENT::Track *trackWithoutTestHit);

    protected:

	/// A pointer to the GEAR TPC parameters
	gear::TPCParameters const *_TPCParameters;

	/// A pointer to the GEAR pad layout
	gear::PadRowLayout2D const * _padLayout;

	/// The type of pad layout (PadRowLayout2D::CARTESIAN or PadRowLayout2D::POLAR)
	int _padLayoutType;
	
	/// Called from both constructors to set the GEAR information
        /// (originally there were two, right now this is obsolete)
	void  initialiseGEARGeometry(gear::TPCParameters const *gearTPCParameters);

	/// Calculate the angular distance on a circle.
	/// The maximum can be \f$\pi\f$.
	double distanceOnCircle(double phi1, double phi2);
    };

}// namespace marlintpc

#endif // TRACK_FITTER_BASE_H
