#ifndef TRACK_FITTER_FACTORY_H
#define TRACK_FITTER_FACTORY_H

namespace EVENT
{
    class LCRunHeader;
    class LCEvent;
}

namespace marlintpc{

    class TrackFitterBase;

/**  The TrackFitterFactory is an interface to handle track fitters. It knows the type of fitter to use.
 *   This is set in he constructor, either directly in case of a TrackFitterProcessor,
 *   or by passing the run header in case of an anlysis processor. The run header
 *   contains the information which fitter type has been used to fit the data.
 *
 *   Analysis processors use the \c calculateResiduals function of the track fitters 
 *   so this code does not have to be implemented twice. Especially for analysis processors the 
 *   abstract interface means that the analysis code is independent from the track fitter used,
 *   but all the specialities of a track fitter like geometry dependence, dependence on the gas mixture
 *   and whatever the special fitter requires are taken into accout.
 *   
 *   The createTrackFitter function is called in every event. The ownership of the TrackFitter stays
 *   with the factory. To avoid memory leaks and for efficiency the factory has a clever
 *   memory handling. It allocates the memory for a fitter only once and initialises every object 
 *   in the same location in case of event-dependent fitters (or not at all in case the fitter does
 *   not depend on changing data). So there always is only one instance of the fitter per fitter factory.
 *   
 */

class TrackFitterFactory
{
 public:

    /** Constructor for use in track fitter processors.
     *  The factory is initialised with the fitter type.
     *
     *  \exception gear::NotImplementedException
     *  Thrown in case the fitter type is not known to the factory.
     */
    TrackFitterFactory(int const &type);

    /** Constructor for use in analysis processors.
     *  The parameters are read from the run header.
     *
     *  NOTICE: not implemented yet, use \c TrackFitterFactory(int const &type); for the time being.
     *
     *  \exception gear::NotImplementedException
     *  Thrown in case the fitter type is not known to the factory.
      */
    TrackFitterFactory(EVENT::LCRunHeader  const *rh);

    /** The destructor
     *  Get rid of the fitte which was allocated
     */
    ~TrackFitterFactory();

    /** Create a track fitter from the heap.
     *  The type of fitter is the one defined in the constructor.
     *
     *  Some fitter types might need information from the conditions data.
     *  This is transferred by giving a pointer to the event.
     *  If the specific fitter type does not need the event it can be ommitted.
     *  Throws !!!FIXME what kind of exception in case the fitter need conditon data 
     *  but \c event = 0.
     */

    TrackFitterBase *createTrackFitter(EVENT::LCEvent const *event = 0 );

 private:
    int _fitterType;  ///< The type of fitter to create
    TrackFitterBase *_theFitter; ///< Pointer to the instance of the fitter

    void setType(int const &type); ///< Only to be called by the constructors to avoid doublication of code
};

} // namespace marlintpc

#endif // TRACK_FITTER_FACTORY_H
