#include "PedestalSubtractorProcessor.h"

// C++
#include <iostream>

// LCIO
#include <EVENT/LCCollection.h>
#include <EVENT/TrackerData.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/TrackerDataImpl.h>



using namespace lcio ;
using namespace marlin ;

namespace marlintpc{


    PedestalSubtractorProcessor aPedestalSubtractorProcessor ;
    

    PedestalSubtractorProcessor::PedestalSubtractorProcessor() : Processor("PedestalSubtractorProcessor") {
	
	/* description of the processor which will be displayed in the
	 * steering file autometicly generated by 
	 * marlin (use MarlinTPC -x (XML-format) or MarlinTPXC -l (text format))
	 */
#ifndef USE_LCCD
	_description = "PedestalSubtractorProcessor correct the TrackerData by subtracting the pedestals provided via LCCD";
#else 
	_description = "PedestalSubtractorProcessor correct the TrackerData by subtracting the pedestals provided via a file";
#endif
	
	
	/* register steering parameters: name, description, class-variable, default value
	 * the type will by definded by the type of the default value
	 * string, double, float and int are possible
	 * plaese note to get a float you must cast in C++: (float) 1.3 
	 */
	registerInputCollection(    LCIO::TRACKERDATA,  
                                    "InputCollectionName" , 
				    "Name of the input collection"  ,
				    _inputColName ,
				    std::string("TPCConvertedRawData") ) ;
	
	registerOutputCollection(   LCIO::TRACKERDATA,
                                    "OutputCollectionName" , 
				    "Name of the output collection"  ,
				    _outputColName ,
				    std::string("TPCData") ) ;
	
	registerProcessorParameter( "SetOutputTransient" , 
				    "if not 0 the output collection is set transient"  ,
				    _outputIsTransient,
				    1);
	
	registerProcessorParameter( "PedestalCollectionName" , 
				    "Name of the collection containing the pedestals",
				    _pedestalColName,
				    PedestalHandler::getDefaultColName() );

/* if MARLIN is compiled without LCCD a file is needed which contains the pedestal information
 */
#ifndef USE_LCCD
	registerProcessorParameter( "PedestalCollectionName" , 
				    "File containing the pedestals",
				    _pedestalFileName,
				    std::string("Pedestal.slcio"));
#endif

    }
    
    
    void PedestalSubtractorProcessor::init() { 
	
	// usually a good idea to
	printParameters() ;
	
	_nRun = 0 ;
	_nEvt = 0 ;

// the pedestal handler with and without LCCD
#ifdef USE_LCCD
	_PedestalHandler = new PedestalHandler( _pedestalColName );
#else
	_PedestalHandler = new PedestalHandler( _pedestalColName, _pedestalFileName );
#endif

    }
    
    void PedestalSubtractorProcessor::processRunHeader( LCRunHeader* run) { 

        run->parameters().setValue( _processorName + "_revision", "$Rev: 762 $");
	for ( ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++ ) {
	    if ( ! i->second->isOptional() || i->second->valueSet() ) {
	        run->parameters().setValue( _processorName + "_" + i->second->name(), i->second->value() );
	    }
	}
	
	_nRun++ ;
	
    } 
    
    void PedestalSubtractorProcessor::processEvent( LCEvent * evt ) { 
	
	// get the input collection form the event
	try{
	    LCCollection* inputCol = evt->getCollection( _inputColName ) ;
	    
	    // create the output collection with a string defining the TYPE (see lcio.h)
	    LCCollectionVec* outputCol = new LCCollectionVec( LCIO::TRACKERDATA ); 
	    // set the transient flag as definded by the processor parameter
	    outputCol->setTransient( (_outputIsTransient != 0) );
	    
	    // get the number of elements in the input collection
	    int nRawData = inputCol->getNumberOfElements();
	    //and loop over it
	    for(int iRawData=0; iRawData< nRawData ; iRawData++){
		
		/* getElement return only a LCObject
		 * you have to cast it the get the right type
		 * please use dynamic_cast which will return 0 if the object can not be casted
		 */
		TrackerData* RawData = 
		    dynamic_cast<TrackerData*>( inputCol->getElementAt( iRawData ) ) ;
		
		if ( RawData != 0 ) {
		    //only poceed if the cast was successful
		    
		    /* create a new object of the TrackerData type
		     * you have to use the IMPL, otherwise you can not set values
		     */
		    TrackerDataImpl* Data = new TrackerDataImpl();
		    
		    // transver the values
		    
		    //first the simple
		    Data->setCellID0( RawData->getCellID0() );
		    Data->setCellID1( RawData->getCellID1() );// normally 0 for TPC data 
		    
		    Data->setTime( RawData->getTime() );
		    
		    // and subtract the pedestal
		    float PedestalVal = _PedestalHandler->getPedestalValue( RawData->getCellID0() );

		    FloatVec RawDataADCVal = RawData->getChargeValues();
		    
		    for (FloatVec::iterator iSlice=RawDataADCVal.begin(); 
			 iSlice!=RawDataADCVal.end(); iSlice++){
			// fill the data vector with 
			(Data->chargeValues()).push_back( (*iSlice)-PedestalVal );
		    }
		    
		    /* fill the TrackerData in the collection
		     * please note: DO NOT DELETE the object this will be done by LCIO
		     */
		    outputCol->addElement( Data );
		    
		}
	    }
	    
	    // save the processor parameters in the first collection stored
	    if (isFirstCollection){
		outputCol->parameters().setValue("InputCollectionName",_inputColName);
		// TO DO: make this sence? 
		outputCol->parameters().setValue("SetOutputTransient",_outputIsTransient);
		outputCol->parameters().setValue("PedestalCollectionName",_pedestalColName);
#ifndef USE_LCCD
		outputCol->parameters().setValue("PedestalFileName",_pedestalFileName);
#endif

		isFirstCollection = false;
	    }
	    
	    
	    /* after convertin all raw data objects
	     * add the collection to the eventand asigning a name to it
	     * please note again: DO NOT DELETE the collection  this will be done by LCIO
	     */
	    evt->addCollection( outputCol, _outputColName );
	    
	}
	catch(DataNotAvailableException){
	    // if there is not input collection give a warning to the user
	    std::cerr<<"Warning of PedestalSubtractorProcessor::processEvent: input collection "
		     <<_inputColName<<" cannot be found in event "<<_nEvt<<std::endl;	
	}
	
	_nEvt ++ ;
    }
    
    
    
    void PedestalSubtractorProcessor::check( LCEvent * evt ) { 
	// nothing to check here - could be used to fill checkplots in reconstruction processor
    }
    
    
    void PedestalSubtractorProcessor::end(){ 
	
	//   std::cout << "PedestalSubtractorProcessor::end()  " << name() 
	// 	    << " processed " << _nEvt << " events in " << _nRun << " runs "
	// 	    << std::endl ;
	
    }
    
}
