#include "PulseFinderProcessor.h"

//C++
#include <iostream>

//LCIO
#include <EVENT/LCCollection.h>
#include <EVENT/TrackerRawData.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/TrackerDataImpl.h>
#include <IMPL/TrackerPulseImpl.h>

#ifdef USE_GEAR
// STUFF needed for GEAR
#include <marlin/Global.h>
#include <gear/GEAR.h>
#include <gear/TPCParameters.h>
#endif

using namespace lcio ;
using namespace marlin ;
using namespace std;

namespace marlintpc{

    PulseFinderProcessor aPulseFinderProcessor;

    PulseFinderProcessor::PulseFinderProcessor() : Processor("PulseFinderProcessor") {
  
	// modify processor description
	_description = "PulseFinderProcessor: find hits out of the raw data and calculate the time and charge information" ;
    
    
	// register steering parameters: name, description, class-variable, default value
    
	registerInputCollection(    LCIO::TRACKERDATA,
	                            "InputCollectionName" , 
				    "Name of the input TrackerData collection"  ,
				    _inputColName ,
				    std::string("TPCData")) ;

	registerOutputCollection(   LCIO::TRACKERDATA,
                                    "OutputTrackerDataCollectionName" , 
				    std::string("Name of the output TrackerData collection")+
				    std::string("only the data related to a pulse will be stored"),
				    _outputDataColName ,
				    std::string("TPCDataInPulses")) ;

	registerOutputCollection(   LCIO::TRACKERPULSE,
                                    "OutputTrackerPulseCollectionName" , 
				    "Name of the output TrackerPulse collection"  ,
				    _outputPulseColName ,
				    std::string("TPCPulses"));
    
	registerProcessorParameter( "SetOutputTransient" , 
				    "if not 0 the output collection is set transient"  ,
				    _outputIsTransient,
				    1);

	registerProcessorParameter( "ForceSpectrumSave" , 
				    "if not 0 the ADC spectrum will be stored for all pulses",
				    _forceSpectrumSave,
				    1);
	
	registerProcessorParameter( "PulseStartThreshold",
				    std::string("a pulse start if ADC value is the given value times width ")+
				    std::string("of the pedestal"),
				    _pulseStartThreshold,
				    (float)5.0);

	registerProcessorParameter( "SaveNBinsBeforeStart",
				    "# bins saved before the start bin (exept there is another pulse)",
				    _saveNBinsBeforeStart,
				    2);

	registerProcessorParameter( "SaveNBinsAfterStart",
				    "# bins saved after the start bin (exept there is another pulse)",
				    _saveNBinsAfterStart,
				    8);

/* if MARLIN is compiled without GEAR the readout frequency must be given via a processor parameter
 */
#ifndef USE_GEAR
	registerProcessorParameter( "ReadoutFrequency",
				    "freqenzy of the readout in Hz",
				    _ReadoutFrequency,
				    1E9);
#endif
    
	registerProcessorParameter( "PedestalCollectionName" , 
				    "Name of the collection containing the pedestal information",
				    _pedestalColName,
				    PedestalHandler::getDefaultColName() );

/* if MARLIN is compiled without LCCD a file is needed which contains the pedestal information
 */
#ifndef USE_LCCD
	registerProcessorParameter( "PedestalCollectionFile" , 
				    "File containing the pedestals",
				    _pedestalFileName,
				    std::string("Pedestal.slcio"));
#endif

    }

void PulseFinderProcessor::init() { 

    // usually a good idea to
    printParameters() ;

    _nRun = 0 ;
    _nEvt = 0 ;

	
#ifdef USE_LCCD
	_PedestalHandler = new PedestalHandler( _pedestalColName );
#else
	_PedestalHandler = new PedestalHandler( _pedestalColName, _pedestalFileName );
#endif
	
// if Marlin is compiled with GEAR the readout frequency will be given via the GEAR steering file
#ifdef USE_GEAR
	_ReadoutFrequency = Global::GEAR->getTPCParameters().getReadoutFrequency();
#endif

    }

    void PulseFinderProcessor::processRunHeader( LCRunHeader* run) { 

        run->parameters().setValue( _processorName + "_revision", "$Rev: 762 $");
	for ( ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++ ) {
	    if ( ! i->second->isOptional() || i->second->valueSet() ) {
	        run->parameters().setValue( _processorName + "_" + i->second->name(), i->second->value() );
	    }
	}

	_nRun++ ;
    } 

    void PulseFinderProcessor::processEvent( LCEvent * evt ) { 

	// give a life sign every 100 event
	_nEvt++;
	if (_nEvt%100 == 0) {
	    std::cout<<_nEvt<<"\r";
	    std::cout.flush();
	}

	try{
	    LCCollection* InputCollection = evt->getCollection( _inputColName );
    
	    // create the output collections for pulses and zero suppressed data (ADC spectra)
	    IMPL::LCCollectionVec* outputColPulse = new IMPL::LCCollectionVec(LCIO::TRACKERPULSE);
	    IMPL::LCCollectionVec* outputColData  = new IMPL::LCCollectionVec(LCIO::TRACKERDATA);

	    // set the transient flag as definded by the processor parameter
	    outputColPulse->setTransient( (_outputIsTransient != 0) );
	    outputColData->setTransient( (_outputIsTransient != 0) );
	

	    // go the input data and do zero suppression using a threshold method

	    /* collection to store the zero suppressed pulses
	     * this extra collectio is needed, because data will only be stored if pulse has quality !=0
	     */
	    IMPL::LCCollectionVec* foundPulses = new IMPL::LCCollectionVec(LCIO::TRACKERDATA);

	    int nData = InputCollection->getNumberOfElements();
	    for(int iData=0; iData< nData ; iData++){
	    
		TrackerData* thisData = 
		    dynamic_cast<TrackerData*>( InputCollection->getElementAt( iData ) ) ;
	    
		if (thisData){// cast is OK
		    float Threshold = _PedestalHandler->getPedestalWidth( thisData->getCellID0() )
			*_pulseStartThreshold;
		    if (Threshold<0) // if threshold is negative the pedestal width is not defined
			Threshold = _pulseStartThreshold; // use pedestal width = 1

		     findPulses( thisData, foundPulses, Threshold);
		}
	    }


	    /* analyze the found oulses calculate time and charge information
	     * and store the pulse and the data (if needed)
	     */
	    bool  isDoublePulseCandidate  = false;
	    bool  isAnomalousShaped       = false;
	    bool  isUnderflow             = false;
	    bool  isOverflow              = false;
	
	    bool  forcePulseSpectrumSave  = true;

	    int   quality  = 0;
	
	    int   nPulse   = foundPulses->getNumberOfElements()  ;
	    for(int iPulse=0; iPulse< nPulse ; iPulse++){
		
	      //	      cout << "foundPulses size" << foundPulses->size() << "\t" << nPulse << endl;
	      TrackerData* thisData = 
		    dynamic_cast<TrackerData*>( foundPulses->getElementAt( iPulse ) ) ;
	    
		if (thisData){
		    float noiseValue = _PedestalHandler->getPedestalWidth( thisData->getCellID0() )
			*_pulseStartThreshold;
		    if (noiseValue<0) // pedestal width is not defined
			noiseValue = _pulseStartThreshold;

		    bool negPolarity = checkForNegPolarity( thisData, noiseValue );

		    quality = //FIXME quality must be analyzed
			+(1<<0)*isDoublePulseCandidate
			+(1<<1)*isAnomalousShaped
			+(1<<2)*isUnderflow
			+(1<<3)*isOverflow
			+(1<<31)*forcePulseSpectrumSave;
		    //save Pulse
		    TrackerPulseImpl* newPulse = new TrackerPulseImpl();
		    newPulse->setCellID0( thisData->getCellID0() );
		    newPulse->setCellID1( thisData->getCellID1() );
		    newPulse->setTime( calcPulseTime( thisData, negPolarity, noiseValue ));
		    newPulse->setCharge( calcPulseCharge( thisData, negPolarity ) );
		    newPulse->setQuality( quality );
		    if ( quality != 0 ) {
			// make a copy of the TrackerData
			TrackerDataImpl* newData = new TrackerDataImpl();
			newData->setCellID0( thisData->getCellID0() );
			newData->setCellID1( thisData->getCellID1() );
			newData->setTime( thisData->getTime() );
			FloatVec thisDataChargeVal = thisData->getChargeValues();
			for (FloatVec::iterator iSlice=thisDataChargeVal.begin(); 
			     iSlice!=thisDataChargeVal.end(); iSlice++){
			    // fill the data vector with 
			    (newData->chargeValues()).push_back( (*iSlice) );
			}
			outputColData->addElement( newData );
		    }
		    outputColPulse->addElement( newPulse );
		
		}// if thisData
		//std::cout<<outputColPulse->getNumberOfElements()<<std::endl;
	    }// for iPulse
	    delete foundPulses;	

	    if (isFirstCollection){
		outputColPulse->parameters().setValue("InputCollectionName",_inputColName);
		// TO DO: make this sence? 
		outputColPulse->parameters().setValue("SetOutputTransient",_outputIsTransient);
		outputColPulse->parameters().setValue("PedestalCollectionName",_pedestalColName);
		outputColPulse->parameters().setValue("OutputTrackerDataCollectionName",_outputDataColName);
		outputColPulse->parameters().setValue("PulseStartThreshold",_pulseStartThreshold);
		outputColPulse->parameters().setValue("SaveNBinsBeforeStart",_saveNBinsBeforeStart);
		outputColPulse->parameters().setValue("SaveNBinsAfterStart",_saveNBinsAfterStart);
		outputColPulse->parameters().setValue("ReadoutFrequency", (float)_ReadoutFrequency);
	    
		isFirstCollection = false;
	    }

	    evt->addCollection(outputColPulse, _outputPulseColName);
	    evt->addCollection(outputColData, _outputDataColName);
	}
	catch(DataNotAvailableException){
	    // if there is not input collection give a warning to the user
	    std::cerr<<"Warning of PulseFinderProcessor::processEvent: input collection "
		     <<_inputColName<<" cannot be found in event "<<_nEvt<<std::endl;	
	}

    }



    void PulseFinderProcessor::check( LCEvent * evt ) { 


    }


    void PulseFinderProcessor::end(){ 
    
    }


    int PulseFinderProcessor::findPulses( TrackerData* thisData, LCCollectionVec* thisCollection,
				 float StartThreshold ){

	int nPulseFound = 0;
	
	//	LCCollection* returnCollection = new IMPL::LCCollectionVec(LCIO::TRACKERDATA);

	FloatVec ADCSpectrum = thisData->getChargeValues();
		    
	TrackerDataImpl* newData = 0;
	int   lastNBinUnderThreshold = 0;
	int   goBack                 = 0;
	int   nBinsOverThreshold     = 0;
	float Time                   = thisData->getTime();
	for (FloatVec::iterator iSlice=ADCSpectrum.begin(); 
	     iSlice<ADCSpectrum.end(); iSlice++){
	
	    if ( ( (*iSlice) >= StartThreshold ) || ( (*iSlice) <= -StartThreshold ) ){
		// pulse detected
	      
	        newData = new TrackerDataImpl;
		newData->setCellID0( thisData->getCellID0() );
		newData->setCellID1( thisData->getCellID1() );

		//go back if possible
		goBack = (lastNBinUnderThreshold<_saveNBinsBeforeStart?
			  lastNBinUnderThreshold:_saveNBinsBeforeStart);

		newData->setTime( Time - goBack );

		for (int iGoBack=goBack; iGoBack>0; --iGoBack){
		    // save time slices before pulse
		    (newData->chargeValues()).push_back( (*(iSlice-iGoBack)) );
 		}

		lastNBinUnderThreshold = 0;
		nBinsOverThreshold     = 0;

		// search for end of pulse
		// at least ? bins unter threshold
		while ( (iSlice<ADCSpectrum.end()) && (lastNBinUnderThreshold<2) ){
		    if ( ( (*iSlice) >= StartThreshold ) || ( (*iSlice) <= -StartThreshold ) ){

			//save the bin(s) 
			for (int iGoBack=lastNBinUnderThreshold; iGoBack>0; iGoBack--){
			    // also the bin between the signals must be stored
			    (newData->chargeValues()).push_back( (*(iSlice-iGoBack)) );
			}
			(newData->chargeValues()).push_back( (*iSlice) );

			lastNBinUnderThreshold = 0;
			nBinsOverThreshold++;
		    
		    }
		    else{
		      lastNBinUnderThreshold++;
		    }

		    Time+=1.; //increment time
		    iSlice++; //next time bin
		}// while !ADCSpectrum.end()
		int saveNBinsAfterPulse = _minPulseDataSize-(newData->chargeValues()).size();

		for (int iBin = 0; iBin<saveNBinsAfterPulse; iBin++) {
		    if ( iSlice==ADCSpectrum.end() ) break; // reached end of pulse
		    (newData->chargeValues()).push_back( (*iSlice) );
		    Time+=1.;
		    iSlice++;
		}
		thisCollection->addElement( newData );

	    }
	    else{
		lastNBinUnderThreshold++;
	    }

	    Time+=1.;

	}

	return nPulseFound;
    }

    bool  PulseFinderProcessor::checkForNegPolarity( TrackerData* thisData, float noiseValue ){
	
	FloatVec ADCSpectrum = thisData->getChargeValues();
	if (ADCSpectrum.size()==0) return false;

	FloatVec::iterator iSlice=ADCSpectrum.begin();
	if (ADCSpectrum.size()==1)
	    // spectrum is only one bit long
	    return ( (*iSlice)<0 );

	// search for first maximum
	float valueBefore = 0.; // value of the bin before
	float valueThis   = (*iSlice); // value of the actual bin
	if (valueThis<0) valueThis*=-1; // value is positve
	do{
	    valueBefore = valueThis;
	    iSlice++;
	    valueThis   = (*iSlice);
	    if (valueThis<0) valueThis*=-1; // value is positve
	}while ((iSlice!=ADCSpectrum.end()) && ( ( valueThis>valueBefore) ) || (valueThis>noiseValue) ); 
        // one to fare, go one back
	iSlice--;
	return ( (*iSlice)<0 );
    }

    float PulseFinderProcessor::calcPulseTime(TrackerData* thisPulse, bool negPolarity, float NoiseValue){
 
	//std::cout<<"calcPulseTime("<<thisPulse<<", "<<NoiseValue<<"):";

	float PulseTime    = 0.0;
	float Derivative   = 0.0; 
	float Norm         = 0.0;
	float SliceTime    = thisPulse->getTime();
	bool  wasOverNoise = false;
	FloatVec ChargeValues = thisPulse->getChargeValues();
	if (ChargeValues.size()>1){
	    for (FloatVec::iterator iSlice = ++ChargeValues.begin(); 
		 iSlice != ChargeValues.end(); iSlice++){
	    
		Derivative  = (*iSlice)-(*(iSlice-1));
		if (negPolarity)
		    Derivative *=-1; // make derivative sign insensitive
		SliceTime +=1.;
		//std::cout<<Derivtive<<" ("<<SliceTime<<");";
		if (Derivative > NoiseValue){
		    wasOverNoise = true;
		}
		Norm     +=Derivative;
		PulseTime+=Derivative*SliceTime;

		//use only the first positve derivative
		if (Derivative< (-NoiseValue) && wasOverNoise ){
		    break;
		}
	    }
	
	    PulseTime /= Norm;
	}
	else{
	    PulseTime = SliceTime;
	}
	//std::cout<<" = "<<PulseTime<<std::endl;

	return PulseTime*1E9/_ReadoutFrequency; //conversion in ns 
    }

    float PulseFinderProcessor::calcPulseCharge(TrackerData* thisPulse, bool negPolarity){

	float    PulseCharge  = 0.0;
	FloatVec ChargeValues = thisPulse->getChargeValues();
	for (FloatVec::iterator iSlice = ChargeValues.begin()++; 
	     iSlice != ChargeValues.end(); iSlice++){
	    PulseCharge+=(*iSlice);
	}

	if (negPolarity) PulseCharge*=-1; /// flip sign if polarity is negative

	return PulseCharge;
    }

}
