#include "TimePixHitCenterCalculatorProcessor.h"
#include "TimePixMapHandlerProcessor.h"
#include "TimePixPixelMode.h"

// C++
#include <iostream>
#include <queue>
#include <string>
#include <sstream>

// LCIO
#include <EVENT/LCCollection.h>
#include <EVENT/TrackerData.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/TrackerHitImpl.h>
#include <IMPL/TrackerDataImpl.h>
#include <IMPL/LCRelationImpl.h>
#include "UTIL/LCTOOLS.h"
#include "UTIL/LCRelationNavigator.h"

// Aida
#ifdef MARLIN_USE_AIDA
#include <AIDA/AIDA.h>
#include <marlin/AIDAProcessor.h>
#include <AIDA/IPlotter.h>
#include <AIDA/IPlotterFactory.h>
#include <RAIDA/RAIDAUtil.h>
#include <RAIDA/IBrushStyleROOT.h>
#include <RAIDA/IPlotterFactoryROOT.h>
#include <AIDA/IMarkerStyle.h>


//marlin
#include <marlin/ConditionsProcessor.h>

//Root
#include "TROOT.h"
#include "TCanvas.h"
#include "TFile.h"
#include "TH2F.h"
#include "THStack.h"
#include "TLine.h"

#endif

using std::string;
using namespace tpcconddata;
using namespace lcio ;
using namespace marlin ;

namespace marlintpc{


    TimePixHitCenterCalculatorProcessor aTimePixHitCenterCalculatorProcessor ;
    

    TimePixHitCenterCalculatorProcessor::TimePixHitCenterCalculatorProcessor() : 
      Processor("TimePixHitCenterCalculatorProcessor") {
	/* the constructor call first the super constructor which the name 
	 * of the processor
	 * this must by the same as the class name.
	 */
	
	/* description of the processor which will be displayed in the
	 * steering file autometicly generated by 
	 * marlin (use MarlinTPC -x (XML-format) or MarlinTPXC -l (text format))
	 */
      _description = "TimePixHitCenterCalculatorProcessor calculates the center of gravity of each cluster" ;
          	
	
	/* register steering parameters: name, description, class-variable, default value
	 * the type will by definded by the type of the default value
	 * string, double, float and int are possible
	 * plaese note to get a float you must cast in C++: (float) 1.3 
	 */
	registerInputCollection( LCIO::TRACKERHIT,
				 "InputCollectionName" , 
				 "Name of the input collection"  ,
				 _inputColName ,
				 std::string("TimePixSepHitCandidates") ) ;
	
	registerOutputCollection( LCIO::TRACKERHIT,
				  "OutputCollectionName" , 
				  "Name of the output collection"  ,
				  _outputColName ,
				  std::string("TimePixHits") ) ;

	registerInputCollection( LCIO::LCGENERICOBJECT,
				  "ModeMapCollectionName" , 
				  "Name of the mode map collection"  ,
				  _modeMapColName ,
				  std::string("TimePixModeMap") ) ;

	registerInputCollection( LCIO::LCGENERICOBJECT,
				 "ModeMapCorrCollectionName" , 
				 "Name of the collection containing the correction to the mode map"  ,
				 _modeMapCorrName ,
				 std::string("TimePixModeMapCorr") ) ;


	registerProcessorParameter( "nColumn" , 
				    "number of columns of the chip"  ,
				    _nColumn ,
				    (int) 256 ) ;

	registerProcessorParameter( "nRow" , 
				    "number of rows of the chip"  ,
				    _nRow ,
				    (int) 256 ) ;

	registerProcessorParameter( "pitch" , 
				    "pitch of the chip in mm"  ,
				    _pitch ,
				    (double) 0.055 ) ;

	registerProcessorParameter( "Shutter" , 
				    "length of shutter signal in µs"  ,
				    _shutter ,
				    (double) 108. ) ;

	registerProcessorParameter( "ReadoutFrequency" , 
				    "readout frequency of TimePix chip in MHz"  ,
				    _readoutFrequency ,
				    (double) 100. ) ;

	registerProcessorParameter( "DriftVelocity" , 
				    "drift velocity of gas in mm/µs"  ,
				    _driftVelocity ,
				    (double) 26.7  ) ;

	
	registerProcessorParameter( "SetOutputTransient" , 
				    "if not 0 the output collection is set transient"  ,
				    _outputIsTransient,
				    0);
	
	_chipID = 0;
	_timePixMap = 0;
	
    }
    
    
    void TimePixHitCenterCalculatorProcessor::init() { 
	
	// usually a good idea to
	printParameters() ;
	
	_nRun = 0 ;

	_timePixMap = TimePixMapHandlerProcessor::getTimePixMap();
	
    }
    
    void TimePixHitCenterCalculatorProcessor::processRunHeader( LCRunHeader* run) { 
	
  
      run->parameters().setValue( _processorName + "_revision", "$Rev: 1317 $");

      for ( ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++ ) {
	if ( ! i->second->isOptional() || i->second->valueSet() ) {
	  run->parameters().setValue( _processorName + "_" + i->second->name(), i->second->value() );
	}
      }

      _nRun++ ;
	
    } 
    
    void TimePixHitCenterCalculatorProcessor::processEvent( LCEvent * evt ) { 

      streamlog_out(MESSAGE1)<<"Event "<<evt->getEventNumber()<<std::endl;

      streamlog_out(DEBUG2)<<"TimePixHitCenterCalculatorProcessor::processEvent( "
			   <<evt->getEventNumber()<<" ) "<<std::endl;

      // get the input collection from the event
      LCCollection* inputCol = 0;
      try{
	inputCol = evt->getCollection( _inputColName ) ;
      } //end of try {

      catch(DataNotAvailableException){
	// if there is not input collection give a warning to the user
	  streamlog_out(WARNING2)<<"processEvent: input collection "
		 <<_inputColName<<" cannot be found in event "<<evt->getEventNumber()<<std::endl;
	  return;
      }

      // create the output collection with a string defining the TYPE (see lcio.h)
      LCCollectionVec* outputCol = new LCCollectionVec( LCIO::TRACKERHIT ); 
 
      // set the transient flag as definded by the processor parameter
      outputCol->setTransient( (_outputIsTransient != 0) );

      LCCollection* oldRelationsCol = 0;
      try {
	  oldRelationsCol = evt->getCollection( "TimePixRelationSepHit" ) ;
      }
      catch(DataNotAvailableException){
	  return;
      }

      LCRelationNavigator* relationNavigator = 0;
      LCCollectionVec* RelationsCol = 0;

      if ( oldRelationsCol != 0 ) {
	  relationNavigator = new LCRelationNavigator( oldRelationsCol );
      	  RelationsCol = new LCCollectionVec( LCIO::LCRELATION );
      }


      for(int iHit=0; iHit < inputCol->getNumberOfElements() ; iHit++){
	
	TrackerHit* RawHit = 
	  dynamic_cast<TrackerHit*>( inputCol->getElementAt( iHit ) ) ;

       
	if( RawHit != 0 ) {
	    

	    TrackerHitImpl* hitCandidate = new TrackerHitImpl();

	    //get original Cluster to this one
	    EVENT::LCObjectVec RawHitRelatedObjects = relationNavigator->getRelatedFromObjects(RawHit);
	    
	    if ( RawHitRelatedObjects.size() != 0 ) {
		
		LCRelationImpl* relation = new LCRelationImpl( RawHit, hitCandidate );
		RelationsCol->addElement( relation );
	    }

	    //hit position x, y in mm, z in time counts	  
	    double pos[3]; 
	    
	    //weighting of pixels to calculate center of gravity
	    double weighingPixel   = 0; 
	    double weighingCluster = 0;
	    
	    
	    // sum of x,y-coordinates weighted by TOT-value or 1 respectively
	    double weightedSumX = 0;
	    double weightedSumY = 0; 
	    
	    // x,y- position of Hit in pixel
	    double posX = 0;
	    double posY = 0; 
	  
	    //sum of TOT/TIME-values in cluster 
	    float sumTOTinCluster  = 0.;
	    float sumTIMEinCluster = 0.;
	    
	    //number of TOT/TIME-pixel in cluster
	    float noTOTPixel  = 0.;
	    float noTIMEPixel = 0.;
	  
	    //TOT value of cluster extrapolated for all pixel in cluster
	    float  TOTCluster = 0;
	    
	    //mean drift time of the cluster in µs
	    float meanDriftTimeCluster = 0;
	    
	    //number of pixel in cluster
	    int clustersize = 0;
	    
	    std::vector<LCObject*> DataVec = RawHit->getRawHits();
	    std::vector<LCObject*>::iterator j;
	    
	    for ( j = DataVec.begin(); j != DataVec.end(); j++) {
		
		TrackerData* Data = dynamic_cast<TrackerData*>( *j ) ;
	    
		if ( Data != 0 ) {
	      

		    int cellid0 = Data->getCellID0();
		    
		    //pixel position in pixel 
		    double x = cellid0%_nColumn; 
		    double y = static_cast<int>((cellid0-x)/_nColumn); 
		    
		    
		    FloatVec ChargeValVec = Data->getChargeValues();
		    
		    FloatVec::iterator iChargeVal;
		    
		    for ( iChargeVal = ChargeValVec.begin(); iChargeVal != ChargeValVec.end(); iChargeVal++ ) {
			
			int mode = 
			    ((_timePixMap->find( TimePixPixelMode::PixelKey( _chipID, cellid0 ) ))
		       ->second)->getPixelMode();
			
			//count all pixels belonging to this cluster
			clustersize++;
			
			
			if ( mode == TimePixPixelMode::TOT ) {
			    weighingPixel = *iChargeVal; 
			    streamlog_out(DEBUG0)<<"TOT = "<<*iChargeVal<<std::endl;
			    sumTOTinCluster += *iChargeVal; 
			    noTOTPixel++;
			}
			else weighingPixel = 1;
			
			if ( mode == TimePixPixelMode::TIMEPIX )  {
			    sumTIMEinCluster += *iChargeVal;		  
			    noTIMEPixel++;
			}
			
			weightedSumX += x*weighingPixel;
			weightedSumY += y*weighingPixel;
			
			weighingCluster += weighingPixel;
			
			
			cellid0++;
			x++;
			if ( x == _nColumn ) {
			    x = 0;
			    y++;
			}
			
		    } //end of for ( iChargeVal = ChargeValVec.begin(); iChargeVal != ChargeValVec.end();  
		    
		    
		    hitCandidate->rawHits().push_back( Data );
		    
		    ChargeValVec.clear();
		    
		} //end of if ( Data != 0 ) {
		
	    } //end of for ( j = DataVec.begin(); j != DataVec.end(); j++
	    
	    
	    posX = weightedSumX/weighingCluster; 
	    posY = weightedSumY/weighingCluster; 
	    
	    if ( noTIMEPixel > 0 ) meanDriftTimeCluster = 
				       _shutter - (sumTIMEinCluster/noTIMEPixel)/_readoutFrequency; 
	    else meanDriftTimeCluster = 0;
	    
	    
	    TOTCluster = sumTOTinCluster*static_cast<float>(clustersize) / noTOTPixel;
	    
	    streamlog_out(DEBUG1)<<"meanDriftTimeCluster = "
				 <<_shutter<<" - ("<<sumTIMEinCluster<<" / "<<noTIMEPixel<<") / "
				 <<_readoutFrequency<<" = "<<meanDriftTimeCluster<<std::endl;
	    streamlog_out(DEBUG1)<<"TOTCluster = "<<sumTOTinCluster<<" * "<<clustersize<<" / "<<noTOTPixel
				 <<" = "<<TOTCluster<<std::endl;
	    
	    pos[0] = posX * _pitch; //x-position in mm
	    pos[1] = posY * _pitch; //y-position in mm
	    pos[2] = meanDriftTimeCluster*_driftVelocity; //drift length in [µs* mm/µs =  mm] 
	    
	    streamlog_out(DEBUG1)<<"pos[2] = "<<meanDriftTimeCluster<<" * " 
				 <<_driftVelocity<<" = "<<pos[2]<<std::endl;
	    
	    streamlog_out(DEBUG1)<<"Position Hit: "<<pos[0]<<" , "<<pos[1]<<" , "<<pos[2]<<std::endl;
	    
	    hitCandidate->setPosition( pos );
	    hitCandidate->setdEdx( TOTCluster );
	    hitCandidate->setTime( meanDriftTimeCluster*1000 ); //ns
	    outputCol->addElement( hitCandidate  );
	    
	    DataVec.clear();		
	    
	} //end of if ( RawHit != 0 ) 
	
      } //end of for(int iHit=0; iHit< nHit ; iHit++){
      
      
      // save the processor parameters in the first collection stored
      if (isFirstCollection){
	  outputCol->parameters().setValue("InputCollectionName",_inputColName);
	  // TO DO: does this make sence? 
	  outputCol->parameters().setValue("SetOutputTransient",_outputIsTransient);
	  isFirstCollection = false;
      }
      
      evt->addCollection( outputCol, _outputColName );
      
      if ( RelationsCol != 0 )
	  evt->addCollection( RelationsCol, "TimePixRelationHit");
      
    }
    
    
    void TimePixHitCenterCalculatorProcessor::check( LCEvent * evt ) { 
    
      LCCollection* HitCol;

      try{
	  HitCol = evt->getCollection( _outputColName );
      }
      catch(DataNotAvailableException)
      {
	  return;
      }

      std::stringstream sEvt;
      sEvt << evt->getEventNumber();

      std::string histoname = "clustersevt" + sEvt.str();
      std::string histonameSP = "clustersevt" + sEvt.str() + "SP";
      std::string histoname3D = "clustersevt" + sEvt.str() + "3D";

      int n = HitCol->getNumberOfElements();

      AIDAProcessor::histogramFactory(this)->destroy( histoCluster );
      histoCluster = AIDAProcessor::histogramFactory(this)->
	createHistogram2D( histoname.c_str(), histoname.c_str(), 
			   256, 0, 256, 256, 0, 256);

      AIDAProcessor::histogramFactory(this)->destroy( histoSP );
      histoSP = AIDAProcessor::histogramFactory(this)->
	createHistogram2D( histonameSP.c_str(), histonameSP.c_str(), 
			   256, 0, 256, 256, 0, 256);

      AIDAProcessor::histogramFactory(this)->destroy( histo3D );
      histo3D = AIDAProcessor::histogramFactory(this)->
	createHistogram3D( histoname3D.c_str(), histoname3D.c_str(), 
			   256, 0, 14.08, 256, 0, 14.08, 40, -20, 20);

      for( int i = 0; i < n; i++ ) {
	TrackerHit* Hit = dynamic_cast<TrackerHit*>( HitCol->getElementAt( i ) ) ;

	//color functionality not yet implemented in RAIDA!
	//=================================================

//	int color = i%10+1;
//	if ( color == 0 || color == 10 || color == 18 ) color = 46;
//	histoCluster->SetLineColor(color);
//	histoSP->SetMarkerStyle( 10 );
//	histoSP->SetMarkerSize(10);
//	histoSP->SetMarkerColor( color + 1 );

	if ( Hit != 0 ) {

	  int xsp = static_cast<int>(Hit->getPosition()[0]/_pitch + 0.5 ); //+0.5 for correct rounding
	  int ysp = static_cast<int>(Hit->getPosition()[1]/_pitch + 0.5 ); //+0.5 for correct rounding

	  histoSP->fill( xsp, ysp );
	  histo3D->fill( Hit->getPosition()[0], Hit->getPosition()[1], Hit->getPosition()[2] );
	  std::vector<LCObject*> DataVec = Hit->getRawHits();
	  std::vector<LCObject*>::iterator j;
	  for ( j = DataVec.begin(); j != DataVec.end(); j++) {

	    TrackerData* Data = dynamic_cast<TrackerData*>( *j ) ;

	    if ( Data!=0 ) {
	      int cellid0 = Data->getCellID0();
	      int x = cellid0%_nColumn;
	      int y = static_cast<int>((cellid0-x)/_nColumn);

	      FloatVec ChargeVec  = Data->getChargeValues();
	      FloatVec::iterator sv;
	      for ( sv = ChargeVec.begin(); sv != ChargeVec.end(); sv++ ) {

		  int mode = 
		      ((_timePixMap->find( TimePixPixelMode::PixelKey( _chipID, cellid0 ) ))
		       ->second)->getPixelMode();
		  
		  float weigh = 0;
		  if ( mode == TimePixPixelMode::TOT ) weigh = *sv;
		  
		  histoCluster->fill( x, y, weigh );
		  x++;
		  
		  if ( x == _nColumn ) {
		      x = 0;
		      y++;
		  }
		  
	      } //end of for ( sv = ChargeVec.begin(); sv != ChargeVec.end(); sv++ ) {
	      
	      ChargeVec.clear();
	      
	    } //end of if ( Data!=0 ) {


	  } //end of for ( j = DataVec.begin(); j != DataVec.end(); j++) {

	} //end of if ( Hit != 0 ) {


       

      } //end of for( int i = 0; i < n; i++ ) {


    }
    
    
  void TimePixHitCenterCalculatorProcessor::end(){ 

      /*	
      streamlog_out(MESSAGE4) << "TimePixHitCenterCalculatorProcessor::end()  " << name() 
		<< " processed " << evt->getEventNumber() << " events in " << _nRun << " runs "
		<< std::endl ;
      */	
    }




}
