//C++
#include <iostream>
#include <stdexcept>
#include <map>
#include <cassert>
#include <cmath>
#include <complex>

//LCIO
#include <EVENT/TrackerPulse.h>
#include <IMPL/TrackerHitImpl.h>

#include "TopoFinder.h"
#include <streamlog/streamlog.h>
#include <gear/PadRowLayout2D.h>

using namespace std;
using namespace lcio;
using namespace gear;

namespace marlintpc{

TopoFinder::TopoFinder() : _maxEmptyPads(1), _minHitSize(1), _maxTimeSpread(100.) //ns
			   , _vDrift(0.), _negativeZDirection(0), _pulsesByIndex(0)
{} // just an empty constructor...

EVENT::TrackerHit* TopoFinder::findHit(std::multimap<int, TrackerPulse*>::iterator currentPulse)
{
    // Check that the multimap has been initialized correctly
    // As it is a programming bug if this is not the case we don't want to waste
    // calculation time if not in debug mode since this function might be called often:
    // use assert!
    assert(_pulsesByIndex);

    	// instantiate a new hitCandidate and add the first pulse to it
	TrackerHitImpl* hitCandidate = new TrackerHitImpl();
	hitCandidate->rawHits().push_back( (*currentPulse).second );

	// A list to remember which pulses are in the current hit candidate
	// After the hit has been added to the collection, these pulses are
	// deleted from the map so they are not considdered any more
	std::list<std::multimap<int, TrackerPulse*>::iterator> pulsesToErase;
	// Add the current pulse to this list
	pulsesToErase.push_back(currentPulse);

	// look for neighbouring pads to the left and to the right of current pad
	// This loop just implements looking in both directions :
	// searchLeftNeighbours = 0 : search to the right
	// searchLeftNeighbours = 1 : search to the left
	for ( uint searchLeftNeighbours = 0 ; searchLeftNeighbours <2 ; searchLeftNeighbours++)
	{
	    uint nEmptyPads=0;

	    // start search at pad index of currentPulse
	    uint lastCheckedPadIndex = currentPulse->first;

	    // loop neighbouring pads to search for pulses
	    // until we are _maxEmptyPads away from the last pad contributing to this hit
	    while ( nEmptyPads < static_cast<uint>(_maxEmptyPads) )
	    {
		// search for next matching pulse
		uint neighbouringPadIndex;
		
		// determine index of pad neighbouring the one which was already checked
		try
		{
		    if ( searchLeftNeighbours )
			neighbouringPadIndex = _padLayout->getLeftNeighbour( lastCheckedPadIndex );
		    else
			neighbouringPadIndex = _padLayout->getRightNeighbour( lastCheckedPadIndex );
		}
		catch ( std::exception &e )
		{
		    streamlog_out(DEBUG0) << e.what() << std::endl; // can be printed for debug
		    
		    // no neighbouring pad in the pad plane, stop searching
		    break;
		}
		
		// remember that this pad has been checked
		lastCheckedPadIndex = neighbouringPadIndex;
		
		// Increase the number of empty pads
		// This is done before the actual search is performed because it avoids
		// introducing an additional variable to remember whether the search was successful.
		// If a matching pulse was found this counter is resert to 0 anyway.
		// If not the counter had been increased.
		nEmptyPads++;
		
		// check if the pad index has been marked as bad ( dead or noisy)
		// tread is as an empty pad and continue
		// !!!FIXME!!! To be implemented
		// if (neighbouringPadIndex is invalid)
		//    continue;
		
		// loop all pulses on the neighbouring pad:
		
		// The pulsesByIndex map provides the first and the last iterator to a matching index as a pair.
		// So this is a pair of iterators to the entries of the multimap, which contains an int and an
		// pointer to a TrackerPulse.
		// The pair variable is called 'neighbours'
		std::pair<std::multimap<int, TrackerPulse*>::iterator, std::multimap<int, TrackerPulse*>::iterator > neighbours;
		
		for(neighbours = _pulsesByIndex->equal_range(neighbouringPadIndex); 
		    neighbours.first != neighbours.second; neighbours.first++) 
		{
		    // neighbours.first is the iterator to the map entry we want to check
		    // the second member in the map entry is the pointer to the TrackerPulse
		    TrackerPulse* pulseCandidate = neighbours.first->second;
		    
		    // check that the pulse is in the correct time window
		    
//	    if (fabs( pulseCandidate->getTime() - - averageTime )
		    if (fabs( pulseCandidate->getTime() - (*currentPulse).second->getTime() )
			< _maxTimeSpread )
		    {
			// pulse belongs to the hit 
			// add it to the hit candidate and the list of iterators to be deleted
			hitCandidate->rawHits().push_back(pulseCandidate);
			pulsesToErase.push_back( neighbours.first );
			
			// reset the counter for invalid pads (empty, noisy or dead)
			nEmptyPads=0;
			streamlog_out(DEBUG2) << "added new pad: padIndex " << neighbouringPadIndex <<std::endl;
		    }
		    
//	    total_charge += pulse_candidate->getCharge();
//	    weighted_charge += pulse_candidate->getCharge() * pulse_candidate->getTime();
		    
		} // for (neighbours)
		
	    }// while ( nInvalidPads < _maxEmptyPads )

	}// for (searchLeftNeighbours)
	    
	// Search is complete.
	// Remove iterators marked for deletion from map.
	// It does not matter whether the hit was valid or not.
	// If the algoritm is working correctly pulses in an invalid
	// candidate can never become part of a valid one.
	for(std::list<std::multimap<int, TrackerPulse*>::iterator>::iterator iter = pulsesToErase.begin();
	    iter != pulsesToErase.end(); iter++)
	{
	    _pulsesByIndex->erase(*iter);
	}

	// check whether the cluster is large enough
	if ( hitCandidate->rawHits().size() >= static_cast<uint>(_minHitSize) )
	{
	    // hitCandidate is valid
	    
	    // calculate the hit from the cluster
	    // the sums needed for centre of gravity:
	    double sum_xphi_q=0; // sum(x*q) or sum(phi*q)
	    double sumzq=0;      // sum(z*q)
	    double sumq=0;  // sum(q)
	    double xphi=0;     // coordinate along pad row
	    double yr=0;     // coordinate orthogonal to pad row

	    // loop over all pulses in cluster and sum up
	    for (EVENT::LCObjectVec::iterator pulseIter = hitCandidate->rawHits().begin();
		 pulseIter < hitCandidate->rawHits().end(); pulseIter++)
	    {
		// Raw hits is a vector of LCObjects.
		// We have to cast the iterator to TrackerPulse to access the member functions.
		TrackerPulse* pulse=dynamic_cast<TrackerPulse*>(*pulseIter);

		// get pad coordinates of current pulse from GEAR layout
		Vector2D padCenter=_padLayout->getPadCenter( pulse->getCellID0() );

		// get x / phi coordinate from first pad
		switch ( _padLayoutType )
		{
		    case PadRowLayout2D::CARTESIAN:
			xphi = padCenter[0]; // x
			break;
		    case PadRowLayout2D::POLAR:
			xphi = padCenter[1]; // phi
			break;
		};

		// take y / r coordinate from first pad, it's the same for all pads
		if ( pulseIter == hitCandidate->rawHits().begin() )
		{
		    switch (_padLayoutType ) 
		    {
			case PadRowLayout2D::CARTESIAN:
			    yr = padCenter[1]; // y
			    break;
			case PadRowLayout2D::POLAR:
			    yr = padCenter[0]; // r
			    break;
		    };
		}

		// summ correctly if handling pads from the first and thr fourth quadrant 
		// in polar coordinates (wrap around from 2 pi to 0)
		if ( (pulseIter != hitCandidate->rawHits().begin()) // for the first pad there is nothing to compare with
		     && (_padLayoutType == PadRowLayout2D::POLAR) ) // problem only occurs in polar coordinates
		{
		    // there can be thwo cases
		    //1.
		    if ( (sum_xphi_q /sumq < M_PI_2) // pulse up to now is in first quadrant (or in fourth, already mapped to <0)
			 && (xphi > 1.5*M_PI) ) // current pulse in fourth quadrant
			xphi-=2*M_PI; // map to range [-pi:pi] so averaging works

		    // 2.
		    // no else needed; statements are mutually excluse because due to first condition
		    // even if xphi was changed in the first statement
		    if ( (sum_xphi_q /sumq > 1.5* M_PI) // pulse up to now is in fourth quadrant (or in first, already mapped to >2*pi)
			 && (xphi < M_PI_2) ) // current pulse in first quadrant
			xphi+=2*M_PI; // map to range [pi:3*pi] so averaging works
		}

		// add to the sums
		sum_xphi_q += xphi * pulse->getCharge();  // x * q or phi * q
		sumzq += pulse->getTime() * _vDrift / 1000 // t * vDrift = z // 1000 ns/us
		         * pulse->getCharge() ;   // * q
		sumq +=pulse->getCharge() ;     // q

	    }// for (pulse)

	    // calculate 3D point
	    double pos[3];
	    switch (_padLayoutType ) 
	    {
		case PadRowLayout2D::CARTESIAN:
		    // in cartesian coordinates xphi is x, yr is y
		    pos[0]= sum_xphi_q / sumq;
		    pos[1]= yr;
		    break;
		case PadRowLayout2D::POLAR:
		    // in polar coordinates xphi is phi, yr is r
		    double phi = sum_xphi_q / sumq;
		    pos[0]= yr * cos( phi );
		    pos[1]= yr * sin( phi );
		    break;
	    };

	    // sumzq contains the weighted drift distance
	    // calculate z coordinate wrt. origin, which is at the cathode
	    pos[2]= _maxDriftDistance - sumzq / sumq;

	    // the second half TPC look in the negative z direction
	    // (rotated by 180 degress around y axis wrt. first endplate)
	    // multiply x and z with -1 in this case
	    if (_negativeZDirection)
	    {
		pos[0] *= -1.;
		pos[2] *= -1.;
	    }

	    // set position and dEdx of the hit
	    hitCandidate->setPosition(pos);
	    hitCandidate->setdEdx(sumq);

	    // return the valid hit
	    return hitCandidate;
	} // ifMinClusterSize
	else
	{
	    streamlog_out(DEBUG3)<<"rejecting hit, too small"<<endl;
	    delete hitCandidate;
	    return 0;
	}

}

EVENT::TrackerHitVec* TopoFinder::findHitsInNeighbouringRows(EVENT::TrackerHit* hit)
{
    // determine x (phi) edge position of the leftmost and the rightmost pad in the hit
    double leftmostPadEdge=0., rightmostPadEdge=0.; // these are the coordinates we are inteseted in
    int currentRow=0;
    //
    LCObjectVec pulseVec = hit->getRawHits();
    streamlog_out(DEBUG1) << "size of pulseVec is " <<pulseVec.size() << endl;
    try{
	double leftmostPadCenter=0, rightmostPadCenter=0; // these are needed for the comparison
	for (LCObjectVec::iterator it = pulseVec.begin(); 
	    it < pulseVec.end(); it++ )
	{
	    TrackerPulse *pulse = dynamic_cast<TrackerPulse*>(*it);
	    int padIndex=pulse->getCellID0();
	    streamlog_out(DEBUG2) << "testing pad index "<< padIndex<< endl;
	    Vector2D position=_padLayout->getPadCenter(padIndex);
	    double x_phi;
	    switch (_padLayoutType)
	    {
		case PadRowLayout2D::CARTESIAN : // x is used in cartesian coordinates
		    x_phi = position[0];
		    streamlog_out(DEBUG1) << "x_phi is "<<x_phi << std::endl;
		    break;
		case PadRowLayout2D::POLAR : // phi is used in polar coordinates, it's the second cordinate
		    x_phi = position[1]  ;
		    break;
		default :
		    throw gear::NotImplementedException(std::string("Unknown pad layout!"));
	    }

	    if( it == pulseVec.begin() )
	    {
		leftmostPadCenter = x_phi;
		leftmostPadEdge = x_phi - _padLayout->getPadWidth(padIndex)/2.;

		rightmostPadCenter = x_phi;
		rightmostPadEdge = x_phi + _padLayout->getPadWidth(padIndex)/2.;

		currentRow=_padLayout->getRowNumber(padIndex) ;
	    }
	    else
	    {
		if ( x_phi < leftmostPadCenter )
		{
		    leftmostPadCenter = x_phi;
		    leftmostPadEdge = x_phi - _padLayout->getPadWidth(padIndex)/2.;
		}
		if ( x_phi > rightmostPadCenter )
		{
		    rightmostPadCenter = x_phi;
		    rightmostPadEdge = x_phi + _padLayout->getPadWidth(padIndex)/2.;
		}
	    }
	}
    }
    catch (std::bad_cast &e) 
    {
	streamlog_out(ERROR4) << "TopoFinder::findHitsInNeighbouringRows: "
			      << "TrackerHit does not contain TrackerPulse as raw hit: " << e.what() << endl;
	streamlog_out(ERROR4)<< "Continuing without searching for neighbouring hits" <<std::endl;
	return 0;
    }
    streamlog_out(DEBUG3)<<"seed hit at x = "<<hit->getPosition()[0]<<" ; leftmostEdge "<< leftmostPadEdge
			 <<" ; rightmostEdge "<< rightmostPadEdge << std::endl;

    TrackerHitVec *neighbouringHits = 0;

    // now search the neighbouring rows, forward and backward
    for (int forward=0; forward <=1 ; forward++)
    {
	for (int deltaRow=1; deltaRow <= _maxSkipRows ; deltaRow++)
	{
	    //calculate y/r coordinate of the investigated row
	    // take pad 0 in this row, all pads have same y/r
	    Vector2D position;
	    try{
		position=_padLayout->getPadCenter( 
		    _padLayout->getPadIndex (currentRow + (forward?+1:-1)*deltaRow, 0));
	    }
	    catch(std::out_of_range &)
	    {   // The requested row does not exist, stop searching
		break;
	    }
	    double y_r;
	    switch (_padLayoutType)
	    {
		case PadRowLayout2D::CARTESIAN : // y is used in cartesian coordinates
		    y_r = position[1];
		    break;
		case PadRowLayout2D::POLAR : // r is used in polar coordinates, it's the second cordinate
		    y_r = position[0]  ;
		    break;
		default :
		    throw gear::NotImplementedException(std::string("Unknown pad layout!"));
	    }
	    
	    // get the pad indices nearest to 
	    int leftmostIndex, rightmostIndex;
	    switch (_padLayoutType)
	    {
		case PadRowLayout2D::CARTESIAN : // x and y is used in cartesian coordinates
		    leftmostIndex = _padLayout->getNearestPad( leftmostPadEdge , y_r );
		    rightmostIndex = _padLayout->getNearestPad( rightmostPadEdge , y_r );
		    break;
		case PadRowLayout2D::POLAR : // r and phi is used in polar coordinates (in this order)
		    leftmostIndex = _padLayout->getNearestPad( y_r , leftmostPadEdge );
		    rightmostIndex = _padLayout->getNearestPad( y_r , rightmostPadEdge );
		    break;
		default :
		    throw gear::NotImplementedException(std::string("Unknown pad layout!"));
	    }
	    
	    TrackerHitVec *hitsInRange = findHitsInRange(leftmostIndex, rightmostIndex, 
							 (_maxDriftDistance + 
							  (_negativeZDirection?1.:-1.)* hit->getPosition()[2])
							   /_vDrift * 1000.);
	                                                 // drift time in ns, calculated from z in mm
	    if (hitsInRange)
	    {
		// recursively search for neighbouring hits
		for (TrackerHitVec::iterator hitIter = hitsInRange->begin(); hitIter < hitsInRange->end(); hitIter++)
		{
		    streamlog_out(DEBUG2)<<"starting recursion..."<<endl;
		    TrackerHitVec *recursivelyFoundHits=findHitsInNeighbouringRows(*hitIter);
		    addTrackerHitVecs(&neighbouringHits,&recursivelyFoundHits);
		}
	    }

	    addTrackerHitVecs(&neighbouringHits,&hitsInRange);
	    
	} //for deltaRow
    }// for forward
    
    if (neighbouringHits)
	streamlog_out(DEBUG2) << "returning "<<neighbouringHits->size()<< " hits"<< std::endl;
    else
    streamlog_out(DEBUG2)<<"returning 0 hits"<<std::endl;
    return neighbouringHits;
}    

EVENT::TrackerHitVec* TopoFinder::findHitsInRange(int index1, int index2, float time)
{
    streamlog_out(DEBUG3) << "TopoFinder::findHitsInRange("<<index1<<","<<index2<<","<<time<<")"<<endl;
    TrackerHitVec *hitsInRange = new  TrackerHitVec;

    // The first pulse with index1
    std::multimap<int, TrackerPulse*>::iterator firstPulse = _pulsesByIndex->lower_bound(index1);
    // The pulse after the last pulse with index2
    std::multimap<int, TrackerPulse*>::iterator lastPulse = _pulsesByIndex->upper_bound(index2);;

    while(firstPulse != lastPulse)
    {
	streamlog_out(DEBUG2) << "testin pulse with index " << firstPulse->first 
	     <<" , time " <<firstPulse->second->getTime() <<endl;
	if ( std::fabs(firstPulse->second->getTime() - time) < _maxTimeSpread )
	{
	    TrackerHit *hit = findHit(firstPulse);
	    if (hit)
		hitsInRange->push_back(hit);

	    // get new values for the iterators, they might have been deleted by findHit
	    firstPulse = _pulsesByIndex->lower_bound(index1);
	    lastPulse = _pulsesByIndex->upper_bound(index2);
	}
	else
	    firstPulse++;
    }

    if (hitsInRange->size() >0)
    {
	streamlog_out(DEBUG3) << "Found "<< hitsInRange->size()<<" hits in range"<< endl;
	return hitsInRange;
    }
    else
    {
	delete hitsInRange;
	return 0;
    }
}

void TopoFinder::addTrackerHitVecs(EVENT::TrackerHitVec **vectorToAddTo, EVENT::TrackerHitVec **vectorToBeAdded)
{
    if ((*vectorToBeAdded) == 0)
	return;

    if ((*vectorToAddTo) == 0)
    {
	*vectorToAddTo = *vectorToBeAdded;
	return;
    }

    for (TrackerHitVec::iterator it = (*vectorToBeAdded)->begin(); it < (*vectorToBeAdded)->end(); it++)
    {
	(*vectorToAddTo)->push_back(*it);
    }
    delete *vectorToBeAdded;
    *vectorToBeAdded=0;
    
}

std::string TopoFinder::getRevision()
{
    return std::string( "$Rev: 248 $" );
}

} // namespace marlintpc
