#include "TrackFitterBase.h"
#include <gear/GEAR.h>
#include <gear/TPCParameters.h>
#include <gear/PadRowLayout2D.h>
#include <CLHEP/Vector/ThreeVector.h>

#include <streamlog/streamlog.h>

#include "IntersectionCalculator.h"

namespace marlintpc{

    // the static variables
    //static const int BASE = 0;
    const int TrackFitterBase::FitterTypes::LIKELIHOOD; ///< Global likelihood method
    const int TrackFitterBase::FitterTypes::CHISQUARE; ///< \f$\chi^2\f$ without geometry information
    const int TrackFitterBase::FitterTypes::CHISQUAREPADS; ///< \f$\chi^2\f$ unsing pad geometry information
    const int TrackFitterBase::FitterTypes::LINEARREGRESSION; ///< Linear regression for a straight line

    EVENT::DoubleVec  TrackFitterBase::calculateResiduals(EVENT::Track *testTrack,
						      uint testHitNumber,
						      EVENT::Track *referenceTrack)
    {
	// get hit testHitNumber from track
	EVENT::TrackerHitVec const &hitVector = testTrack->getTrackerHits();

	if ( testHitNumber >= hitVector.size() )
	{
	    streamlog_out(ERROR1) << "Cannot calulate residual, not hit number " 
				  << testHitNumber <<" in track."<< std::endl;
	    throw EVENT::DataNotAvailableException("Could not find hit in track.");
	}
	
	EVENT::TrackerHit *hit = hitVector[testHitNumber];

	// now simply call calculateResidual(testHit, trackWithoutTestHit),
	// functionality is the same for this basic implementation

	if (referenceTrack) // residual wrt. referenceTrack has been requested
	    return calculateResiduals( hit, referenceTrack );

	else // return residual wrt. track which contains the hit
	    return calculateResiduals( hit, testTrack );
	
    }

    EVENT::DoubleVec TrackFitterBase::calculateResiduals(EVENT::TrackerHit *testHit,
				     EVENT::Track *trackWithoutTestHit)
    {
	// just for convenience: alias for trackWithoutTestHit
	EVENT::Track *track = trackWithoutTestHit;

	// Description of the algorithm:
	// - Calculate the centre of the pad row ( y or r ). Do this by determining the
	//   nearest pad and getting its centre.
	// - Calculate the intersection of the track (helix or straight line) with
	//   the pad row centre (straight line or circle)
	// - Calculate distance of measured coordinate and intersection point along the pad row (x or r*pi)

	// As the implementations have different syntax for polar and cartesian coordinates (or other pad layouts to come)
	// the code is separated in two blocks rather than introducing a switch statement every time the layout differ.

	// But first calculate some track parameters from track:

	// unit vector from reference point to point of closest approach (pca)
	CLHEP::Hep3Vector n_pca( -sin( track->getPhi() ), cos ( track->getPhi() ) );

	// reference point in xy projection
	CLHEP::Hep3Vector reference_point( track->getReferencePoint ()[0], 
					   track->getReferencePoint ()[1] );

	// point of closest approach
	CLHEP::Hep3Vector pca =  reference_point +  track->getD0() *  n_pca ;

	// centre of the helix projection
	// (n_pca is unit vercot pointing outwards)
	CLHEP::Hep3Vector helix_centre; // is initialised with 0., 0. by default
	// can only be calculated if curcature != 0 
	// leave it at the origin otherwise
	if (  track->getOmega() != 0. ) 
	    helix_centre = pca - 1./ track->getOmega() * n_pca ;
	streamlog_out(DEBUG2) <<"helix_centre" << helix_centre << std::endl;
	
	/// A Hep3Vector for the test hit
	CLHEP::Hep3Vector hit(testHit->getPosition()[0],testHit->getPosition()[1] );
	
	// Calculate pca wrt the hit.
	CLHEP::Hep3Vector pcaWrtHit;
	// the tangent at the pcaWrtHit (unit vector)
	CLHEP::Hep3Vector tangent_pcaWrtHit;
	if (  track->getOmega() != 0. ) 
	{
	    CLHEP::Hep3Vector n_pcaWrtHit =  (helix_centre - hit).unit();
	    streamlog_out(DEBUG2) << "n_pcaWrtHit" << n_pcaWrtHit << "\t mag:" 
				  <<  (helix_centre - hit).mag() << "\t 1/omega"
				  << 1./ fabs( track->getOmega() )<< std::endl;
	    pcaWrtHit = helix_centre - n_pcaWrtHit / fabs( track->getOmega() );
	    streamlog_out(DEBUG2) << "pcaWrtHit " << pcaWrtHit << std::endl;

	    // this is n_pcaWrtHit (cross) e_z  * sign of the curvature
	    // ( the cross product always gives clockwise rotation )
	    tangent_pcaWrtHit = CLHEP::Hep3Vector( -n_pcaWrtHit[1], n_pcaWrtHit[0] ) 
		                    * (track->getOmega()<0 ? -1 : 1) ;
	}
	else
	{
	    // project vector from hit to pca (wrt ref point) onto n_pca
	    // add this vector to hit to get the pca wrt. the hit
	    pcaWrtHit = hit + n_pca*( n_pca.dot( hit - pca ) ); 

	    // the tangent is the direction of the track itself (it's a straight line, omega=0 )
	    tangent_pcaWrtHit = CLHEP::Hep3Vector( cos( track->getPhi() ) , sin( track->getPhi() ) );
	}

	// the z-component of the cross  product of (pcaWrtHit - hit) with tangent_pcaWrtHit,
	// which is a unit vector,
	// gives the magnitude of (pcaWrtHit - hit) since the vectors are perpendicular,
	// including the correct sign
	CLHEP::Hep3Vector pcaWrtHit_minus_hit = (pcaWrtHit - hit);
	double residualInReadoutPlane = pcaWrtHit_minus_hit[0]*tangent_pcaWrtHit[1]
	                                - pcaWrtHit_minus_hit[1]*tangent_pcaWrtHit[0];
	streamlog_out(DEBUG4) << "residual xy is " <<  residualInReadoutPlane 
			      << "(pcaWrtHit - hit).mag() is "<< (pcaWrtHit - hit).mag() << std::endl;

	double s ; // track length on trajectory from pca to reference point

 	if (  track->getOmega() != 0. ) 
	{
	    double angle_pcaWrtHit = (pcaWrtHit - helix_centre).phi();
	    double angle_pca = (pca - helix_centre).phi();
	    s = (angle_pcaWrtHit - angle_pca)/track->getOmega() ; // use "signed radius" to get correct orientation
	}
	else
	{
	    s = (pcaWrtHit - pca).mag();
	}

	double z_track = s * track->getTanLambda();
	streamlog_out(DEBUG2) << "z_track + z0 = " <<  z_track + track->getZ0()  << std::endl;

	double residualZ = testHit->getPosition()[2]  - z_track - track->getZ0();
	streamlog_out(DEBUG2) << "residualZ = " <<  residualZ  << std::endl;

	// s is only known mod(2 pi R), so it might be on the wrong loop of the curler
	// maximum distance in z to the next loop: pi*r*tan(lanbda)
 	if (  track->getOmega() != 0. ) 
	{
	    // first wrap to [- 2pi r tan(lambda) : 2pi r tan(lambda)]  (!not to  pi*r*tan(lanbda) !)
	    residualZ = fmod(residualZ , 2*M_PI*track->getTanLambda()/fabs( track->getOmega())  ) ;

	    // now check for the correct distance to the next loop
	    if ( fabs(residualZ) >  M_PI*track->getTanLambda()/fabs( track->getOmega())  )
		residualZ = fmod(2*M_PI*track->getTanLambda()/fabs( track->getOmega()) - residualZ,  M_PI*track->getTanLambda()/fabs( track->getOmega()) );
	}

	EVENT::DoubleVec retVal;
	retVal.push_back(residualInReadoutPlane);
	retVal.push_back( residualZ );

	return retVal;
	
    }// calculateResidual( testHit, trackWithoutTestHit )

    std::string TrackFitterBase::getRevision()
    {
	return std::string( "$Rev: 778 $" );
    }

    // calculate angular distance on a circle, max angle can be pi
    double TrackFitterBase::distanceOnCircle(double phi1, double phi2)
	{
	    double deltaPhi=std::fmod( std::fabs(phi1-phi2) , 2*M_PI); 
	    if ( deltaPhi > M_PI )
		deltaPhi = 2 * M_PI - deltaPhi;
	    
	    return deltaPhi;
	}
    
}//namespace marlintpc
