
// STL
#include <vector>
#include <map>
#include <algorithm>
#include <iomanip>
#include <cassert>

// LCIO
#include "EVENT/LCIO.h"
#include "EVENT/LCCollection.h"
#include "EVENT/TrackerHit.h"
#include "IMPL/LCCollectionVec.h"
#include "IMPL/TrackImpl.h"
#include <IMPL/LCFlagImpl.h>

// GEAR
#include "gear/GEAR.h"
#include "gear/TPCParameters.h"
#include "gear/PadRowLayout2D.h"
#include "gearxml/GearXML.h"

// CLHEP
#include "CLHEP/Matrix/Matrix.h"
#include "CLHEP/Matrix/Vector.h"
#include "CLHEP/Vector/LorentzVector.h"
#include "CLHEP/Vector/TwoVector.h"

// tpcconddata
#include "TPCConditions.h"

// AIDA
#include <AIDA/AIDA.h>

// Marlin
#include "marlin/Global.h"
#include <marlin/AIDAProcessor.h>
#include "TrackSeederProcessor.h"

using namespace marlin;
using namespace lcio;

namespace marlintpc
{

TrackSeederProcessor aTrackSeederProcessor;

TrackSeederProcessor::TrackSeederProcessor() : Processor("TrackSeederProcessor")
{

	_description = "Calculates seed track parameters from  in the TrackerHits in the track candidates collection!";

	registerInputCollection(    LCIO::TRACK,
				    "InputTrackCandidates",
				    "The the name of the input collection of track candidates (default: TPCTrackCandidates)",
				    _input_track_candidates_collection_name,
				    std::string("TPCTrackCandidates"));

	registerOutputCollection(    LCIO::TRACK,
				    "OutputSeedTracks",
				    "The name of the collection the seed tracks be stored under (default: TPCSeedTracks)",
				    _output_seed_tracks_collection_name,
				    std::string("TPCSeedTracks"));
	registerProcessorParameter( "SetOutputTransient" , 
				    "if not 0 the output hits collection is set transient (default: 0)"  ,
				    _outputIsTransient,
				    int(0));

	registerOptionalParameter( "ProduceSZTuples",
							   "Controls whether tuples are produced for this processsor. Currently tuples are produced for the (s,z) location of a single track.",
							   _produce_s_z_tuples,
							   false );

	registerOptionalParameter( "TPCConditionsCollection" ,
								"Name of the collection containing the TPC conditions",
								_tpc_conditions_collection_name,
								tpcconddata::TPCConditions::getDefaultColName() );
}


TrackSeederProcessor::~TrackSeederProcessor() {

}


void TrackSeederProcessor::init() {

  // FIXME!!! This if statement will become obsolete / should be removed 
  //          once the new logging mechanism is implemented in printParameters()
    if( streamlog::out.write< streamlog::MESSAGE3 >() )
	printParameters();

  // get the pad layout from gear
//  try
//  {
//    _padLayoutType = Global::GEAR->getTPCParameters().getPadLayout().getPadLayoutType();
//    switch ( _padLayoutType ) {
//    case PadRowLayout2D::CARTESIAN:
//      _padLayout = &dynamic_cast<const RectangularPadRowLayout&>( Global::GEAR->getTPCParameters().getPadLayout() );
//      break;
//    case PadRowLayout2D::POLAR:
//      _padLayout = &dynamic_cast<const FixedPadSizeDiskLayout&>( Global::GEAR->getTPCParameters().getPadLayout() );
//      break;
//    default:
//	throw gear::Exception( "Unknown pad layout type" );
//    };
////    _maxDriftDistance = Global::GEAR->getTPCParameters().getMaxDriftLength ();
//  }
//  catch(std::exception& e)
//  {
//      streamlog_out(ERROR)<< e.what() << std::endl;
//    throw e;
//  }
//
}


void TrackSeederProcessor::processRunHeader(LCRunHeader *run) {

    run->parameters().setValue( _processorName + "_revision", "$Rev: 1259 $");

    for ( ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++ ) {
        if ( ! i->second->isOptional() || i->second->valueSet() ) {
            run->parameters().setValue( _processorName + "_" + i->second->name(), i->second->value() );
        }
    }

}

void TrackSeederProcessor::processEvent(LCEvent *evt) {

    streamlog_out(DEBUG4) << "TrackSeederProcessor"<<std::endl;
    
    IMPL::LCCollectionVec* trackCandidates=0;
    try {
	// the tracker hit collection
	trackCandidates = dynamic_cast<IMPL::LCCollectionVec*>(evt->getCollection(_input_track_candidates_collection_name));
    }
    catch (DataNotAvailableException &) {
	streamlog_out(MESSAGE3) << "No seed tracks collection \""<<_input_track_candidates_collection_name
				<<"\" in event "<< evt->getEventNumber()<<std::endl;
	return;
    }
    catch (std::bad_cast)
    {
	streamlog_out(WARNING1) << "Could not convert collection \""<<_input_track_candidates_collection_name
				<<"\" to LCCollectionVec, skipping event "<< evt->getEventNumber()<<std::endl;
	return;
    }
    
//		const double innermost_electronics_z = tpc_parameters.getMaxDriftLength();

    // a collection to hold the new track
    IMPL::LCCollectionVec* seedTrackCollection = new IMPL::LCCollectionVec(EVENT::LCIO::TRACK);

    LCFlagImpl trkFlag(0) ;
    trkFlag.setBit( LCIO::TRBIT_HITS ) ;
    seedTrackCollection->setFlag( trkFlag.getFlag() );

    //loop all tracks
    for (IMPL::LCCollectionVec::iterator trackIter=trackCandidates->begin(); 
	 trackIter < trackCandidates->end() ; trackIter++)
    {
	EVENT::Track *track = 0;
	try{
	    track = dynamic_cast<EVENT::Track *>(*trackIter);
	}
	catch(std::bad_cast &)
	{
	    streamlog_out(ERROR3) << "bad_cast: Track candidates collection"
				  << _input_track_candidates_collection_name
				  << " does not contain EVENT::Tracks!"<< std::endl;
            throw DataNotAvailableException("No track candidates found.");
	}
	
	// The TrackerHitVec is a vector of pointers to TrackerHits:  std::vector< TrackerHit * >
	// trackerHits is a copy (!) of the TrackerHitVec in the track, so we can resort the pointers
	// using the helper classes defined in the header file
	EVENT::TrackerHitVec trackerHits = track->getTrackerHits();

	// check whether there are at least 4 hits stored in the track, otherwise continue with next track
	if (trackerHits.size() < 4)
	{
	    streamlog_out(WARNING2) << "Less than 4 hits in track candidates. "
				    << "Check your settings of the track finder!"<< std::endl;
	    continue;
	}

	// the actual track estimate
	IMPL::TrackImpl* seedTrack = new IMPL::TrackImpl();

	// fill all hits into the seed track:
	for (EVENT::TrackerHitVec::iterator hit = trackerHits.begin();
	     hit < trackerHits.end(); hit++)
	{
	    seedTrack->addHit(*hit);
	}

	// set the origin as the reference point
	const CLHEP::Hep3Vector reference_point(0., 0., 0.);
	{
	    float refPoint[3];
	    refPoint[0] = reference_point.x();
	    refPoint[1] = reference_point.y();
	    refPoint[2] = reference_point.z();
	    seedTrack->setReferencePoint(refPoint);
	}

	// ok, up to now there only was getting the data from lcio
	// now the real work starts, calculations of the parameters
	
	// step one: calculate omega, phi and d_0 and the coordinates in s-z assuming the track is moving clockwise
	// it is Jasons code ;-)

	// estimate the curvature using least squares
	std::vector< std::pair<double,double> > xy_points;
	for( EVENT::TrackerHitVec::iterator hit = trackerHits.begin(); hit < trackerHits.end(); hit++ ) {

		double const *hit_position = (*hit)->getPosition();
		xy_points.push_back( std::pair<double,double>(hit_position[0], hit_position[1]) );
	}
// 	xy_points.push_back( std::pair<double,double>( -1, 2 ) );
// 	xy_points.push_back( std::pair<double,double>( -2.,1.) );
// 	xy_points.push_back( std::pair<double,double>( -1., 0.) );

	// least squares results: circle_fit[0] = helix_centre_x, circle_fit[1] = helix_centre_y, circle_fit[2] = radius
	const std::vector<double> circle_fit = calculate_least_squares_circle( xy_points );

	double estimated_omega = 0.;
	double estimated_phi = 0.;
	double estimated_d0 = 0.;
	std::vector< std::pair<double, double> > s_z_pairs;

	CLHEP::Hep3Vector point_of_closest_approach;

	// check to see how large the radius is... if it's bigger than 1E03 meters assume it's a straight line
	if(circle_fit[2] > 1E08) {

		streamlog_out(DEBUG2) << "curvature is very small, assuming straight line" << std::endl;

		// estimate Omega
		estimated_omega = 0.;

		// estimate phi (could this line be any more confusing? ;)
		estimated_phi = atan2(xy_points.back().second - xy_points.front().second, xy_points.back().first - xy_points.front().first);

		// estimate d0
		const double m = tan( estimated_phi );
		const double b = xy_points.front().second - (m * xy_points.front().first);
		const double x0 = (reference_point[0] - m*(b - reference_point[1])) / (m*m + 1);
		const double y0 = m*x0 + b;

		point_of_closest_approach = CLHEP::Hep3Vector( x0, y0, 0 );
		CLHEP::Hep3Vector n_pca( -1 * sin(estimated_phi), cos(estimated_phi) );

		estimated_d0 = n_pca.dot(point_of_closest_approach);

		streamlog_out(DEBUG3) << "pca is " << point_of_closest_approach  << std::endl;

		const CLHEP::Hep3Vector n_phi(cos(estimated_phi),sin(estimated_phi));

		for( EVENT::TrackerHitVec::iterator hitIter = trackerHits.begin(); hitIter < trackerHits.end(); hitIter++ ) {

			double const *hit_position = (*hitIter)->getPosition();
			const CLHEP::Hep3Vector hit(hit_position[0], hit_position[1], hit_position[2]);
			const CLHEP::Hep3Vector pca_to_hit = hit - point_of_closest_approach;

			const double s = pca_to_hit.dot(n_phi); // the distance in the xy plane
			s_z_pairs.push_back( std::pair<double,double>(s, hit_position[2]) );
		}

	}
	else {

	    // calculate the 1 / radius
	    const double estimated_curvature = 1. / circle_fit[2];

	    streamlog_out(DEBUG2)<< "estimating curvature as " << estimated_curvature << "mm^-1 (radius = " << 1 / estimated_curvature << "mm)"<< std::endl;

	    // calculate the centre (in x,y - relative to global (0,0)
	    const CLHEP::Hep3Vector helix_centre( circle_fit[0], circle_fit[1], 0 );

	    streamlog_out(DEBUG2) << "estimating helix centre at " << helix_centre << "mm." <<std::endl;

		// calculate the p.c.a.
		point_of_closest_approach = helix_centre + (reference_point - helix_centre).unit() * circle_fit[2];

	    streamlog_out(DEBUG2) << "calculated the point of closest approach at " << point_of_closest_approach << std::endl;

		// estimate the direction of the track
		// find the closest hit and assume it is moving away from the origin
		bool track_moving_clockwise;
		std::stable_sort( trackerHits.begin(), trackerHits.end(), sort_by_distance_from_point(0., 0., 0.) );
		double const *closest_position = (*trackerHits.begin())->getPosition();
		const CLHEP::Hep3Vector closest_point(closest_position[0], closest_position[1], 0);
		const CLHEP::Hep3Vector n_helix_at_closest_point = (helix_centre - closest_point).unit();
		const CLHEP::Hep3Vector z_axis = CLHEP::Hep3Vector(0,0,1);

		CLHEP::Hep3Vector n_phi1 = n_helix_at_closest_point;
		n_phi1.rotate(M_PI/2, z_axis);
		CLHEP::Hep3Vector n_phi2 = n_helix_at_closest_point;
		n_phi2.rotate(-1.*M_PI/2, z_axis);
		
		streamlog_out(DEBUG2) << "closest point: " << closest_point << ", n_phi1: " << n_phi1 << ", n_phi2: " << n_phi2 << std::endl;

		if( n_phi1.dot(closest_point) > n_phi2.dot(closest_point) ) {
			
			estimated_phi = (point_of_closest_approach - helix_centre).rotate(-1*M_PI/2, z_axis).phi();
			track_moving_clockwise = true;
			estimated_omega = estimated_curvature;
		}
		else {
			
			estimated_phi = (point_of_closest_approach - helix_centre).rotate(M_PI/2, z_axis).phi();
			track_moving_clockwise = false;
			estimated_omega = -1 * estimated_curvature;
		}

		// record the phi/z for each hit in the collection
		// phi will be the angular displacement of the hit from the p.c.a.
		// z is the z-position
		std::vector< std::pair<double, double> > phi_z_pairs;

		const double centre_to_pca_phi = (point_of_closest_approach - helix_centre).phi();
		for( EVENT::TrackerHitVec::iterator hitIter = trackerHits.begin(); hitIter < trackerHits.end(); hitIter++ ) {

			double const *hit_position = (*hitIter)->getPosition();
			const CLHEP::Hep3Vector hit(hit_position[0], hit_position[1], hit_position[2]);
			const double centre_to_hit_phi = (hit - helix_centre).phi();

			const double phi = (track_moving_clockwise)? centre_to_pca_phi - centre_to_hit_phi: centre_to_hit_phi - centre_to_pca_phi;
			phi_z_pairs.push_back( std::pair<double,double>(phi, hit_position[2]) );
		}

		// FIXME analyze the phi-z points and look for curlers, etc

		for( std::vector< std::pair<double, double> >::const_iterator phi_z = phi_z_pairs.begin(); phi_z != phi_z_pairs.end(); phi_z++ ) {

			const double s = fmod((*phi_z).first + 2*M_PI, 2*M_PI) / estimated_curvature;
			s_z_pairs.push_back( std::pair<double,double>(s, (*phi_z).second ) );
		}

		// estimate d0
		const CLHEP::Hep3Vector d = (point_of_closest_approach - reference_point);
		const CLHEP::Hep3Vector n_pca( -1 * sin( estimated_phi ) , cos( estimated_phi ) );
		estimated_d0 = n_pca.dot( d );
	}
	
	// estimate tan(lambda) and z0
	const std::vector<double> line_fit = calculate_least_squares_line( s_z_pairs );

	const double estimated_z0 = line_fit.at(0);
	const double estimated_tan_lambda = line_fit.at(1);

	// now if tan_lambda does not share the same sign as
	
	// produce tuples of the angular and 'z' position of the hits about the centre of the helix
	if( _produce_s_z_tuples ) {

		AIDA::ITuple *_s_z_tuples = AIDAProcessor::tupleFactory(this)->create("TPC Hit Position", "Path integral and 'z' position of the hits produced for this Track", "double s, double z");
		assert( _s_z_tuples );

		for( std::vector< std::pair<double, double> >::const_iterator s_z = s_z_pairs.begin(); s_z != s_z_pairs.end(); s_z++ ) {
			_s_z_tuples->fill(0, (*s_z).first);
			_s_z_tuples->fill(1, (*s_z).second);
			_s_z_tuples->addRow();
		}

		delete _s_z_tuples;
	}

	// now set the estimated values in the seed track
	seedTrack->setOmega( estimated_omega );
	seedTrack->setPhi( estimated_phi );
	seedTrack->setD0( estimated_d0 );
	seedTrack->setTanLambda( estimated_tan_lambda );
	seedTrack->setZ0( estimated_z0 );

	seedTrackCollection->push_back(seedTrack);

	
	streamlog_out(MESSAGE3) << std::setprecision(5) << "\tSeedTrack phi:" << std::setw(8) << seedTrack->getPhi() << " rads ,\tOmega: " 
				<< std::setw(8) << seedTrack->getOmega() << " mm^-1 ,\td0: " << std::setw(8) << seedTrack->getD0() 
				<< " mm ,\ttan(lambda): " << std::setw(8) << seedTrack->getTanLambda() 
				<< " ,\tz0: " << std::setw(8) << seedTrack->getZ0() << "mm." << std::endl;;

    } // loop tracks

    // now add the collection to the event!
    try {
	
	evt->addCollection(seedTrackCollection, _output_seed_tracks_collection_name);
    }
    catch (EVENT::EventException e) {
	
		streamlog_out(WARNING) << e.what() << "Collection " << _output_seed_tracks_collection_name << " is being deleted." << std::endl;
	
		std::vector<EVENT::LCObject*>::iterator iter = seedTrackCollection->begin();
		
		for(; iter != seedTrackCollection->end(); iter++) {
			
			if(*iter) {
			
			delete (*iter);
			(*iter) = NULL;
			}
		}
		
		seedTrackCollection->clear();
		delete seedTrackCollection;
    }
}// processEvent


std::vector<double> TrackSeederProcessor::calculate_least_squares_circle(std::vector< std::pair<double,double> > points) {
	
	CLHEP::HepMatrix A(3, 3, 0);
	CLHEP::HepVector b(3, 0);

	for( std::vector< std::pair<double, double> >::const_iterator point = points.begin(); point != points.end(); point++ ) {

		const double &x = (*point).first;
		const double x2 = x*x;

		const double &y = (*point).second;
		const double y2 = y*y;

		const double x2_plus_y2 = x2 + y2;

		A[0][0] += x2;
		A[0][1] += x*y;
		A[0][2] += x;

		A[1][0] += x*y;
		A[1][1] += y2;
		A[1][2] += y;

		A[2][0] += x;
		A[2][1] += y;
		A[2][2] += 1;

		b[0] += x*x2_plus_y2;
		b[1] += y*x2_plus_y2;
		b[2] += x2_plus_y2;
	}

	int singular;
	const CLHEP::HepMatrix a = A.inverse(singular) * b;

	streamlog_out(DEBUG) << "A: " << A << "b: " << b << "a: " << a;

	std::vector<double> x(4);
	x[0] = a[0][0] / 2;		// x
	x[1] = a[1][0] / 2;		// y
	x[2] = sqrt(x[0]*x[0] + x[1]*x[1] + a[2][0]);	// r

	return x;
}


std::vector<double> TrackSeederProcessor::calculate_least_squares_line(std::vector< std::pair<double,double> > points) {

	CLHEP::HepMatrix A(2, 2, 0);
	CLHEP::HepVector b(2, 0);
	
	for( std::vector< std::pair<double, double> >::const_iterator point = points.begin(); point != points.end(); point++ ) {
		
		const double &s = (*point).first;
		const double &z = (*point).second;

		A[0][1] += s;
		A[1][1] += s*s;

		b[0] += z;
		b[1] += s*z;
	}
	A[0][0] = points.size();
	A[1][0] = A[0][1];

	int singular;
	const CLHEP::HepMatrix a = A.inverse(singular) * b;
	
	std::vector<double> x(3);
	x[0] = a[0][0];	// z0
	x[1] = a[1][0];	// slope (tanlambda)
	x[2] = 0.;		// sum of residuals (eventually)

	return x;
}


Processor* TrackSeederProcessor::newProcessor() {

	return new TrackSeederProcessor();
}


void TrackSeederProcessor::check(LCEvent *evt) {


}


void TrackSeederProcessor::end() {


}

} // namespace marlintpc
