
// LCIO
#include <EVENT/LCIO.h>
#include <EVENT/LCCollection.h>
#include <IMPL/LCCollectionVec.h>

// CLHEP
#include <CLHEP/Random/RandGauss.h>

#include <cassert>

// gear
#include <gear/GEAR.h>
#include <gear/TPCParameters.h>
#include <gear/PadRowLayout2D.h>

// stl
#include <map>

// Marlin
#include <marlin/Global.h>
#include <marlin/Exceptions.h>

// MarlinTPC
#include "TPCConditions.h"
#include "ElectronCloud.h"
#include "ChargeSignal.h"
#include "ElectronCloudChargeDepositProcessor.h"

using namespace std;
using namespace EVENT;
using namespace marlin;
using namespace CLHEP;
using namespace gear;
using namespace tpcconddata;

namespace marlintpc {

// create one to register it with Marlin
ElectronCloudChargeDepositProcessor anElectronCloudChargeDepositProcessor;


ElectronCloudChargeDepositProcessor::ElectronCloudChargeDepositProcessor():
	Processor("ElectronCloudChargeDepositProcessor"),
	_input_collection_name("TPCElectronClouds"),
	_output_collection_name("TPCChargeSignals"),
	_tpc_conditions_collection_name(TPCConditions::getDefaultColName()) {

	registerInputCollection(
		LCIO::LCGENERICOBJECT,
		"InputTPCElectronClouds",
		"Name of the input electron cloud collection",
		_input_collection_name,
		string("TPCElectronClouds") );

	registerOutputCollection(
		LCIO::LCGENERICOBJECT,
		"OutputTPCChargeSignals",
		"Name of the output charge signal collection",
		_output_collection_name,
		string("TPCChargeSignals") );

	registerOptionalParameter(
		"TPCConditionsName",
		"The name to use for the tpc conditions collection.",
		_tpc_conditions_collection_name,
		TPCConditions::getDefaultColName());
}


ElectronCloudChargeDepositProcessor::~ElectronCloudChargeDepositProcessor() {}


void ElectronCloudChargeDepositProcessor::processRunHeader( LCRunHeader* run) {

	// store the processor parameters in the event
    for ( ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++ ) {

        if ( ! i->second->isOptional() || i->second->valueSet() ) {

            run->parameters().setValue( _processorName + "_" + i->second->name(), i->second->value() );
        }
    }
}


void ElectronCloudChargeDepositProcessor::processEvent( LCEvent * evt ) {

	// grab the collection of electron clouds
	LCCollection *electron_cloud_collection = NULL;
	try {

		electron_cloud_collection = evt->getCollection( _input_collection_name );
	}
	catch (DataNotAvailableException &e) {

		// the collection of electron clouds with that name wasn't in the LCIO event file, skip simulating this event
		throw marlin::SkipEventException( this );
	}

	// grab the drift velocity from the tpc conditions data
	LCCollection *tpc_conditions_collection = NULL;
	try {

		tpc_conditions_collection = evt->getCollection( _tpc_conditions_collection_name );
	}
	catch (DataNotAvailableException &e) {

		throw marlin::SkipEventException( this );
	}
	TPCConditions tpc_conditions( tpc_conditions_collection->getElementAt(0) );

	const double drift_velocity = tpc_conditions.getDriftVelocity();

	streamlog_out( MESSAGE0 ) << "Depositing " << electron_cloud_collection->getNumberOfElements() << " electron clouds on the readout electronics\n";

	// and grab the GEAR description of the TPC
	const gear::PadRowLayout2D& pad_layout = Global::GEAR->getTPCParameters().getPadLayout();

	// and create a collection for the charge signals
	LCCollectionVec *new_charge_signal_collection = new LCCollectionVec( LCIO::LCGENERICOBJECT );
	new_charge_signal_collection->setTransient(true);
	new_charge_signal_collection->setSubset(false);

	// create a map, one charge signal for an individual deposit (ie for each track or each segment of a curler)
	map<ChargeSignalTag, ChargeSignal*> signals_by_tag;

	// now loop through the clouds of electrons depositing energy on the readout electronics
	// FIXME this does not use the new GEAR description of the readout modules as it is not finished yet
	for( int cloud_number = 0; cloud_number < electron_cloud_collection->getNumberOfElements(); cloud_number++ ) {

		streamlog_out( DEBUG3 ) << "Depositing cloud number " << cloud_number << "\n";

		ElectronCloud *current_cloud = dynamic_cast<ElectronCloud*>( electron_cloud_collection->getElementAt( cloud_number ) );
		assert(current_cloud);

		ChargeSignalTag tag;

		tag.cloud_id = current_cloud->get_cloud_id();

		// get the number of electrons in the cloud
		const int number_of_electrons = current_cloud->get_electron_count();

		for( int i = 0; i < number_of_electrons; i++ ) {

			// figure out which coordinate system we're using...
			const double x = current_cloud->get_x_position() + (RandGauss::shoot() * current_cloud->get_transverse_stdev());
			const double y = current_cloud->get_y_position() + (RandGauss::shoot() * current_cloud->get_transverse_stdev());
			const double time = current_cloud->get_time() + (RandGauss::shoot() * current_cloud->get_longitudinal_stdev() / drift_velocity * 1.E03);

			streamlog_out( DEBUG2 ) << "(\t" << x << ",\t" << y << ",\t" << time << ")" << std::endl; //, cloud: " << tag.cloud_id << ", pad: " << tag.pad_id << std::endl;

			double coord_one = x;
			double coord_two = y;

			if( pad_layout.getPadLayoutType() == PadRowLayout2D::POLAR ) {

				coord_one = sqrt( (x*x) + (y*y) );
				coord_two = atan2(y, x);

				// FIXME a bug in gear isn't 'normalizing' the phi angle for getNearestPad
				if( coord_two < 0 ) {

					coord_two += 2.*M_PI;
				}
			}

			// get the nearest index
			tag.pad_id = pad_layout.getNearestPad(coord_one, coord_two);

			// if it's in the layout...
			if( !pad_layout.isInsidePad(coord_one, coord_two, tag.pad_id) ) {

				continue;
			}

			// FIXME this isn't used yet
			tag.module_id = 0;

			// add this electron to the charge signal
			ChargeSignal *&signal = signals_by_tag[tag];

			if( !signal ) {

				signal = new ChargeSignal(tag.cloud_id, tag.pad_id, tag.module_id);

				// and add this new signal to the collection
				new_charge_signal_collection->addElement( signal );
			}

			signal->add_electron( time );
		}

	}

	streamlog_out( DEBUG3 ) << "ECHO DONE" << endl;

	streamlog_out( MESSAGE3 ) << "There are " << signals_by_tag.size() << " signals, " << new_charge_signal_collection->getNumberOfElements() << " are in the collection." << endl;
	
	// now add the collection to the event so other processors may access it
	try {

		// if they have the same name try replacing the input collection (experimental)
		if( this->_output_collection_name == this->_input_collection_name ) {

			evt->removeCollection( this->_input_collection_name );
		}
		evt->addCollection( new_charge_signal_collection, _output_collection_name );
	}
	catch( ... ) {

		throw SkipEventException( this );
	}
	
}


void ElectronCloudChargeDepositProcessor::init() {

	// usually a good idea to
	printParameters();
}


void ElectronCloudChargeDepositProcessor::check( LCEvent * evt ) { 

}


void ElectronCloudChargeDepositProcessor::end() { 

}

}
