
// CLHEP
#include <CLHEP/Random/RandGauss.h>
#include <CLHEP/Matrix/Vector.h>
#include <CLHEP/Vector/ThreeVector.h>

// LCIO
#include <EVENT/LCIO.h>
#include <EVENT/LCCollection.h>
#include <IMPL/LCCollectionVec.h>

// Marlin
#include <marlin/Exceptions.h>

// GSL
#include <gsl/gsl_errno.h>
#include <gsl/gsl_odeiv.h>

// MarlinTPC
#include "GlobalFieldProcessor.h"
#include "TPCConditions.h"
#include "ElectronCloud.h"
#include "ElectronCloudDrifterProcessor.h"

#include <cassert>

using namespace std;
using namespace EVENT;
using namespace IMPL;
using namespace marlin;
using namespace CLHEP;
using namespace tpcconddata;

namespace marlintpc {

// static variables
/** the ODE solving functions from the GSL
*/
gsl_odeiv_step *ElectronCloudDrifterProcessor::stepper = 0;

gsl_odeiv_control *ElectronCloudDrifterProcessor::step_control = 0;

gsl_odeiv_evolve *ElectronCloudDrifterProcessor::evolver = 0;


// create one to register it with Marlin
ElectronCloudDrifterProcessor anElectronCloudDrifterProcessor;


ElectronCloudDrifterProcessor::ElectronCloudDrifterProcessor():
	Processor("ElectronCloudDrifterProcessor"),
	_input_collection_name("TPCElectronClouds"),
	_tpc_conditions_collection_name(TPCConditions::getDefaultColName()),
	_tpc_section_name(""),
	_primary_drift_section(false),
	_omega_tau(10.),
	_drift_velocity(40.),
	_longitudinal_diffusion(200.),
	_transverse_diffusion(50.),
	_tpc_outer_radius(1500.),
	_tpc_inner_radius(300.),
	_section_outer_z(100.),
	_section_inner_z(10.) {

	registerInputCollection(
		LCIO::LCGENERICOBJECT,
		"InputTPCElectronClouds",
		"Name of the input electron cloud collection.",
		_input_collection_name,
		string("TPCElectronClouds") );

	registerOptionalParameter(
		"TPCConditionsName",
		"The name to use for the tpc conditions collection.",
		_tpc_conditions_collection_name,
		TPCConditions::getDefaultColName());

	registerOptionalParameter(
		"TPCSectionName",
		"The name of this section of the TPC.",
		_tpc_section_name,
		string(""));

	registerOptionalParameter(
		"PrimaryDriftSection?",
		"Set as true if this drift section is the main drift section of the TPC.",
		_primary_drift_section,
		false);

	registerProcessorParameter(
		"OmegaTau",
		"The gas parameterization for this section of the TPC",
		_omega_tau,
		10. );

	registerProcessorParameter(
		"DriftVelocity",
		"The parameterized drift velocity in this section of the TPC (mm/us)",
		_drift_velocity,
		40. );

	registerProcessorParameter(
		"LongitudinalDiffusion",
		"The parameterized longitudinal diffusion in this section of the TPC (um/sqrt(cm))",
		_longitudinal_diffusion,
		200. );

	registerProcessorParameter(
		"TransverseDiffusion",
		"The parameterized transverse diffusion in this section of the TPC (um/sqrt(cm))",
		_transverse_diffusion,
		50. );

	registerProcessorParameter(
		"TPCOuterRadius",
		"The outer radius of the TPC (mm)",
		_tpc_outer_radius,
		1500.);

	registerProcessorParameter(
		"TPCInnerRadius",
		"The inner radius of the TPC (mm)",
		_tpc_inner_radius,
		300.);

	registerProcessorParameter(
		"SectionOuterZ",
		"The absolute value of the z coordinate of the outermost position of this section of the TPC (mm)",
		_section_outer_z,
		100.);

	registerProcessorParameter(
		"SectionInnerZ",
		"The absolute value of the z coordinate of the innermost position of this section of the TPC (mm)",
		_section_inner_z,
		10.);

	if( !stepper ) {

		stepper = gsl_odeiv_step_alloc (gsl_odeiv_step_rkf45, 4);
	}

	if( !step_control ) {

		step_control = gsl_odeiv_control_y_new(1.E-03, 0.0);
	}

	if( !evolver ) {

		evolver = gsl_odeiv_evolve_alloc(4);
	}
}


ElectronCloudDrifterProcessor::~ElectronCloudDrifterProcessor() {

	if( evolver ) {

		gsl_odeiv_evolve_free(evolver);
	}

	if( step_control ) {

		gsl_odeiv_control_free(step_control);
	}

	if( stepper ) {

		gsl_odeiv_step_free(stepper);
	}
}


void ElectronCloudDrifterProcessor::processRunHeader( LCRunHeader* run) {

	// store the processor parameters in the event
    for ( ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++ ) {

        if ( ! i->second->isOptional() || i->second->valueSet() ) {

            run->parameters().setValue( _processorName + "_" + i->second->name(), i->second->value() );
        }
    }
}


void ElectronCloudDrifterProcessor::processEvent( LCEvent * evt ) {

	// grab the collection of electron clouds
	LCCollection *electron_cloud_collection = NULL;
	try {

		electron_cloud_collection = evt->getCollection( _input_collection_name );
	}
	catch (DataNotAvailableException &e) {

		// the collection of electron clouds with that name wasn't in the LCIO event file, skip simulating this event
		throw marlin::SkipEventException( this );
	}

	streamlog_out( MESSAGE0 ) << "Drifting " << electron_cloud_collection->getNumberOfElements() << " electron clouds through TPC section " << _tpc_section_name << "\n";
	streamlog_out( DEBUG4 ) << "Gas Parameters; drift velocity: " << _drift_velocity << "mm/us, long diffusion: " << _longitudinal_diffusion << "sqrt(mm), trans diffusion: " << _transverse_diffusion << "sqrt(mm), omega*tau: " << _omega_tau << "\n";

	// if this is the primary drift section of the TPC then add the parameters to the TPC conditions class
	if( _primary_drift_section ) {

		TPCConditions* tpc_conditions = NULL;

		try {
			LCCollection *tpc_conditions_collection = evt->getCollection( _tpc_conditions_collection_name );
			tpc_conditions = dynamic_cast<TPCConditions*>( tpc_conditions_collection->getElementAt(0) );
			assert(tpc_conditions);
		}
		catch (DataNotAvailableException &e) {

			LCCollectionVec *tpc_conditions_collection = new LCCollectionVec( LCIO::LCGENERICOBJECT );
			tpc_conditions_collection->setTransient(false);
			tpc_conditions_collection->setSubset(false);
			tpc_conditions = new TPCConditions();
			tpc_conditions_collection->addElement( tpc_conditions );
			evt->addCollection( tpc_conditions_collection, _tpc_conditions_collection_name );
		}

		tpc_conditions->setDriftVelocity( _drift_velocity );

		// tpcconddata uses sqrt(mm) for the diffusion constants... 1sqrt(mm) = 10^-3.5 um/sqrt(cm)
		tpc_conditions->setLongDiffusionCoef( _longitudinal_diffusion );
		tpc_conditions->setTransDiffusionCoef( _transverse_diffusion );
	}

	// now loop through the energy deposits creating clouds of electrons
	for( int cloud_number = 0; cloud_number < electron_cloud_collection->getNumberOfElements(); cloud_number++ ) {

		ElectronCloud *current_cloud = dynamic_cast<ElectronCloud*>( electron_cloud_collection->getElementAt( cloud_number ) );
		assert(current_cloud);

		// initialize the ode state
		double state[4];
		state[0] = current_cloud->get_x_position();
		state[1] = current_cloud->get_y_position();
		state[2] = current_cloud->get_z_position();
		state[3] = current_cloud->get_time()*1.E-03;	// convert the time to us

		// now transport the cloud to the end
		drift_through_cylinder( state, _tpc_inner_radius, _tpc_outer_radius, _section_inner_z, _section_outer_z, _drift_velocity, _omega_tau );

		const double z_drift_distance = fabs(this->_section_outer_z - current_cloud->get_z_position());

		// now smear the cloud position a bit
		const double sxy = sqrt(z_drift_distance) * this->_transverse_diffusion;
		const double sz = sqrt(z_drift_distance) * this->_longitudinal_diffusion;

		streamlog_out( DEBUG3 ) << "Drifted a cloud from (x,y,z,t): (" << current_cloud->get_x_position() << ", " << current_cloud->get_y_position() << ", " << current_cloud->get_z_position() << ", " << current_cloud->get_time() << ") to (x,y,z,t): (" << state[0] << ", " << state[1] << ", " << state[2] << ", " << state[3]*1.E03 << ")\n";

		current_cloud->set_x_position( state[0] );
		current_cloud->set_y_position( state[1] );
		current_cloud->set_z_position( state[2] );
		current_cloud->set_transverse_stdev(  sqrt( (current_cloud->get_transverse_stdev()*current_cloud->get_transverse_stdev()) + (sxy*sxy) ) );
		current_cloud->set_longitudinal_stdev(  sqrt( (current_cloud->get_longitudinal_stdev()*current_cloud->get_longitudinal_stdev()) + (sz*sz) ) );
		current_cloud->set_time( state[3]*1.E03 );						// convert the time back to ns

		streamlog_out( DEBUG1 ) << "Cloud trans stdev: " << current_cloud->get_transverse_stdev() << "mm, long stddev: " << current_cloud->get_longitudinal_stdev() << "mm\n";
	}
	
}


void ElectronCloudDrifterProcessor::init() {

	// convert the diffusion constants into sqrt(mm)
	// tpcconddata uses sqrt(mm) for the diffusion constants... 1sqrt(mm) = 10^-3.5 um/sqrt(cm)
	const double conversion = pow(10., -3.5);
	_longitudinal_diffusion *= conversion;
	_transverse_diffusion *= conversion;

	// usually a good idea to
	printParameters();
}


void ElectronCloudDrifterProcessor::check( LCEvent * evt ) { 

}


void ElectronCloudDrifterProcessor::end() {


}


int ElectronCloudDrifterProcessor::ion_drift_derivative( double t, const double y[], double f[], void *params ) {

	// put the position in a vector...
	HepVector position(3);
	position[0] = y[0];
	position[1] = y[1];
	position[2] = y[2];

	// calculate B and E (E is constant for now)
	const HepVector b = GlobalFieldProcessor::get_instance().get_global_magnetic_field(position);
	const Hep3Vector B(b[0], b[1], b[2]);

	const Hep3Vector E(0., 0., 1.);

	DriftParameters &drift_parameters = *( (DriftParameters*)(params) );
	
	const double ot = drift_parameters._omega_tau;
	const double ot_2 = ot*ot;

	const double prefix = drift_parameters._drift_velocity / (1. + ot_2);

	const double e_dot_b = E.unit().dot(B.unit());					// unit(E) . unit(B)
	const CLHEP::Hep3Vector e_cross_b = E.unit().cross(B.unit());	// unit(E) X unit(B)

	const CLHEP::Hep3Vector velocity = prefix * ( E.unit() + (ot*e_cross_b) + (ot_2*e_dot_b*B.unit()) );

	f[0] = velocity.x();
	f[1] = velocity.y();
	f[2] = velocity.z();
	f[3] = 1.;

	return GSL_SUCCESS;
}


int ElectronCloudDrifterProcessor::drift_through_cylinder( double (&state)[4], const double min_radius, const double max_radius, const double min_z, const double max_z, const double drift_velocity, const double omega_tau ) {

	DriftParameters drift_parameters;

	drift_parameters._drift_velocity = drift_velocity;
	drift_parameters._omega_tau = omega_tau;

	// reset the solvers
	gsl_odeiv_step_reset(stepper);
	gsl_odeiv_control_init(step_control, 1.E-03, 0., 1., 0.);
	gsl_odeiv_evolve_reset(evolver);

	gsl_odeiv_system system = {ion_drift_derivative, 0, 4, &drift_parameters};

	double step_size = 1.;

	double time = state[3];

	// drift the cloud further towards the electronics
	while( fabs(state[2]) < max_z ) {

		if( gsl_odeiv_evolve_apply(evolver, step_control, stepper, &system, &time, time + 1000., &step_size, state) != GSL_SUCCESS ) {

			break;
		}
	}

	// now move the cloud backwards to hit the end of this section
	double derivative[4];

	ion_drift_derivative( 0., state, derivative, (void*)&drift_parameters );
	const double back_distance = state[2] - max_z;

	// move the cloud back onto the z boundary ... ugh, have to do this one axis at a time
	state[0] -= (back_distance / derivative[2]) * derivative[0];
	state[1] -= (back_distance / derivative[2]) * derivative[1];
	state[2] -= back_distance;
	state[3] -= (back_distance / derivative[2]) * derivative[3];

	return GSL_SUCCESS;
}


}
