#ifndef PrimaryIonisationProcessor_h
#define PrimaryIonisationProcessor_h 1

#include <EVENT/MCParticle.h>
#include "marlin/Processor.h"
#include "lcio.h"
#include <string>
//#include <gsl/gsl_rng.h>
//#include <gsl/gsl_randist.h>

/** \mainpage <a href="http://www.physik.rwth-aachen.de/group/IIIphys/TPC/en/simulation/doc_lcio/html/index.html">TPCGEMSimLCIO</a> (v0.01) 
 *This simulation package is dedicated to detailed studies of a Time 
 *Projection Chamber (TPC) with a Gas Electron Multiplier (GEM) readout.
 * \section Overview
 *It consists of for Modules <br>
 *1. Primary ionisation <br>
 *2. Drifting the electrons including diffusion <br>
 *3. Amplification in a triple GEM structure <br>
 *4. Shaping and binning of charge in ADC counts <br>
 * \section Usage
 * The <a href="./annotated.html">Class List</a> gives an overview of available classes
 * and their functionality.
 * For a detailed description refer to the 
 *<a href="http://www.physik.rwth-aachen.de/group/IIIphys/TPC/en/simulation/UserManual/UserManual_lcio.html">user manual</a>.
 */

/** Processors thats creates primary ionisation out of MCParticles
 */
class PrimaryIonisationProcessor : public marlin::Processor {
  
 public:
  
  virtual Processor*  newProcessor() { return new PrimaryIonisationProcessor ; }
  
  
  PrimaryIonisationProcessor() ;
  virtual  ~PrimaryIonisationProcessor(){ }
  
/** Called at the begin of the job before anything is read.
   */
  virtual void init() ;
  
  /** Called for every run.
   */
  virtual void processRunHeader( lcio::LCRunHeader* run ) ;
  
  /** Called for every event - the working horse.
   */
  virtual void processEvent( lcio::LCEvent * evt ) ; 
  
  
  virtual void check( lcio::LCEvent * evt ) ; 
  
  
  /** Called after data processing for clean up.
   */
  virtual void end() ;
  
  /*! Reads in values from generator particle needed for track, like vertex, energy, momentum etc.*/
  virtual void SetParticleParameters(lcio::MCParticle* mcparticle);
  /*! Calculates parameters for helix like radius, so tracks can be curled if magnetic field is present. */
  virtual void InitializeTrackPoint();
  /*! Determines number of electrons on track segment using parametrisation obtained with HEED. */
  virtual  void CalculateNumberOfElectrons();
  virtual void CalculateNumberOfClusters();
  virtual void ReadClusterInfoFile();
  /*! Calculates end point of track segment with magnetic field. */
  virtual void CalculateTrackPointwithBfield();
  /*! Calculates end point of track segment without magnetic field. */
  virtual void CalculateTrackPointwithoutBfield();
  int GetTotalNumberOfElectrons(){return InTPC;}
  /*! distributes number of electrons randomly along track segment. */
  virtual  void CalculateClusterDistance();
  /*! Calculates energy loss for helix. */
  virtual  void CalculateEnergyLoss();
  /*! Rotation for delta rays. */
  virtual  void RotateToTrack(double x, double y, double z);
  /*! Rotation for delta rays. */
  virtual  void RotateToDelta(double x, double y, double z);
  /*! Multiple scattering for delta rays. */
  virtual  void MultipleScattering(double phi, double theta);
  virtual  void SetNextStartPoint();
  
  


 protected:

  /** Input/output collection names.
   */
  std::string _inputCollectionName;
  std::string _outputCollectionName;

  /*!File with parametrisation of clustering processes. */
  std::string clusterfile;
  int _nRun ;
  int _nEvt ;
 
  float _TPCRadius; 
  float _TPCInnerRadius; 
  float _BField;
  float _distance;
  /*! Switch transient mode on and off, determines if output collection is written to file or not. */
  int is_transient;
  /*! For full simulation of Curler set to 0, otherwise standard 1 means simplified
    treatment of low momentum particles.*/
  int Curler;
  /*! Random generator for landau and gaussian. */
  //  gsl_rng *r; 
  /*! Fitparameters for calculation of # clusters from Parameterisation from HEED. */
  double a0,a1,a2,a3,a4,a5,a6,a7;
  int numberofelectrons;
  double numberofclusters;
  double clusterdistance;
  double electrons[7633][2];
  /*! #e- in one event */
  int InTPC; 
  // coordinates and variables for track
  /*! Start point of track segment. */
  double xa,ya,za; 
  /*! Vertex of particle. */
  double xv,yv,zv; 
  /*! End point of track segment. */
  double xend,yend,zend, tend; 
  /*! e- coordinate on track segment. */
  double x_e,y_e,z_e,t_e; 
  /*! Length of track segment. */
  double dx,dy,dz,dr; 
  /*! Center of helix . */
  double xm,ym; 
  /*! Scale factor for radius in case of dE/dx.*/
  double dp;
  /*! Angle corresponding to track segment on helix. */
  double alpha;
  // particle variables
  /*! Particle impulses. */
  double px,py,pz,pt,p,p_const; 
  double energy;
  double mass;
  double charge;
  int pdg;
  double delta_x,delta_y,delta_z;
} ;

#endif
