
//*****************************************************
// Part of program to simulate a TPC with GEM readout *
//                                                    *
// Author: Astrid Muennich                            *
//                                                    *
//   Astrid Muennich                                  *
//   RWTH Aachen, III. Phys. Inst. B                  * 
//   Office: 28 A 112, Phone: +49/(0)241/8            *
//   Email: muennich@physik.rwth-aachen.de            * 
//                                                    *
//*****************************************************

/*! Programm to create a primary electrons along the particle track.  */

#include "PrimaryIonisationProcessor.h"
#include <iostream>
#include <cmath>
#include "IO/LCWriter.h"
#include "IMPL/LCRunHeaderImpl.h"
#include <EVENT/LCCollection.h>
#include <EVENT/MCParticle.h>
#include "IMPL/SimTrackerHitImpl.h" 
#include "IMPL/LCEventImpl.h" 
#include "IMPL/LCCollectionVec.h"
#include <EVENT/LCParameters.h>
#include <gear/TPCParameters.h>
#include <marlin/Global.h>

#include <gearimpl/RectangularPadRowLayout.h>
#include <gearimpl/FixedPadSizeDiskLayout.h>


#include <CLHEP/Random/RandExponential.h>

using namespace lcio ;
using namespace marlin ;
using namespace std;

#define C 0.299792458 // speed of light / Giga

PrimaryIonisationProcessor aPrimaryIonisationProcessor ;

PrimaryIonisationProcessor::PrimaryIonisationProcessor() : Processor("PrimaryIonisationProcessor") {
  
  _description = "PrimaryIonisationProcessor creates electrons along mcparticle track ..." ;
  
  
  // register steering parameters: name, description, class-variable, default value

  registerInputCollection(    LCIO::MCPARTICLE,
                              "InputCollectionName" , 
			      "Name of the input collection with MCParticles"  ,
			      _inputCollectionName ,
			      string("MCParticle") ) ;

  registerOutputCollection(   LCIO::SIMTRACKERHIT ,
                              "OutputCollectionName" ,
                              "Name of the output collection with primary electrons" ,
                              _outputCollectionName ,
                              string("PrimaryElectrons") ) ;

  registerProcessorParameter( "ClusterInfo" , 
			      "File with Information about #e- in Cluster"  ,
			       clusterfile,
			      std::string("clusterinfofile")) ;

   registerProcessorParameter( "FullCurler" , 
			      "switch on simulation of low momentum particles "  ,
			      Curler,
			      (int) 1) ;

  registerProcessorParameter( "Transient" , 
			      "switch off writing of output collection"  ,
			      is_transient,
			      (int) 0) ;
}


void PrimaryIonisationProcessor::init() { 
  cout<<_description<<endl;

  ReadClusterInfoFile();

  //read TPC parameters from gear
  _BField         =Global::GEAR->getTPCParameters().getDoubleVal("BField") ;
  _TPCInnerRadius =Global::GEAR->getTPCParameters().getDoubleVal("TPCInnerRadius") ;
  _TPCRadius      =Global::GEAR->getTPCParameters().getDoubleVal("TPCRadius") ;
  _distance       =Global::GEAR->getTPCParameters().getMaxDriftLength() ;

  printParameters() ;

  _nRun = 0 ;
  _nEvt = 0 ;
  
  //set and initialize other private members
  //  r = gsl_rng_alloc(gsl_rng_mt19937);
  //parameters for HEED Parametrisation
  // for number of clusters per mm
  a0 = 41.1221      ;
  a1 = -30.6686     ;
  a2 = 22.839       ;
  a3 = -7.92683     ;
  a4 = 1.57031      ;
  a5 = -0.182929    ;
  a6 = 0.0116357    ;
  a7 = -0.000310538 ;
  clusterdistance=0.;
  numberofelectrons=0;
  numberofclusters=0.;
  InTPC=0;
  // coordinates on particle track
  xa=0.,ya=0.,za=0.;
  xv=0.,yv=0.,zv=0.;
  xend=0.,yend=0., zend=0., tend=0.;
  dx=0., dy=0., dz=0.,dr=0.; 
  xm=0., ym=0.; 
  
  dp=1.; 
  alpha=0; 
  px=0., py=0., pz=0.,pt=0.,p=0., p_const=0.;
  energy=0.;
  mass=0.;
  charge=0.;
  pdg=0;
  delta_x=0.,delta_y=0.,delta_z=0.;


}

void PrimaryIonisationProcessor::processRunHeader( LCRunHeader* run) { 
 
  _nRun++ ;
} 

void PrimaryIonisationProcessor::processEvent( LCEvent * evt ) { 
  
  LCCollection* mcParticles = 0;
  int nMCs=1;
  bool fullCurler=false;
  if(Curler)
    fullCurler=true;
  try {
    mcParticles = evt->getCollection( _inputCollectionName  ) ;
  } catch(EventException &exception){
    nMCs=0;
  }
  if(nMCs!=0){
    nMCs = mcParticles->getNumberOfElements() ;
    streamlog_out(DEBUG2) << nMCs << " MCParticles in file! " << endl;
  }
  else
    streamlog_out(DEBUG2) << "No objects in input file!" << endl;
  
  LCCollectionVec* primelVec = new LCCollectionVec( LCIO::SIMTRACKERHIT )  ;
  if(is_transient)
    primelVec->setTransient(true);
  
    if(_nEvt==0)
      {
	// set simulation parameter. These will be needed by following processors
	primelVec->parameters().setValue("TPCRadius" , _TPCRadius ) ;
	primelVec->parameters().setValue("TPCInnerRadius" , _TPCInnerRadius ) ;
	primelVec->parameters().setValue("TPCLength" ,_distance ) ;
	primelVec->parameters().setValue("BField" , _BField ) ;
	primelVec->parameters().setValue("ClusterInfoFile" , clusterfile ) ;
      }
    
    for(int j=0;j<nMCs;j++)
      {
	MCParticle* mcpart = dynamic_cast<MCParticle*> (mcParticles->getElementAt(j) );
	SetParticleParameters(mcpart);
	bool isCurler=false;
	 if(!fullCurler)
	   {
	     if(sqrt(px*px+py*py)<1 && 
		atan(sqrt(px*px+py*py)/(fabs(pz)))*180/M_PI > 88)
	       isCurler=true;
	   }
	 streamlog_out(DEBUG2) << mcpart->getGeneratorStatus() << endl;
	if ((mcpart->getGeneratorStatus()==1) && ((mcpart->getCharge())!=0))
	  {
	    InitializeTrackPoint();
	    InTPC=0;
	    int e_curlercounter=0;
	    CalculateNumberOfClusters();
	    // as long as cluster in TPC continue track
	    while((xend+xv)*(xend+xv)+(yend+yv)*(yend+yv)<=_TPCRadius*_TPCRadius && 
		  fabs(zend+zv)<fabs(_distance) && p>0.0)
	      {
		
		if(energy-mass>0.)
		  {   
		    CalculateClusterDistance();
		   
		    double delta_cut=0;
		    if(_BField==0)
		      {
			CalculateTrackPointwithoutBfield();
			delta_cut=60; // equals range of 100 mum
		      }
		    if(_BField!=0) 
		      {
			CalculateTrackPointwithBfield();
			delta_cut=1200; //equals 100 mum at B=1T
		      }
		    if(dr>10 && dp>1) cout<<"Warning: high dE/dx "<<endl; 
		    CalculateNumberOfElectrons();
		   
		    if(numberofelectrons > delta_cut) //delta electron that moves away from track
		      {
			double delta_range=0;
			double delta_energy=(0.026*numberofelectrons-0.017)*pow(10.,-6.);//in GeV
			//angle to track in most cases 90 degrees
			double delta_angle_track=acos(sqrt(delta_energy/energy));// cos of angle
			double delta_angle_plane=2*M_PI*drand48();
			
			if(_BField==0)
			  delta_range=8.23935e-05*pow(numberofelectrons,1.7293); //mm
			if(_BField!=0)
			  delta_range=(delta_energy/(0.3*_BField))*1000; //in mm
			
			double step=delta_range/(double)numberofelectrons;
			
			double xe=xend+xv;
			double ye=yend+yv;
			double ze=zend+zv;
			
			//direction if track would be in 0,1,0
			double deltax=sin(delta_angle_track)*cos(delta_angle_plane);
			double deltaz=cos(delta_angle_track);
			double deltay=sin(delta_angle_track)*sin(delta_angle_plane);
			//transform to track coordinates   
			RotateToTrack(deltax,deltay,deltaz);
			
			
			for(int enumber=0;enumber<numberofelectrons;enumber++)
			  {
			    
			    //Multiple Scattering
			    if(!(enumber%30) && !(enumber==0))
			      {
				double scatter_phi=2*M_PI*drand48();
				double scatter_theta=1./6*M_PI*drand48();
				// new direction
				MultipleScattering(scatter_phi,scatter_theta);
			      }
			    
			    if(sqrt((xe+delta_x*step)*(xe+delta_x*step)+(ye+delta_y*step)*(ye+delta_y*step))>=_TPCInnerRadius)
			      {
				if(isCurler)
				  {
				    if(e_curlercounter%10==0)
				      {
					SimTrackerHitImpl* primel = new SimTrackerHitImpl ;
					double pos[3]={xe+delta_x*step,ye+delta_y*step,ze+delta_z*step};
					primel->setPosition(pos);
					primel->setTime(tend);
					primel->setMCParticle(mcpart);
					primel->setdEdx(0.000000026);
					primelVec->addElement( primel ) ;
					InTPC++;  // counter for # e-
				      }
				    e_curlercounter++;
				  }
				else
				  {
				    SimTrackerHitImpl* primel = new SimTrackerHitImpl ;
				    double pos[3]={xe+delta_x*step,ye+delta_y*step,ze+delta_z*step};
				    primel->setPosition(pos);
				    primel->setTime(tend);
				    primel->setMCParticle(mcpart);
				    primel->setdEdx(0.000000026);
				    primelVec->addElement( primel ) ;
				    InTPC++;  // counter for # e-
				  }
				
			      }
			    xe+=delta_x*step;
			    ye+=delta_y*step;
			    ze+=delta_z*step;
			  }
		      } //delta e-
		    else
		      {
			for(int enumber=0;enumber<numberofelectrons;enumber++)
			  {
			    
			    if(sqrt((xend+xv)*(xend+xv)+(yend+yv)*(yend+yv))>=_TPCInnerRadius)
			      {
				if(isCurler)
				  {
				    if(e_curlercounter%10==0)
				      {
					SimTrackerHitImpl* primel2 = new SimTrackerHitImpl ;
					double pos[3]={xend+xv,yend+yv,zend+zv};
					primel2->setPosition(pos);
					primel2->setTime(tend);
					primel2->setMCParticle(mcpart);
					primel2->setdEdx(0.000000026);
					primelVec->addElement( primel2 ) ;
					InTPC++;  // counter for # e-
				      }
				    e_curlercounter++;
				  }
				else
				  {
				    SimTrackerHitImpl* primel2 = new SimTrackerHitImpl ;
				    double pos[3]={xend+xv,yend+yv,zend+zv};
				    primel2->setPosition(pos);
				    primel2->setTime(tend);
				    primel2->setMCParticle(mcpart);
				    primel2->setdEdx(0.000000026);
				    primelVec->addElement( primel2 ) ;
				    InTPC++;  // counter for # e-
				  }
			      }
			  }
		      }
		    CalculateEnergyLoss();
		    SetNextStartPoint();
		    CalculateNumberOfClusters();
		  }
		else
		  {//Energy lower than mass -> Particle stopped in TPC
		    break;}
		
	      } //loop over clusters
	    
	  }//if particle ok
      }// particle loop
    
    if (primelVec->empty() && _nEvt!=0) 
      delete primelVec;
    else
      evt->addCollection( primelVec , _outputCollectionName ) ;
      
    if(_nEvt%100==0)
    cout<<"PrimaryIonisationProcessor: Event "<<_nEvt<<" done!"<<endl;
    _nEvt ++ ;
    
}



void PrimaryIonisationProcessor::check( LCEvent * evt ) { 

}


void PrimaryIonisationProcessor::end(){ 
  
  std::cout <<  name() 
 	    << " processed " << _nEvt << " events in " << _nRun << " runs "
	    << std::endl ;
  
}

void PrimaryIonisationProcessor::SetParticleParameters(MCParticle* mcparticle)
{
  const double *v;
  const double *parray;
  energy=mcparticle->getEnergy();
  mass=mcparticle->getMass();
  pdg=mcparticle->getPDG();
  charge=mcparticle->getCharge();
  v=(mcparticle->getVertex());
  xv=v[0]; // Vx() is in mm
  yv=v[1]; 
  zv=v[2];  
  parray=mcparticle->getMomentum();
  px=parray[0];
  py=parray[1];
  pz=parray[2];
  pt=sqrt(px*px+py*py);
  p=sqrt(pt*pt+pz*pz);
  p_const=p;
}

void PrimaryIonisationProcessor::InitializeTrackPoint()
{
  xa=0.; ya=0.; za=0.;
  xend=0.;yend=0.;zend=0.;
}

void PrimaryIonisationProcessor::CalculateTrackPointwithBfield()
{
  // rotation angle
  double factor = charge*_BField*C;
  alpha=factor/p*clusterdistance/1000.; //[1/mm]
  // center of circle
  xm=1000*py*dp/factor; //[mm]
  ym=-1000*px*dp/factor;
  //coordinates on helix
  xend=cos(alpha)*(xa-xm)*dp+sin(alpha)*(ya-ym)*dp+xm;
  yend=-sin(alpha)*(xa-xm)*dp+cos(alpha)*(ya-ym)*dp+ym;
  zend=za+pz/p*clusterdistance;
    
}

void PrimaryIonisationProcessor::CalculateTrackPointwithoutBfield()
{
  //coordinates on straight line along p
  dr=clusterdistance;
  dx=px/p_const;
  dy=py/p_const;
  dz=pz/p_const;
  xend=xa+dr*dx;
  yend=ya+dr*dy;
  zend=za+dr*dz;
}

void PrimaryIonisationProcessor::CalculateNumberOfClusters()
{
  //from parametrisation one knows function (polynom) and constants
  double x=sqrt((energy*energy)/(mass*mass)-1);
  double logx = std::log(x);
  double log10000 = std::log(10000.0);
  if(x<10000)
    numberofclusters = a0+a1*logx+a2*logx*logx+a3*pow(logx,3)
			+a4*pow(logx,4)+a5*pow(logx,5)
			+a6*pow(logx,6)+a7*pow(logx,7);
  else
    numberofclusters = a0+a1*log10000+a2*log10000*log10000+a3*pow(log10000,3)
			+a4*pow(log10000,4)+a5*pow(log10000,5)
			+a6*pow(log10000,6)+a7*pow(log10000,7);
  // number of cluster per cm constant after p/m>10.000
}

void PrimaryIonisationProcessor::CalculateClusterDistance()
{
  //  clusterdistance =10.*(gsl_ran_exponential(r,1./numberofclusters)); 
  clusterdistance =10.*(CLHEP::RandExponential::shoot(1./numberofclusters)); 
  //*10 because distance needed in mm
}

void PrimaryIonisationProcessor::CalculateNumberOfElectrons()
{
  //pick a random value between 0 and 1 and check file if random number is bigger than
  //integrated probability
  double r=0.998177337261218*drand48(); 
  //cryptic factor because of cut off on distribution at n=7632 comming from HEED
  int i=0;
  while(r>electrons[i][1] && i<7632)
    i++;
  numberofelectrons=(int)electrons[i][0];
}

void  PrimaryIonisationProcessor::ReadClusterInfoFile()
{
  int lines=7632;
  FILE* fin = fopen(clusterfile.c_str(),"r");
  if(! fin) {                             
       printf("File with cluster information could not be opened.\n");
       exit(1);                           
   }
   /* read in events */
   int events = 0;                  /* evetn counter */               
   while(!feof(fin)) {              /* until end of file read one line*/    
                                          
      fscanf(fin,"%lf %lf " ,
             &electrons[events][0],
	     &electrons[events][1]);
      events++;                            
      
      if(events >lines) {             /* check for array overflow */
         printf("More events than array range (%i) allows.\n", lines);
         break;                            
      }
   }
   
   printf("%i Events read. \n", events);
   fclose(fin);            /* close file */                
}

void PrimaryIonisationProcessor::CalculateEnergyLoss()
{
  // assume an energy loss of 26 eV for each produced electron in argon
  energy=energy-numberofelectrons*0.000000026; 
  dp=(p-numberofelectrons*0.000000026)/p;
  if(energy-mass<=0)
    streamlog_out(DEBUG3)<<"Particle stopped in TPC "<< endl;
  else
    p=sqrt(energy*energy-mass*mass);
}


void PrimaryIonisationProcessor::RotateToTrack(double x, double y, double z)
{
  double phi;
  if(px>=0)
    phi=atan(py/px);
  else
    phi=atan(py/px)+M_PI;
  if(px==0 && py<0)
    phi=-phi;
  double theta=acos(pz);
  delta_x=cos(phi)*cos(theta)*x-sin(phi)*y+cos(phi)*sin(theta)*z;
  delta_y=sin(phi)*cos(theta)*x+cos(phi)*y+sin(phi)*sin(theta)*z;
  delta_z=-sin(theta)*x+cos(theta)*z;
}

void  PrimaryIonisationProcessor::RotateToDelta(double x, double y, double z)
{
  double phi;
  if(delta_x>0)
    phi=atan(delta_y/delta_x);
  else
    phi=atan(delta_y/delta_x)+M_PI;
  if(delta_x==0 && delta_y<0)
    phi=-phi;
  double theta=acos(delta_z);
  delta_x=cos(phi)*cos(theta)*x-sin(phi)*y+cos(phi)*sin(theta)*z;
  delta_y=sin(phi)*cos(theta)*x+cos(phi)*y+sin(phi)*sin(theta)*z;
  delta_z=-sin(theta)*x+cos(theta)*z;
}

void  PrimaryIonisationProcessor::MultipleScattering(double phi, double theta)
{
  double x_s=sin(theta)*cos(phi);
  double z_s=cos(theta);
  double y_s=sin(theta)*sin(phi);
  RotateToDelta(x_s,y_s,z_s);
}


void PrimaryIonisationProcessor::SetNextStartPoint()
{
  xa=xend;
  ya=yend;
  za=zend;
}

