
// STL
#include <vector>
#include <algorithm>
#include <cmath>
#include <iomanip>
#include <cassert>

// LCIO
#include "EVENT/LCIO.h"
#include "EVENT/LCCollection.h"
#include "EVENT/SimTrackerHit.h"
#include "EVENT/TrackerHit.h"
#include "EVENT/TrackerPulse.h"
#include "IMPL/LCCollectionVec.h"
#include "IMPL/TrackImpl.h"

// GEAR
#include "gear/GEAR.h"
#include "gear/TPCParameters.h"
#include "gear/PadRowLayout2D.h"
#include "gearimpl/Vectors.h"

// CLHEP
#include "CLHEP/Matrix/Matrix.h"
#include "CLHEP/Vector/LorentzVector.h"
#include "CLHEP/Vector/TwoVector.h"
#include "CLHEP/Geometry/Transform3D.h"

// Marlin
#include "marlin/Global.h"
#include "HepRepXMLWriter.h"
#include "HepRepOutputProcessor.h"

using namespace std;

HepRepOutputProcessor aHepRepOutputProcessor;

HepRepOutputProcessor::HepRepOutputProcessor():
	Processor("HepRepOutputProcessor") {
	//_input_tracker_hits_collection_name("TPCHits"),
	//_output_seed_track_collection_name("TPCSeedTracks") {

	_description = "Finds a seed track in the input TrackerHit collection!";

	registerOptionalParameter("SimTrackerHitCollections",
				"A list of SimTrackerHit collections to draw",
				_simtrackerhit_collection_names,
				_simtrackerhit_collection_names);

	registerOptionalParameter("TrackerDataCollections",
				"A list of TrackerPulse collections to draw",
				_trackerdata_collection_names,
				_trackerdata_collection_names);

	registerOptionalParameter("TrackerHitCollections",
				"A list of TrackerHit collections to draw.",
				_trackerhit_collection_names,
				_trackerhit_collection_names);

	registerOptionalParameter("TrackerPulseCollections",
				"A list of TrackerPulse collections to draw",
				_trackerpulse_collection_names,
				std::string(""));
				
	registerOptionalParameter("TrackCollections",
				"A list of Track collections to draw",
				_track_collection_names,
				std::string(""));

	registerOptionalParameter("EventNumbersToDraw",
				"A list of event numbers to display",
				_events_to_display,
				_events_to_display);
}


HepRepOutputProcessor::~HepRepOutputProcessor() {


}


void HepRepOutputProcessor::init() {

	printParameters();

	//const gear::TPCParameters& tpc_parameters = Global::GEAR->getTPCParameters();

	_heprep_writer.open("last_run.heprep");
}


void HepRepOutputProcessor::processRunHeader(LCRunHeader *run) {

	_heprep_writer.addType("Run", 0);

	_heprep_writer.addAttDef("Run Number", "The run number", "Physics", "");
	
	_heprep_writer.addInstance();

	try {

		const gear::TPCParameters& tpc_parameters = Global::GEAR->getTPCParameters();
	
		const double max_drift_length = tpc_parameters.getMaxDriftLength();
	
		// DRAW THE TPC
	
		// try and get some extra parameters from MOKKA_GEAR
		const double tpc_inner_radius = tpc_parameters.getDoubleVal( "TPC_inner_radius" ); // mm
		const double tpc_outer_radius = tpc_parameters.getDoubleVal( "TPC_outer_radius" ); // mm

		
		_heprep_writer.addType( "TPC_OuterCylinder", 0 );
		_heprep_writer.addAttValue( "DrawAs", "Cylinder" );

		_heprep_writer.addInstance();
		_heprep_writer.addAttValue( "Radius", tpc_outer_radius );
		_heprep_writer.addPrimitive();
		_heprep_writer.addPoint( 0., 0., -1 * max_drift_length );
		_heprep_writer.addPoint( 0., 0.,  max_drift_length );

		_heprep_writer.addType( "TPC_InnerCylinder", 0 );
		_heprep_writer.addAttValue( "DrawAs", "Cylinder" );

		_heprep_writer.addInstance();
		_heprep_writer.addAttValue( "Radius", tpc_inner_radius );
		_heprep_writer.addPrimitive();
		_heprep_writer.addPoint( 0., 0., -1 * max_drift_length );
		_heprep_writer.addPoint( 0., 0.,  max_drift_length );
	}
	catch( ... ) {
		
	}

}


void HepRepOutputProcessor::processEvent(LCEvent *evt) {

	if( !_heprep_writer.isOpen ) {
		
		streamlog_out( ERROR ) << "An output file is not open";
		return;
	}

	if( _events_to_display.size() > 0 ) {

		if( find(_events_to_display.begin(), _events_to_display.end(), evt->getEventNumber()) == _events_to_display.end() ) {

			return;
		}
	}

	_heprep_writer.addAttDef( "omega", "signed inverse curvature", "Physics", "" );

	_heprep_writer.addType("Event", 0);
	_heprep_writer.addAttDef( "Event Number", "The event number", "Physics", "" );
	_heprep_writer.addAttValue("LineWidth",2);
	
	_heprep_writer.addInstance();

	_heprep_writer.addAttValue( "Event Number", evt->getEventNumber() );

	const gear::TPCParameters& tpc_parameters = Global::GEAR->getTPCParameters();
	const double max_drift_length = tpc_parameters.getMaxDriftLength();

	try {

		/* DRAW THE HIT COLLECTIONS */

		std::vector< std::string >::const_iterator hit_collection_name = _trackerhit_collection_names.begin();

		_heprep_writer.addType( "TPC_TrackerHit", 0 );
		_heprep_writer.addAttValue( "Color", 1., 0., 0. );
		_heprep_writer.addAttValue( "DrawAs", "Point" );
		_heprep_writer.addAttValue( "MarkName", "Circle" );
		_heprep_writer.addAttValue( "MarkSize", 2. );

		for( ; hit_collection_name != _trackerhit_collection_names.end(); hit_collection_name++ ) {

			EVENT::LCCollection* tracker_hits = evt->getCollection(*hit_collection_name);

			streamlog_out( MESSAGE3 ) << "Drawing TrackerHit collection " << *hit_collection_name << ".\n";

			for( int hit_counter = 0; hit_counter < tracker_hits->getNumberOfElements(); hit_counter++ ) {

				TrackerHit* current_hit = dynamic_cast<TrackerHit*>( tracker_hits->getElementAt(hit_counter) );

				assert(current_hit);
				
				const double* position = current_hit->getPosition();

				_heprep_writer.addInstance();
				_heprep_writer.addPrimitive();
				_heprep_writer.addPoint( position[0], position[1], position[2] );
			}
		}
	}
	catch (DataNotAvailableException &e) {

		streamlog_out( WARNING ) << e.what() << std::endl;
	}
	
	try {

		/* DRAW THE SIMTRACKERHIT COLLECTIONS */

		std::vector< std::string >::const_iterator sim_hit_collection_name = _simtrackerhit_collection_names.begin();

		_heprep_writer.addType( "TPC_SimTrackerHit", 0 );
		_heprep_writer.addAttDef( "dEdx", "Energy Desposit", "Physics", "" );
		_heprep_writer.addAttValue( "Color", 0., 1., 0. );
		_heprep_writer.addAttValue( "DrawAs", "Point" );
		_heprep_writer.addAttValue( "MarkName", "Circle" );
		_heprep_writer.addAttValue( "MarkSize", 2. );

		for( ; sim_hit_collection_name != _simtrackerhit_collection_names.end(); sim_hit_collection_name++ ) {

			EVENT::LCCollection* tracker_hits = evt->getCollection(*sim_hit_collection_name);

			streamlog_out( MESSAGE3 ) << "Drawing SimTrackerHit collection " << *sim_hit_collection_name << ".\n";

			for( int hit_counter = 0; hit_counter < tracker_hits->getNumberOfElements(); hit_counter++ ) {

				SimTrackerHit* current_hit = dynamic_cast<SimTrackerHit*>( tracker_hits->getElementAt(hit_counter) );

				assert(current_hit);

				const double* position = current_hit->getPosition();

				streamlog_out( DEBUG2 ) << "\tHit: " << hit_counter << " (x,y,z): " << position[0] << ", " << position[1] << ", " << position[2] << endl;

				_heprep_writer.addInstance();
				_heprep_writer.addPrimitive();
				_heprep_writer.addPoint( position[0], position[1], position[2] );
			}
		}
	}
	catch (DataNotAvailableException &e) {

		streamlog_out( WARNING ) << e.what() << std::endl;
	}

	try {
	
		/* DRAW THE PULSE COLLECTIONS */

		const gear::PadRowLayout2D& pad_layout = Global::GEAR->getTPCParameters().getPadLayout();

		EVENT::LCCollection* tracker_pulses = evt->getCollection(_trackerpulse_collection_names);

		streamlog_out( MESSAGE3 ) << "Representing " << tracker_pulses->getNumberOfElements() << " " << tracker_pulses->getTypeName() << " from collection " << _trackerpulse_collection_names << ".\n";

		_heprep_writer.addType( "TPC_TrackerPulse", 0 );
		_heprep_writer.addAttDef( "Primary_Electrons", "Primary Electrons", "Physics", "electrons" );
		_heprep_writer.addAttValue( "DrawAs", "Polygon" );
		_heprep_writer.addAttValue( "Fill", "True" );
		_heprep_writer.addAttValue( "FillStyle", "NonZero" );

		if( pad_layout.getPadShape() != pad_layout.RECTANGLE ) {

			throw Exception("HepRepOutputProcessor only supports RECTANGULAR shaped pads at the moment.");
		}

		// determine the largest charge on the pads
		float max_charge = 0.;
		float min_charge = 0.;
		for( int pulse_counter = 0; pulse_counter < tracker_pulses->getNumberOfElements(); pulse_counter++ ) {

			TrackerPulse* current_pulse = dynamic_cast<TrackerPulse*>( tracker_pulses->getElementAt(pulse_counter) );

			assert(current_pulse);

			min_charge = std::min(current_pulse->getCharge(), min_charge);
			max_charge = std::max(current_pulse->getCharge(), max_charge);
		}

		for( int pulse_counter = 0; pulse_counter < tracker_pulses->getNumberOfElements(); pulse_counter++ ) {

			TrackerPulse* current_pulse = dynamic_cast<TrackerPulse*>( tracker_pulses->getElementAt(pulse_counter) );

			assert(current_pulse);
			
			_heprep_writer.addInstance();

			const int pad_index = current_pulse->getCellID0();
			const double pad_charge = current_pulse->getCharge();

			_heprep_writer.addAttValue( "Primary_Electrons", pad_charge );

			if(pad_charge > 0) {

				const double red = pad_charge / max_charge;
				_heprep_writer.addAttValue( "FillColor", red, 1. - red, 1. - red );
			}
			else {

				const double blue = pad_charge / max_charge;
				_heprep_writer.addAttValue( "FillColor", 1. - blue, 1. - blue, blue );
			}
			
			const gear::Vector2D pad_centre = pad_layout.getPadCenter( pad_index );

			const double half_pad_height = pad_layout.getPadHeight( pad_index ) / 2.;
			
			const double half_pad_width = (pad_layout.getPadLayoutType() == pad_layout.CARTESIAN)? (pad_layout.getPadWidth( pad_index ) / 2.): (pad_layout.getPadWidth(pad_index) * pad_centre[0] / 2.);

			HepGeom::Vector3D<double> point_1( -half_pad_width, -half_pad_height, max_drift_length );
			HepGeom::Vector3D<double> point_2( -half_pad_width,  half_pad_height, max_drift_length );
			HepGeom::Vector3D<double> point_3(  half_pad_width,  half_pad_height, max_drift_length );
			HepGeom::Vector3D<double> point_4(  half_pad_width, -half_pad_height, max_drift_length );

			switch( pad_layout.getPadLayoutType() ) {

				case gear::PadRowLayout2D::CARTESIAN: {

					HepGeom::Vector3D<double> translate(pad_centre[0], pad_centre[1], 0);
					point_1 += translate;
					point_2 += translate;
					point_3 += translate;
					point_4 += translate;
				}
				break;

				case gear::PadRowLayout2D::POLAR: {

					//HepGeom::Transform3D transform( HepGeom::Rotate3D(pad_centre[1], HepGeom::Vector3D(0., 0., 1.)), HepGeom::Vector3D(0., pad_centre[0], 0.) );
					const HepGeom::Rotate3D rotate( pad_centre[1] - M_PI/2, HepGeom::Vector3D<double>(0., 0., 1.) );
					const HepGeom::Translate3D translate( 0., pad_centre[0], 0. );
					//HepGeom::Transform3D transform( rotate, translate );
					
					point_1[1] += pad_centre[0];
					point_2[1] += pad_centre[0];
					point_3[1] += pad_centre[0];
					point_4[1] += pad_centre[0];
					
					point_1 = rotate * point_1;
					point_2 = rotate * point_2;
					point_3 = rotate * point_3;
					point_4 = rotate * point_4;
				}
				break;
			}

			// setup the pad as upright at the origin
			// set the points
			_heprep_writer.addPrimitive();
			_heprep_writer.addPoint( point_1[0], point_1[1], point_1[2] );
			_heprep_writer.addPoint( point_2[0], point_2[1], point_2[2] );
			_heprep_writer.addPoint( point_3[0], point_3[1], point_3[2] );
			_heprep_writer.addPoint( point_4[0], point_4[1], point_4[2] );
		}
	}
	catch (DataNotAvailableException &e) {

		streamlog_out( WARNING ) << e.what() << std::endl;
	}

	try {
	
		/* DRAW THE TRACK COLLECTIONS */

		EVENT::LCCollection* tracker_tracks = evt->getCollection(_track_collection_names);

		_heprep_writer.addType( "TPC_Track", 0 );
		_heprep_writer.addAttValue( "DrawAs", "Line" );
		_heprep_writer.addAttValue( "Color", 0., 1., 0. );

		streamlog_out( MESSAGE3 ) << "Drawing collection " << _track_collection_names << ".\n";

		for( int track_counter = 0; track_counter < tracker_tracks->getNumberOfElements(); track_counter++ ) {

			Track* current_track = dynamic_cast<Track*>( tracker_tracks->getElementAt(track_counter) );

			assert(current_track);

			const double d0 = current_track->getD0();
			const double phi0 = current_track->getPhi();
			const double z0 = current_track->getZ0();
			const double omega = current_track->getOmega();
			const double tanLambda = current_track->getTanLambda();

			const CLHEP::Hep3Vector n_pca( -1 * sin(phi0), cos(phi0) );
			const CLHEP::Hep3Vector pca = n_pca * d0;

			_heprep_writer.addInstance();
			_heprep_writer.addPrimitive();
			_heprep_writer.addAttValue( "Curvature", fabs(omega) );
			_heprep_writer.addAttValue( "Phi", phi0 );
			_heprep_writer.addAttValue( "Theta", tan(tanLambda) );

			// draw 5m of track
			if( fabs(omega) >= 0.0000001 ) {
				
				const CLHEP::Hep3Vector helix_centre = -1 * (1./omega - d0) * n_pca;
				const double angular_distance = 5000 * omega;

				for( double phi = 0.; phi <= angular_distance; phi += (M_PI/180) ) {

					const double angle = (omega < 0)? phi0 + phi: phi0 - phi;
					const CLHEP::Hep3Vector n1( -1 * sin(phi+phi0), cos(angle) );
					const CLHEP::Hep3Vector p1 = helix_centre + (1/omega - d0)*n1;

					_heprep_writer.addPoint( p1.x(), p1.y(), z0 + phi*tanLambda );
				}
			}
			else {
				
				_heprep_writer.addPoint( pca.x(), pca.y(), z0 );
				_heprep_writer.addPoint( pca.x() + (5000 * cos(phi0)), pca.y() + (5000*sin(phi0)), z0 + 5000*tanLambda );
			}

			// draw the hits that created this track
			const EVENT::TrackerHitVec& tracker_hits = current_track->getTrackerHits();


			for(EVENT::TrackerHitVec::const_iterator current_hit = tracker_hits.begin(); current_hit != tracker_hits.end(); current_hit++) {

				assert(*current_hit);
				
				const double* position = (*current_hit)->getPosition();

				_heprep_writer.addType( "TPC_SimTrackerHit", 0 );
				_heprep_writer.addAttDef( "dEdx", "Energy Deposit", "Physics", "electrons" );
				_heprep_writer.addAttValue( "Color", 0., 1., 0. );
				_heprep_writer.addAttValue( "DrawAs", "Point" );
				_heprep_writer.addAttValue( "MarkName", "Circle" );
				_heprep_writer.addAttValue( "MarkSize", 2. );
				_heprep_writer.addInstance();
				_heprep_writer.addAttValue( "dEdx", (*current_hit)->getdEdx() );
				_heprep_writer.addPrimitive();
				_heprep_writer.addPoint( position[0], position[1], position[2] );
			}
		}
	}
	catch (DataNotAvailableException &e) {

		streamlog_out( WARNING ) << e.what() << std::endl;
	}

}


Processor* HepRepOutputProcessor::newProcessor() {

	return new HepRepOutputProcessor();
}


void HepRepOutputProcessor::check(LCEvent *evt) {


}


void HepRepOutputProcessor::end() {

	_heprep_writer.close();
}

