#ifndef TPCHitToTrackerDataConverterProcessor_h
#define TPCHitToTrackerDataConverterProcessor_h 1

// C++
#include <string>

// LCIO
#include "lcio.h"

// Marlin
#include "marlin/Processor.h"

namespace marlintpc
{

    /* the following comment is tranfered to doxygen documentation
     * which begins with a second asterix, or in case of on line comment an extra third slash
     */
    /**  TPCHit data converting Processor of the MarlinTPC package .
     *
     *   TPCHitToTrackerDataConverterProcessor converts data in the (deprecated) TPCHit format to TrackerData. By setting the parameter IgnoreFirstNBins to a non-zero integer value N, the first N bins will be disregarded.
     *
     *  <h4>Input - Prerequisites</h4>
     *  Collection containing TPCHits
     *
     *  <h4>Output</h4>
     *  Collection containing TrackerData.
     *
     * @param InputCollectionName Name of raw data collection
     * @param OutputCollectionName Name of data collection
     * @param SetOutputTransient If not 0 the output collection is set transient
     *
     * @author R. Diener, Uni Hamburg
     * @version $Id: TPCHitToTrackerDataConverterProcessor.h,v 0.1 2008/02/01 11:34:04 rdiener Exp $
     */
    class TPCHitToTrackerDataConverterProcessor : public marlin::Processor
    {

        public:

            /* This method will be called by the marlin package
             * It returns a processor of the currend type (here TrackerRawDataToDataConverter)
             */
            virtual Processor*  newProcessor() { return new TPCHitToTrackerDataConverterProcessor ; }

            /* the default constructor
             * here the processor parameters are registered to the marlin package
             * other initialisation should be placed in the init method
             */
            TPCHitToTrackerDataConverterProcessor() ;

            /* Called at the begin of the job before anything is read.
             * Use to initialize the processor, e.g. book histograms
             */
            virtual void init() ;

            /* Called for every run.
             * in this processor it is not used
             */
            virtual void processRunHeader( lcio::LCRunHeader* run ) ;

            /* Called for every event - the working horse.
             * Here the real conversion take place
             */
            virtual void processEvent( lcio::LCEvent * evt ) ;

            /* This method is only called if the check flag is set (default) in the main processor
             * (where the input files are given)
             * It should be used to create check plots which are not needed in a complete analysis
             * (where normaly the check flag is deactivated)
             */
            virtual void check( lcio::LCEvent * evt ) ;

            /* Called after data processing for clean up.
             * e.g. saving contol histograms, delete created objects
             * (which are not stored in a collection/event), etc.
             */
            virtual void end() ;

        protected:

            /* normally the doxygen documentation is placed before the method/variable
             * but for one line comments in case of variables is may be more conviened to you
             * to place it after the variable. In this case you use an arrow to tell doxygen
             * to which variable the commet belong to.
             */

                                 ///<- the name of the input collection
            std::string _inputColName ;
                                 ///<- the name of the output collection
            std::string _outputColName ;

            /** give the status which is set to the transient flag of the output collection
             * it is an int instead of a boolean, because the processor parameter cannot be a boolean
             */
            int _outputIsTransient;

            /** is true till the first output collection is written to an event
             *  used to to store parameters only in the first collection
             *  (also done, if collection is transient)
             */
            bool isFirstCollection;

            int _ignoreFirstNBins; ///<- # of bins at the beginning of the pulse that should not be stored
            
            int _nRun ;          ///<- counts the runs processed
            int _nEvt ;          ///<- counts the events processed

    } ;

}
#endif
