#ifndef TimePixMaskCreaterProcessor_h
#define TimePixMaskCreaterProcessor_h 1

#include "marlin/DataSourceProcessor.h"
#include <fstream>

#include <TROOT.h>
#include <TH2D.h>

namespace marlin{
  
  /** Creates an lcio file containing the masking of dead and noisy pixel
   *  Decides which pixels to mask with the help of an occupancy histogram
   *
   *  With sufficient statistics dead pixels show zero occupancy and
   *  noisy pixel show maximum occupancy 
   *  
   *
   *  <h4>Input - Prerequisites</h4>
   *  TimePix occupancy histogram.
   *
   *  <h4>Output</h4>
   *  \li A runheader
   *  \li Events with TimePixRawData (lcio::TrackerRawData)
   *  \li First event contains TimePixModeMap collection
   *
   * @param FileListName   Text file with list of input TimePix files
   * \alert{TimePix file names must be absolute pathes!}
   * @param OutPutCollectionName Name of the output status map collection, default \c TimePixStatusMap
   * @param DetectorName Detector name for run header, default "TPC"
   * @param Description  Data descripton for run header, default "no description given"
   * @param SubDetectorName Sub-detector name for run header, default "TimePix"
   * @param ChipID Interer ID number of the chip (lcio::TrackerRawData::CellID0), 
   *   should be 0 as it's only one chip in the current version
   * @param MaxRecordNumber Maximum number of events 100000
   * @param SetOutputTransient If not 0 the output collection is set transient
   *
   * @author Simone Zimmermann(University of Bonn) 
   */
  
  class TimePixMaskCreaterProcessor : public DataSourceProcessor {
    
  public:

   TimePixMaskCreaterProcessor() ;

    virtual TimePixMaskCreaterProcessor*  newProcessor() ;


    /** Creates events with TrackerRawData collections from the TimePix raw data input files and
     *  calls all active processors' processEvent() and processRunHeader Method.
     *
     */
    virtual void readDataSource( int numEvents ) ;
    
    
    virtual void init() ;
    virtual void end() ;

    
  protected:
    
    std::string _outputColName;

    bool _isFirstEvent;
    std::string _histoFileName ;
    std::string _histoDirectoryName ;
    std::string _histoName ;

    std::ifstream _infile;
    int _chipID;
    int _runNr;
    int _evtNr;
    int _maxRecordN;
    int _nColumn;
    double _noiseThreshold;

    std::string _detName;
    std::string _description;
    std::string _subDetName;
    std::string _statusMapColName;
    TFile* _file;
    TH2D* _histoOccupancy;

/** give the status which is set to the transient flag of the output collection
 * it is an int instead of a boolean, because the processor parameter cannot be a boolean
 */
    int _outputIsTransient;
    
  };

 
} // end namespace marlin 

#endif
