#include "lcio.h"

#include "IO/LCWriter.h"
#include "EVENT/LCIO.h"

#include "IMPL/LCEventImpl.h" 
#include "IMPL/LCRunHeaderImpl.h" 
#include "IMPL/LCCollectionVec.h"
#include "IMPL/LCTOOLS.h"
#include "IMPL/TrackerRawDataImpl.h"

#include "gearimpl/RectangularPadRowLayout.h"
#include "gearimpl/Util.h"
#include "gearxml/GearXML.h"
#include "gear/GearMgr.h"
#include "gear/GEAR.h"

#include "ADCChannelMapping.h"
#include "Pedestal.h"

#include <iostream>
#include <sstream>
#include <string>
#include <cstdio>

using namespace std ;
using namespace lcio ;
using namespace gear;
using namespace tpcconddata;

void printUsage()
{
  cout << "generateRawData gearfile.xml [outfile.slcio]" <<endl;
    cout <<" outfile name is optional, "
	 << "default is TPCPulses.slcio"<< endl;
}

// return true if hit is on pad plane, false otherwise
bool placeHit(float x, float y, float z, TPCParameters const &TPCParams, float padPitch, LCCollectionVec *trackerRawDataVec)
{
           RectangularPadRowLayout  const *padLayout=NULL;
	   try{
	     padLayout = 
	       &dynamic_cast<const RectangularPadRowLayout&>( TPCParams.getPadLayout() );
	   }
	   catch(std::bad_cast& e){
	     std::cout << "  wrong type of layout - expected RectangularPadRowLayout ! " << std::endl ;
	   }

	  // assume rectangular charge distribution
	  // with width=padPitch and charge = 100
	  // charge is distributed onto (max) two pads
          int padNr; // the pseudo hardware channel number, continuous numbering within plane
	  int padIndex; // the gear pad index
	  try
	  {
	    padIndex=padLayout->getNearestPad(x,y);
	    // getPadsInRow(0) assumes that the number of pads per row is the same for all rows
	    padNr=(padLayout->getRowNumber(padIndex)*padLayout->getPadsInRow(0).size() 
		   + padLayout->getPadNumber(padIndex) );
	  }
	  catch(exception& e)
	  {
	    // pad not on the pad plane
	    // break the loop
	    cout << "Track left pad plane" << endl;
	    return false;
	  }
	  
	  Vector2D padCenter= padLayout->getPadCenter(padIndex);

	  TrackerRawDataImpl* rawData = new TrackerRawDataImpl; 
	  rawData->setCellID0(padNr);

	  float charge=(255. *(1.- fabs(x-padCenter[0])/padPitch) );
	  EVENT::ShortVec ADCValues;
	  ADCValues.push_back(static_cast<short>(charge*0.1));
	  ADCValues.push_back(static_cast<short>(charge*0.7));
	  ADCValues.push_back(static_cast<short>(charge*0.2));
	  rawData->setADCValues(ADCValues);
	  rawData->setTime(static_cast<int>(z*TPCParams.getReadoutFrequency()/TPCParams.getDriftVelocity()));

	  // debug:
	  cout << "padIndex: " << padIndex 
	       << "\t x/y: "<< x<<"/"<<y
	       << "\t center: "<< padCenter[0]<<"/"
	       << padCenter[1] 
	       << "\t charge: "<< charge
	       << endl;


	  // add the  rawData to the collection
	  trackerRawDataVec->push_back(rawData);

	   // in case of a central hit these is almost  no charge 
	  // on the neighbouring pad, skip it
	  if (charge>253.)
	    return true;

	  // try to find neighbour which also gets charge
	  int neighbouringPadNr;
	  int neighbouringPadIndex;
	  try
	  {
	    if (x-padCenter[0] < 0)
	    {
	      neighbouringPadIndex = padLayout->getLeftNeighbour(padIndex);
	    }
	    else
	    {
	      neighbouringPadIndex = padLayout->getRightNeighbour(padIndex);
	    }
	    neighbouringPadNr = (padLayout->getRowNumber(neighbouringPadIndex)*padLayout->getPadsInRow(0).size() 
		   + padLayout->getPadNumber(neighbouringPadIndex) ); 
	  }
	  catch(exception& e)
	  {
	    //no neighbour found, must be edge pad
	    return true;
	  }

	  // if there is a neighbour calculate charge and add it to collection
	  padCenter= padLayout->getPadCenter(neighbouringPadIndex);
	  
	  rawData = new TrackerRawDataImpl; 
	  rawData->setCellID0(neighbouringPadNr);
	  charge=(255. *(1.- fabs(x-padCenter[0])/padPitch) );

	  EVENT::ShortVec ADCValues2;
	  ADCValues2.push_back(static_cast<short>(charge*0.1));
	  ADCValues2.push_back(static_cast<short>(charge*0.7));
	  ADCValues2.push_back(static_cast<short>(charge*0.2));
	  rawData->setADCValues(ADCValues2);
	  rawData->setTime(static_cast<int>(z*TPCParams.getReadoutFrequency()/TPCParams.getDriftVelocity()));

	  // debug:
	  cout << "   nb: " << neighbouringPadNr
	       << "\t x/y: "<< x<<"/"<<y
	       << "\t center: "<< padCenter[0]<<"/"
	       << padCenter[1] 
	       << "\t charge: "<< charge
	       << "\t z: " << z << " = "<<rawData->getTime()
	       << endl;

	  // add the  rawData to the collection
	  trackerRawDataVec->push_back(rawData);

	  return true;
}

int main(int argc, char **argv)
{

  if (argc!=2 && argc!=3 )
  {
    printUsage();
    return(1);
  }

  
  // read pad layout from gear
  GearXML gearXML( argv[1] ) ;
  GearMgr* gearMgr = gearXML.createGearMgr() ;
  TPCParameters const &TPCParams=gearMgr->getTPCParameters();
  
  RectangularPadRowLayout const *padLayout=NULL;
  try{
    padLayout = 
      &dynamic_cast<const RectangularPadRowLayout&>( gearMgr->getTPCParameters().getPadLayout() );
  }
  catch(std::bad_cast& e){
    std::cout << "  wrong type of layout - expected RectangularPadRowLayout ! " << std::endl ;
  }
 
  // get outfile name
  string outFileName;
  if (argc==3)
    outFileName=argv[2];
  else
    outFileName="TPCRawData.slcio";

  // open outfile			     
  try{
    // create sio writer
    LCWriter* lcWrt = LCFactory::getInstance()->createLCWriter() ;
    
    try{  lcWrt->open( outFileName, LCIO::WRITE_NEW ) ;
    }
    catch( IOException& e ){    
      cout << e.what() << endl ;
      return 0 ;
    }
      
      // create runheader and write it
      // the information should come from a file
      // or be given interactively in a future version
      LCRunHeaderImpl* runHdr = new LCRunHeaderImpl ; 
      int runNr=0;
      runHdr->setRunNumber( runNr) ;
      
      string detName("demo")  ;
      runHdr->setDetectorName( detName ) ;
      
      stringstream description ; 
      description << "demo tpc rawData along straight line";
      runHdr->setDescription( description.str()  ) ;
      
      string tpcName("demotpc") ;
      runHdr->addActiveSubdetector( tpcName ) ;
      
      lcWrt->writeRunHeader( runHdr ) ;
  
      
      // loop the events, generate 11 events
      int maxEvents=11;
      for (int eventNr=0; eventNr<maxEvents; eventNr++)
      {
	// create new lcio event
	LCEventImpl*  evt = new LCEventImpl() ;
	
	evt->setRunNumber(  runNr   ) ;
	evt->setEventNumber( eventNr ) ;
	evt->setDetectorName( detName ) ;
      
	// we want to store a collection of trackerRawData
	LCCollectionVec* trackerRawDataVec = new LCCollectionVec( LCIO::TRACKERRAWDATA )  ;
	
	const std::vector< RectangularPadRowLayout::Row> rows=
	  padLayout->rows();

	// place one hit per row
	for (int row=0; row < padLayout->getNRows(); row++)
	{
	  // the position of the hit
	  // x = a * y + b
	  // z = c * y + d 

	  // y positon is center of pad row
	  float y = rows[row].Center;
	  
	  // one track for the first 10 events
	  if (eventNr<(maxEvents-1))
	  {
	    // the x position 
	    // starting position grows with event number
	    // angle grows with event number, max 45 deg
	    // direction alternates from event to event
	    float x =  1.8*eventNr
	               + y*eventNr/(maxEvents-2) // maximum angle = 45 deg for evt9, 10 has special treatment
                       *(eventNr%2?1:-1);
	    
	    // the z position
	    // three cases: no Angle, positive, negative dZ
	    float z;
	    switch (eventNr%3)
	    {
	      case 0:  z = 130.;
		       break;
	      case 1:  z = 130. + y;
		       break;
	      case 2:  
	      default: z = 130. - y;
		       break;
	    }
	  
	    // exit loop if placing of hit was not successful, i. e. track left pad plane
	    if(! placeHit(x, y, z, TPCParams, rows[row].WidthPerPad, trackerRawDataVec) )
	      //                                   \\  padPitch   //
	      break;
	  }
	  else // place three hits per row, another at the same z, the third at the same x
	  {
	    float x = 4.3 + 0.2 * y ;
	    float z = 130. + y;
	    
	    // do not check whether hit could be placed. This is inefficient, but it would be
	    // to complicated to disantangle which track left the plane and which one did not
	    placeHit(x,     y, z    , TPCParams, rows[row].WidthPerPad, trackerRawDataVec);
	    placeHit(x+10., y, z    , TPCParams, rows[row].WidthPerPad, trackerRawDataVec);
	    placeHit(x,     y, z+100., TPCParams, rows[row].WidthPerPad, trackerRawDataVec);
	  }

	}// for row

	// add the rawData collection to the event
	evt->addCollection( trackerRawDataVec , "TPCRawData" ) ;

	// in the first event: write the mapping and the pedestals
	if (eventNr==0)
	{
	  // create a collections of generic objects
	  LCCollectionVec* mappingCol = new LCCollectionVec( LCIO::LCGENERICOBJECT )  ;
	  LCCollectionVec* pedestalCol = new LCCollectionVec( LCIO::LCGENERICOBJECT )  ;
	  
	  //loop all pads and create a conditions objects for it

	  //outer loop: the rows
	  for (int row = 0; row < padLayout->getNRows(); row++)
	  {
	    // vector of the pad indices in this row
	    std::vector<int> padsInRowVector = padLayout->getPadsInRow(row);
	    
	    // nnow loop all pads in row
	    for (vector<int>::iterator padIndex=padsInRowVector.begin();
		 padIndex < padsInRowVector.end(); padIndex++  )
	    {
	      int padNr=(row*padsInRowVector.size() + padLayout->getPadNumber(*padIndex));
	      mappingCol->push_back( new ADCChannelMapping(padNr, *padIndex ) );
	      // create pedestal with 0 as value and width, data are already zero-suppressed
	      pedestalCol->push_back( new Pedestal(padNr)) ; 
	    }
	  }

	  // add the pe collection to the event
	  evt->addCollection( mappingCol , "TPCChannelMapping" ) ;
	  evt->addCollection( pedestalCol , "TPCPedestal" ) ;

	  
	}

	// write the event to the file
	lcWrt->writeEvent( evt ) ;
	
	// dump the event to the screen 
	LCTOOLS::dumpEvent( evt ) ;

	// ------------ IMPORTANT ------------- !
	// we created the event so we need to delete it ...
	delete evt ;
	// -------------------------------------

      } // for eventNr

      delete runHdr ;
      
      lcWrt->close() ;
      
  } catch( std::exception& ex){
    
    cout << " an excpetion occured: " << endl ;
    cout << "  " << ex.what() << endl ;
    return 1 ;
  }
     
return 0;

}
