#include "lcio.h"

#include "IO/LCWriter.h"
#include "EVENT/LCIO.h"

#include "IMPL/LCEventImpl.h" 
#include "IMPL/LCRunHeaderImpl.h" 
#include "IMPL/LCCollectionVec.h"
#include "IMPL/LCTOOLS.h"
#include "IMPL/TrackerRawDataImpl.h"

#include "gearimpl/FixedPadSizeDiskLayout.h"
#include "gearimpl/Util.h"
#include "gearxml/GearXML.h"
#include "gear/GearMgr.h"
#include "gear/GEAR.h"

#include "ADCChannelMapping.h"
#include "Pedestal.h"

#include "curlerGenerator.h"

#include <iostream>
#include <sstream>
#include <string>
#include <cstdio>
#include <cmath>
#include <complex>

using namespace std ;
using namespace lcio ;
using namespace gear;
using namespace tpcconddata;

void printUsage()
{
  cout << "generateRawData2 gearfile.xml [outfile.slcio]" <<endl;
    cout <<" outfile name is optional, "
	 << "default is TPCPulses.slcio"<< endl;
}

TrackerRawDataImpl* newRawHit(int padIndex, double time)
{
    TrackerRawDataImpl* rawData = new TrackerRawDataImpl; 
    rawData->setCellID0(padIndex);
    EVENT::ShortVec ADCValues;
    ADCValues.push_back(static_cast<short>(10));
    ADCValues.push_back(static_cast<short>(70));
    ADCValues.push_back(static_cast<short>(20));
    rawData->setADCValues(ADCValues);
    rawData->setTime(static_cast<int>(time));
    
    cout << "time is " << time << endl;
    return rawData;
}

// return true if hit is on pad plane, false otherwise 
bool placeChargeOnPads(float phiMin, float phiMax, int row, TPCParameters const &TPCParams, curlerGenerator &cg,
		       LCCollectionVec *trackerRawDataVec)
{
    cout << "placing charge in row "<< row << " angle " << phiMin << " to "<< phiMax << endl;

    // always place charge in a range < pi 
    // this means : if there is a wrap around at +-pi exchange phiMin and phiMax if phiMin < phiMax
    // if there is no wrap around exchange phiMin and phiMax if phiMin > phiMax

    // Attention: Does not handle the unlikely case that the center of the helix is very near
    // the center of the coordinate system and charge really is in an area > pi!

    if ( ( ( fabs(phiMax - phiMin) > M_PI ) &&  (phiMin < phiMax) ) || 
	 ( ( fabs(phiMax - phiMin) < M_PI ) &&  (phiMin > phiMax) ) )
    { //exchange phiMin and phiMax
	double tmp=phiMax;
	phiMax=phiMin;
	phiMin=tmp;
    }


    FixedPadSizeDiskLayout  const *padLayout=NULL;
    try{
	padLayout = 
	    &dynamic_cast<const FixedPadSizeDiskLayout&>( TPCParams.getPadLayout() );
    }
    catch(std::bad_cast& e){
	std::cout << "  wrong type of layout - expected FixedPadSizeDiskLayout ! " << std::endl ;
	return false;
    }
    
    double rowCenter= padLayout->getPadCenter(padLayout->getPadIndex(row,0))[0];
    // left means lower phi
    int leftmostPad = padLayout->getNearestPad(rowCenter,phiMin);
    leftmostPad = padLayout->getLeftNeighbour(leftmostPad);

    int rightmostPad = padLayout->getNearestPad(rowCenter,phiMax);
    rightmostPad = padLayout->getRightNeighbour(rightmostPad);

    // calculate correct time of the hit
    // first calculate the angle wrt center of curler
    double phiMean=(phiMin+phiMax)/2.;
    double x = rowCenter*cos(phiMean);
    double y = rowCenter*sin(phiMean);

    std::complex<double> positionInCurler(x - cg.geta(), y - cg.getb() );

    double windingPitch=80.;
    for ( double z = windingPitch*std::arg(positionInCurler)/2./M_PI; z <  TPCParams.getMaxDriftLength(); z+=windingPitch)
    {
	double time = z * TPCParams.getReadoutFrequency()/TPCParams.getDriftVelocity();

	if (leftmostPad <= rightmostPad)
	{ // no wrap aroud numbering border
	    for (int padIndex=leftmostPad; padIndex <=rightmostPad; padIndex++)
	    {
//	    TrackerRawDataImpl* rawData =  newRawHit(padIndex, time);
		// add the  rawData to the collection
		trackerRawDataVec->push_back( newRawHit(padIndex, time)) ;
		cout << "1. adding pad index " << padIndex << endl;
	    }
	}
	else
	{
	    const std::vector<int> rowVec =  padLayout->getPadsInRow(row);
	    for (int padIndex=rowVec.front(); padIndex <=rightmostPad ; padIndex++)
	    {
		trackerRawDataVec->push_back( newRawHit(padIndex, time)) ;
		cout << "2. adding pad index " << padIndex << endl;
	}
	    for (int padIndex=leftmostPad; padIndex <= rowVec.back(); padIndex++)
	    {
		trackerRawDataVec->push_back( newRawHit(padIndex, time)) ;
		cout << "3. adding pad index " << padIndex << endl;
	    }
	    
	}    
    } // for time
    
    return true;
}

// calculate angular distance on a circle, max angle can be pi
double angleDisance(double phi1, double phi2)
{
    // phi1 and phi2 are expected to be in [-pi:pi]
    double deltaPhi=std::fabs(phi1-phi2);
    if ( deltaPhi > M_PI )
	deltaPhi = 2 * M_PI - deltaPhi;
    
    return deltaPhi;
}

int main(int argc, char **argv)
{

  if (argc!=2 && argc!=3 )
  {
    printUsage();
    return(1);
  }

  
  // read pad layout from gear
  GearXML gearXML( argv[1] ) ;
  GearMgr* gearMgr = gearXML.createGearMgr() ;
  TPCParameters const &TPCParams=gearMgr->getTPCParameters();
  
  FixedPadSizeDiskLayout  const *padLayout=NULL;
  try{
      padLayout = 
	  &dynamic_cast<const FixedPadSizeDiskLayout&>( TPCParams.getPadLayout() );
  }
  catch(std::bad_cast& e){
      std::cout << "  wrong type of layout - expected FixedPadSizeDiskLayout ! " << std::endl ;
      return 0;
    }
    
  // get outfile name
  string outFileName;
  if (argc==3)
    outFileName=argv[2];
  else
    outFileName="TPCRawData2.slcio";

  // open outfile			     
  try{
    // create sio writer
    LCWriter* lcWrt = LCFactory::getInstance()->createLCWriter() ;
    
    try{  lcWrt->open( outFileName, LCIO::WRITE_NEW ) ;
    }
    catch( IOException& e ){    
      cout << e.what() << endl ;
      return 0 ;
    }
      
      // create runheader and write it
      // the information should come from a file
      // or be given interactively in a future version
      LCRunHeaderImpl* runHdr = new LCRunHeaderImpl ; 
      int runNr=0;
      runHdr->setRunNumber( runNr) ;
      
      string detName("demo")  ;
      runHdr->setDetectorName( detName ) ;
      
      stringstream description ; 
      description << "demo tpc rawData along straight line";
      runHdr->setDescription( description.str()  ) ;
      
      string tpcName("demotpc") ;
      runHdr->addActiveSubdetector( tpcName ) ;
      
      lcWrt->writeRunHeader( runHdr ) ;
  
 	// create new lcio event
	LCEventImpl*  evt = new LCEventImpl() ;
	
	evt->setRunNumber(  runNr   ) ;
	evt->setEventNumber( 0 ) ;
//	evt->setEventNumber( eventNr ) ;
	evt->setDetectorName( detName ) ;
      
	// we want to store a collection of trackerRawData
	LCCollectionVec* trackerRawDataVec = new LCCollectionVec( LCIO::TRACKERRAWDATA )  ;
     
      // we have a curler with centre at 40,20 and a radius of 28
      curlerGenerator cg(40,20,28);

      // determine the first and the last row (border) which are hit:

      // start with invalid range
      int innermostRow=padLayout->getNRows();
      int outermostRow=0;

      double minRadius =  ( cg.getM() - cg.getr() < 0 ? 0 : cg.getM() - cg.getr()) ;
      double maxRadius =  cg.getM() + cg.getr() ;
      for (int row=0; row < padLayout->getNRows(); row++)
      {
	  double rowCenter= padLayout->getPadCenter(padLayout->getPadIndex(row,0))[0];
	  
	  if ( (minRadius < rowCenter + padLayout->getRowHeight(row)/2.)
	       &&  (minRadius >= rowCenter - padLayout->getRowHeight(row)/2.) )
	       innermostRow=row;
	  if ( (maxRadius < rowCenter + padLayout->getRowHeight(row)/2.)
	       &&  (maxRadius >= rowCenter - padLayout->getRowHeight(row)/2.) )
	       outermostRow=row;
      }

      if (innermostRow < outermostRow)
      {
	  // the innermost row only has one hit
	  double rowCenter= padLayout->getPadCenter(padLayout->getPadIndex(innermostRow,0))[0];
	  
	  std::pair<double,double> phiMinMax=cg.calculateIntersectionAngles(rowCenter+
									    padLayout->getRowHeight(innermostRow)/2.);

          placeChargeOnPads(phiMinMax.first, phiMinMax.second, innermostRow, TPCParams, cg, trackerRawDataVec);
	  
      }

      for (int row=innermostRow+1; row <outermostRow ; row ++)
      {
	  double rowCenter= padLayout->getPadCenter(padLayout->getPadIndex(row,0))[0];

	  // get the intersection angles for upper and lower egde radius of this row
	  std::pair<double,double> phiMinMaxInner=cg.calculateIntersectionAngles(rowCenter - 
									    padLayout->getRowHeight(outermostRow)/2.);

	  std::pair<double,double> phiMinMaxOuter=cg.calculateIntersectionAngles(rowCenter +
									    padLayout->getRowHeight(outermostRow)/2.);
	  
	  // find correct pairing of angles on inner and outer row
	  // the matching pair is the one with the smalest difference
	  double angle1=angleDisance( phiMinMaxInner.first , phiMinMaxOuter.first );
	  double angle2=angleDisance( phiMinMaxInner.first , phiMinMaxOuter.second );

	  if (angle1 < angle2)
	  {
	      // the pairing is (inner.first, outer.first) ; (inner.second, outer.second)
	      placeChargeOnPads(phiMinMaxInner.first, phiMinMaxOuter.first, row, TPCParams, cg, trackerRawDataVec);
	      placeChargeOnPads(phiMinMaxInner.second, phiMinMaxOuter.second, row, TPCParams, cg, trackerRawDataVec);
	  }
	  else
	  {
	      // the pairing is (inner.first, outer.second) ; (inner.second, outer.first)
	      placeChargeOnPads(phiMinMaxInner.first, phiMinMaxOuter.second, row, TPCParams, cg, trackerRawDataVec);
	      placeChargeOnPads(phiMinMaxInner.second, phiMinMaxOuter.first, row, TPCParams, cg, trackerRawDataVec);
	  }
      }

      if (outermostRow > innermostRow)
      {
	  // the innermost row only has one hit
	  double rowCenter= padLayout->getPadCenter(padLayout->getPadIndex(outermostRow,0))[0];
	  
	  std::pair<double,double> phiMinMax=cg.calculateIntersectionAngles(rowCenter - 
									    padLayout->getRowHeight(outermostRow)/2.);

          placeChargeOnPads(phiMinMax.first, phiMinMax.second, outermostRow, TPCParams, cg, trackerRawDataVec);
	  
      }

      // add the rawData collection to the event
      evt->addCollection( trackerRawDataVec , "TPCRawData" ) ;
      
      // only in first event, currently there is only one...
      {
	  // create a collections of generic objects
	  LCCollectionVec* mappingCol = new LCCollectionVec( LCIO::LCGENERICOBJECT )  ;
	  LCCollectionVec* pedestalCol = new LCCollectionVec( LCIO::LCGENERICOBJECT )  ;
	  
	  //loop all pads and create a conditions objects for it

	  //outer loop: the rows
	  for (int row = 0; row < padLayout->getNRows(); row++)
	  {
	    // vector of the pad indices in this row
	    std::vector<int> padsInRowVector = padLayout->getPadsInRow(row);
	    
	    // nnow loop all pads in row
	    for (vector<int>::iterator padIndex=padsInRowVector.begin();
		 padIndex < padsInRowVector.end(); padIndex++  )
	    {
//	      int padNr=(row*padsInRowVector.size() + padLayout->getPadNumber(*padIndex));
	      int padNr=*padIndex; // trivial mapping
	      mappingCol->push_back( new ADCChannelMapping(padNr, *padIndex ) );
	      // create pedestal with 0 as value and width, data are already zero-suppressed
	      pedestalCol->push_back( new Pedestal(padNr)) ; 
	    }
	  }

	  // add the pe collection to the event
	  evt->addCollection( mappingCol , "TPCChannelMapping" ) ;
	  evt->addCollection( pedestalCol , "TPCPedestal" ) ;

	  
      }

      // write the event to the file
      lcWrt->writeEvent( evt ) ;
	
	// dump the event to the screen 
	LCTOOLS::dumpEvent( evt ) ;

	// ------------ IMPORTANT ------------- !
	// we created the event so we need to delete it ...
	delete evt ;
	// -------------------------------------

//      } // for eventNr

      delete runHdr ;
      
      lcWrt->close() ;
      
  } catch( std::exception& ex){
    
    cout << " an excpetion occured: " << endl ;
    cout << "  " << ex.what() << endl ;
    return 1 ;
  }
     
return 0;

}
