// My Stuff
#include "MyPFOAnalysis.h"
#include "MyRecoDisplay.h"
#include "MyMCTree.h"
// headers for io,std etc
#include <iostream>
#include <cstdlib>
// LCIO stuff
#include <EVENT/LCCollection.h>
#include <EVENT/LCRelation.h>
#include <EVENT/MCParticle.h>
#include <EVENT/Track.h>
#include "IMPL/ClusterImpl.h"
#include <EVENT/SimCalorimeterHit.h>
#include <EVENT/CalorimeterHit.h>
#include "IMPL/ReconstructedParticleImpl.h"
#include "IMPL/LCFlagImpl.h"
// GEAR
#include <gear/GEAR.h>
#include <gear/CalorimeterParameters.h>
#include <gear/TPCParameters.h>
#include <gear/PadRowLayout2D.h>
#include <gear/LayerLayout.h>
// Marlin stuff
#include <marlin/Global.h>
// root stuff for histograms
#include "TLorentzVector.h"
#include "TFile.h"
#include "TH1.h"
#include "TH2.h"
#include "TFile.h"

using namespace lcio ;
using namespace marlin ;

const float pi = acos(-1.0);
const float twopi = 2.0*pi;

MyPFOAnalysis aMyPFOAnalysis;

MyPFOAnalysis::MyPFOAnalysis() : Processor("MyPFOAnalysis") {
  
  // modify processor description
  _description = "MyPFOAnalysis analyses output of PandoraPFA" ;
  

  // register steering parameters: name, description, class-variable, default value


  std::string rootFile;
  rootFile = "MyPFOAnalysis.root";
  registerProcessorParameter( "RootFile" , 
			      "Name of the Track collection used for clustering"  ,
			      _rootFile,
			      rootFile ) ;

  std::string particleCollectionName;
  particleCollectionName = "PandoraPFOs";
  registerProcessorParameter( "ParticleCollectionName" , 
			      "Particle Collection Name "  ,
			      _particleCollectionName,
			      particleCollectionName);



  std::string clusterCollectionName;
  particleCollectionName = "PandoraClusters";
  registerProcessorParameter( "ClusterCollectionName" , 
			      "Cluster Collection Name "  ,
			      _clusterCollectionName,
			      clusterCollectionName);


  // Name of relation between calorimeter hit and sim calorimeter hit
  std::string caloRel;
  caloRel = "RelationCaloHit"; 
  registerProcessorParameter( "RelationCaloString" , 
			      "Name of the calo hit <-> sim calo hit relation"  ,
			      _relationCaloString,
			      caloRel) ;


  // Name of relation between true and MC tracks
  std::string trackRel;
  trackRel = "TrueTracksMCP";
  registerProcessorParameter( "RelationTrackString" , 
			      "Name of the track hit <-> sim track hit relation"  ,
			      _relationTrackString,
			      trackRel) ;

  registerProcessorParameter( "RecoDisplay" ,
                              "Draw the display",
			      _display,
			      (int)0 );

  registerProcessorParameter( "Printing" ,
                              "Set the debig print level",
			      _printing,
			      (int)0 );


  registerProcessorParameter( "ECutLow" ,
                              "Low E cut",
			      _eCutLow,
			      (float)10000. );

  registerProcessorParameter( "ECutHigh" ,
                              "High E cut",
			      _eCutHigh,
			      (float)0. );


  registerProcessorParameter( "CosQQCut" ,
                              "Low qq cut",
			      _cosQQCut,
			      (float)1.1 );

  registerProcessorParameter( "ConfusionAnalysis" ,
                              "Run Confusion Analysis",
			      _confusionAnalysis,
			      (int)0 );

  registerProcessorParameter( "PostRecoFix" ,
                              "Apply corrections to recoed PFOs",
			      _postRecoFix,
			      (int)0 );

}


void MyPFOAnalysis::init() { 

  // usually a good idea to
  printParameters() ;

  _recoDisplay = NULL;
  _clusterNavigator = NULL;
  _trackNavigator = NULL;
  _myMcTree = NULL;
  
  _nRun = 0 ;
  _nEvt = 0 ;

  if(_recoDisplay==NULL && _display != 0)_recoDisplay = new MyRecoDisplay(1);

  fPFA      = new TH1F("fPFA", "total energy",2500, 0., 1250.0);
  fMcPfoPdgCode = new TH1F("fMcPfoPdgCode", "IPDG MC PFO",10000, 0., 10000.0);
  fPfoPdgCode = new TH1F("fPfoPdgCode", "IPDG  PFO",10000, 0., 10000.0);


  fPFAhcal  = new TH1F("fPFAhcal", "total energy hcal only events",2500, 0., 1250.0);
  fPFAnu    = new TH1F("fPFAnu", "total energy + nu",2500, 0., 1250.0);
  fPFAnufwd = new TH1F("fPFAnufwd", "total energy + nu + fwd",2500, 0., 1250.0);  fPFAudscb = new TH1F("fPFAudscb", "total energy",2500, 0., 1250.0);
  fPFAuds   = new TH1F("fPFAuds", "total energy",2500, 0., 1250.0);
  fPFAudsHM20   = new TH1F("fPFAudsHM20", "total energy",2500, 0., 1250.0);
  fPFAudsHM10   = new TH1F("fPFAudsHM10", "total energy",2500, 0., 1250.0);
  fPFAudsHP10   = new TH1F("fPFAudsHP10", "total energy",2500, 0., 1250.0);
  fPFAudsHP20   = new TH1F("fPFAudsHP20", "total energy",2500, 0., 1250.0);
  fPFAcb    = new TH1F("fPFAcb", "total energy",2500, 0., 1250.0);
  fPFA1  = new TH1F("fPFA1", "total energy 0.0-0.1",2500, 0., 1250.0);
  fPFA2  = new TH1F("fPFA2", "total energy 0.1-0.2",2500, 0., 1250.0);
  fPFA3  = new TH1F("fPFA3", "total energy 0.2-0.3",2500, 0., 1250.0);
  fPFA4  = new TH1F("fPFA4", "total energy 0.3-0.4",2500, 0., 1250.0);
  fPFA5  = new TH1F("fPFA5", "total energy 0.4-0.5",2500, 0., 1250.0);
  fPFA6  = new TH1F("fPFA6", "total energy 0.5-0.6",2500, 0., 1250.0);
  fPFA7  = new TH1F("fPFA7", "total energy 0.6-0.7",2500, 0., 1250.0);
  fPFAL7A  = new TH1F("fPFAL7A", "total energy <0.7",2500, 0., 1250.0);
  fPFAL7Aud= new TH1F("fPFAL7Aud", "total energy <0.7 ud",2500, 0., 1250.0);
  fPFAL7As = new TH1F("fPFAL7As", "total energy <0.7 s",2500, 0., 1250.0);
  fPFAL7Ac = new TH1F("fPFAL7Ac", "total energy <0.7 c",2500, 0., 1250.0);
  fPFAL7Ab = new TH1F("fPFAL7Ab", "total energy <0.7 b",2500, 0., 1250.0);
  fPFAL7B  = new TH1F("fPFAL7B", "total energy <0.7 - bad",2500, 0., 1250.0);
  fPFAQQ   = new TH1F("fPFAQQ", "total energy - true E",2000, -500., 500.0);
  fPFAQQ8  = new TH1F("fPFAQQ8", "total energy - true E 8",2000, -500., 500.0);
  fPFA8  = new TH1F("fPFA8", "total energy 0.7-0.8",2500, 0., 1250.0);
  fPFA9  = new TH1F("fPFA9", "total energy 0.8-0.9",2500, 0., 1250.0);
  fPFA10 = new TH1F("fPFA10","total energy 0.9-1.0",2500, 0., 1250.0);
  fPFA11 = new TH1F("fPFA11","total energy 0.9-0.925",2500, 0., 1250.0);
  fPFA12 = new TH1F("fPFA12","total energy 0.925-0.95",2500, 0., 1250.0);
  fPFA13 = new TH1F("fPFA13","total energy 0.95-0.975",2500, 0., 1250.0);
  fPFA14 = new TH1F("fPFA14","total energy 0.975-1.0",2500, 0., 1250.0);



  fPFADMZ    = new TH1F("fPFADMZ"    , "Delta Mz",200, -50.0, 50.0);
  fPFADMZ8   = new TH1F("fPFADMZ8"   , "Delta Mz",200, -50.0, 50.0);
  fPFADMZQQ8 = new TH1F("fPFADMZQQ8" , "Delta Mz",200, -50.0, 50.0);
  fPFADMZP8  = new TH1F("fPFADMZP8"  , "Delta Mz",200, -50.0, 50.0);
  fPFADMZOMZ = new TH1F("fPFADMZOMZ" , "Delta Mz / Mz",200, -25.0, 25.0);
  fPFADMZOMZQQ8 = new TH1F("fPFADMZOMZQQ8" , "Delta Mz / Mz",200, -25.0, 25.0);


  fETrackud   = new TH1F("fETrackud","Track energy ud",200, 0., 100.0);
  fETracks    = new TH1F("fETracks" ,"Track energy  s",200, 0., 100.0);
  fENeutralud = new TH1F("fENeutralud","Neutral energy ud",200, 0., 100.0);
  fEGammaud   = new TH1F("fEGammaud","Gamma energy ud",200, 0., 100.0);
  fENeutrals  = new TH1F("fENeutrals","Neutral energy s",200, 0., 100.0);
  fEGammas    = new TH1F("fEGammas","Gamma energy s",200, 0., 100.0);

  fxproton    = new TH1F("fxproton","proton frag",500, 0., 1.0);

  // MIP Calibration

  fEcalMIP = new TH1F("fEcalMIP", "ecal MIP peak ",100, 0., 5.0);
  fEcal1MIP = new TH1F("fEcal1MIP", "ecal1 MIP peak ",100, 0., 5.0);
  fHcalMIP = new TH1F("fHcalMIP", "hcal MIP peak ",100, 0., 5.0);
  fEcalMIPcorr = new TH1F("fEcalMIPcorr", "ecal MIP peak ",100, 0., 5.0);
  fEcal1MIPcorr = new TH1F("fEcal1MIPcorr", "ecal1 MIP peak ",100, 0., 5.0);
  fHcalMIPcorr = new TH1F("fHcalMIPcorr", "hcal MIP peak ",100, 0., 5.0);

  fFragmentEnergy = new TH1F("fFragmentEnergy","frag energy",200,0.,50.);
  fConfusedHadron = new TH1F("fConfusedHadron","conf hadron",200,0.,50.);
  fConfusedPhoton = new TH1F("fConfusedPhoton","frag photon",200,0.,50.);
  fConfusion = new TH1F("fConfusion","total confusion",200,-25.,25.);

  fConfused = new TH2F("fConfused","2d confusion",3,0.,3.,3,0.,3.);

  fNPFO = new TH1F("fNPFO","N PFO",100,0.,100.);
  fNTrackPFO = new TH1F("fNTrackPFO","N Track PFO",100,0.,100.);
  fNClusterPFO = new TH1F("fNClusterPFO","N Cluster PFO",100,0.,100.);

}



void MyPFOAnalysis::processRunHeader( LCRunHeader* run) { 

  _nRun++ ;
  _detectorName = run->getDetectorName();

  // determine detector (still some hard-coded parameters)
  std::cout << " DETECTOR : " << _detectorName << std::endl;

  // Calorimeter geometry from GEAR

  const gear::CalorimeterParameters& pEcalEndcap = Global::GEAR->getEcalEndcapParameters();
  //  const gear::LayerLayout& ecalEndcapLayout = pEcalEndcap.getLayerLayout();
  _zOfEndcap = (float)pEcalEndcap.getExtent()[2];


} 


void MyPFOAnalysis::processEvent( LCEvent * evt ) { 

  _nEvt ++ ;

  // clear the event/display
  this->Clear();

  typedef const std::vector<std::string> StringVec ;
  _mz = 0.;
  _ez = 0.;
  _costz = 0.;


  StringVec* strVec = evt->getCollectionNames() ;
  for(StringVec::const_iterator name=strVec->begin(); name!=strVec->end(); name++){    

    LCCollection* col = evt->getCollection(*name);
    int nelem = col->getNumberOfElements();
    if(_nEvt==1)std::cout << *name << " " << nelem << std::endl;
    // find the MC tree and form collection of PerfectPFOs
    if (col->getTypeName() == LCIO::MCPARTICLE ) {
      _myMcTree = new MyMCTree(col);
      _myMcTree->PerfectPFO(0);
      _pMcPFOs = *(_myMcTree->getMCPFOs());
      std::vector<MCParticle*> mcQuarks = *(_myMcTree->getQuarks());
      if(mcQuarks.size()==2){
	TLorentzVector q1(mcQuarks[0]->getMomentum()[0] ,mcQuarks[0]->getMomentum()[1] ,mcQuarks[0]->getMomentum()[2], mcQuarks[0]->getEnergy()  );
	TLorentzVector q2(mcQuarks[1]->getMomentum()[0] ,mcQuarks[1]->getMomentum()[1] ,mcQuarks[1]->getMomentum()[2], mcQuarks[1]->getEnergy()  );
	TLorentzVector z = q1+q2;
	_mz = z.M();
	float pz[3];
	float pq1[3];
	float pq2[3];
	_ez = mcQuarks[0]->getEnergy()  + mcQuarks[1]->getEnergy(); 
	pq1[0] = mcQuarks[0]->getMomentum()[0];
	pq1[1] = mcQuarks[0]->getMomentum()[1];
	pq1[2] = mcQuarks[0]->getMomentum()[2];
	pq2[0] = mcQuarks[1]->getMomentum()[0];
	pq2[1] = mcQuarks[1]->getMomentum()[1];
	pq2[2] = mcQuarks[1]->getMomentum()[2];
	pz[0] = mcQuarks[0]->getMomentum()[0] + mcQuarks[1]->getMomentum()[0];
	pz[1] = mcQuarks[0]->getMomentum()[1] + mcQuarks[1]->getMomentum()[1];
	pz[2] = mcQuarks[0]->getMomentum()[2] + mcQuarks[1]->getMomentum()[2];
	float pq1tot = sqrt(pq1[0]*pq1[0]+pq1[1]*pq1[1]+pq1[2]*pq1[2]);
	float pq2tot = sqrt(pq2[0]*pq2[0]+pq2[1]*pq2[1]+pq2[2]*pq2[2]);
	float pztot  = sqrt(pz[0]*pz[0]+pz[1]*pz[1]+pz[2]*pz[2]);
	_costz  = pz[2]/pztot;
	_costq1 = pq1[2]/pq1tot;
	_costq2 = pq2[2]/pq2tot;
	std::cout << " Eqq    = " << _ez << std::endl;
	std::cout << " Costq1 = " << _costq1 << std::endl;
	std::cout << " Costq2 = " << _costq2 << std::endl;
	std::cout << " Costz  = " << _costz << std::endl;
	std::cout << " Mqq    = " << _mz  << std::endl;
      }
    }

    if (col->getTypeName() == LCIO::RECONSTRUCTEDPARTICLE) {  
      if(*name==_particleCollectionName){
	for (int j=0; j < nelem; ++j) {
	  ReconstructedParticle* recoPart = dynamic_cast<ReconstructedParticle*>(col->getElementAt(j));
	  _pfovec.push_back(recoPart);
	}
      }
    }
  }  
  
  // create navigators for matching of tracks and clusters and MC particles
  _relationCaloCollection = NULL;
  try {
    _relationCaloCollection = evt->getCollection(_relationCaloString.c_str());
  }
  catch(DataNotAvailableException &e){
    std::cout << "MyPFOAnalysis : " 
	      << "No relation exists " 
	      << "between Sim and Calo hits" << std::endl; 	
  }
  
  if(_relationCaloCollection!=NULL)_clusterNavigator = new LCRelationNavigator(_relationCaloCollection); 
  
  _relationTrackCollection = NULL;
  try {
    _relationTrackCollection = evt->getCollection(_relationTrackString.c_str());	
  }
  catch(DataNotAvailableException &e){
    std::cout << "MyPFOAnalysis : " 
	      << "No relation exists " 
	      << "between Sim and Track hits" << std::endl; 	
  }
  
  if(_relationTrackCollection!=NULL)_trackNavigator = new LCRelationNavigator(_relationTrackCollection); 
  
  AnalysePFAPerformance();
  if(_confusionAnalysis>0)this->ConfusionAnalysis();

}



void MyPFOAnalysis::AnalysePFAPerformance(){ 

  float totenergy = 0;
  float fwdenergy = 0;
  float totgammaenergy = 0;
  //  float tottrackenergy = 0;
  float tothadenergy = 0;
  float totmom[3]={0.,0.,0.};
  float enu = 0;
  float efw = 0.;
  float ebad = 0.;
  float ntrack = 0.;
  float nclust = 0.;
  float npfo   = 0.;

  //int ntracks = 0;
  //int nclusters = 0;
  float ehcaltot = 0.;
  float eecaltot = 0.;
  float eecal = 0;
  float ehcal = 0;

  // Loop over reconstructed PFOs
  bool dodgy = false;


  float mcet = 0.;
  float mcen = 0.;
  float mceg = 0.;
 
  for(unsigned int i=0;i<_pMcPFOs.size();i++){
    fMcPfoPdgCode->Fill(abs(_pMcPFOs[i]->getPDG()) );
    switch( _pMcPFOs[i]->getPFAType()){
    case MC_PFA_TYPE_TRACK:
    case MC_PFA_TYPE_ELECTRON:
    case MC_PFA_TYPE_MUON:
      mcet+=_pMcPFOs[i]->getMomentum()[3];
      break;
    case MC_PFA_TYPE_PHOTON:
      mceg+=_pMcPFOs[i]->getMomentum()[3];
      break;
    case MC_PFA_TYPE_NEUTRAL_HADRON:
      mcen+=_pMcPFOs[i]->getMomentum()[3];
      break;
    default:
      break;
    }
    if( abs(_pMcPFOs[i]->getPDG())==2212){
      float p=sqrt(_pMcPFOs[i]->getMomentum()[0]*_pMcPFOs[i]->getMomentum()[0]
		 + _pMcPFOs[i]->getMomentum()[1]*_pMcPFOs[i]->getMomentum()[1]
		   + _pMcPFOs[i]->getMomentum()[1]*_pMcPFOs[i]->getMomentum()[1]);
      float x = p/45.6;
      fxproton->Fill(x);
    }
  }

  int ipdg = _myMcTree->getPrimary();
  float thrust = _myMcTree->getThrust();
  if(ipdg<=2){
    fETrackud->Fill(mcet,1.);
    fENeutralud->Fill(mcen,1.);
    fEGammaud->Fill(mceg,1.);
  }
  if(ipdg==3){
    fETracks->Fill(mcet,1.);
    fENeutrals->Fill(mcen,1.);
    fEGammas->Fill(mceg,1.);
  }
  

  for(unsigned int i=0;i<_pfovec.size();i++){
    npfo++;
    TrackVec     tracks = _pfovec[i]->getTracks();
    ClusterVec clusters = _pfovec[i]->getClusters();
    if(tracks.size()>0)ntrack++;
    if(tracks.size()==0)nclust++;
    fPfoPdgCode->Fill(abs(_pfovec[i]->getType()));

    float correction = 1.0;
    if(_postRecoFix!=0){
      // hadron
      if(_pfovec[i]->getType()!=22&&tracks.size()==0){
	float p[3];
	p[0] = _pfovec[i]->getMomentum()[0];
	p[1] = _pfovec[i]->getMomentum()[1];
	p[2] = _pfovec[i]->getMomentum()[2];
	float ptot = sqrt(p[0]*p[0]+p[1]*p[1]+p[2]*p[2]);
	float x = fabs(p[2])/ptot;
	float e = _pfovec[i]->getEnergy();
	float factor = 0.16*(1+e/100);
	if(factor>0.25)factor=0.25;
	if(x<0.2)correction =  (1.0-factor* exp( -80.*x ) );
	if(x>0.60 && x<0.77)correction = (1-0.08*(x-0.60)/0.17);
	if(x>0.77 && x<0.80)correction = (1-0.08+0.054*(x-0.77)/0.03);
	if(x>0.80)correction = 0.97;
      }
    }

    totenergy += _pfovec[i]->getEnergy()/correction;
    totmom[0]+=_pfovec[i]->getMomentum()[0]/correction;
    totmom[1]+=_pfovec[i]->getMomentum()[1]/correction;
    totmom[2]+=_pfovec[i]->getMomentum()[2]/correction;
    if(tracks.size()==1 && clusters.size()==1){
      float sigpop = sqrt(tracks[0]->getCovMatrix()[5])/fabs(tracks[0]->getOmega());
      float sigeoe = 0.55/sqrt(clusters[0]->getEnergy());
      float a  = 1.0/sigpop/sigpop;
      float b  = 1.0/sigeoe/sigeoe;
      float tw = a/(a+b); 
      float cw = b/(a+b); 
      float ecor = cw*clusters[0]->getEnergy() + tw*_pfovec[i]->getEnergy();
      if(cw>0.05){
	std::cout << " *****************************************************************************************Track vs Cluster : " << sigpop << " : " << sigeoe << std::endl;
	std::cout << " cw = " << cw << std::endl;
	std::cout << _pfovec[i]->getEnergy() << " -> " << ecor << std::endl;
      }
    }



    if(_printing>0)std::cout << " RECOPFO : " << i << " " << _pfovec[i]->getType() << " " << _pfovec[i]->getEnergy() << "  " << tracks.size() << ":" << clusters.size() << " charge : " << _pfovec[i]->getCharge() << std::endl;

    if(_pfovec[i]->getType()==999)dodgy = true;

    if(_pfovec[i]->getType()!=22&&tracks.size()==0){
      tothadenergy += _pfovec[i]->getEnergy();
    }
    if(_pfovec[i]->getType()==22&&tracks.size()==0){
      totgammaenergy += _pfovec[i]->getEnergy();
    }
    for(unsigned int icluster=0;icluster<clusters.size();icluster++){
      const CalorimeterHitVec hits   = clusters[icluster]->getCalorimeterHits();
      for(unsigned int ihit=0;ihit<hits.size();ihit++){
	if(hits[ihit]->getType()==0)eecal+=1.;
	if(hits[ihit]->getType()==1)eecal+=1.;
	if(hits[ihit]->getType()==2)ehcal+=1.;
      }
      ehcaltot += ehcal;
      eecaltot += eecal;
    }
  }


  float totmass = sqrt(totenergy*totenergy-
                       totmom[0]*totmom[0]-
		       totmom[1]*totmom[1]-
                       totmom[2]*totmom[2]);

  // loop over MC PFOs and find energy in "primary" neutrinos
  for(unsigned int imc=0;imc<_pMcPFOs.size();imc++){
    MCParticle* mcpart = _pMcPFOs[imc]->getMCParticle();
    if(_pMcPFOs[imc]->isNeutrino())enu+= mcpart->getEnergy();
    // find polar angle of MC PFO
    float px = _pMcPFOs[imc]->getMomentum()[0];
    float py = _pMcPFOs[imc]->getMomentum()[1];
    float pz = _pMcPFOs[imc]->getMomentum()[2];
    if(fabs(pz)/sqrt(px*px+py*py+pz*pz)>0.98)fwdenergy+= mcpart->getEnergy();
  }

  fNPFO->Fill(npfo);

  if(_printing>-1){
    std::cout << " EVENT                : " <<_nEvt << std::endl;
    if(dodgy)std::cout << " DETECTED DODGY CLUSTER " << std::endl;
    std::cout << " RECONSTRUCTED MASS   : " << totmass << std::endl; 
    if(_mz>0.1)std::cout << " DELTA MASS/MASS      : " << 100*(totmass-_mz)/_mz << " % " << std::endl; 
    std::cout << " RECONSTRUCTED ENERGY : " << totenergy << std::endl; 
    std::cout << " RECONSTRUCTED HAD EN : " << tothadenergy-ebad << std::endl; 
    std::cout << " RECONSTRUCTED ENERGY*: " << totenergy-ebad << std::endl; 
    std::cout << " RECO ENERGY + ENU    : " << totenergy+enu << std::endl; 
    std::cout << " FWD ENERGY           : " << fwdenergy << std::endl; 
  }
  if(dodgy)return;
  if(_mz>0.1 && fabs(_costz)<0.95){
    fPFAQQ->Fill(totenergy-_ez);
    fPFADMZ->Fill(totmass-_mz);
    if(_mz>75)fPFADMZOMZ->Fill(100*(totmass-_mz)/totmass);
    if(fabs(_costz)<0.8)fPFADMZ8->Fill(totmass-_mz);
    if(fabs(_costz)>0.8)fPFADMZP8->Fill(totmass-_mz);
    if(fabs(_costq1)<0.8 && fabs(_costq2)<0.8){
      fPFADMZQQ8->Fill(totmass-_mz);
      fPFADMZOMZQQ8->Fill( 100*(totmass-_mz)/totmass);
      fPFAQQ8->Fill(totenergy-_ez);
    }
  }    


  fPFA->Fill(totenergy,1.);
  fPFAnu->Fill(totenergy+enu,1.);
  fPFAnufwd->Fill(totenergy+enu+efw,1.);
  if(ipdg>=1&&ipdg<=3)fPFAuds->Fill(totenergy,1.);
  if(ipdg>=4&&ipdg<=5)fPFAcb->Fill(totenergy,1.);
  if(ipdg>=1&&ipdg<=5)fPFAudscb->Fill(totenergy,1.);
  if(ipdg>=1&&ipdg<=3){
    std::cout << " Thrust : " << thrust << std::endl;
    if(thrust<=0.1                 )fPFA1->Fill(totenergy,1.);
    if(thrust>0.1  && thrust<=0.2  )fPFA2->Fill(totenergy,1.);
    if(thrust>0.2  && thrust<=0.3  )fPFA3->Fill(totenergy,1.);
    if(thrust>0.3  && thrust<=0.4  )fPFA4->Fill(totenergy,1.);
    if(thrust>0.4  && thrust<=0.5  )fPFA5->Fill(totenergy,1.);
    if(thrust>0.5  && thrust<=0.6  )fPFA6->Fill(totenergy,1.);
    if(thrust<=0.7  ){
      std::cout << "Filling L7A : " << totenergy+enu << std::endl; 
      fPFAL7A->Fill(totenergy+enu,1.);
      if(ipdg<=2)fPFAL7Aud->Fill(totenergy+enu,1.);
      if(ipdg==3)fPFAL7As->Fill(totenergy+enu,1.);
      if(ipdg==4)fPFAL7Ac->Fill(totenergy+enu,1.);
      if(ipdg==5)fPFAL7Ab->Fill(totenergy+enu,1.);
      fPFAudsHM20->Fill(totenergy+enu-tothadenergy*0.1,1.);
      fPFAudsHM10->Fill(totenergy+enu-tothadenergy*0.05,1.);
      fPFAudsHP10->Fill(totenergy+enu+tothadenergy*0.05,1.);
      fPFAudsHP20->Fill(totenergy+enu+tothadenergy*0.1,1.);
    }
    if(thrust<=0.7 && fwdenergy/totenergy<0.01 ){
      fPFAL7B->Fill(totenergy+enu, 1.);
      std::cout << "Filling L7B : " << totenergy+enu << std::endl; 
    }
    if(thrust>0.6  && thrust<=0.7  )fPFA7->Fill(totenergy,1.);
    if(thrust>0.7  && thrust<=0.8  )fPFA8->Fill(totenergy,1.);
    if(thrust>0.8  && thrust<=0.9  )fPFA9->Fill(totenergy,1.);
    if(thrust>0.9                  )fPFA10->Fill(totenergy,1.); 
    if(thrust>0.9  && thrust<=0.925)fPFA11->Fill(totenergy,1.);
    if(thrust>0.925&& thrust<=0.95 )fPFA12->Fill(totenergy,1.);
    if(thrust>0.95 && thrust<=0.975)fPFA13->Fill(totenergy,1.);
    if(thrust>0.975                )fPFA14->Fill(totenergy,1.);
  } 



  bool _hcalOnly = false;
  if(eecaltot+ehcaltot>0){
    if(ehcaltot/(eecaltot+ehcaltot)>0.90)_hcalOnly=true;
  }
  if(_hcalOnly)fPFAhcal->Fill(totenergy);


  bool pass = false;
  if(totenergy<_eCutLow || totenergy>_eCutHigh)pass = true;
  if(thrust>_cosQQCut)pass = false;
  this->setReturnValue(pass);

  std::cout << " DONE " << std::endl;
  
  return;

}


void MyPFOAnalysis::ConfusionAnalysis(){ 

  Track* mcpToRecoTrack[1000];
  std::map<MCParticle*,unsigned int>pmcToiMcPFO;
  
  // Zero pointers from mcpfo to reconstructed track
  for(unsigned int imc=0;imc<_pMcPFOs.size();imc++)mcpToRecoTrack[imc]=NULL;	      

  float totenergy = 0;
  //  float fwdenergy = 0;
  float totgammaenergy = 0;
  //float tottrackenergy = 0;
  float tothadenergy = 0;
  float totmom[3]={0.,0.,0.};
  //float enu = 0;
  //float efw = 0.;
  //float ebad = 0.;
  float matches[1000];
  float totalmce[1000];
  float tt =0.;
  float tg =0.;
  float th =0.;
  float gt =0.;
  float gg =0.;
  float gh =0.;
  float ht =0.;
  float hg =0.;
  float hh =0.;

  // Loop over reconstructed PFOs
  for(unsigned int i=0;i<_pfovec.size();i++){
    totenergy += _pfovec[i]->getEnergy();
    totmom[0]+=_pfovec[i]->getMomentum()[0];
    totmom[1]+=_pfovec[i]->getMomentum()[1];
    totmom[2]+=_pfovec[i]->getMomentum()[2];
    TrackVec     tracks = _pfovec[i]->getTracks();
    ClusterVec clusters = _pfovec[i]->getClusters();
    if(_printing>0){
      std::cout << _pfovec[i]->getType() << " : " << tracks.size() << "," << clusters.size() << " -> " << _pfovec[i]->getEnergy() << std::endl;
    }
    if(_pfovec[i]->getType()!=22&&tracks.size()==0){
      tothadenergy += _pfovec[i]->getEnergy();
    }
    if(_pfovec[i]->getType()==22&&tracks.size()==0){
      totgammaenergy += _pfovec[i]->getEnergy();
    }
    for(unsigned int it=0;it<tracks.size();it++){
      LCObjectVec objectVec = _trackNavigator->getRelatedToObjects(tracks[it]);
      if (objectVec.size() > 0){ 
	MCParticle* navpart = dynamic_cast<MCParticle*>(objectVec[0]);
	for(unsigned int imc=0;imc<_pMcPFOs.size();imc++){
	  MCParticle* mcpart = _pMcPFOs[imc]->getMCParticle();
	  if(mcpart==navpart)mcpToRecoTrack[imc]=tracks[it];
	}
      }
    }
  }


  // find total recorded energy for MC particles
  for(unsigned int imc=0;imc<_pMcPFOs.size();imc++)totalmce[imc] = 0.;

  for(unsigned int i=0;i<_pfovec.size();i++){
    ClusterVec clusters = _pfovec[i]->getClusters();
    for(unsigned int icluster=0;icluster<clusters.size();++icluster){
      CalorimeterHitVec hits = clusters[icluster]->getCalorimeterHits();
      for(unsigned int ihit=0;ihit<hits.size();++ihit){
	LCObjectVec objectVec = _clusterNavigator->getRelatedToObjects(hits[ihit]);
	if (objectVec.size() > 0) {
	  SimCalorimeterHit * simhit =
	    dynamic_cast<SimCalorimeterHit*>(objectVec[0]);
	  if (simhit->getNMCContributions() > 0) {
	    MCParticle * part = simhit->getParticleCont(0);
	    unsigned int jmc = _myMcTree->GetMCPFOIndex(part);
	    if(jmc<_pMcPFOs.size())totalmce[jmc]+=hits[ihit]->getEnergy();
	  }
	}
      }
    }
  }




  // TRY TO MATCH TRUE PFOs with RECONSTRUCTED PFOs
  // loop over particles and match CALO hits to MC PFOs 
  for(unsigned int i=0;i<_pfovec.size();i++){
    bool isTrack=false;
    bool isGamma=false;
    bool isHad=false;
    float thisgg= 0.;
    float thishh= 0.;
    float thistg= 0.;
    float thisth= 0.;
    float thistot = 0.;
    
    ClusterVec clusters = _pfovec[i]->getClusters();
    TrackVec     tracks = _pfovec[i]->getTracks();
    if(tracks.size()>=1)isTrack=true;
    if(_pfovec[i]->getType()==22){
      isGamma=true;
      isTrack=false;
    }
    if(!isGamma && !isTrack)isHad = true;
    for(unsigned int imc=0;imc<_pMcPFOs.size();imc++)matches[imc]=0.;	      
    if(_printing>2)std::cout << " MCPFO " << _pfovec[i]->getType() <<  " " << tracks.size() << " : " << clusters.size() << std::endl;
    
    for(unsigned int icluster=0;icluster<clusters.size();++icluster){
      CalorimeterHitVec hits = clusters[icluster]->getCalorimeterHits();
      for(unsigned int ihit=0;ihit<hits.size();++ihit){
	LCObjectVec objectVec = _clusterNavigator->getRelatedToObjects(hits[ihit]);
	if (objectVec.size() > 0) {
	  SimCalorimeterHit * simhit = 
	    dynamic_cast<SimCalorimeterHit*>(objectVec[0]);
	  if (simhit->getNMCContributions() > 0) {
	    MCParticle * part = simhit->getParticleCont(0);
	    unsigned int jmc = _myMcTree->GetMCPFOIndex(part);
	    if(jmc<_pMcPFOs.size())matches[jmc]+=hits[ihit]->getEnergy();
	  }
	}
      }
      
      
      
    }
    
    for(unsigned int imc=0;imc<_pMcPFOs.size();imc++){
      if(matches[imc]>0.){
	thistot +=matches[imc];
	MCParticle* mcpart = _pMcPFOs[imc]->getMCParticle();
	if(_printing>2)std::cout << " ASSOC : " << imc << _pMcPFOs[imc]->getName() << " e = " << mcpart->getEnergy() << " -> " << matches[imc]<<std::endl;
	if(_pMcPFOs[imc]->isTrack()){
	  if(isTrack){
	    tt+=matches[imc];
	  }else{
	    if(mcpToRecoTrack[imc]==NULL){
	      if(_printing>2)std::cout << " xxx no reco track " << std::endl;
	      if(isGamma)hg+=matches[imc];
	      if(isHad){
		hh+=matches[imc];
	      }
	    }else{
	      if(_printing>2)std::cout << " *** reco track " << std::endl;
	      if(isGamma){
		tg+=matches[imc];
		thistg+=matches[imc];
	      }
	      if(isHad){
		th+=matches[imc];
		thisth+=matches[imc];
	      }
	    }
	  }
	}
	if(_pMcPFOs[imc]->isPhoton()){
	  if(isTrack)gt+=matches[imc];
	  if(isGamma){
	    gg+=matches[imc];
	    thisgg+=matches[imc];
	  }
	  if(isHad)gh+=matches[imc];
	}
	if(_pMcPFOs[imc]->isNeutralHadron()){
	  if(isTrack)ht+=matches[imc];
	  if(isGamma)hg+=matches[imc];
	  if(isHad){
	    hh+=matches[imc];
	    thishh += matches[imc];
	  }
	}
      }
    }
      
    // found a fragment
    bool badcluster = false;
    bool goodcluster = false;
    bool okcluster = false;
    float completeness = 0.;
    float purity=0.;
    if((thisth+thistg>0.5*thistot))badcluster = true;
    if((thishh>0.9*thistot))goodcluster = true;
    if((thisgg>0.9*thistot))goodcluster = true;
    int imost = -1;
    float most = 0.;
    float total = 0.;
    for(unsigned int imc=0;imc<_pMcPFOs.size();imc++){
      total += matches[imc];
      if(matches[imc]>most){
	most = matches[imc];
	imost = imc;
      }
    }
    if(imost>-1)completeness = most/totalmce[imost];
    purity = most/total;
    if(goodcluster){
      if(purity > 0.9){
	if(completeness<0.75){
	  goodcluster = false;
	  okcluster   = true;
	}
      }else{
	goodcluster = false;
      }
    } 
  }
  

  if(_printing>-1){
    std::cout << " TRACKS : " << tt << ":" << tg << ":" << th << std::endl;
    std::cout << " GAMMAS : " << gt << ":" << gg << ":" << gh << std::endl;
    std::cout << " N HADS : " << ht << ":" << hg << ":" << hh << std::endl;
  }

  fFragmentEnergy->Fill(tg+th);
  fConfusedHadron->Fill(ht);
  fConfusedPhoton->Fill(gt);
  fConfusion->Fill(tg+th-ht-gt);

  float tote = tt+tg+th+gt+gg+gh+ht+hg+hh;

  fConfused->Fill(0.5,0.5,ht/tote);
  fConfused->Fill(1.5,0.5,hg/tote);
  fConfused->Fill(2.5,0.5,hh/tote);
  fConfused->Fill(0.5,1.5,gt/tote);
  fConfused->Fill(1.5,1.5,gg/tote);
  fConfused->Fill(2.5,1.5,gh/tote);
  fConfused->Fill(0.5,2.5,tt/tote);
  fConfused->Fill(1.5,2.5,tg/tote);
  fConfused->Fill(2.5,2.5,th/tote);

  return;
}

void MyPFOAnalysis::check( LCEvent * evt ) { 
  // nothing to check here - could be used to fill checkplots in reconstruction processor
}


void MyPFOAnalysis::end(){ 
  
  std::cout << "MyPFOAnalysis::end()  " << name() 
 	    << " processed " << _nEvt << " events in " << _nRun << " runs "
 	    << std::endl ;

  TFile *hfile = new TFile(_rootFile.c_str(),"recreate");
  fPFA->TH1F::Write();
  fxproton->TH1F::Write();
  fMcPfoPdgCode->TH1F::Write();
  fPfoPdgCode->TH1F::Write();
  fPFAhcal->TH1F::Write();
  fPFAnu->TH1F::Write();
  fPFAnufwd->TH1F::Write();
  fPFAuds->TH1F::Write();
  fPFAudsHP20->TH1F::Write();
  fPFAudsHP10->TH1F::Write();
  fPFAudsHM10->TH1F::Write();
  fPFAudsHM20->TH1F::Write();
  fPFAcb->TH1F::Write();
  fPFAudscb->TH1F::Write();
  fPFA1->TH1F::Write();
  fPFA2->TH1F::Write();
  fPFA3->TH1F::Write();
  fPFA4->TH1F::Write();
  fPFA5->TH1F::Write();
  fPFA6->TH1F::Write();
  fPFA7->TH1F::Write();
  fPFAL7A->TH1F::Write();
  fPFAQQ->TH1F::Write();
  fPFAQQ8->TH1F::Write();
  fPFAL7Aud->TH1F::Write();
  fPFAL7As->TH1F::Write();
  fPFAL7Ac->TH1F::Write();
  fPFAL7Ab->TH1F::Write();
  fPFAL7B->TH1F::Write();
  fPFA8->TH1F::Write();
  fPFA9->TH1F::Write();
  fPFA10->TH1F::Write();
  fPFA11->TH1F::Write();
  fPFA12->TH1F::Write();
  fPFA13->TH1F::Write();
  fPFA14->TH1F::Write();
  fPFADMZ->TH1F::Write();
  fPFADMZOMZ->TH1F::Write();
  fPFADMZQQ8->TH1F::Write();
  fPFADMZOMZQQ8->TH1F::Write();
  fPFADMZ8->TH1F::Write();
  fPFADMZP8->TH1F::Write();


  fENeutralud->TH1F::Write();
  fETrackud->TH1F::Write();
  fEGammaud->TH1F::Write();
  fENeutrals->TH1F::Write();
  fETracks->TH1F::Write();
  fEGammas->TH1F::Write();

  fNPFO->TH1F::Write();
  fNTrackPFO->TH1F::Write();
  fNClusterPFO->TH1F::Write();

  fEcalMIP->TH1F::Write();
  fEcal1MIP->TH1F::Write();
  fHcalMIP->TH1F::Write();
  fEcalMIPcorr->TH1F::Write();
  fEcal1MIPcorr->TH1F::Write();
  fHcalMIPcorr->TH1F::Write();

  fFragmentEnergy->TH1F::Write();
  fConfusedHadron->TH1F::Write();
  fConfusedPhoton->TH1F::Write();
  fConfusion->TH1F::Write();
  fConfused->TH2F::Write();

  hfile->Close();
  delete hfile;

}

void MyPFOAnalysis::Clear(){

  _pfovec.clear();

  if(_clusterNavigator!=NULL)delete _clusterNavigator;
  if(_trackNavigator!=NULL)delete _trackNavigator;
  _clusterNavigator = NULL;
  _trackNavigator = NULL;

  if(_myMcTree!=NULL)delete _myMcTree;
  _myMcTree = NULL;

  if(_recoDisplay!=NULL){
    _recoDisplay->Clear();
    _recoDisplay->CleanUp();
  }

}
