// My Stuff
#include "PandoraPFACalibrator.h"
#include "MyRecoDisplay.h"
#include "MyMCTree.h"
// headers for io,std etc
#include <iostream>
// LCIO stuff
#include <EVENT/LCCollection.h>
#include <EVENT/LCRelation.h>
#include <EVENT/MCParticle.h>
#include <EVENT/Track.h>
#include "IMPL/ClusterImpl.h"
#include <EVENT/SimCalorimeterHit.h>
#include <EVENT/CalorimeterHit.h>
#include "IMPL/ReconstructedParticleImpl.h"
#include "IMPL/LCFlagImpl.h"
// GEAR
#include <gear/GEAR.h>
#include <gear/CalorimeterParameters.h>
#include <gear/TPCParameters.h>
#include <gear/PadRowLayout2D.h>
#include <gear/LayerLayout.h>
// Marlin stuff
#include <marlin/Global.h>
#include "UTIL/CellIDDecoder.h"
// root stuff for histograms
#include "TH1.h"
#include "TH2.h"
#include "TFile.h"

using namespace lcio ;
using namespace marlin ;

const float pi = acos(-1.0);
const float twopi = 2.0*pi;

PandoraPFACalibrator aPandoraPFACalibrator;

PandoraPFACalibrator::PandoraPFACalibrator() : Processor("PandoraPFACalibrator") {
  
  // modify processor description
  _description = "PandoraPFACalibrator for calibration of of PandoraPFA" ;
  

  // register steering parameters: name, description, class-variable, default value


  std::string rootFile;
  rootFile = "PandoraPFACalibrator.root";
  registerProcessorParameter( "RootFile" , 
			      "Name of the Track collection used for clustering"  ,
			      _rootFile,
			      rootFile ) ;

  std::string particleCollectionName;
  particleCollectionName = "PandoraPFOs";
  registerProcessorParameter( "ParticleCollectionName" , 
			      "Particle Collection Name "  ,
			      _particleCollectionName,
			      particleCollectionName);



  std::string clusterCollectionName;
  clusterCollectionName = "PandoraClusters";
  registerProcessorParameter( "ClusterCollectionName" , 
			      "Cluster Collection Name "  ,
			      _clusterCollectionName,
			      clusterCollectionName);


  std::vector<std::string>  ecalCollections  ;
  ecalCollections.push_back(std::string("ECAL"));
  registerInputCollections( LCIO::CALORIMETERHIT,
			   "ECALcollections" , 
			   "Name of the ECAL collection used to form clusters"  ,
			   _ecalCollections,
			   ecalCollections ) ;

  std::vector<std::string>  hcalCollections  ;
  hcalCollections.push_back(std::string("HCAL"));
  registerInputCollections( LCIO::CALORIMETERHIT,
			   "HCALcollections" , 
			   "Name of the HCAL collection used to form clusters"  ,
			   _hcalCollections,
			   hcalCollections ) ;

  std::vector<std::string>  bcalCollections  ;
  bcalCollections.push_back(std::string("BCAL"));
  registerInputCollections( LCIO::CALORIMETERHIT,
			   "BCALcollections" , 
			   "Name of the BCAL collection used to form clusters"  ,
			   _bcalCollections,
			   bcalCollections ) ;


  std::vector<std::string>  lhcalCollections  ;
  lhcalCollections.push_back(std::string("LHCAL"));
  registerInputCollections( LCIO::CALORIMETERHIT,
			   "LHCALcollections" , 
			   "Name of the LHCAL collection used to form clusters"  ,
			   _lhcalCollections,
			   lhcalCollections ) ;


  std::vector<std::string>  lcalCollections  ;
  lcalCollections.push_back(std::string("LCAL"));
  registerInputCollections( LCIO::CALORIMETERHIT,
			   "LCALcollections" , 
			   "Name of the LCAL collection used to form clusters"  ,
			   _lcalCollections,
			   lcalCollections ) ;


  //********************************************************************
  //**************************** Calibration ***************************
  //********************************************************************
  // Copy from Pandora PFA
  // Calibration constants
  // NOTE THE CALIBRATION SCHEME USED HERE IS NON-STANDARD
  // First hits in ECAL/HCAL are converted to normal incident MIP equivalents
  // Thresholds are applied at the MIP level
  // MIP equivalent signals are converted to EM or Hadronic energy 
  //   i.e. different calibrations for EM and hadronic showers

  registerProcessorParameter( "ECALMIPcalibration" , 
			      "Calibration from deposited ECAL energy to MIP"  ,
			      _ecalToMIP,
			       (float)200.0 ) ;

  registerProcessorParameter( "HCALMIPcalibration" , 
			      "Calibration from deposited HCAL energy to MIP"  ,
			      _hcalToMIP,
			       (float)35. ) ;

  registerProcessorParameter( "ECALThreshold" , 
			      "ECAL Threshold in MIPS"  ,
			      _ecalMIPThreshold,
			       (float)0.0 ) ;

  registerProcessorParameter( "HCALThreshold" , 
			      "HCAL Threshold in MIPS"  ,
			      _hcalMIPThreshold,
			       (float)0.0 ) ;


  registerProcessorParameter( "HCALEMMIPToGeV" , 
			      "Calibration from deposited HCAL MIP to EM energy"  ,
			      _hcalEMMIPToGeV,
			       (float)0.0233 ) ;

  registerProcessorParameter( "HCALHadMIPToGeV" , 
			      "Calibration from deposited HCAL MIP to Hadronic energy"  ,
			      _hcalHadMIPToGeV,
			       (float)0.02 ) ;

  registerProcessorParameter( "ECALEMMIPToGeV" , 
			      "Calibration from deposited ECAL MIP to EM energy"  ,
			      _ecalEMMIPToGeV,
			       (float)0.005076 ) ;

  registerProcessorParameter( "ECALHadMIPToGeV" , 
			      "Calibration from deposited ECAL MIP to Hadronic energy"  ,
			      _ecalHadMIPToGeV,
			       (float)0.005 ) ;

  CellIDDecoder<CalorimeterHit>::setDefaultEncoding("M:3,S-1:3,I:9,J:9,K-1:6");

}


void PandoraPFACalibrator::init() { 

  // usually a good idea to
  printParameters() ;

  _nRun = 0 ;
  _nEvt = 0 ;

  // PFA total energy
  fPFA      = new TH1F("fPFA", "total energy",1000, 0., 250.0);
  fPFAB      = new TH1F("fPFAB", "total energy barrel",1000, 0., 250.0);
  fPFAVsCosTheta      = new TH2F("fPFAVsCosTheta", "total energy vs CosTheta",500,0.,1.,1000, 0., 250.0);
  fPFAVsCosThetaR     = new TH2F("fPFAVsCosThetaR", "total energy vs CosThetaReco",500,0.,1.,1000, 0., 250.0);
  fPFAVsZCoG     = new TH2F("fPFAVsZCoG", "total energy vs zCog",600,0.,3000.,1000, 0., 200.0);
  fPFAVsCosThetaX      = new TH2F("fPFAVsCosThetaX", "total energy vs CosTheta",200,-1.,1.,1000, 0., 250.0);

  fXvsY      = new TH2F("fXvsY", "x vs y",1000,-2500.,2500.,1000, -2500., 2500.0);

  fPFAVsX1      = new TH2F("fPFAVsX1", "total energy vs x +z endcap",1000,-2000.,2000.,1000, 0., 250.0);
  fPFAVsX2      = new TH2F("fPFAVsX2", "total energy vs x -z endcap",1000,-2000.,2000.,1000, 0., 250.0);
  fPFAVsY1      = new TH2F("fPFAVsY1", "total energy vs y +z endcap",1000,-2000.,2000.,1000, 0., 250.0);
  fPFAVsY2      = new TH2F("fPFAVsY2", "total energy vs y -z endcap",1000,-2000.,2000.,1000, 0., 250.0);

  // PFA (hcal only)
  fPFAH     = new TH1F("fPFAH", "total energy HCAL only events",1000, 0., 250.0);

  // ECAL total energy
  fEcalEnergy      = new TH1F("fEcalEnergy", "total ecal energy",1000, 0., 250.0);
  fEcalHcalEnergy  = new TH2F("fEcalHcalEnergy", "ecal vs hcal energy",1000, 0., 250.0,1000,0.,250.);
  fHcalEnergy      = new TH1F("fHcalEnergy", "total hcal energy",1000, 0., 250.0);
  fLcalEnergy      = new TH1F("fLcalEnergy", "total lcal energy",1000, 0., 250.0);
  fCalEnergy       = new TH1F("fCalEnergy" , "total cal energy",1000, 0., 250.0);
  fCalEnergyH       = new TH1F("fCalEnergyH" , "total cal energy H",1000, 0., 250.0);
  fCalEnergyVsCosTheta  = new TH2F("fCalEnergyVsCosTheta" , "total cal energy vs cos theta",500,0.,1.,1000, 0., 250.0);
  fCalEnergyVsCosThetaR  = new TH2F("fCalEnergyVsCosThetaR" , "total cal energy vs cos theta reco",500,0.,1.,1000, 0., 250.0);
  fEnergyByLayer   = new TH1F("fEnergyByLayer", "ecal energy profile",100, 0., 100.0);

 
  // MIP Calibration

  fEcalMIP = new TH1F("fEcalMIP", "ecal MIP peak ",100, 0., 5.0);
  fEcal1MIP = new TH1F("fEcal1MIP", "ecal1 MIP peak ",100, 0., 5.0);
  fHcalMIP = new TH1F("fHcalMIP", "hcal MIP peak ",100, 0., 5.0);
  fEcalMIPcorr = new TH1F("fEcalMIPcorr", "ecal MIP peak ",100, 0., 5.0);
  fEcal1MIPcorr = new TH1F("fEcal1MIPcorr", "ecal1 MIP peak ",100, 0., 5.0);
  fHcalMIPcorr = new TH1F("fHcalMIPcorr", "hcal MIP peak ",100, 0., 5.0);


  fCosT = new TH1F("fCosT", "cosTheta ",100, 0., 1.0);
  fPhotonCosT = new TH1F("fPhotonCosT", "cosTheta Photon",100, 0., 1.0);



  _myMcTree= NULL;

}

void PandoraPFACalibrator::processRunHeader( LCRunHeader* run) { 

  _nRun++ ;
  _detectorName = run->getDetectorName();

  // determine detector (still some hard-coded parameters)
  std::cout << " DETECTOR : " << _detectorName << std::endl;

  // Calorimeter geometry from GEAR

  const gear::CalorimeterParameters& pEcalEndcap = Global::GEAR->getEcalEndcapParameters();
  //const gear::LayerLayout& ecalEndcapLayout = pEcalEndcap.getLayerLayout();
  _zOfEndcap = (float)pEcalEndcap.getExtent()[2];


} 


void PandoraPFACalibrator::processEvent( LCEvent * evt ) { 

  _nEvt ++ ;
  float ecalEnergy=0.;
  float hcalEnergy=0.;
  float lcalEnergy=0.;
  float bcalEnergy=0.;
  float lhcalEnergy=0.;
  float totEnergy=0.;
  float pfoEnergy=0.;


  // clear the event/display
  this->Clear();
 
  typedef const std::vector<std::string> StringVec ;
  StringVec* strVec = evt->getCollectionNames() ;
  for(StringVec::const_iterator name=strVec->begin(); name!=strVec->end(); name++){    
    LCCollection* col = evt->getCollection(*name);
    if (col->getTypeName() == LCIO::MCPARTICLE ) {
      _myMcTree = new MyMCTree(col);
      _myMcTree->PerfectPFO(1);
      MCParticle* part =  dynamic_cast<MCParticle*>( col->getElementAt(0) ) ;
      float x =  part->getMomentum()[0];
      float y =  part->getMomentum()[1];
      float z =  part->getMomentum()[2];
      float p =  sqrt(x*x+y*y+z*z);
      _cosTheta = 0;
      
      _x = 0.;
      _y = 0.;
      if(p>0){
	_cosThetaX = z/sqrt(x*x+y*y+z*z);
	_cosTheta = fabs(z)/sqrt(x*x+y*y+z*z);
	if(fabs(z)>0){
	  _x = _zOfEndcap*x/z;
	  _y = _zOfEndcap*y/z;
	}
      }
      float energy = part->getEnergy();
      std::cout << " First MC Particle : " << energy << " GeV " << part->getPDG() << " at " << x << "," << y << "," << z << " (|cosTheta| = " << _cosTheta << ")" << std::endl;
    }
  }  

  typedef const std::vector<std::string> StringVec ;
  
  // read ECAL hits
  _zmean = 0;
  float nz = 0;
  for (unsigned int i=0; i < _ecalCollections.size(); ++i) {
    try {
      LCCollection * col = evt->getCollection( _ecalCollections.at(i).c_str() );
      if (col != 0) {
	CellIDDecoder<CalorimeterHit> id( col ) ;
	int nelem = col->getNumberOfElements();
	for(int ihit=0;ihit<nelem;ihit++){
	  CalorimeterHit * hit = dynamic_cast<CalorimeterHit*>(col->getElementAt(ihit));
	  float hitEnergy = hit->getEnergy();
	  ecalEnergy+=hitEnergy;
	  int layerNumber = id( hit )["K-1"] + 1 ;
	  fEnergyByLayer->Fill(layerNumber,hitEnergy);
	  float scale = 1.0;
	  if(hit->getType()==1)scale =2.0;
	  float energyInMips=0;
	  float x = hit->getPosition()[0];
	  float y = hit->getPosition()[1];
	  float z = hit->getPosition()[2];
	  float r = sqrt(x*x+y*y+z*z);
	  _zmean += z;
	  nz    +=1.;
	  float correction  =1.;
	  if(fabs(z)<_zOfEndcap){
	    correction = r/sqrt(x*x+y*y);
	  }else{
	    correction = r/z;
	  }
	  energyInMips = hit->getEnergy()*_ecalToMIP/scale;
	  fEcalMIP->Fill(energyInMips);
	  fEcalMIPcorr->Fill(energyInMips/correction);
	}
	std::cout << " ECAL hits : " << nelem << " energy = " << ecalEnergy << std::endl;
      }
    }
    catch(DataNotAvailableException &e){
      std::cout << "No Collection : " <<  _ecalCollections[i] << std::endl; 	
    }
  }
  if(nz>0)_zmean=_zmean/nz;


  // read HCAL hits
  for (unsigned int i=0; i < _hcalCollections.size(); ++i) {
    try {
      LCCollection * col = evt->getCollection( _hcalCollections.at(i).c_str() );
      if (col != 0) {
	int nelem = col->getNumberOfElements();
	for(int ihit=0;ihit<nelem;ihit++){
	  CalorimeterHit * hit = dynamic_cast<CalorimeterHit*>(col->getElementAt(ihit));
	  hcalEnergy+=hit->getEnergy();
	  float energyInMips=0;
	  float x = hit->getPosition()[0];
	  float y = hit->getPosition()[1];
	  float z = hit->getPosition()[2];
	  float r = sqrt(x*x+y*y+z*z);
	  float correction  =1.;
	  if(fabs(z)<_zOfEndcap){
	    correction = r/sqrt(x*x+y*y);
	  }else{
	    correction = r/z;
	  }
	  energyInMips = hit->getEnergy()*_hcalToMIP;
	  fHcalMIP->Fill(energyInMips);
	  fHcalMIPcorr->Fill(energyInMips/correction);
	}
	std::cout << " HCAL hits : " << nelem << " energy = " << hcalEnergy << std::endl;
      }
    }
    catch(DataNotAvailableException &e){
      std::cout << "No Collection : " <<  _hcalCollections[i] << std::endl; 	
    }
  }


  // read LCAL hits
  for (unsigned int i=0; i < _lcalCollections.size(); ++i) {
    try {
      LCCollection * col = evt->getCollection( _lcalCollections.at(i).c_str() );
      if (col != 0) {
	int nelem = col->getNumberOfElements();
	for(int ihit=0;ihit<nelem;ihit++){
	  CalorimeterHit * hit = dynamic_cast<CalorimeterHit*>(col->getElementAt(ihit));
	  lcalEnergy+=hit->getEnergy();
	}
	std::cout << " LCAL hits : " << nelem << " energy = " << lcalEnergy << std::endl;
      }
    }
    catch(DataNotAvailableException &e){
      std::cout << "No Collection : " <<  _lcalCollections[i] << std::endl; 	
    }
  }


  // read BCAL hits
  for (unsigned int i=0; i < _bcalCollections.size(); ++i) {
    try {
      LCCollection * col = evt->getCollection( _bcalCollections.at(i).c_str() );
      if (col != 0) {
	int nelem = col->getNumberOfElements();
	for(int ihit=0;ihit<nelem;ihit++){
	  CalorimeterHit * hit = dynamic_cast<CalorimeterHit*>(col->getElementAt(ihit));
	  bcalEnergy+=hit->getEnergy();
	}
	std::cout << " BCAL hits : " << nelem << " energy = " << bcalEnergy << std::endl;
      }
    }

    catch(DataNotAvailableException &e){
      std::cout << "No Collection : " <<  _bcalCollections[i] << std::endl; 	
    }
  }



  // read BCAL hits
  for (unsigned int i=0; i < _lhcalCollections.size(); ++i) {
    try {
      LCCollection * col = evt->getCollection( _lhcalCollections.at(i).c_str() );
      if (col != 0) {
	int nelem = col->getNumberOfElements();
	for(int ihit=0;ihit<nelem;ihit++){
	  CalorimeterHit * hit = dynamic_cast<CalorimeterHit*>(col->getElementAt(ihit));
	  lhcalEnergy+=hit->getEnergy();
	}
	std::cout << " LHCAL hits : " << nelem << " energy = " << lhcalEnergy << std::endl;
      }
    }

    catch(DataNotAvailableException &e){
      std::cout << "No Collection : " <<  _lhcalCollections[i] << std::endl; 	
    }
  }



  totEnergy = ecalEnergy+hcalEnergy+lcalEnergy + bcalEnergy + lhcalEnergy;
  std::cout << " Total Calorimetric Energy : " << totEnergy << std::endl;

  if(_cosTheta<0.95){
    fEcalEnergy->Fill(ecalEnergy,1.);
    fEcalHcalEnergy->Fill(ecalEnergy*_ecalHadMIPToGeV/_ecalEMMIPToGeV,hcalEnergy,1.);
    fHcalEnergy->Fill(hcalEnergy,1.);
    fLcalEnergy->Fill(lcalEnergy,1.);
    fCalEnergy->Fill(totEnergy,1.);
    if(totEnergy>0 && hcalEnergy/totEnergy>0.95)fCalEnergyH->Fill(totEnergy);
  }
  fCalEnergyVsCosTheta->Fill(_cosTheta,totEnergy,1.);

  // read reconstructed PFOs
  try {
    LCCollection * col = evt->getCollection( _particleCollectionName.c_str() );
    if (col != 0) {
      int nelem = col->getNumberOfElements();
      bool check=true;
      float photonE = 0;
      float p[3] = {0.,0.,0.};
      _zCoG=0.;
      float esum =0;
      for(int j=0;j<nelem;j++){
	ReconstructedParticle* recoPart = dynamic_cast<ReconstructedParticle*>(col->getElementAt(j));
	pfoEnergy+=recoPart->getEnergy();
	TrackVec tracks = recoPart->getTracks();
	if(tracks.size()!=0)check=false;
	if(recoPart->getType()==22)photonE+=recoPart->getEnergy();
	p[0] += recoPart->getMomentum()[0];
	p[1] += recoPart->getMomentum()[1];
	p[2] += recoPart->getMomentum()[2];
	ClusterVec clusters = recoPart->getClusters();
	for(unsigned int icluster = 0; icluster < clusters.size(); icluster++){
	  _zCoG+= clusters[icluster]->getPosition()[2]*clusters[icluster]->getEnergy();
	  esum += clusters[icluster]->getEnergy();
	}
      }
      if(esum>0){
	_zCoG=_zCoG/esum;
      }else{
	_zCoG = 9999.;
      }
      _cosThetaR = fabs(p[2])/sqrt(p[0]*p[0]+p[1]*p[1]+p[2]*p[2]);
      std::cout << " PFO objects : " << nelem << " energy = " << pfoEnergy << std::endl;

      if(check){
	fCosT->Fill(_cosTheta,1.);
	if(pfoEnergy>0 && photonE>0.5*pfoEnergy)fPhotonCosT->Fill(_cosTheta,1.);
      }

      fPFAVsCosTheta->Fill(_cosTheta,pfoEnergy,1.);
      fPFAVsCosThetaR->Fill(_cosThetaR,pfoEnergy,1.);
      fPFAVsZCoG->Fill(fabs(_zCoG),pfoEnergy,1.);
      fPFAVsCosThetaX->Fill(_cosThetaX,pfoEnergy,1.);
      if(_cosTheta<0.95){
	fPFA->Fill(pfoEnergy,1.);
	if(_cosTheta<0.7)fPFAB->Fill(pfoEnergy,1.);
	if(totEnergy>0 && hcalEnergy/totEnergy>0.95)fPFAH->Fill(pfoEnergy,1.);
      }
      fXvsY->Fill(_x,_y,1.);

      if(_zmean>_zOfEndcap){
	fPFAVsX1->Fill(_x,totEnergy,1.);
	fPFAVsY1->Fill(_y,totEnergy,1.);
      }
      if(_zmean<-_zOfEndcap){
	fPFAVsX2->Fill(_x,totEnergy,1.);
	fPFAVsY2->Fill(_y,totEnergy,1.);
      }
 
    }
  }
  catch(DataNotAvailableException &e){
    std::cout << "No Collection : " <<  _particleCollectionName << std::endl; 	
  }


  fCalEnergyVsCosThetaR->Fill(_cosThetaR,totEnergy,1.);

}



void PandoraPFACalibrator::check( LCEvent * evt ) { 
  // nothing to check here - could be used to fill checkplots in reconstruction processor
}


void PandoraPFACalibrator::end(){ 
  
  std::cout << "PandoraPFACalibrator::end()  " << name() 
 	    << " processed " << _nEvt << " events in " << _nRun << " runs "
 	    << std::endl ;

  TFile *hfile = new TFile(_rootFile.c_str(),"recreate");
  fPFA->TH1F::Write();
  fCosT->TH1F::Write();
  fPhotonCosT->TH1F::Write();
  fPFAB->TH1F::Write();
  fPFAVsCosTheta->TH2F::Write();
  fPFAVsCosThetaR->TH2F::Write();
  fPFAVsZCoG->TH2F::Write();
  fPFAVsCosThetaX->TH2F::Write();
  fPFAH->TH1F::Write();
  fEcalMIP->TH1F::Write();
  fEcal1MIP->TH1F::Write();
  fHcalMIP->TH1F::Write();
  fEcalMIPcorr->TH1F::Write();
  fEcal1MIPcorr->TH1F::Write();
  fHcalMIPcorr->TH1F::Write();
  fEnergyByLayer->TH1F::Write();
  fEcalEnergy->TH1F::Write();
  fEcalHcalEnergy->TH2F::Write();
  fHcalEnergy->TH1F::Write();
  fLcalEnergy->TH1F::Write();
  fCalEnergy->TH1F::Write();
  fCalEnergyH->TH1F::Write();
  fXvsY->TH2F::Write();
  fPFAVsX1->TH2F::Write();
  fPFAVsX2->TH2F::Write();
  fPFAVsY1->TH2F::Write();
  fPFAVsY2->TH2F::Write();
  fCalEnergyVsCosTheta->TH2F::Write();
  fCalEnergyVsCosThetaR->TH2F::Write();
  hfile->Close();
  delete hfile;

}

void PandoraPFACalibrator::Clear(){

  _pfovec.clear();
  if(_myMcTree!=NULL)delete _myMcTree;
  _myMcTree = NULL;


}
