#ifndef PixelReco_Cluster_H
#define PixelReco_Cluster_H

#include "PixelPos.h"
#include "Digi.h"

#include <vector>

namespace SICLU {

/** ORCA, CMS software, 2006
 *
 *  Contains all the topological information of pixel clusters:
 *  charge, global size, size and the barycenter in x and y
 *  local directions. It builds a vector of Pixel (which is
 *  an internal public class) and a container of channels. 
 *
 *  Adapted in MarlinReco : S Shulga 2007/06/06
 *  @version $Id: CPEFromDetPosition.h, v 1 2007/06/06 shulga
 */

class Cluster {
public:

  Cluster( const PixelPos& pix, float adc) ;

  void add( const PixelPos& pix, float adc) ;

  /// Analog linear average position (barycenter) 
  float x() const { return theSumX / theCharge ;  }
  float y() const { return theSumY / theCharge ;}

  /// Return number of pixels.
  int size() const { return thePixels.size() ; }

  /// Return cluster dimension in the x direction.
  int sizeX() const {return theMaxPixelRow - theMinPixelRow +1 ; }

  /// Return cluster dimension in the y direction.
  int sizeY() const {return theMaxPixelCol - theMinPixelCol +1 ; }

  /// Detect clusters at the edge of the detector
  bool edgeHitX() const ;
  bool edgeHitY() const ;

  inline float charge() const { return theCharge ; } // Return total cluster charge.
  inline int minPixelRow() const { return theMinPixelRow ; } // The min x index.
  inline int maxPixelRow() const { return theMaxPixelRow ; } // The max x index.
  inline int minPixelCol() const { return theMinPixelCol ; } // The min y index.
  inline int maxPixelCol() const { return theMaxPixelCol ; } // The max y index.

  class Pixel {
  public:
    Pixel( float pix_x, float pix_y, float pix_adc) :
      x(pix_x), y(pix_y), adc(pix_adc) {}
    float x;
    float y;
    float adc;
  } ; // class Pixel.

  const vector<Pixel>& pixels() const { return thePixels;}

  typedef vector<SIUTIL::Digi::ChannelType>    ChannelContainer;
  ChannelContainer  channels() const ;

private:

  vector<Pixel> thePixels;

  float theSumX ;  // Sum of charge weighted pixel positions.
  float theSumY ;
  float theCharge ;  // Total charge
  int theMinPixelRow ; // Minimum pixel index in the x direction (low edge).
  int theMaxPixelRow ; // Maximum pixel index in the x direction (top edge).
  int theMinPixelCol ; // Minimum pixel index in the y direction (left edge).
  int theMaxPixelCol ; // Maximum pixel index in the y direction (right edge).

} ; // class Cluster.
 
}  // namespace SICLU.

#endif




