#include "CPEFromDetPosition.h"
#include "DetGeometry/GeomVector/LocalVector.h"

const float PI = 3.141593;
const float degsPerRad = 57.29578;

//using namespace PixelReco;

namespace SICLU{

CPEFromDetPosition::CPEFromDetPosition( const SIDETUNIT::DetUnit& det )
           : theDet(&det) {
  thePart = barrel ;
  theDetZ = theDet->get_ZdetUnit_refPoint() ;
  theDetR = theDet->get_RdetUnit_refPoint() ;

  //define parameters for chargewidth calculation

  theThickness = theDet->getThickness() ;

  theNumOfRow = theDet->topology().nrows() ;
  theNumOfCol = theDet->topology().ncolumns() ;

  pair<float,float> pitchxy = theDet->topology().pitch() ;

  thePitchX = pitchxy.first ;
  thePitchY = pitchxy.second ;

  // Find offset
  SIUTIL::MeasurementPoint  offset = 
    theDet->topology().measurementPosition( SIUTIL::LocalPoint( 0., 0. ) ) ;

  theOffsetX = offset.x();
  theOffsetY = offset.y();

  // Find if the E field is flipped
  theSign = 1;
  if ( isFlipped() ) theSign = -1;

  theLShift = lorentzShift() ;

//#ifdef _DEBUG_CPEFROMDETPOSITION_
#ifdef _DEBUG_1_
  cout << "c-tor CPEFromDetPosition: ***** PIXEL LAYOUT *****" << endl;
  cout << " theThickness = " << theThickness << endl;
  cout << " thePitchX  = " << thePitchX << endl;
  cout << " thePitchY  = " << thePitchY << endl;
  cout << " theOffsetX  = " << theOffsetX << endl;
  cout << " theOffsetY  = " << theOffsetY << endl;
#endif

}



SIUTIL::MeasurementError 
CPEFromDetPosition::measurementError( const Cluster & cluster ) const {
  return theDet->topology().measurementError( localPosition( cluster ), 
					      localError( cluster ) ) ;
}




SIUTIL::LocalError  
CPEFromDetPosition::localError( const Cluster & cluster ) const {
  int sizex = cluster.sizeX() ;
  int sizey = cluster.sizeY() ;
  bool edgex = (cluster.edgeHitX()) || (cluster.maxPixelRow()> theNumOfRow); 
  bool edgey = (cluster.edgeHitY()) || (cluster.maxPixelCol() > theNumOfCol); 
  return SIUTIL::LocalError( err2X(edgex, sizex), 0, err2Y(edgey, sizey) );
}




SIUTIL::MeasurementPoint 
CPEFromDetPosition::measurementPosition( const Cluster & cluster ) const {

  return SIUTIL::MeasurementPoint( xpos( cluster ) - theLShift, 
  			   ypos( cluster ) ) ;
}





SIUTIL::LocalPoint
CPEFromDetPosition::localPosition( const Cluster & cluster ) const {
  return theDet->topology().localPosition( measurementPosition(cluster) ) ; 
}





float CPEFromDetPosition::err2X(bool & edgex, int & sizex) const {
  // Assign maximum error Pitch = 0.0150 cm = 150 mu
  //  if edge cluster the maximum error is assigned: Pitch/sqrt(12)=43mu
  //  float xerr = 0.0043; [cm]

  float xerr = thePitchX/3.464 ; // [mm]

  if (!edgex){

   if ( thePitchX < 0.021 ) {     //[mm] 20mu pixel size

      if ( thePart == barrel ) {xerr = 0.01698 ;
      } else { //forward
	if ( sizex == 1)       xerr = 0.01 ;     //[mm]
	else if ( sizex == 2)  xerr = (0.05351 - atan(fabs(theDetZ/theDetR)) * 0.03291); //[mm] 
	else xerr = (0.03094 - atan(fabs(theDetZ/theDetR)) * 0.01854); //[mm] 
      } // if ( thePart.

   } else if ( thePitchX < 0.026 ) {     //[mm] 25mu pixel size

                                                   // p4,s5,c5  p4,s5,c6 
      if ( thePart == barrel ) { xerr = 0.0211*0.2032*0.9862   *0.9777 ;
      } else { //forward
	if ( sizex == 1)       xerr = 0.01 ;     //[mm]
	else if ( sizex == 2)  xerr = (0.05351 - atan(fabs(theDetZ/theDetR)) * 0.03291); //[mm] 
	else xerr = (0.03094 - atan(fabs(theDetZ/theDetR)) * 0.01854); //[mm] 
	
      } // if ( thePart.

   } else if ( thePitchX < 0.051 ) {     //[mm] 50mu pixel size
                                                      // 4-5-6
      if ( thePart == barrel ) { xerr = 0.0211*0.495 * 1.018 * 0.9683;
      } else { //forward
	if ( sizex == 1)       xerr = 0.01 ;     //[mm]
	else if ( sizex == 2)  xerr = (0.05351 - atan(fabs(theDetZ/theDetR)) * 0.03291); //[mm] 
	else xerr = (0.03094 - atan(fabs(theDetZ/theDetR)) * 0.01854); //[mm] 
	
      } // if ( thePart.

   }
   else if ( thePitchX < 0.101 ) {     //[mm] 100mu pixel size

      if ( thePart == barrel ) { xerr = 0.029 * 0.7795 ;
      } else { //forward
	if ( sizex == 1)       xerr = 0.01 ;     //[mm]
	else if ( sizex == 2)  xerr = (0.05351 - atan(fabs(theDetZ/theDetR)) * 0.03291); //[mm] 
	else xerr = (0.03094 - atan(fabs(theDetZ/theDetR)) * 0.01854); //[mm] 
	
      } // if ( thePart.

    } else if ( thePitchX < 0.151 ) {  //[mm] 150 mu pixel size
 
      if (thePart == barrel) { xerr = 0.036 ; //* 1.024 ;
      } else { //forward
	if ( sizex == 1 ) xerr = (-0.0385 + atan(fabs(theDetZ/theDetR)) * 0.0407);//[mm]
	else xerr = ( 0.0366 - atan(fabs(theDetZ/theDetR)) * 0.0209 ) ;  //[mm]
      } // if (thePart.
      
    } // if ( fabs( thePitchX.

  } // if (!edgex).

  return xerr*xerr;

} // err2X.






float CPEFromDetPosition::err2Y(bool& edgey, int& sizey) const {
// Assign maximum error
// if edge cluster the maximum error is assigned: Pitch/sqrt(12)=43mu
//  float yerr = 0.0043; [cm]
  float yerr = thePitchY/3.464;  // [mm]
  if (!edgey){

   if ( thePitchY < 0.021 ) {     //[mm] 20mu pixel size

    if (thePart == barrel) { yerr = 0.01808;
    } else { //forward
      if ( sizey == 1) yerr = 0.022;  // [mm] 22mu
      else             yerr = 0.008;  // [mm] 7mu
    }

   } else if ( thePitchY < 0.026 ) {     //[mm] 100mu pixel size
                                             // p4,s5,c5  p4,s5,c6 
    if (thePart == barrel) {yerr = 0.007*0.5725*0.8697    *0.8675 ;
    } else { //forward
      if ( sizey == 1) yerr = 0.022;  // [mm] 22mu
      else             yerr = 0.008;  // [mm] 7mu
    }

   } else if ( thePitchY < 0.051 ) {     //[mm] 50mu pixel size

    if (thePart == barrel) {yerr = 0.007* 1.489 ; 
    } else { //forward
      if ( sizey == 1) yerr = 0.022;  // [mm] 22mu
      else             yerr = 0.008;  // [mm] 7mu
    }

   } else if ( thePitchY < 0.101 ) {     //[mm] 100mu pixel size

    if (thePart == barrel) { yerr = 0.030 * 0.7513;
    } else { //forward
      if ( sizey == 1) yerr = 0.022;  // [mm] 22mu
      else             yerr = 0.008;  // [mm] 7mu
    }

   } else if ( thePitchY < 0.151 ) {

      if (thePart == barrel) {yerr = 0.0376 *0.965;
      } else { //forward
       if ( sizey == 1) yerr = 0.022;  // [mm] 22mu
       else             yerr = 0.008;  // [mm] 7mu
      }
   
   }

  } // if (!edgey.

  return yerr*yerr;
}





float CPEFromDetPosition::xpos(const Cluster& cluster) const {

  float xcluster = 0 ;

  int size = cluster.sizeX() ;

  const vector<Cluster::Pixel>& pixelsVec = cluster.pixels();

  float baryc = cluster.x() ;

  if (size == 1) // the middle of only one pixel is equivalent to the baryc.

    xcluster = baryc;

  else {

    //calculate center
    float xmin = float( cluster.minPixelRow() ) + 0.5 ;
    float xmax = float( cluster.maxPixelRow() ) + 0.5 ;
    float xcenter = ( xmin + xmax ) / 2;

    vector<float> xChargeVec = xCharge(pixelsVec, xmin, xmax) ;
    float q1 = xChargeVec[0] ;
    float q2 = xChargeVec[1] ;
    float chargeWX = chargeWidthX() + theSign * geomCorrection() * xcenter;
    float effchargeWX = fabs(chargeWX) - ( float(size) - 2 ) ;

    // truncated charge width only if it greather than the cluster size
    if ( fabs( effchargeWX ) > 2 ) effchargeWX = 1 ;

    xcluster = xcenter + ( q2 - q1 ) * effchargeWX / ( q1 + q2 ) / 2. ;


    float alpha = estimatedAlphaForBarrel( xcenter ) ;

    if (alpha < 1.53) {
      float etashift = 0 ;
      float charatio = q1/(q1 + q2) ;
      etashift = theEtaFunc.xEtaShift(size, thePitchX, 
				      charatio, alpha ) ;
      xcluster = xcluster - etashift;
    }

  } // if (size.

  return xcluster;

} // xpos.






float CPEFromDetPosition::ypos(const Cluster& cluster) const
{
  float ycluster = 0;
  const vector<Cluster::Pixel>& pixelsVec = cluster.pixels();
  int size = cluster.sizeY();
  float baryc = cluster.y();
  if (size == 1) {
    ycluster = baryc;
  } else if (size < 4) {

    // Calculate center
    float ymin = float(cluster.minPixelCol()) + 0.5;
    float ymax = float(cluster.maxPixelCol()) + 0.5;
    float ycenter = ( ymin + ymax ) / 2;

    //calculate charge width
    float chargeWY = chargeWidthY() + geomCorrection() * ycenter;
    float effchargeWY = fabs(chargeWY) - (float(size)-2);

    // truncate charge width when it is > 2
    if ( (effchargeWY < 0) || (effchargeWY > 1.) ) effchargeWY = 1;

    //calculate charge of first, last and inner pixels of cluster
    vector<float> yChargeVec = yCharge(pixelsVec, ymin, ymax);
    float q1 = yChargeVec[0];
    float q2 = yChargeVec[1];
    // float qm = yChargeVec[2];
    // float charatio = q1/(q1+q2);

    ycluster = ycenter + (q2-q1) * effchargeWY / (q1+q2) / 2.;

  } else {    //  Use always the edge method
    // Calculate center
    float ymin = float(cluster.minPixelCol()) + 0.5;
    float ymax = float(cluster.maxPixelCol()) + 0.5;
    float ycenter = ( ymin + ymax ) / 2;
    
    //calculate charge of first, last and inner pixels of cluster
    vector<float> yChargeVec = yCharge(pixelsVec, ymin, ymax);
    float q1 = yChargeVec[0];
    float q2 = yChargeVec[1];
    
    float shift = (q2 - q1) / (q2 + q1) / 2.; // Single edge 

    ycluster = ycenter + shift;

  }
  return ycluster;
}







//SSh dummy:
bool CPEFromDetPosition::isFlipped() const {
  return false;
}


/*//SSh

bool CPEFromDetPosition::isFlipped() const
{
  float tmp1 = theDet->toGlobal( Local3DPoint(0.,0.,0.) ).perp();
  float tmp2 = theDet->toGlobal( Local3DPoint(0.,0.,1.) ).perp();
  if ( tmp2<tmp1 ) return true;
  else return false;
}

*/

float CPEFromDetPosition::chargeWidthX() const {
  float chargeW = 0 ;
  float lorentzWidth = 2 * theLShift ;
  if (thePart == barrel) {
    // Redefine the charge width to include the offset
    chargeW = lorentzWidth - theSign * geomCorrection() * theOffsetX;
  } else { // forward
    chargeW = fabs(lorentzWidth) 
                + theThickness * fabs(theDetR/theDetZ) / thePitchX;
  }

  return chargeW;

} // chargeWidthX.




float CPEFromDetPosition::chargeWidthY() const {
  float chargeW = 0;  
  if (thePart == barrel) {
    chargeW =   theThickness * fabs(theDetZ/theDetR) / thePitchY;
    chargeW -= (geomCorrection() * theOffsetY);
  } else { //forward
    // Width comes from geometry only, fixed by the tilt angle
   chargeW = theThickness * tan(20./degsPerRad) / thePitchY; 
  }
  return chargeW;

} // chargeWidthY.




float CPEFromDetPosition::geomCorrection() const { 
  //@@ the geometrical correction are calculated only
  //@@ for the barrel part (am I right?) 
  if (thePart == forward) return 0;
  else return theThickness / theDetR;
}


float CPEFromDetPosition::lorentzShift() const {
  // Implement only the x direction shift for now (OK for barrel)
  SIUTIL::LocalVector dir = SIUTIL::LocalVector( (theDet->driftDirection()).x(), 
  				 (theDet->driftDirection()).y(),
				 (theDet->driftDirection()).z() ) ;
  // max shift in cm 
  float xdrift = dir.x()/dir.z() * theThickness;  
  // express the shift in units of pitch, 
  // divide by 2 to get the average correction 
  // (SSh: max=xdrift, min=0, average=xdrift/2)
  float lshift = xdrift / thePitchX / 2; 

#ifdef _DEBUG_1_
  cout << "Lorentz Drift = " << xdrift << endl;
  cout << "X Drift = " << dir.x() << endl;
  cout << "Z Drift = " << dir.z() << endl;
#endif 

  return lshift ;

}




float CPEFromDetPosition::chaWidth2X( const float & centerx ) const {
  float chargeWX = chargeWidthX() + theSign * geomCorrection() * centerx;
  if ( chargeWX > 1. || chargeWX <= 0. ) chargeWX = 1. ;
  return chargeWX ;
}





float CPEFromDetPosition::estimatedAlphaForBarrel(const float& centerx) const {
  float tanalpha = theSign * ( centerx - theOffsetX ) / theDetR * thePitchX ;
  return PI/2 - atan( tanalpha ) ;
}





vector<float> 
CPEFromDetPosition::xCharge(const vector<Cluster::Pixel>& pixelsVec, 
			    const float& xmin, const float& xmax ) const {
  vector<float> charge ;

  //calculate charge in the first and last pixel in y
  // and the total (SSh: inner?, total!) cluster charge
  float q1 = 0, q2 = 0, qm = 0 ;
  for (int i = 0 ;  i < int( pixelsVec.size() ) ; i++ ) {
    if      ( (int)( pixelsVec[i].x - xmin )== 0 ) q1 += pixelsVec[i].adc ;
    else if ( (int)( pixelsVec[i].x - xmax )== 0 ) q2 += pixelsVec[i].adc ;
    else qm += pixelsVec[i].adc;
  }
  charge.clear();
  charge.push_back(q1); 
  charge.push_back(q2); 
  charge.push_back(qm);
  return charge;

} // xCharge. 




vector<float> 
CPEFromDetPosition::yCharge(const vector<Cluster::Pixel>& pixelsVec,
			    const float& ymin, const float& ymax) const {
  vector<float> charge ;

  //calculate charge in the first and last pixel in y
  // and the inner cluster charge
  float q1 = 0, q2 = 0, qm = 0 ;
  for (int i = 0;  i < int(pixelsVec.size()); i++) {
    if      ( (int)( pixelsVec[i].y - ymin ) == 0 ) q1 += pixelsVec[i].adc;
    else if ( (int)( pixelsVec[i].y - ymax ) == 0 ) q2 += pixelsVec[i].adc;
    else if (pixelsVec[i].y < ymax && 
	     pixelsVec[i].y > ymin ) qm += pixelsVec[i].adc;
  }
  charge.clear();
  charge.push_back(q1); 
  charge.push_back(q2); 
  charge.push_back(qm);

  return charge;

} // yCharge.


} // namespace SICLU.

