#include "IMPL/LCCollectionVec.h"

#include "SiClusterizer.h"
#include "ClusterParameterEstimator.h"
#include "Pixel/BarrelDetUnit.h"

namespace SICLU {


/**  General logic of SiTrkDetUnitClusterizer:
 *
 *   The clusterization is performed on a matrix with size
 *   equal to the size of the pixel detector, each cell containing 
 *   the ADC count of the corresponding pixel.
 *   The matrix is reset after each clusterization.
 *
 *   The search starts from seed pixels, i.e. pixels with sufficiently
 *   large amplitudes, found at the time of filling of the matrix
 *   and stored in a ...
 */


SiTrkDetUnitClusterizer::SiTrkDetUnitClusterizer() 
  : theNumOfRows(0), theNumOfCols(0), bufferAlreadySet(false) {

  theEstimatorFactory = new ClusterParameterEstimatorFactory ;

#ifdef _DEBUG_CLUSTER_
cout << " Constructor SiTrkDetUnitClusterizer ended. "<< endl;
#endif

 // to define encoder
 // encoder need to define raw hits for TrackerHit collection
 // rawVec itself does not used 
 LCCollectionVec * rawVec = new LCCollectionVec( LCIO::TRACKERRAWDATA );

 _encoderTrackerRawData 
   = new CellIDEncoder<TrackerRawDataImpl>( SIDETUNIT::BarrelDetUnit::_cellIDEncoding, rawVec ) ;
 
 _bf64 = new UTIL::BitField64( SIDETUNIT::BarrelDetUnit::_cellIDEncoding ) ;
 
} // c-tor SiTrkDetUnitClusterizer.




SiTrkDetUnitClusterizer::~SiTrkDetUnitClusterizer() {  
  delete theEstimatorFactory ;
  delete _encoderTrackerRawData ;
 // delete _bf64 ;
}




//-----------------------------------------------------------------
bool SiTrkDetUnitClusterizer::setup( SIDETUNIT::DetUnit * readout) throw(string) {

  if ( readout->getNumberOfRawHits() == 0 ) return false ;

  // Check if topology has changed
  int nrows = readout->topology().nrows() ;    // rows in x
  int ncols = readout->topology().ncolumns() ; // cols in y

  if(  (theNumOfRows != nrows)||(theNumOfCols != ncols) ){
   theNumOfRows = nrows;
   theNumOfCols = ncols;

   theBuffer.setSize(nrows,ncols);
   bufferAlreadySet = true ; 

#ifdef _DEBUG_CLUSTER_
cout << "SiTrkDetUnitClusterizer::setup constructed for barrel(SSh), pixels " << endl;
cout << readout->topology().pitch().first << " ";  // pitch size in x
cout << readout->topology().pitch().second << " "; // pitch size in y
cout << readout->topology().nrows() << " " ;       // rows in x
cout << readout->topology().ncolumns() << endl;
cout << " Thresholds " << readout->get_thresholdInNoiseUnits() << " " 
     << readout->get_seedThresholdInNoiseUnits() << " "
     << readout->get_clusterThresholdInNoiseUnits() << endl;
cout << " SiTrkDetUnitClusterizer: pixel buffer redefined " 
     << nrows << " " << ncols << endl;
#endif

  } else throw(string("SiTrkDetUnitClusterizer::setup: (theNumOfRows != nrows)||(theNumOfCols != ncols)"
                + SIUTIL::i2s(theNumOfRows) + " != " + SIUTIL::i2s(nrows) + " || " + SIUTIL::i2s(theNumOfCols) + " != " +SIUTIL::i2s(ncols)
	       ) ) ;

  // Get all noise/threshold parameters 
//  double noise = readout->get_noiseInAdcCounts() ;  // Get noise in adc units
    
  // Convert thresholds to adc units.
  // Single pixel thr. 

  thePixelThreshold = readout->get_pixelThresholdInADC() ;
  // To start the cluster search
  theSeedThreshold  = readout->get_seedThresholdInADC() ;
  // Full cluster thr.
  theClusterThreshold = readout->get_clusterThresholdInADC() ;

  return true ;

} // setup.






vector<Cluster>
SiTrkDetUnitClusterizer::clusters( SIDETUNIT::DetUnit * readout ) {

  //  vector<Cluster> result;
  theClusters.clear();
  // select seeds and copy digis to buffer()
  copy_to_buffer( readout );  // copy digis to a local pixel array(buffer)
    
  // Loop over all seeds
  for ( int i = 0; i < (int)(theSeeds.size()); i++) {
      
    if ( theBuffer(theSeeds[i]) != 0 ) {  // Is this seed still valid?
	
      // Make a cluster around this seed
      Cluster cluster = make_cluster( theSeeds[i]);
	
      // Check if the cluster is above threshold 
      if ( cluster.charge() >= theClusterThreshold )
	                           theClusters.push_back( cluster);

    } // if ( theBuffer.

  } // for ( int i .

  // Some clearing
  theSeeds.clear();  // erase seeds
  // Need to clean unused pixels.
  clear_buffer( readout );  // clear pixel array(buffer)

  return theClusters;
}  // clusters.











//-----------------------------------------------------------------------
// pixels which are not part of recognized clusters are NOT ERASED 
// during the cluster finding. Erase them now.
void SiTrkDetUnitClusterizer::clear_buffer( SIDETUNIT::DetUnit * readout ) {

#ifdef _DEBUG_CLUSTER_
cout << " SiTrkDetUnitClusterizer::clear_buffer start. "<< endl;
#endif

 vector<TrackerRawDataImpl *> rawhits = readout->getRawHits() ;

 for(vector<TrackerRawDataImpl *>::const_iterator di  = rawhits.begin() ; 
    						  di != rawhits.end() ;
						  ++di ) {
     (*_bf64) = ( *(SIDETUNIT::DetUnit::_decoder) )( *di ) ;

     int row = (int)( (*_bf64)["pixelX"] ) ;
     int col = (int)( (*_bf64)["pixelY"] ) ;

     theBuffer.set_adc( row, col, 0 ) ;

 } // for( int i.


} // clear_buffer.






//-----------------------------------------------------------------------
void SiTrkDetUnitClusterizer::copy_to_buffer( SIDETUNIT::DetUnit * readout ) throw(string) {

#ifdef _DEBUG_CLUSTER_
cout << " SiTrkDetUnitClusterizer::copy_to_buffer start:" << endl;
#endif

 // Loop over raw hits : 
 
 vector<TrackerRawDataImpl *> rawhits = readout->getRawHits() ;

 int iPixBelowThreshold = 0, iPixAboveThreshold = 0  ;

#ifdef _DEBUG_CLUSTER_
cout << " Loop start rawhits.size() = "<< rawhits.size() << endl;
#endif

 for(vector<TrackerRawDataImpl *>::const_iterator di  = rawhits.begin() ; 
    						  di != rawhits.end() ;
						  ++di ) {
   EVENT::ShortVec adcVec = (*di)->getADCValues();
   if( adcVec.size() != 1 )
     throw(string(" SiTrkDetUnitClusterizer::copy_to_buffer adcVec.size() != 1 : " + SIUTIL::i2s(adcVec.size()) ) ) ;

   int adc = adcVec[0] ;

//   if( adc <= 0 )
//    throw(string(" SiTrkDetUnitClusterizer::copy_to_buffer adc <= 0 : " + SITRK::i2s(adc) )) ;
   if( adc <= 0 ){
#ifdef _DEBUG_CLUSTER_
      cout<<"WARNING: SiTrkDetUnitClusterizer::copy_to_buffer adc <= 0 : " + SIUTIL::i2s(adc) << endl ;
#endif
      continue ;
   }

   if ( adc >= thePixelThreshold ) {

     iPixAboveThreshold++ ;

     (*_bf64) =  ( *(SIDETUNIT::DetUnit::_decoder) )( *di ) ;

     int row = (int)( (*_bf64)["pixelX"] ) ;
     int col = (int)( (*_bf64)["pixelY"] ) ;

     theBuffer.set_adc( row, col, adc) ;
//cout << " SiTrkDetUnitClusterizer::copy_to_buffer : adc = " << adc << endl;

     if ( adc >= theSeedThreshold )
	              theSeeds.push_back( PixelPos( row, col ) ) ;

   } else
      iPixBelowThreshold++ ;

 } // for( di.

#ifdef _DEBUG_CLUSTER_
cout << " Loop ended. iPixAboveThreshold = "<<iPixAboveThreshold << ", iPixBelowThreshold = " <<iPixBelowThreshold << endl ;
cout << " thePixelThreshold = " << thePixelThreshold << endl ;
cout << " SiTrkDetUnitClusterizer::copy_to_buffer end." << endl << endl;
#endif

} // copy_to_buffer.





//------------------------------------------------------------------------
Cluster 
SiTrkDetUnitClusterizer::make_cluster( const PixelPos & pix ) {

  Cluster cluster( pix, theBuffer( pix.row(), pix.col()) ) ;

  stack<PixelPos> pixel_stack ;

  theBuffer.set_adc( pix, 0);

  pixel_stack.push( pix);

  while ( ! pixel_stack.empty()) {

    PixelPos curpix = pixel_stack.top() ; pixel_stack.pop() ;

    for ( int r = curpix.row()-1; r <= curpix.row()+1; r++) {
      for ( int c = curpix.col()-1; c <= curpix.col()+1; c++) {
	if ( theBuffer( r, c ) >= thePixelThreshold ) {

	  PixelPos newpix( r, c ) ;
	  cluster.add( newpix, theBuffer(r,c) ) ;
	  theBuffer.set_adc( newpix, 0 ) ;
	  pixel_stack.push( newpix ) ;

	} // if ( theBuffer(r,c).

      } // for ( int c.

    } // for ( int r .

  } // while ( ! pixel_stack.empty.

  return cluster;

} // make_cluster.




void SiTrkDetUnitClusterizer::addRawHitsToCollection
            ( const Cluster & cl, 
	      TrackerHitImpl * trkHit, 
	      SIDETUNIT::DetUnit * rd  ){

  const vector<Cluster::Pixel> pixs =  cl.pixels() ;
  
  for(vector<Cluster::Pixel>::const_iterator ip  = pixs.begin();
                                             ip != pixs.end(); ++ip ){
  
    (*_bf64)["technology"] = (unsigned int)rd->getDetUnitType() ;
    (*_bf64)["subdet"    ] = 0 ; // VXD
    (*_bf64)["layer"     ] = rd->getLayerN() ;
    (*_bf64)["ladder"    ] = rd->getDetUnitN() ;
    (*_bf64)["pixelX"    ] = (unsigned int)(*ip).x ;
    (*_bf64)["pixelY"    ] = (unsigned int)(*ip).y ;

    SIUTIL::TcellID cellid = _bf64->getValue() ; // lcio::long64, def. in SiTrk.h

    // create elements of collections  (code copied from BarrelDetUnit::make_digis):
    TrackerRawDataImpl* rawhit = new TrackerRawDataImpl() ;

    _encoderTrackerRawData->setValue( cellid ) ;
    
    _encoderTrackerRawData->setCellID( rawhit ) ;

    rawhit->setTime( 9999 ) ;

    // according to lcio/v01-08/src/cpp/include/IMPL/TrackerRawDataImpl.h: 
    rawhit->adcValues().clear() ; rawhit->adcValues().resize( 1 ) ;
    rawhit->adcValues()[0] = (short int)( (*ip).adc ) ;
    //  or  const EVENT::ShortVec AdC; AdC.push_back( adc );// rawhit->setADCValues( AdC );
    
    trkHit->rawHits().push_back( rawhit ) ;

  } // for ip.

} // addRawHitsToCollection.






void SiTrkDetUnitClusterizer::clusterize( SIDETUNIT::DetUnit * rd ) {
 
  theRecHits.clear();
  rd->clearRecHits() ;
//  rd->clearClusters() ;

  if ( !setup( rd ) ) return ;

  ClusterParameterEstimator* parEst = theEstimatorFactory->get( *rd ) ;

  vector<Cluster> clust = clusters( rd ) ;

#ifdef _DEBUG_CLUSTER_
cout << " SiTrkDetUnitClusterizer::clusterize start:" << endl 
     <<" raw hits Nb "<< rd->getNumberOfRawHits() 
     <<" clusters Nb "<<  clust.size() << endl ;
#endif

  for (vector<Cluster>::const_iterator i=clust.begin(); i!=clust.end(); i++) {


    TrackerHitImpl* trkHit = new TrackerHitImpl() ;

    SIUTIL::LocalPoint lpos = parEst->localPosition( *i ) ;
    Hep3Vector glPos = Hep3Vector( lpos.x(), lpos.y(), lpos.z() ) ;
    rd->getGlobalPos( glPos ) ; // convert to global coordinate system

    SIUTIL::LocalError lerr = parEst->localError( *i ) ;
    float locErr[3] = { lerr.xx(), lerr.xy(), lerr.yy() } ;
    float gloErr[TRKHITNCOVMATRIX] ; // see lcio: TrackerHitImpl.h: '#define TRKHITNCOVMATRIX 6'

    rd->getGlobalError( locErr, gloErr ) ; // convert to global coordinate system

    trkHit->setCovMatrix( gloErr ) ;
    
    addRawHitsToCollection ( *i, trkHit, rd ) ;
   
    double recoHitPos[3] ; 
    recoHitPos[0] = glPos.x() ;
    recoHitPos[1] = glPos.y() ;
    recoHitPos[2] = glPos.z() ;
    trkHit->setPosition( recoHitPos ) ;

    trkHit->setdEdx( (double)( (*i).charge() ) ) ;


    theRecHits.push_back( trkHit ) ;
    
    rd->addRecHit( trkHit ) ;

  }
  delete parEst;
  return ;

} // clusterize.


}  // namespace SICLU.




