#include "Pixel/BarrelDetUnit.h"
#include "CLHEP/Units/PhysicalConstants.h"


namespace SIDETUNIT{

// Used in make_digi() ( make_digi.cc )

string BarrelDetUnit::_cellIDEncoding = 
  "technology:3,subdet:3,layer:8,ladder:10,pixelX:20,pixelY:20" ;

BarrelDetUnit::~BarrelDetUnit(){
 //delete _bf64 ;
}

BarrelDetUnit::BarrelDetUnit( SIUTIL::DetUnitType  detUnitType,
  	   unsigned int layerN, // 0,1,... (see SiTracker/src/gearInitVXD.cc)
	   unsigned int detUnitsInLayer,  // Ndu
	   unsigned int detUnitN, // 0,1,..., Ndu-1
	   double Rmin, double Rmax, 
	   double Zmin, double Zmax,
	   double pixelSizeX_duFrame, double pixelSizeY_duFrame,
	   double vxdBField,
	   double tanLorentzAnglePerTesla,
	   double pixThresholdInNoiseUnits,
	   double seedThresholdInNoiseUnits,
	   double clusterThresholdInNoiseUnits )

	      : DetUnit(detUnitType, 
	     		Rmin, Rmax, Zmin, Zmax,
 // half-sizes of detUnit in det.Unit frame, see function detUnitBarrelFrame( double * v )
			Rmin*tan( CLHEP::pi/detUnitsInLayer ), 
			Zmax, 
			0.5*( Rmax - Rmin ),
			pixelSizeX_duFrame, pixelSizeY_duFrame,
			vxdBField,
			tanLorentzAnglePerTesla,
			pixThresholdInNoiseUnits,
			seedThresholdInNoiseUnits,
			clusterThresholdInNoiseUnits ),
  	        _layerN(layerN), 
		_detUnitsInLayer(detUnitsInLayer), 
		_detUnitN(detUnitN),
		_thickness(Rmax - Rmin),
  		_RdetUnit_refPoint( Rmin ),
		_PhiDetUnit_refPoint( (detUnitN + 0.5)*(2.*CLHEP::pi/detUnitsInLayer) ),
		_DR_detUnit( Rmax-Rmin ),
		_DPhi_detUnit( 2.*CLHEP::pi/detUnitsInLayer ) {


    if( fabs( get_globalBField() ) < SIUTIL::get_epsil() )
      cout<< "WARNING: constructor BarrelDetUnit: small BField = " << get_globalBField() << endl ;

    // _bf64 used in make_digi for all events.
    _bf64 = new UTIL::BitField64( BarrelDetUnit::_cellIDEncoding ) ;

    // _cellID_bf64 reset in setRawHit for clustering.
    _cellID_bf64 = new UTIL::BitField64("") ;

#ifdef _DEBUG_DU_
  cout<< " Barrel DU before _BFieldLocal " << endl ;
#endif
    _BFieldLocal.set( 0., 0., get_globalBField() ) ;
#ifdef _DEBUG_DU_
  cout<< " Barrel DU after _BFieldLocal " << endl ;
#endif

    _BFieldLocal.rotateZ( -_PhiDetUnit_refPoint ) ;
    
    SIUTIL::detUnitBarrelFrame( _BFieldLocal ) ;

    if( (fabs(_BFieldLocal.x()) > SIUTIL::get_epsil())
     || (fabs(_BFieldLocal.z()) > SIUTIL::get_epsil())
     || (_BFieldLocal.y() < 0. )  )
       throw(string("constructor BarrelDetUnit: It is impossible: BField in local d.u. frame: x,y,z = "
         + SIUTIL::d2s(_BFieldLocal.x()) +", " + SIUTIL::d2s(_BFieldLocal.y()) +", " + SIUTIL::d2s(_BFieldLocal.z()) ) );

    double dir_x = -tanLorentzAnglePerTesla * _BFieldLocal.y() ;//!=0
    double dir_y =  tanLorentzAnglePerTesla * _BFieldLocal.x() ;// =0
    double dir_z =  1. ; 
    // E field always in -z direction, F_E in +z direction =>  e- move in +z direction

    _theDriftDirection.set( dir_x, dir_y, dir_z ) ;  // Drift direction

    set_isRefPointDone( true ) ;

    double s = sin( _PhiDetUnit_refPoint ), c = cos( _PhiDetUnit_refPoint ) ;

    _g2l = HepMatrix( 3, 3 ) ;
    _g2l(1,1) = -s  ;
    _g2l(1,2) =  c  ;
    _g2l(1,3) =  0. ;

    _g2l(2,1) =  0. ;
    _g2l(2,2) =  0. ;
    _g2l(2,3) =  1. ;

    _g2l(3,1) =  c  ;
    _g2l(3,2) =  s  ;
    _g2l(3,3) =  0. ;

    _l2g = HepMatrix( 3, 3 ) ;
    _l2g(1,1) =  -s ;
    _l2g(1,2) =  0. ;
    _l2g(1,3) =   c ;

    _l2g(2,1) =  c  ;
    _l2g(2,2) =  0. ;
    _l2g(2,3) =  s  ;

    _l2g(3,1) =  0. ;
    _l2g(3,2) =  1. ;
    _l2g(3,3) =  0. ;

    _refPoint = HepVector(3) ;
    _refPoint(1) = _RdetUnit_refPoint * c ;
    _refPoint(2) = _RdetUnit_refPoint * s ;
    _refPoint(3) = 0. ;

} // c-tor.





void BarrelDetUnit::getGlobalPos(  Hep3Vector &local_Pos ) const {

 SIUTIL::detUnitBarrelFrame_back( local_Pos ) ;

 local_Pos = local_Pos + Hep3Vector( _RdetUnit_refPoint + _thickness/2., 0., 0. ) ;

 local_Pos.rotateZ( _PhiDetUnit_refPoint ) ;
    
 return ;

} // .


void BarrelDetUnit::getGlobalError( float * locErr, float * gloErr ) const {
// locErr[3], gloErr[6].
 HepMatrix ge( 3, 3, 0 ) ;
 HepMatrix le( 3, 3, 0 ) ;
 le(1,1) = locErr[0] ; // xx
 le(1,2) = locErr[1] ; // xy
 le(1,3) = 0.        ; // xz

 le(2,1) = locErr[1] ; // yx
 le(2,2) = locErr[2] ; // yy
 le(2,3) = 0. ;        // yz

 le(3,1) = 0. ;        // zx
 le(3,2) = 0. ;        // zy
 le(3,3) = _thickness/3.464 ;  // zz  _thickness/sqrt(12)

 ge = _l2g * (le * _g2l) ; 
 
 gloErr[0] = ge(1,1) ;  // xx
 gloErr[1] = ge(1,2) ;  // xy
 gloErr[2] = ge(1,3) ;  // xz
 gloErr[3] = ge(2,2) ;  // yy
 gloErr[4] = ge(2,3) ;  // yz
 gloErr[5] = ge(3,3) ;  // zz


} // getGlobalError.



void BarrelDetUnit::getLocalError( const EVENT::FloatVec gloErr, 
                                   float * locErr ) const {

 HepMatrix ge( 3, 3, 0 ) ;
 HepMatrix le( 3, 3, 0 ) ;

 ge(1,1) = gloErr[0] ;  // xx
 ge(1,2) = gloErr[1] ;  // xy
 ge(1,3) = gloErr[2] ;  // xz

 ge(2,1) = gloErr[1] ;  // yy
 ge(2,2) = gloErr[3] ;  // yy
 ge(2,3) = gloErr[4] ;  // yz

 ge(3,1) = gloErr[2] ;  // zx
 ge(3,2) = gloErr[4] ;  // zy
 ge(3,3) = gloErr[5] ;  // zz

 le = _g2l * ( ge * _l2g ) ;

 locErr[0] = le(1,1) ;  // xx
 locErr[1] = le(1,2) ;  // xy
 locErr[2] = le(1,3) ;  // xz
 locErr[3] = le(2,2) ;  // yy
 locErr[4] = le(2,3) ;      // yz
 locErr[5] = le(3,3) ;  // zz

} // getLocalError.




void BarrelDetUnit::getLocalPos(  Hep3Vector &global_Pos ) const {

 global_Pos.rotateZ( -_PhiDetUnit_refPoint ) ;

 global_Pos = global_Pos - Hep3Vector( _RdetUnit_refPoint + _thickness/2., 0., 0. ) ;

 SIUTIL::detUnitBarrelFrame( global_Pos ) ;

 return ;

} // .



void BarrelDetUnit::getGlobalVector(  Hep3Vector &local_Vec ) const {

 SIUTIL::detUnitBarrelFrame_back( local_Vec ) ;

 local_Vec.rotateZ( _PhiDetUnit_refPoint ) ;

 return ;

} // .


void BarrelDetUnit::getLocalVector(  Hep3Vector &global_Vec ) const {

 global_Vec.rotateZ( -_PhiDetUnit_refPoint ) ;

 SIUTIL::detUnitBarrelFrame( global_Vec ) ;

 return ;

} // .






void BarrelDetUnit::getPixelPos( unsigned int pixelX,
				 unsigned int pixelY,
				 double * recoHitPos ) throw(string) {

 SIUTIL::LocalPoint lp =_topology.localPosition( SIUTIL::MeasurementPoint(pixelX,pixelY) ) ;

 Hep3Vector localPos ;
 localPos[0] = lp.x() + pixelSizeX()/2. ;
 localPos[1] = lp.y() + pixelSizeY()/2. ;
 localPos[2] = 0. ;

 getGlobalPos( localPos ) ;

 recoHitPos[0] = localPos[0];
 recoHitPos[1] = localPos[1];
 recoHitPos[2] = localPos[2];
 return ;

} // getPixelPos.






bool BarrelDetUnit::setRawHit( TrackerRawDataImpl *rawHit )
		    throw(string) {
// Set raw hit for clustering (see Detector::fillDetUnits( TrackerRawDataImpl *rawHit )).

   *_cellID_bf64 = (*_decoder)( rawHit ) ;

   int technology = (*_cellID_bf64)["technology"] ;
   int layer = (*_cellID_bf64)["layer"] ;
   int ladder = (*_cellID_bf64)["ladder"] ;

#ifdef _DEBUG_DU_20_
cout << "BarrelDetUnit::setRawHit technology ="<<technology <<", layer="<< layer<<", ladder="<<ladder << endl ;
cout << "BarrelDetUnit::setRawHit getDetUnitType() = "<< getDetUnitType() << endl ;
#endif

   if( technology != (int)( getDetUnitType() ) )
     throw(string("BarrelDetUnit::setRawHit: I do not know such technology: "
            + SIUTIL::i2s(technology)) ) ;

   if( (layer  == (int)getLayerN()) 
    && (ladder == (int)getDetUnitN())    ){

         _rawHits.push_back( rawHit ) ; 
	 return true ;

   }

   return false ;

} //setRawHit .




bool BarrelDetUnit::checkHit( Hep3Vector &hitGlobalPos, 
			      Hep3Vector &hitGlobalMom ) throw(string)  {

  double Rmin, Rmax, Zmin, Zmax ;
  getRZ( Rmin, Rmax, Zmin, Zmax ) ;
  double d_err = Rmax - Rmin;

  if( d_err < SIUTIL::get_epsil() ) throw(string("BarrelDetUnit::checkHit: To small d_err=" 
    + SIUTIL::d2s(d_err)+" for DetUnitType = " + SIUTIL::sDetUnitType(_detUnitType) )) ;

  double leftZ  = Zmin - d_err ;
  double rightZ = Zmax + d_err ;

  if( ( hitGlobalPos.z() < leftZ  )||( hitGlobalPos.z() > rightZ ) ) 
    throw(string("BarrelDetUnit::setHit: hit Z = "
          + SIUTIL::d2s(hitGlobalPos.z()) + " < " + SIUTIL::d2s( leftZ)
	  + " or " + SIUTIL::d2s(hitGlobalPos.z()) + " > " + SIUTIL::d2s( rightZ ) +" : outside layer Zmin, Zmax.")) ;

  return true ;

}  // checkHit.






void BarrelDetUnit::getDetUnit_refPoint( double &RdetUnit_refPoint, 
				   	 double &PhiDetUnit_refPoint )
                                     const throw(string) {
 if( !isRefPointDone() ) throw(string("BarrelDetUnit::getDetUnit_refPoint: ref.point is not set."));
 RdetUnit_refPoint = _RdetUnit_refPoint; 
 PhiDetUnit_refPoint = _PhiDetUnit_refPoint;
}




} //namespace SIDETUNIT{
