#include "Pixel/GaussianTailNoiseGenerator.h"

namespace SIDETUNIT{

void GaussianTailNoiseGenerator::generate(int NumberOfchannels, 
					  double threshold, 
					  double noiseRMS, 
					  map<int,double, less<int> >& theMap ) {

  // Compute number of channels with noise above threshold

  /**    gsl-1.8/doc/gsl-ref.info-1
   *
   *   Function: int gsl_sf_erf_Q_e (double X, gsl_sf_result * RESULT)
   *   These routine compute the upper tail of the Gaussian probability
   *   function Q(x) = (1/\sqrt{2\pi}) \int_x^\infty dt \exp(-t^2/2).
   *
   *  Gaussian tail probability
   */
#ifdef _DEBUG_NOISE_
  cout << " GaussianTailNoiseGenerator::generate : NumberOfchannels, threshold, noiseRMS =" 
      << NumberOfchannels <<", "<< threshold<<", "<< noiseRMS << endl;
#endif

  gsl_sf_result result;
  int status = gsl_sf_erf_Q_e(threshold, &result);
  if (status != 0) throw(string("GaussianTailNoiseGenerator::could not compute gaussian tail probability for the threshold chosen"));
  double probabilityLeft = result.val;
  
  double meanNumberOfNoisyChannels = probabilityLeft * NumberOfchannels;
  int numberOfNoisyChannels = RandPoisson::shoot(meanNumberOfNoisyChannels);
#ifdef _DEBUG_NOISE_
  cout << "  numberOfNoisyChannels = " << numberOfNoisyChannels  ;
#endif

  // draw noise at random according to Gaussian tail

  // initialise default gsl uniform generator engine
  static gsl_rng * mt19937 = gsl_rng_alloc (gsl_rng_mt19937);

  double lowLimit = threshold * noiseRMS;
#ifdef _DEBUG_NOISE_
  cout << "   lowLimit = " << lowLimit<< endl;
#endif

  for (int i = 0; i < numberOfNoisyChannels; i++) {

    // Find a random channel number    
    int theChannelNumber = (int) RandFlat::shootInt(NumberOfchannels);
    
    // Find random noise value
    double noise = gsl_ran_gaussian_tail(mt19937, lowLimit, noiseRMS);

#ifdef _DEBUG_NOISE_
  cout << " theChannelNumber = " << theChannelNumber <<" noise = "<< noise << endl;
#endif
						          
    // Fill in map
    theMap[theChannelNumber] = noise;
    
  } //  for (int i.

}  // generate.

} //namespace SIDETUNIT{

/**  gsl-1.8/doc/gsl-ref.info-2
 *
 * - Function: double gsl_ran_gaussian_tail (const gsl_rng * R, double A,
 *          double SIGMA)
 *     This function provides random variates from the upper tail of a
 *     Gaussian distribution with standard deviation SIGMA.  The values
 *     returned are larger than the lower limit A, which must be
 *     positive.  The method is based on Marsaglia's famous
 *     rectangle-wedge-tail algorithm (Ann.  Math. Stat. 32, 894-899
 *     (1961)), with this aspect explained in Knuth, v2, 3rd ed, p139,586
 *     (exercise 11).
 *
 *     The probability distribution for Gaussian tail random variates is,
 *
 *          p(x) dx = {1 \over N(a;\sigma) \sqrt{2 \pi \sigma^2}} \exp (- x^2/(2 \sigma^2)) dx
 *
 *     for x > a where N(a;\sigma) is the normalization constant,
 *
 *          N(a;\sigma) = (1/2) erfc(a / sqrt(2 sigma^2)).
 */
