#ifndef Geom_LocalError_H
#define Geom_LocalError_H

/** CMS software, ORCA.
 *  A very simple class for 2D error matrix components,
 *  used for the local frame.
 *
 *  Adapted in MarlinReco : S Shulga 2007/06/06
 *  @version $Id: LocalError.h, v 1 2007/06/06 shulga
 */

#include <cmath>
#include <iosfwd>

namespace SIUTIL{

class LocalError {
public:

  LocalError() : thexx(0), thexy(0), theyy(0) {}

  LocalError( double xx, double xy, double yy) :
    thexx(xx), thexy(xy), theyy(yy) {}

  double xx() const { return thexx;}
  double xy() const { return thexy;}
  double yy() const { return theyy;}

  /** Return a new LocalError, scaled by a factor interpreted as a 
   *  number of sigmas (standard deviations).
   *  The error matrix components are actually multiplied with the square 
   *  of the factor.
   */
  LocalError scale(double s) { 
    double s2 = s*s;
    return LocalError(s2*xx(), s2*xy(), s2*yy());
  }

  /// Return a new LocalError, rotated by an angle defined by the direction (x,y)
  LocalError rotate(double x, double y) { 
    double mag = sqrt(x*x+y*y);
    return rotateCosSin( x/mag, y/mag);
  }

  /// Return a new LocalError, rotated by an angle phi
  LocalError rotate(double phi) { 
    return rotateCosSin( cos(phi), sin(phi));
  }

  /// Return a new LocalError, rotated by an angle defined by it's cosine and sine
  LocalError rotateCosSin( double c, double s) {
    return LocalError( c*c*xx() - 2*c*s*xy() + s*s*yy(),
		       c*s*xx() + (c*c-s*s)*xy() - c*s*yy(),
		       s*s*xx() + 2*c*s*xy() + c*c*yy());
  }

private:

  double thexx;
  double thexy;
  double theyy;

};  

std::ostream & operator<<( std::ostream& s, const LocalError& err) ;

} // namespace SIUTIL.

#endif // LocalError_H
