#ifndef TP_PIXELCHIPINDICES_H
#define TP_PIXELCHIPINDICES_H

#include <iostream>

/** ORCA, CMS software, 2006
 *  Adapted in MarlinReco, Sergey Shulga, May, 2007
 *
 * Numbering of the pixels inside the readout chip (ROC).
 * There is a column index and a row index.
 * In the barrel the row index is in the global rfi direction (local X) and
 * the column index is in the global z (local Y) direction.
 * In the endcaps the row index is in the global r direction (local X) and
 * the column index in the gloabl rfi (local Y) direction.
 * The defaults are specific to 100*150 micron pixels.
 * Needs fixing: 
 *   - no limit checking (some added 2/03).
 *   - the 2nd ROC row has inverted indices.
 *   - the 2nd column of each dcol needs reversed pixe counting.
 * 
 * DO NOT USE YET FOR ROC DATA TRANSFORMATION, CAN ONLY BE USED 
 * FOR RATE STUDIES!
 * d.k. 1/2003
 */

namespace SIUTIL{

// A few constants just for error checking
// The maximum number of ROCs in the X (row) direction per sensor.
const int maxROCsInX_0 = 2;  //  
// The maximum number of ROCs in the Y (column) direction per sensor.
const int maxROCsInY_0 = 8;  //
// The nominal number of double columns per ROC is 26. Very unlikely
// to ever change, but who knows?
const int nominalDColsPerROC_0 = 26; 
// Default ROC size 
const int defaultROCSizeInX_0 = 80;  // ROC row size in pixels 
const int defaultROCSizeInY_0 = 52;  // ROC col size in pixels 
// Default DET barrel size 
//const int defaultDetSizeInX = 106;  // Det barrel row size in pixels 
const int defaultDetSizeInX_0 = 160;  // Det barrel row size in pixels 
const int defaultDetSizeInY_0 = 416;  // Det barrel col size in pixels 

// Check the limits
#define TP_CHECK_LIMITS

//#define _DEBUG_PIXELCHIPS_

class PixelChipIndices {

 public:

  /**
   * Constructor with all 4 parameters defines explicitly:
   * pixsInChipY = number of column's pixels in a ROC
   * pixsInChipX = number of row's pixels in a ROC
   * pixsInDetY = number of column's pixels in a detector module.
   * pixsInDetX = number if row's pixels in a detector module.
   */
  PixelChipIndices(const int pixsInChipY, const int pixsInChipX, 
		   const int pixsInDetY,  const int pixsInDetX ) throw(string) :
    _pixsInChipY(pixsInChipY), _pixsInChipX(pixsInChipX), 
    _pixsInDetY(pixsInDetY), _pixsInDetX (pixsInDetX) {

    _chipsInX = _pixsInDetX / _pixsInChipX; // number of ROCs in X (Rows)
    _chipsInY = _pixsInDetY / _pixsInChipY; // number of ROCs in Y (Cols)

    // A double column consists of 2 columns
    _DColsInChip = _pixsInChipY / 2 ;

    _maxROCsInX = _chipsInX ;   // 2
    _maxROCsInY = _chipsInY ;   // 8
    _nominalDColsPerROC = _DColsInChip ;   // 26
    _defaultROCSizeInX =  _pixsInChipX ;   // 80
    _defaultROCSizeInY =  _pixsInChipY ;   // 52
    _defaultDetSizeInX =  _pixsInDetX ;    // 160
    _defaultDetSizeInY =  _pixsInDetY ;    // 416

#ifdef TP_CHECK_LIMITS
if( _chipsInX < 1 || _chipsInX > _maxROCsInX ) throw(string("c-tor PixelChipIndices(int,int,int,int): Error in _chipsInX = ") + SIUTIL::i2s(_chipsInX) + "  _pixsInDetX=" + SIUTIL::i2s(_pixsInDetX) + "  _pixsInChipX=" + SIUTIL::i2s(_pixsInChipX) ) ;
if( _chipsInY < 1 || _chipsInY > _maxROCsInY ) throw(string("c-tor PixelChipIndices(int,int,int,int): Error in _chipsInY = ") + SIUTIL::i2s(_chipsInY) + "  _pixsInDetY=" + SIUTIL::i2s(_pixsInDetY) + "  _pixsInChipY=" + SIUTIL::i2s(_pixsInChipY) ) ;
if( _DColsInChip != _nominalDColsPerROC ) throw(string("c-tor PixelChipIndices(int,int,int,int): Error in DCOLsInChip = ")    + SIUTIL::i2s(_DColsInChip) ) ;
#endif

#ifdef _DEBUG_PIXELCHIPS_
cout << "C-tor PixelChipIndices(int,int,int,int): Pixel det initialized with " << _chipsInX 
     << " chips in x and " << _chipsInY << " in y " << endl ;
#endif
  } // Constructor PixelChipIndices. 



  // Constructor with the ROC size fixed to default.
  PixelChipIndices(const int pixsInDetY,  const int pixsInDetX ) throw(string) : 
    _pixsInChipY(defaultROCSizeInY_0), _pixsInChipX(defaultROCSizeInX_0), 
    _pixsInDetY(pixsInDetY), _pixsInDetX (pixsInDetX) {

    _chipsInX = _pixsInDetX / _pixsInChipX; // number of ROCs in X
    _chipsInY = _pixsInDetY / _pixsInChipY;    // number of ROCs in Y

    // A double column consists of 2 columns
    _DColsInChip = _pixsInChipY / 2 ;  

#ifdef TP_CHECK_LIMITS
if( _chipsInX < 1 || _chipsInX > _maxROCsInX ) throw(string("c-tor PixelChipIndices(int,int): Error in _chipsInX = ") + SIUTIL::i2s(_chipsInX) ) ;
if( _chipsInY < 1 || _chipsInY > _maxROCsInY ) throw(string("c-tor PixelChipIndices(int,int): Error in _chipsInY = ") + SIUTIL::i2s(_chipsInY) ) ;
if( _DColsInChip != _nominalDColsPerROC ) throw(string("c-tor PixelChipIndices(int,int): Error in DCOLsInChip = ")    + SIUTIL::i2s(_DColsInChip) ) ;
#endif

#ifdef _DEBUG_PIXELCHIPS_
cout << "C-tor PixelChipIndices(int,int): Pixel det initialized with " << _chipsInX 
     << " chips in x and " << _chipsInY << " in y " << endl ;
#endif
  }  // Constructor PixelChipIndices.




  // Constructor with all dimensioned fixed to BARREL default ROC.
  // Usefull only for barrel.
  PixelChipIndices( ) throw(string) : 
    _pixsInChipY(defaultROCSizeInY_0), _pixsInChipX(defaultROCSizeInX_0), 
    _pixsInDetY(defaultDetSizeInY_0), _pixsInDetX (defaultDetSizeInX_0) {

    _chipsInX = _pixsInDetX / _pixsInChipX;  // number of ROCs in X
    _chipsInY = _pixsInDetY / _pixsInChipY;  // number of ROCs in Y

    // A double column consists of 2 columns
    _DColsInChip = _pixsInChipY / 2 ;

#ifdef TP_CHECK_LIMITS
if( _chipsInX < 1 || _chipsInX > _maxROCsInX ) throw(string("c-tor PixelChipIndices(): Error in _chipsInX = ") + SIUTIL::i2s(_chipsInX) ) ;
if( _chipsInY < 1 || _chipsInY > _maxROCsInY ) throw(string("c-tor PixelChipIndices(): Error in _chipsInY = ") + SIUTIL::i2s(_chipsInY) ) ;
if( _DColsInChip != _nominalDColsPerROC ) throw(string("c-tor PixelChipIndices(): Error in DCOLsInChip = ")    + SIUTIL::i2s(_DColsInChip) ) ;
#endif

#ifdef _DEBUG_PIXELCHIPS_
cout << "C-tor PixelChipIndices(): Pixel det initialized with " << _chipsInX 
     << " chips in x and " << _chipsInY << " in y " << endl ;
#endif
  } //  Constructor PixelChipIndices.

  //************************************************************************
  ~PixelChipIndices() {}
  //***********************************************************************

  void print(void) const {
    cout << " Pixel det with " << _chipsInX << " chips in x and "
	 << _chipsInY << " in y " << endl; 
    cout << " Pixel pixels in row " << _pixsInDetX << " and in column " 
	 << _pixsInDetY << endl;  
    cout << " Pixels in row of one chip " << _pixsInChipX << " and in column " 
	 << _pixsInChipY << endl;  
    cout << " Double columns per ROC " << _DColsInChip << endl;
  } 
  

  /**  ORCA comments:
   *  Convert detector indices to chip indices. 
   *  These are SINGLE COLUMN indices!
   *  IN: pixelX - detector index in the X (row) direction from 1 to max
   *  IN: pixelY - detector index in the Y (col) direction from 1 to max
   *  OUT: chipX - ROC index in the X direction, goes from 0 to 1
   *  OUT: chipY - ROC index in the Y direction, goes from 0 to 7 
   *  OUT: row - row (X) index within a ROC, goes from 1 to max(80)
   *  OUT: col - column (Y) index within a ROC, goes from 1 to max(52) 
   *  WARNING: the pixel indices within a ROC are calulated from the WRONG
   *           side. CHECK THIS LATER => Needs to be fixed.
   */
  bool chipIndices(const int pixelX, const int pixelY, int & chipX,
		   int & chipY, int & row, int & col) const {

    int pixX = pixelX - 1 ;
    int pixY = pixelY - 1 ;

    chipX = pixX / _pixsInChipX; // row index of the chip 0-1
    chipY = pixY / _pixsInChipY; // col index of the chip 0-7

    if( (chipX > _chipsInX - 1)
      ||(chipY > _chipsInY - 1) ) return false ; // hit pixel outside the det.unit

    row = ( pixX % _pixsInChipX ) + 1; // row in chip from 1 to max(80)
    col = ( pixY % _pixsInChipY ) + 1; // col in chip from 1 to max(52)

#ifdef TP_CHECK_LIMITS
    if( chipX < 0 || chipX > (_chipsInX-1) ){ 
      cout << " chipIndices: Error in chipX " << chipX << endl; exit(0) ; }
    if( chipY < 0 || chipY > (_chipsInY-1) ){ 
      cout << " chipIndices: Error in chipY " << chipY << endl; exit(0) ; }
    if( row < 1 || row > _pixsInChipX ){ 
      cout << " chipIndices: Error in ROC row# " << row << endl; exit(0) ; }
    if( col < 1 || col > _pixsInChipY ){ 
      cout << " chipIndices: Error in ROC col# " << col << endl; exit(0) ; }
#endif

    return true ;

  } // chipIndices.





  /**
   *  Calcilate a single number ROC index from the 2 ROC indices (coordinates)
   *  chipX and chipY.  Goes from 1 to 16.
   *  chipX - ROC index in the X direction, goes from 0 to maxX-1
   *  chipY - ROC index in the Y direction, goes from 0 to maxY-1
   *	const int maxROCsInX = 2
   *	const int maxROCsInY = 8
   */
  inline int chipIndex(const int chipX, const int chipY) {
    //SSh    int chipIndex = (chipX * 8) + chipY + 1 ;  // index 1-16
    int chipIndex = (chipX * _maxROCsInY) + chipY + 1 ; // index 1 - (_maxROCsInX * _maxROCsInY)

#ifdef TP_CHECK_LIMITS
if( chipIndex < 1 || chipIndex > (_maxROCsInX * _maxROCsInY) ) {
  cout << " PixelChipIndices::chipIndex: Error in ROC index " << chipIndex << endl;
  exit(0) ; }
#endif

    return chipIndex;

  } // chipIndex.


  /**
   *  Calulate the DCOL indices from single column indices.
   *  IN: row = row index witin a single column , goes 1 to max (80)
   *  IN: col = single column index, goes 1 to max (52)
   *  Returns a pair<pixId, dColumnId>
   *  pixId = pixel index within 1 double column, goes from 1 to max (106)
   *  dColumnId = double column (DCOL) index, goes from 1 to max (26)
   */
  pair<int,int> dColumn(const int row, const int col) const {

    // double column 1-26 :
    int dColumnId = (col - 1)/2 + 1;

     //id in dCol 1-106 :
    int pixId = row + (col - dColumnId * 2 + 1) * _pixsInChipX ;

#ifdef TP_CHECK_LIMITS
if(dColumnId < 1 || dColumnId > _DColsInChip ) { cout << " PixelChipIndices: Error in DColumn Id " << dColumnId << endl;
 exit(0) ; }
if(pixId < 1 || pixId > (2*_pixsInChipX) ){ cout << " PixelChipIndices: Error in pix Id " << pixId << endl;
 exit(0) ; }
#endif

    return pair<int,int>( pixId, dColumnId ) ;

  } //  dColumn.

  //************************************************************************
  // Return a unique double column index with the whole detector module.
  // Usefull only for diagnostics.
  // IN: dcol = DCOL index within a ROC, goes from 1 to max 
  // IN: chipIndex = ROC index, goes from 1 to max .
  // Return a 3 digit number XYZ, where X is the ROC index and YZ is 
  // the DCOL index. 
  inline int dColumnIdInDet(const int dcol, const int chipIndex) {
    int id = dcol + ( 100 * ( chipIndex - 1 ) ) ;
    if( dcol > 100 ) { cout <<"PixelChipIndicesd::ColumnIdInDet: error : dcol > 100 : "<< dcol << endl ; exit(0); }

#ifdef TP_CHECK_LIMITS
int maxChipIndex = ( _maxROCsInX * _maxROCsInY) + _maxROCsInY + 1 ;
if( id < 1 || id > ( ( maxChipIndex - 1 )*100 + _nominalDColsPerROC) ) {
      cout << " PixelChipIndices: Error in det dcol id " << id << endl;
      exit(0) ; }
#endif

    return id;
  } 
  
  //***********************************************************************
 private:

    int _pixsInChipY;   // Columns per chip
    int _pixsInChipX;   // Rows per chip
    int _pixsInDetY;    // Columns per Det
    int _pixsInDetX;    // Rows per Det
    int _chipsInX;      // Chips in det in X (column direction)
    int _chipsInY;      // Chips in det in Y (row direction)
    int _DColsInChip;   // Number of double column per ROC

// introduce members for compatibility with ORCA definitions (see #define TP_CHECK_LIMITS)
    int _maxROCsInX ; 
    int _maxROCsInY ;
    int _nominalDColsPerROC ; 
    int _defaultROCSizeInX ;
    int _defaultROCSizeInY ;
    int _defaultDetSizeInX ;
    int _defaultDetSizeInY ;

} ; // class PixelChipIndices.

}// namespace SIUTIL{

#endif
